(*
Copyright 2009 Christian Sternagel, René Thiemann, Sarah Winkler, Harald Zankl

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory ArgumentFilter
imports TermOrder
begin

datatype af_entry = Collapse nat | AFList "nat list"
types 'f afs = "'f \<Rightarrow> nat \<Rightarrow> af_entry"
datatype 'f filtered = FPair 'f  nat
(* the main issue to not use pairs is that we can provide an alternative show function *)

fun
  filtered_fun :: "'f filtered \<Rightarrow> 'f"
where
  "filtered_fun (FPair f n) = f"

instantiation filtered :: ("show") "show" begin
  definition "shows_prec (d::nat) ff = shows (filtered_fun ff)"
  definition "shows_list_filtered (ffs::('f::show) filtered list) = shows_list_aux shows ffs"
  instance by (intro_classes, auto simp: shows_prec_filtered_def shows_list_filtered_def)
end

fun apply_af_entry :: "'f \<Rightarrow> af_entry \<Rightarrow> ('f,'v)term list \<Rightarrow> ('f,'v)term" where
  "apply_af_entry _ (Collapse i) ts = ts ! i"
| "apply_af_entry f (AFList is) ts = Fun f (map (\<lambda> i. ts ! i) is)"

fun wf_af_entry :: "nat \<Rightarrow> af_entry \<Rightarrow> bool"
where "wf_af_entry n (Collapse i) = (i < n)"
  |   "wf_af_entry n (AFList is) = forall (\<lambda> i. i < n) is"

definition wf_afs :: "'f afs \<Rightarrow> bool"
where "wf_afs \<pi> \<equiv> \<forall> f n. wf_af_entry n (\<pi> f n)"

fun afs_term :: "'f afs \<Rightarrow> ('f,'v)term \<Rightarrow> ('f filtered,'v)term" where 
  "afs_term \<pi> (Fun f ts) = (let l = length ts in apply_af_entry (FPair f l) (\<pi> f l) (map (afs_term \<pi>) ts))"
| "afs_term \<pi> (Var x) = Var x"

definition afs_rule :: "'f afs \<Rightarrow> ('f,'v)rule \<Rightarrow> ('f filtered,'v)rule"
where "afs_rule \<pi> lr \<equiv> (afs_term \<pi> (fst lr), afs_term \<pi> (snd lr))"

definition afs_rules :: "'f afs \<Rightarrow> ('f,'v)rules \<Rightarrow> ('f filtered,'v)rules"
where "afs_rules \<pi> R = map (afs_rule \<pi>) R"

definition
  afs_check ::
    "shows \<Rightarrow> 'f::show afs \<Rightarrow> (('f filtered,'v::show)rule \<Rightarrow> shows check) \<Rightarrow> ('f,'v)rule  \<Rightarrow> shows check"
where
  "afs_check rel \<pi> g lr \<equiv> let pl = afs_term \<pi> (fst lr); pr = afs_term \<pi> (snd lr) in g (pl,pr) <+? (\<lambda>s. 
  ''could not orient '' +#+ shows(fst lr) +@+ '' '' +#+ rel +@+ '' '' +#+ shows(snd lr) +@+ shows_nl +@+
  ''pi( '' +#+ shows(fst lr) +@+ '' ) = '' +#+ shows pl +@+ shows_nl +@+
  ''pi( '' +#+ shows(snd lr) +@+ '' ) = '' +#+ shows pr +@+ shows_nl +@+ s
)"
 
locale afs =
  fixes \<pi> :: "'f afs"
  assumes wf_afs: "wf_afs \<pi>"
begin

fun afs_subst :: "('f,'v)sub \<Rightarrow> ('f filtered,'v)sub"
where "afs_subst (Sub sigma) = Sub (\<lambda> x. afs_term \<pi> (sigma x))"

definition afs_rel :: "('f filtered,'v)term ars \<Rightarrow> ('f,'v)term ars"
where "afs_rel R = {lr | lr. afs_rule \<pi> lr \<in> R}"

lemma afs_subst[simp]: "afs_term \<pi> (t \<cdot> \<sigma>) = afs_term \<pi> t \<cdot> afs_subst \<sigma>"
proof (cases \<sigma>)
  case (Sub sig)
  have "afs_term \<pi> (t \<cdot> Sub sig) = afs_term \<pi> t \<cdot> afs_subst (Sub sig)"
  proof (induct t, simp)
    case (Fun f ss)
    let ?l = "length ss"
    let ?e = "\<pi> f ?l"
    from wf_afs have wf_e: "wf_af_entry ?l ?e" unfolding wf_afs_def by auto
    show ?case
    proof (cases ?e)
      case (Collapse i)
      with wf_e have "i < ?l" by simp
      with Collapse Fun(1)[of "ss ! i"] show ?thesis
	by (auto simp: Let_def)
    next
      case (AFList "is")
      with wf_e have wf: "\<forall> i \<in> set is. i < ?l" by auto 
      with AFList Fun(1) show ?thesis 
	by auto
    qed
  qed
  with Sub show ?thesis by simp
qed

lemma afs_sub_closed : 
  assumes "sub_closed R"
  shows "sub_closed (afs_rel R)"
unfolding sub_closed_def
proof (rule subsetI, simp add: split_paired_all)
  fix ss ts
  assume "(ss,ts) \<in> CS (afs_rel R)"
  thus "(ss,ts) \<in> afs_rel R"
  proof (induct)
    case (sub_closureI s t \<sigma>)
    from this have "(afs_term \<pi> s, afs_term \<pi> t) \<in> R" unfolding afs_rel_def afs_rule_def by simp
    with assms have "(afs_term \<pi> s \<cdot> afs_subst \<sigma>, afs_term \<pi> t \<cdot> afs_subst \<sigma>) \<in> R" unfolding sub_closed_def 
      using sub_closure.sub_closureI by auto    
    thus ?case unfolding afs_rel_def afs_rule_def by auto
  qed
qed

lemma afs_SN : 
  assumes "SN R"
  shows "SN (afs_rel R)"
proof (rule ccontr)
  assume "\<not> ?thesis"
  from this obtain f where nSN: "\<forall> i. (f i, f (Suc i)) \<in> afs_rel R" unfolding SN_defs by blast
  obtain g where "g = (\<lambda> i. afs_term \<pi> (f i))" by blast
  with nSN have "\<forall> i. (g i, g (Suc i)) \<in> R" unfolding afs_rel_def afs_rule_def by auto
  hence "\<not> (SN R)" unfolding SN_defs by auto
  with assms show False by blast
qed

lemma afs_compat : 
  assumes "NS O S \<subseteq> S"
  shows "afs_rel NS O afs_rel S \<subseteq> afs_rel S"
using assms
  by (unfold afs_rel_def afs_rule_def, clarify, auto)

lemma afs_NS_mono : 
  assumes acc: "all_ctxt_closed R"
  and refl: "refl R"
  shows "one_step_ctxt_closed (afs_rel R)"
unfolding one_step_ctxt_closed_def
proof (intro allI impI)
  fix f bef s t aft
  assume st: "(s,t) \<in> afs_rel R"
  let ?ps = "afs_term \<pi> s"
  let ?pt = "afs_term \<pi> t"
  let ?bsa = "bef @ s # aft"
  let ?bta = "bef @ t # aft"
  from st have pst: "(?ps,?pt) \<in> R" unfolding afs_rel_def afs_rule_def by auto
  let ?l = "Suc (length bef + length aft)"
  let ?ls = "length (bef @ s # aft)"
  let ?lt = "length (bef @ t # aft)"
  let ?af = "\<pi> f ?l"
  from wf_afs have wf_af: "wf_af_entry ?l ?af" unfolding wf_afs_def by simp
  let ?msa = "map (afs_term \<pi>) ?bsa"
  let ?mta = "map (afs_term \<pi>) ?bta"
  let ?pfs = "apply_af_entry (FPair f ?l) ?af ?msa"
  let ?pft = "apply_af_entry (FPair f ?l) ?af ?mta"
  have all2: "\<forall> i < ?l. (?msa ! i, ?mta ! i) \<in> R"
  proof (intro allI impI)
    fix i
    assume "i < ?l"
    hence i_s: "i < ?ls" and i_t: "i < ?lt" by simp_all
    {
      assume bef: "i < length bef"
      with refl have "(?msa ! i, ?mta ! i) \<in> R" unfolding refl_on_def 
	by (simp only: nth_map[OF i_s], simp add: nth_append)
    }
    moreover
    {
      assume "i = length bef"
      with pst have "(?msa ! i, ?mta ! i) \<in> R" unfolding refl_on_def
	by (simp only: nth_map[OF i_s], simp add: nth_append)
    }
    moreover
    {
      assume i: "i > length bef"
      from this obtain k where k1: "i - length bef = Suc k" by (cases "i - length bef", arith+)
      with i_s have k2: "k < length aft" by auto
      from i refl have "(?msa ! i, ?mta ! i) \<in> R" unfolding refl_on_def 
	by (simp only: nth_map[OF i_s] nth_map[OF i_t], simp add: nth_append k1)
    }
    moreover
    {
      have "i < length bef \<or> i = length bef \<or> i > length bef" by auto
    }      
    ultimately
    show "(?msa ! i, ?mta ! i) \<in> R" by blast
  qed
  have "(?pfs,?pft) \<in> R"
  proof (cases ?af)
    case (Collapse i)
    with wf_af have "i < ?l"  by simp
    with all2 Collapse show ?thesis by auto
  next
    case (AFList ids)
    with wf_af have ids: "\<forall> i \<in> set ids. i < ?l" by simp
    show ?thesis 
    proof (simp only: AFList apply_af_entry.simps)
      let ?msi = "map (op ! (map (afs_term \<pi>) ?bsa)) ids"
      let ?mti = "map (op ! (map (afs_term \<pi>) ?bta)) ids"
      from ids all2 have "\<forall> i < length ids. (?msi ! i, ?mti ! i) \<in> R" by auto
      with acc show "(Fun (FPair f ?l) ?msi, Fun (FPair f ?l) ?mti) \<in> R"
	unfolding all_ctxt_closed_def by force
    qed
  qed
  thus "(Fun f ?bsa, Fun f ?bta) \<in> afs_rel R" unfolding afs_rel_def afs_rule_def by simp
qed
end  


locale afs_redpair_impl = afs \<pi> + redpair_order S NS + redpair_impl S NS cS cNS
  for \<pi> :: "('f::show)afs" and S NS :: "('f filtered,'v::show)trs" and cS cNS

locale mono_afs_redpair_impl = afs \<pi> + redpair_order S NS + mono_redpair_impl S NS cS cNS cm
  for \<pi> :: "('f::show)afs" and S NS :: "('f filtered,'v::show)trs" and cS cNS cm

sublocale mono_afs_redpair_impl \<subseteq> afs_redpair_impl ..

sublocale afs_redpair_impl \<subseteq> redpair_impl "afs_rel S" "afs_rel NS" "afs_check rS \<pi> cS" "afs_check rNS \<pi> cNS"
proof -
  let ?S = "afs_rel S"
  let ?NS = "afs_rel NS"
  let ?cS = "afs_check rS \<pi> cS"
  let ?cNS = "afs_check rNS \<pi> cNS"
  show "redpair_impl ?S ?NS ?cS ?cNS"
  proof
    show "SN ?S" 
      by (rule afs_SN, rule SN)
  next
    show "?NS O ?S \<subseteq> ?S"
      by (rule afs_compat, rule compat)
  next
    show "sub_closed ?S"
      by (rule afs_sub_closed, rule subS)
  next
    show "sub_closed ?NS"
      by (rule afs_sub_closed, rule subNS)
  next
    show "one_step_ctxt_closed ?NS"
      by (rule afs_NS_mono, rule all_ctxt_closed, rule refl_NS) 
  next
    fix lr
    assume "isOK (afs_check rS \<pi> cS lr)"
    thus "lr \<in> ?S" unfolding afs_check_def afs_rel_def 
      using checkSSound by (auto simp: split_paired_all Let_def afs_rule_def)
  next
    fix lr
    assume "isOK (afs_check rNS \<pi> cNS lr)"
    thus "lr \<in> ?NS" unfolding afs_check_def afs_rel_def 
      using checkNSSound by (auto simp: split_paired_all Let_def afs_rule_def)
  qed
qed

fun mono_af_entry :: "nat \<Rightarrow> af_entry \<Rightarrow> bool"
where "mono_af_entry n (Collapse i) = (n \<le> 1)"
  | "mono_af_entry n (AFList ids) = forall (\<lambda> i. i mem ids) [0 ..< n]"

definition mono_afs :: "'f afs \<Rightarrow> bool"
where "mono_afs \<pi> \<equiv> \<forall> f n. mono_af_entry n (\<pi> f n)"

sublocale mono_afs_redpair_impl \<subseteq> mono_redpair_impl "afs_rel S" "afs_rel NS" "afs_check rS \<pi> cS" "afs_check rNS \<pi> cNS" "cm \<guillemotright> check (mono_afs \<pi>) mess"
proof
  let ?S = "afs_rel S"
  let ?pi = "afs_term \<pi>"
  assume "isOK (cm \<guillemotright> check (mono_afs \<pi>) mess)"
  with checkMonoSound have monoS: "one_step_ctxt_closed S" and mono_pi: "mono_afs \<pi>" by auto
  show "one_step_ctxt_closed ?S"
    unfolding afs_rel_def afs_rule_def one_step_ctxt_closed_def
  proof (intro allI impI, simp del: afs_term.simps)
    fix f bef s t aft
    assume st: "(?pi s, ?pi t) \<in> S"
    let ?bsa = "bef @ s # aft"
    let ?bta = "bef @ t # aft"
    let ?fs = "Fun f ?bsa"
    let ?ft = "Fun f ?bta"
    let ?l = "Suc (length bef + length aft)"
    let ?lba = "length bef + length aft"
    let ?e = "\<pi> f ?l"
    from mono_pi have mono_e: "mono_af_entry ?l ?e" unfolding mono_afs_def by auto
    from wf_afs have wf_e: "wf_af_entry ?l ?e" unfolding wf_afs_def by auto
    show "(?pi ?fs, ?pi ?ft) \<in> S"
    proof (cases ?e)
      case (Collapse i)
      with mono_e wf_e show ?thesis using st by auto
    next
      case (AFList ids)
      thus ?thesis 
      proof simp
	let ?ids = "{n | n. n < ?l}"
	let ?f = "FPair f ?l"
	obtain l where l: "?l = l" by auto
	from AFList wf_e have ids1: "set ids \<subseteq> ?ids" by auto
	from AFList mono_e have ids2: "?ids \<subseteq> set ids" 	
	  by (simp only: l, auto simp: mem_iff)
	from ids1 ids2 have ids: "set ids = ?ids" by auto
	let ?eq = "\<lambda> i. ?pi (?bsa ! i) = ?pi (?bta ! i)"
	let ?gr = "\<lambda> i. (?pi (?bsa ! i), ?pi (?bta ! i)) \<in> S"
	let ?lb = "length bef"
	have geq: "\<forall> i. ?eq i \<or> ?gr i"
	proof
	  fix i
	  {
	    assume "i < ?lb" hence "?eq i" by (simp add: nth_append)
	  }
	  moreover
	  {
	    assume "i = ?lb"
	    with st have "?gr i" by (simp add: nth_append)
	  }
	  moreover 
	  {
	    assume i2: "i > ?lb"
	    then obtain k where k1: "i - length bef = Suc k" by (cases "i - length bef", arith+)
	    have "?eq i" by (simp only: nth_append, simp add: i2, simp add: k1)
	  }
	  moreover
	  {
	    have "i < ?lb \<or> i = ?lb \<or> i > ?lb" by arith
	  }
	  ultimately show "?eq i \<or> ?gr i" by blast
	qed
	from ids have lb: "?lb \<in> set ids" by auto
	from st have "?gr ?lb" by (simp add: nth_append)
	with lb have gr: "\<exists> i \<in> set ids. ?gr i \<and> i = ?lb" by auto
	obtain ss ts pi g where g: "g = ?f" and ss: "ss = ?bsa" and ts: "ts = ?bta" and pi: "pi = (?pi :: ('f,'v)term \<Rightarrow> ('f filtered,'v)term)" by auto
	let ?eq = "\<lambda> i. pi (ss ! i) = pi (ts ! i)"
	let ?gr = "\<lambda> i. (pi (ss ! i), pi (ts ! i)) \<in> S"
	let ?ss = "map (op ! (map pi ss))"
	let ?ts = "map (op ! (map pi ts))"
	let ?both = "\<lambda> n. ?ts (take n ids) @ ?ss (drop n ids)"
	let ?gss = "Fun g (?ss ids)"
	let ?gboth = "\<lambda> n. Fun g (?both n)"
	from geq ss ts pi have geq: "\<forall> i. ?eq i \<or> ?gr i" by auto
	from gr ss ts pi have "\<exists> i \<in> set ids. ?gr i \<and> i = ?lb" by auto
	from this obtain i where i: "?gr i" and iids: "i \<in> set ids" and idef: "i = ?lb" by auto
	have "\<forall> n. (i \<in> set (take n ids) \<longrightarrow> (?gss, ?gboth n) \<in> S) \<and> (i \<notin> set (take n ids) \<longrightarrow> ?gss = ?gboth n)" (is "\<forall> n. ?p n") 
	proof (intro allI)
	  fix n
	  show "?p n"
	  proof (induct n, simp)
	    case (Suc n)
	    show ?case
	    proof (cases "n \<ge> length ids")
	      case True
	      with Suc show ?thesis by simp
	    next
	      case False 
	      hence n: "n < length ids" by auto
	      let ?bef = "?ts (take n ids)"
	      let ?aft = "?ss (drop (Suc n) ids)"
	      let ?ssn = "pi (ss ! (ids ! n))"
	      let ?tsn = "pi (ts ! (ids ! n))"
	      from n ids have ids_n: "ids ! n < ?l" by force
	      have gn: "?both n = ?bef @ ?ssn # ?aft"
		by (simp only: drop_Suc_conv_tl[OF n, symmetric], simp, simp add: nth_map[of "ids ! n" ?bsa pi, simplified, OF ids_n] ss)
	      have gsn: "?both (Suc n) = ?bef @ ?tsn # ?aft"
		by (simp only: take_Suc_conv_app_nth[OF n], simp, simp add: nth_map[of "ids ! n" ?bta pi, simplified, OF ids_n] ts)
	      {
		assume "?ssn \<noteq> ?tsn \<or> ids ! n = i"
		with geq i have "(?ssn, ?tsn) \<in> S" by blast		
		with monoS[unfolded one_step_ctxt_closed_def] 
		have gnsn: "(?gboth n, ?gboth (Suc n)) \<in> S"
		  by (simp only: gn gsn)
	      } note noteq_i = this
	      {
		assume "ids ! n \<noteq> i"
		{
		  assume "ids ! n < ?lb"
		  hence "?gboth (Suc n) = ?gboth n" by (simp add: gn gsn, simp add: ss ts, simp add: nth_append)
		}
		moreover
		{
		  assume i2: "ids ! n > ?lb"
		  then obtain k where k1: "ids ! n - length bef = Suc k" by (cases "ids ! n - length bef", arith+)
		  have "?gboth (Suc n) = ?gboth n" by (simp add: gn gsn, simp add: ss ts, simp only: nth_append, simp add: i2, simp add: k1)
		}
		moreover
		{
		  from `ids ! n \<noteq> i` have "ids ! n < ?lb \<or> ids ! n > ?lb" by (simp only: idef, arith)
		}
		ultimately have "?gboth (Suc n) = ?gboth n" by blast
	      } note eq = this	      
	      show ?thesis 
	      proof (cases "i \<in> set (take (Suc n) ids)")
		case False 
		hence i_not: "i \<notin> set (take n ids)" and i_not2: "ids ! n \<noteq> i" using take_Suc_conv_app_nth[OF n] by auto
		from i_not Suc have rec: "?gss = ?gboth n" by blast 
		with False eq[OF i_not2] rec show ?thesis by simp 
	      next
		case True note oTrue = this
		show ?thesis
		proof (cases "i \<in> set (take n ids)")
		  case False
		  with True have id: "ids ! n = i" using take_Suc_conv_app_nth[OF n] by auto
		  from False Suc have rec: "?gss = ?gboth n" by blast
		  from noteq_i id have "(?gboth n, ?gboth (Suc n)) \<in> S" by blast
		  with True rec show ?thesis by simp
		next
		  case True
		  with Suc have rec: "(?gss, ?gboth n) \<in> S" by blast
		  show ?thesis
		  proof (cases "ids ! n = i")
		    case False
		    from rec eq[OF False] oTrue show ?thesis by simp
		  next
		    case True
		    with noteq_i have "(?gboth n, ?gboth (Suc n)) \<in> S" by blast
		    from rec this have "(?gss, ?gboth (Suc n)) \<in> S" using trans_S unfolding trans_def by blast
		    with oTrue show ?thesis by simp
		  qed
		qed
	      qed
	    qed
	  qed
	qed
	with spec[OF this, of "length ids"] iids have "(Fun g (?ss ids), Fun g (?ts ids)) \<in> S" by auto
	with g ss ts pi show "(Fun ?f (map (op ! (map ?pi bef @ ?pi s # map ?pi aft)) ids),
	  Fun ?f (map (op ! (map ?pi bef @ ?pi t # map ?pi aft)) ids)) \<in> S"
	  by auto 
      qed
    qed
  qed
qed
  

types 'f afs_list = "(('f \<times> nat) \<times> af_entry)list"

definition default_af_entry :: "nat \<Rightarrow> af_entry"
  where "default_af_entry n \<equiv> AFList [0..<n]"

definition afs_list_to_afs :: "'f afs_list \<Rightarrow> 'f afs"
  where "afs_list_to_afs \<pi> f n \<equiv> case lookup (f,n) \<pi> of Some entry \<Rightarrow> entry | None \<Rightarrow> default_af_entry n"

definition afs_to_unused_arity :: "'f afs_list \<Rightarrow> nat"
  where "afs_to_unused_arity \<pi> \<equiv> foldr (\<lambda> ((f,n),e) m. max n m) \<pi> 0"

lemma afs_to_unused_arity_sound : shows "\<forall> e \<in> set pi. snd (fst e) \<le> afs_to_unused_arity pi"
proof (induct pi, simp)
  case (Cons e es)
  show ?case
  proof(clarify, simp)
    fix f n en
    assume "((f,n),en) = e \<or> ((f,n),en) \<in> set es" (is "?one \<or> ?two")
    thus "n \<le> afs_to_unused_arity (e # es)"
    proof
      assume ?one thus ?thesis unfolding afs_to_unused_arity_def by auto
    next
      assume ?two with Cons have "n \<le> afs_to_unused_arity es" by auto
      also have "\<dots> \<le> afs_to_unused_arity (e # es)" unfolding afs_to_unused_arity_def by (induct es arbitrary: e, auto)
      finally show ?thesis .
    qed
  qed
qed

lemma default_af_entry_id: 
  assumes "\<pi> f (length ts) = default_af_entry (length ts)"
  shows "afs_term \<pi> (Fun f ts) = Fun (FPair f (length ts)) (map (afs_term \<pi>) ts)"
  by (simp add: default_af_entry_def assms Let_def, simp only: list_eq_iff_nth_eq, simp)

lemma wf_default_af_entry: 
  "wf_af_entry n (default_af_entry n)"
  unfolding default_af_entry_def by auto

definition check_wf_afs :: "('f::show)afs_list \<Rightarrow> shows check"
where "check_wf_afs pi = check_allm (\<lambda>((f,n),e). (case e of
  Collapse i \<Rightarrow> check (i < n) i
| AFList ids \<Rightarrow> check_all (\<lambda> i. i < n) ids
) <+? (\<lambda>i. ''index '' +#+ shows i +@+ '' too high for AF-entry of'' +#+ shows(f,n) +@+ shows_nl)
) pi"

lemma check_wf_afs_sound: assumes check: "isOK(check_wf_afs pi)"
  shows "afs (afs_list_to_afs pi)"
proof (intro afs.intro, unfold wf_afs_def afs_list_to_afs_def, intro allI impI)
  fix f n
  show "wf_af_entry n (case lookup (f, n) pi of None \<Rightarrow> default_af_entry n | Some en \<Rightarrow> en)"
  proof (cases "lookup (f,n) pi", simp add: wf_default_af_entry)
    case (Some en)
    hence mem: "((f,n),en) \<in> set pi" by (rule lookup_sound)
    show ?thesis 
    proof (simp add: Some, cases en)
      case (Collapse i)
      with Some mem check show "wf_af_entry n en" unfolding check_wf_afs_def by auto
    next
      case (AFList ids)
      with Some mem check show "wf_af_entry n en" unfolding check_wf_afs_def by auto
    qed
  qed
qed

(* check_mono_afs_help cannot be inlined since then the code-generator will produce 
   code where two parenthesis are missing *) 
fun check_mono_afs_help where "check_mono_afs_help ((f,a),e) = (''entry for '' +#+ shows f +@+ '' not monotonic'' +#+ shows_nl)"

definition
  check_mono_afs :: "('f::show)afs_list \<Rightarrow> shows check"
where
  "check_mono_afs pi = (check_all (\<lambda>x . mono_af_entry (snd(fst x)) (snd x)) pi <+? check_mono_afs_help)"

lemma check_mono_afs_sound: assumes check: "isOK(check_mono_afs pi)"
  and check2: "wf_afs (afs_list_to_afs pi)"
  shows "mono_afs (afs_list_to_afs pi)"
proof (unfold mono_afs_def afs_list_to_afs_def, intro allI impI)
  fix f n
  show "mono_af_entry n (case lookup (f, n) pi of None \<Rightarrow> default_af_entry n | Some en \<Rightarrow> en)"
  proof (cases "lookup (f,n) pi", simp add: wf_default_af_entry default_af_entry_def, force simp: mem_iff)
    case (Some en)
    hence mem: "((f,n),en) \<in> set pi" by (rule lookup_sound)
    show ?thesis 
    proof (simp add: Some, cases en)
      case (Collapse i)
      with Some mem check check2 show "mono_af_entry n en" unfolding check_mono_afs_def mono_af_entry_def wf_afs_def by auto
    next
      case (AFList ids)
      with Some mem check check2 show "mono_af_entry n en" unfolding check_mono_afs_def mono_af_entry_def wf_afs_def by auto
    qed
  qed
qed

definition afs_list_to_af :: "'f afs_list \<Rightarrow> 'f af"
where "afs_list_to_af pi f n \<equiv> \<lambda> i. case lookup (f,n) pi of Some e \<Rightarrow> (case e of Collapse j \<Rightarrow> i = j | AFList ids \<Rightarrow> i mem ids)  | None \<Rightarrow> True"
   
context afs
begin

lemma apply_af_entry_const_is_id : "apply_af_entry (FPair f (0 :: nat)) (\<pi> f 0) [] = Fun (FPair f 0) []"
proof -
  let ?e = "\<pi> f 0"
  from wf_afs have "wf_af_entry 0 ?e" unfolding wf_afs_def by auto
  thus ?thesis by (cases ?e, simp_all)
qed
end

locale afs_ce_redpair_impl = afs_redpair_impl "afs_list_to_afs pi" S NS cS cNS + filtered : ce_redpair_impl S NS cS cNS  
  for pi and S NS :: "(('f::show)filtered,'v::show)trs" and cS cNS 

locale afs_ce_mono_redpair_impl = mono_afs_redpair_impl "afs_list_to_afs pi" S NS cS cNS cm + filtered : ce_mono_redpair_impl S NS cS cNS cm 
  for pi and S NS :: "(('f::show)filtered,'v::show)trs" and cS cNS cm

sublocale afs_ce_mono_redpair_impl \<subseteq> afs_ce_redpair_impl ..

sublocale afs_ce_redpair_impl \<subseteq> ce_af_redpair_impl "afs_rel S" "afs_rel NS" "afs_check rS (afs_list_to_afs pi) cS" "afs_check rNS (afs_list_to_afs pi) cNS" "afs_list_to_af pi"
proof -
  let ?pi = "afs_list_to_afs pi"
  let ?S = "afs_rel S"
  let ?NS = "afs_rel NS"
  let ?cS = "afs_check rS ?pi cS"
  let ?cNS = "afs_check rNS ?pi cNS"
  let ?ar = "afs_to_unused_arity pi"
  let ?sar = "Suc (Suc ?ar)"
  let ?af = "afs_list_to_af pi"
  show "ce_af_redpair_impl ?S ?NS ?cS ?cNS ?af"
  proof
    show "af_compatible ?af ?NS"
      unfolding af_compatible_def
    proof (intro allI)      
      fix f :: 'f and bef :: "('f,'v)term list" and s t :: "('f,'v)term" and  aft
      let ?bsa = "bef @ s # aft"
      let ?bta = "bef @ t # aft"
      let ?lb = "length bef"
      let ?l = "Suc (?lb + length aft)"    
      let ?s = "Fun f ?bsa"
      let ?t = "Fun f ?bta"
      show "?af f ?l ?lb \<or> (?s,?t) \<in> ?NS"
      proof (cases "?af f ?l ?lb", simp)
	case False
	let ?e = "?pi f ?l"
	from wf_afs have wf: "wf_af_entry ?l ?e" unfolding wf_afs_def by auto
	show ?thesis
	proof (cases "lookup (f,?l) pi")
	  case None
	  with False show ?thesis unfolding afs_list_to_af_def by auto
	next
	  case (Some e)
	  hence e: "?e = e" by (simp add: afs_list_to_afs_def)
	  let ?ls = "length (bef @ s # aft)"
	  let ?lt = "length (bef @ t # aft)"
	  let ?msa = "map (afs_term ?pi) ?bsa"
	  let ?mta = "map (afs_term ?pi) ?bta"
	  let ?pfs = "apply_af_entry (FPair f ?l) ?e ?msa"
	  let ?pft = "apply_af_entry (FPair f ?l) ?e ?mta"
	  have all2: "\<forall> i < ?l. i \<noteq> ?lb \<longrightarrow> (?msa ! i, ?mta ! i) \<in> NS" 
	  proof (intro allI impI)
	    fix i
	    assume "i < ?l" and ib: "i \<noteq> ?lb"
	    hence i_s: "i < ?ls" and i_t: "i < ?lt" by simp_all
	    {
	      assume bef: "i < length bef"
	      with refl_NS have "(?msa ! i, ?mta ! i) \<in> NS" unfolding refl_on_def 
		by (simp only: nth_map[OF i_s], simp add: nth_append)
	    }
	    moreover
	    {
	      assume i: "i > length bef"
	      from this obtain k where k1: "i - length bef = Suc k" by (cases "i - length bef", arith+)
	      with i_s have k2: "k < length aft" by auto
	      from i refl_NS have "(?msa ! i, ?mta ! i) \<in> NS" unfolding refl_on_def 
		by (simp only: nth_map[OF i_s] nth_map[OF i_t], simp add: nth_append k1)
	    } 
	    moreover
	    {
	      have "i < length bef \<or> i = length bef \<or> i > length bef" by auto
	    }      
	    ultimately
	    show "(?msa ! i, ?mta ! i) \<in> NS" using ib by blast
	  qed 
	  have "(?pfs,?pft) \<in> NS" 
	  proof (cases ?e)
	    case (Collapse i)
	    with wf have len: "i < ?l"  by simp
	    from Collapse[simplified] Some False have "i \<noteq> ?lb" unfolding afs_list_to_afs_def afs_list_to_af_def by auto
	    with all2 len Collapse show ?thesis by auto
	  next
	    case (AFList ids)
	    with wf have ids: "\<forall> i \<in> set ids. i < ?l" by simp
	    show ?thesis 
	    proof (simp only: AFList apply_af_entry.simps)
	      let ?msi = "map (op ! (map (afs_term ?pi) ?bsa)) ids"
	      let ?mti = "map (op ! (map (afs_term ?pi) ?bta)) ids"
	      from ids False AFList[simplified] Some have "\<forall> i \<in> set ids. i \<noteq> ?lb" unfolding afs_list_to_afs_def afs_list_to_af_def 
		by (auto simp: mem_iff)
	      with ids all2 have "\<forall> i < length ids. (?msi ! i, ?mti ! i) \<in> NS" by auto
	      with all_ctxt_closed show "(Fun (FPair f ?l) ?msi, Fun (FPair f ?l) ?mti) \<in> NS"
		unfolding all_ctxt_closed_def by force
	    qed
	  qed
	  thus "afs_list_to_af pi f ?l ?lb \<or> (Fun f ?bsa, Fun f ?bta) \<in> afs_rel NS" unfolding afs_rel_def afs_rule_def by simp
	qed
      qed
    qed
  next
    have pi2: "\<And> m c. m \<ge> ?sar \<Longrightarrow> ?pi c m = default_af_entry m"
    proof -
      fix m c
      assume m: "m \<ge> ?sar"
      show "?pi c m = default_af_entry m"
      proof (unfold afs_list_to_afs_def, cases "lookup (c, m) pi", simp)
        case (Some foo)
        hence "((c,m),foo) \<in> set pi" by (rule lookup_sound)
        from bspec[OF afs_to_unused_arity_sound this] m
        show "(case lookup (c, m) pi of None \<Rightarrow> default_af_entry m | Some foo \<Rightarrow> foo) = (default_af_entry m)" by simp
      qed
    qed
    from NS_ce_compat obtain n where ce_NS: "\<And> m c d . m \<ge> n \<Longrightarrow> ce_trs (c,d,m) \<subseteq> NS" by blast
    let ?n = "max n ?sar"
    show "\<exists> n. \<forall> m \<ge> n. \<forall> c d. ce_trs (c,d,m) \<subseteq> ?NS"
    proof (rule exI[of _ ?n], intro allI impI)
      fix m c d
      assume m: "m \<ge> ?n"
      hence mn: "m \<ge> n" by simp
      show "ce_trs (c,d,m) \<subseteq> ?NS"
      proof (simp add: ce_trs.simps afs_rel_def afs_rule_def, auto simp: pi2 simp del: afs_term.simps)      
        fix t s :: "('f,'v)term"
        let ?list = "t # s # replicate m (Fun d [])"
        from pi2 m have id: "?pi c (length ?list) = default_af_entry (length ?list)" by simp
        show "(afs_term ?pi (Fun c ?list), afs_term ?pi t) \<in> NS"
 	  using ce_NS[OF mn, simplified ce_trs.simps, of "FPair c (Suc (Suc m))" "FPair d 0"]  
          by (simp only: default_af_entry_id[of ?pi c ?list, OF id], auto simp: apply_af_entry_const_is_id)
      next
        fix t s :: "('f,'v)term"
        let ?list = "t # s # replicate m (Fun d [])"
        from pi2 m have id: "?pi c (length ?list) = default_af_entry (length ?list)" by simp
        show "(afs_term ?pi (Fun c ?list), afs_term ?pi s) \<in> NS"
 	  using ce_NS[OF mn, simplified ce_trs.simps, of "FPair c (Suc (Suc m))" "FPair d 0"]  
          by (simp only: default_af_entry_id[of ?pi c ?list, OF id], auto simp: apply_af_entry_const_is_id)
      qed
    qed
  qed
qed


sublocale afs_ce_mono_redpair_impl \<subseteq> ce_mono_redpair_impl "afs_rel S" "afs_rel NS" "afs_check rS (afs_list_to_afs pi) cS" "afs_check rNS (afs_list_to_afs pi) cNS" "cm \<guillemotright> check_mono_afs pi"
proof 
  assume "isOK (cm \<guillemotright> check_mono_afs pi)"
  hence c1: "isOK(cm)" and c2: "isOK(check_mono_afs pi)" by auto
  hence check: "isOK(cm \<guillemotright> check (mono_afs (afs_list_to_afs pi)) mess)"
    using check_mono_afs_sound[OF c2 wf_afs] by auto 
  show "one_step_ctxt_closed (afs.afs_rel (afs_list_to_afs pi) S)"
    by (rule checkMonoSound[OF check])
next
  let ?pi = "afs_list_to_afs pi"
  let ?S = "afs_rel S"
  let ?cS = "afs_check rS ?pi cS"
  let ?ar = "afs_to_unused_arity pi"
  let ?sar = "Suc (Suc ?ar)"
  let ?af = "afs_list_to_af pi"
  have pi2: "\<And> m c. m \<ge> ?sar \<Longrightarrow> ?pi c m = default_af_entry m"
  proof -
    fix m c
    assume m: "m \<ge> ?sar"
    show "?pi c m = default_af_entry m"
    proof (unfold afs_list_to_afs_def, cases "lookup (c, m) pi", simp)
      case (Some foo)
      hence "((c,m),foo) \<in> set pi" by (rule lookup_sound)
      from bspec[OF afs_to_unused_arity_sound this] m
      show "(case lookup (c, m) pi of None \<Rightarrow> default_af_entry m | Some foo \<Rightarrow> foo) = (default_af_entry m)" by simp
    qed
  qed
  from S_ce_compat obtain n where ce_S: "\<And> m c d . m \<ge> n \<Longrightarrow> ce_trs (c,d,m) \<subseteq> S" by blast
  let ?n = "max n ?sar"
  show "\<exists> n. \<forall> m \<ge> n. \<forall> c d. ce_trs (c,d,m) \<subseteq> ?S"
  proof (rule exI[of _ ?n], intro allI impI)
    fix m c d
    assume m: "m \<ge> ?n"
    hence mn: "m \<ge> n" by simp
    show "ce_trs (c,d,m) \<subseteq> ?S"
    proof (simp add: ce_trs.simps afs_rel_def afs_rule_def, auto simp: pi2 simp del: afs_term.simps)      
      fix t s :: "('f,'v)term"
      let ?list = "t # s # replicate m (Fun d [])"
      from pi2 m have id: "?pi c (length ?list) = default_af_entry (length ?list)" by simp
      show "(afs_term ?pi (Fun c ?list), afs_term ?pi t) \<in> S"
 	using ce_S[OF mn, simplified ce_trs.simps, of "FPair c (Suc (Suc m))" "FPair d 0"]  
        by (simp only: default_af_entry_id[of ?pi c ?list, OF id], auto simp: apply_af_entry_const_is_id)
    next
      fix t s :: "('f,'v)term"
      let ?list = "t # s # replicate m (Fun d [])"
      from pi2 m have id: "?pi c (length ?list) = default_af_entry (length ?list)" by simp
      show "(afs_term ?pi (Fun c ?list), afs_term ?pi s) \<in> S"
 	using ce_S[OF mn, simplified ce_trs.simps, of "FPair c (Suc (Suc m))" "FPair d 0"]  
        by (simp only: default_af_entry_id[of ?pi c ?list, OF id], auto simp: apply_af_entry_const_is_id)
    qed
  qed
qed

(*
lemma redpair_all_impl_to_afs_redpair: assumes wf_afs: "isOK(check_wf_afs pi)" and redpair: "redpair_all_impl check_strict check_non_strict check_mono pi'"
  shows "redpair_all_impl (afs_check rS (afs_list_to_afs pi) check_strict) (afs_check rNS (afs_list_to_afs pi) check_non_strict) (check_mono \<guillemotright> check_mono_afs pi) (afs_list_to_af pi)"
proof
  fix s_list ns_list
  assume stri: "isOK (checkAllM (afs_check rS (afs_list_to_afs pi) check_strict) s_list)"
  assume nstri: "isOK (checkAllM (afs_check rNS (afs_list_to_afs pi) check_non_strict) ns_list)"
  let ?pi = "afs_list_to_afs pi"
  let ?ar = "afs_to_unused_arity pi"
  let ?sar = "Suc (Suc ?ar)"
  let ?af = "afs_list_to_af pi"
  let ?af_term = "afs_term ?pi"
  let ?af_list = "map (\<lambda> (s,t). (?af_term s, ?af_term t))"
  let ?as_list = "?af_list s_list"
  let ?ans_list = "?af_list ns_list"
  from wf_afs have afs: "afs ?pi" by (rule check_wf_afs_sound)
  from stri have stri: "isOK(checkAllM check_strict ?as_list)" by (auto simp: Let_def afs_check_def)
  from nstri have nstri: "isOK(checkAllM check_non_strict ?ans_list)" by (auto simp: Let_def afs_check_def)    
  from redpair_all_impl.check[OF redpair stri nstri]
  obtain S NS n c d where mono: "mono_ce_redpair S NS n c d"
     and ce: "ce_af_redpair S NS n c d pi'"
     and stri: "set ?as_list \<subseteq> S"
     and nstri: "set ?ans_list \<subseteq> NS"
     and cmono: "isOK check_mono \<longrightarrow> one_step_ctxt_closed S" by blast  
  let ?aS = "afs.afs_rel ?pi S"
  let ?aNS = "afs.afs_rel ?pi NS"
  show "\<exists>S NS n c d.
             mono_ce_redpair S NS n c d \<and>
             ce_af_redpair S NS n c d (afs_list_to_af pi) \<and> set s_list \<subseteq> S \<and> set ns_list \<subseteq> NS \<and> (isOK (check_mono \<guillemotright> check_mono_afs pi) \<longrightarrow> one_step_ctxt_closed S)"
  proof (intro exI conjI impI)
    from stri show "set s_list \<subseteq> ?aS" by  (auto simp: afs.afs_rel_def[OF afs] afs_rule_def)
  next
    from nstri show "set ns_list \<subseteq> ?aNS" by  (auto simp: afs.afs_rel_def[OF afs] afs_rule_def)
  next
    show "mono_ce_redpair ?aS ?aNS (afs_to_unused_arity pi) c d"
    assume "isOK(check_mono \<guillemotright> check_mono_afs pi)"
    hence cm: "isOK(check_mono)" and cma: "isOK(check_mono_afs pi)" by auto
    show "one_step_ctxt_closed ?aS"
    proof (rule mono_redpair_impl.checkMonoSound[OF _ cm])
*)
end
