(* Author: Alexander Krauss, TU Muenchen
   Author: Christian Sternagel, University of Innsbruck
*)

header {* Monad notation for arbitrary types *}

theory Monad_Notation
imports Main
uses
  ("adhoc_overloading.ML")
  ("monad_notation.ML")
begin

nonterminals
  do_expr

syntax
  "_do_expr" :: "do_expr \<Rightarrow> 'a"
    ("(2do {//_)//}" [12] 100)
  "_do_bind" :: "pttrn \<Rightarrow> 'a \<Rightarrow> do_expr \<Rightarrow> do_expr"
    ("(2_ <-/ _);//_" [1000, 13, 12] 12)
  "_do_then" :: "'a \<Rightarrow> do_expr \<Rightarrow> do_expr"
    ("_;//_" [13, 12] 12)
  "_do_let" :: "pttrn \<Rightarrow> 'a \<Rightarrow> do_expr \<Rightarrow> do_expr"
    ("(2let _ =/ _);//_" [1000, 13, 12] 12)
  "_do_done" :: "'a \<Rightarrow> do_expr"
    ("_" [12] 12)

syntax (xsymbols)
  "_do_bind" :: "pttrn \<Rightarrow> 'a \<Rightarrow> do_expr \<Rightarrow> do_expr"
    ("(2_ \<leftarrow>/ _);//_" [1000, 13, 12] 12)

consts
  bindM :: "'a \<Rightarrow> ('b \<Rightarrow> 'c) \<Rightarrow> 'c" (infixl ">>=" 54)

syntax "_thenM" :: "'a \<Rightarrow> 'b \<Rightarrow> 'b" (infixl ">>" 54)
translations "(m >> n)" == "(m >>= (\<lambda>_. n))"

notation (xsymbols) bindM (infixl "\<guillemotright>=" 54)
syntax (xsymbols) "_thenM" :: "'a \<Rightarrow> 'b \<Rightarrow> 'b" (infixl "\<guillemotright>" 54)

translations
  "_do_expr f" => "f"
  "_do_done f" => "f"
  "_do_bind x m n" => "CONST bindM m (\<lambda>x. n)"
  "_do_then m n" => "CONST bindM m (\<lambda>_. n)"
  "_do_let x t f" => "CONST Let t (\<lambda>x. f)"

use "adhoc_overloading.ML"
setup Adhoc_Overloading.setup

use "monad_notation.ML"

print_translation {* Monad_Notation.print_translations *}

setup {*
  Adhoc_Overloading.add_overloaded @{const_name bindM}
*}
end
