(*
Copyright 2009 Christian Sternagel, René Thiemann, Sarah Winkler, Harald Zankl

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory NonTermination
imports Loop StringReversal DpFramework SubtermCriterionImpl
begin

types ('f,'v)subL = "('v \<times> ('f,'v)term)list"

definition subst_of :: "('f,'v)subL \<Rightarrow> ('f,'v)sub"
where "subst_of xts \<equiv> Sub (\<lambda> x. case lookup x xts of None \<Rightarrow> Var x | Some t \<Rightarrow> t)"

datatype ('f, 'v) DPDisproof =
  DPLoop "('f,'v)term" "('f,'v)prseq" "('f,'v)subL" "('f,'v)ctxt" |
  DPRuleRemoval "('f, 'v) rules" "('f, 'v) rules" "('f, 'v) DPDisproof"


definition check_loop :: "('f :: show,'v :: show)term \<Rightarrow> ('f,'v)rseq \<Rightarrow> ('f,'v)subL \<Rightarrow> ('f,'v)ctxt \<Rightarrow> ('f,'v)rules \<Rightarrow> shows check"
where "check_loop s rseq \<sigma> C R \<equiv> do {
           check (rseq \<noteq> []) (shows ''rewrite sequence must be non-empty'');
           check_rsteps R rseq s (C\<langle>s \<cdot> subst_of \<sigma>\<rangle>)
        }"

lemma check_loop_sound : assumes check: "isOK(check_loop s rseq \<sigma> C R)"
  shows "\<not> SN (rstep (set R))"
proof -
  obtain sig where sig: "sig = subst_of \<sigma>" by auto
  with check have check: "isOK(check_rsteps R rseq s (C\<langle>s \<cdot> sig\<rangle>))" and nonempty: "rseq \<noteq> []" by (auto simp: Let_def check_loop_def)
  from nonempty obtain n where rseq: "length rseq = Suc n" by (cases rseq, auto)
  with check_rsteps_sound[OF check] rseq have "(s,C\<langle>s \<cdot> sig\<rangle>) \<in> (rstep (set R)) ^^ (Suc n)" by auto
  hence step: "(s,C\<langle>s \<cdot> sig\<rangle>) \<in> (rstep (set R))^+" using pow_Suc_subset_trancl by auto
  from loop_imp_not_SN[OF step] show ?thesis unfolding SN_def by auto
qed


definition check_dp_loop :: "('f :: show,'v :: show)term \<Rightarrow> ('f,'v)prseq \<Rightarrow> ('f,'v)subL \<Rightarrow> ('f,'v)ctxt \<Rightarrow> ('f,'v)rules \<Rightarrow> ('f,'v)rules \<Rightarrow> shows check"
where "check_dp_loop s prseq \<sigma> C P R \<equiv> if C = \<box> then do {
             check (prseq \<noteq> []) (shows ''rewrite sequence must be non-empty'');
             check_steps P R prseq s (s \<cdot> subst_of \<sigma>)
        } else do {
             check_loop s (map (\<lambda> (x,y,_,z). (x,y,z)) prseq) \<sigma> C R
        } "
  
lemma check_dp_loop: assumes check: "isOK (check_dp_loop s prseq \<sigma> C P R)"
  shows "infinite_dpp (set P, set R)"
proof -
  obtain sig where sig: "sig = subst_of \<sigma>" by auto
  show ?thesis
  proof (cases C)
    case Hole
    with sig check have check: "isOK(check_steps P R prseq s (s \<cdot> sig))" and nonempty: "prseq \<noteq> []" by (auto simp: check_dp_loop_def)
    from nonempty obtain n where prseq: "length prseq = Suc n" by (cases prseq, auto)
    obtain PR where PR: "PR = rrstep (set P) \<union> rstep (set R)" by auto
    from check_steps_sound[OF check] prseq PR have "(s,s \<cdot> sig) \<in> PR ^^ (Suc n)" by simp
    hence step: "(s,s \<cdot> sig) \<in> PR^+" using pow_Suc_subset_trancl by auto
    from loop_imp_infinite[OF step[simplified PR]] show ?thesis .
  next
    case (More f bef D aft)
    with sig check have check: "isOK(check_loop s (map (\<lambda> (x,y,_,z). (x,y,z)) prseq) \<sigma> C R)" by (auto simp: check_dp_loop_def)
    from check_loop_sound[OF this] SN_subset have "\<not> SN(rrstep (set P) \<union> rstep (set R))" unfolding SN_def by blast
    thus ?thesis by (simp add: infinite_dpp_alt)
  qed
qed
    
    
fun
  check_dp_disproof :: "shows \<Rightarrow> ('f,'v)rules \<Rightarrow> ('f,'v)rules \<Rightarrow> ('f :: show,'v :: show)DPDisproof \<Rightarrow> shows check"
where
  "check_dp_disproof i p r (DPLoop s rseq \<sigma> c) = debug (i []) ''Loop'' (check_dp_loop s rseq \<sigma> c p r <+? (\<lambda>s. i +@+ '': error in loop proof'' +#+ shows_nl +@+ s))" |
  "check_dp_disproof i p r (DPRuleRemoval p' r' prf) = debug (i []) ''Rule Removal'' (do {
     check_subseteq p' p
       <+? (\<lambda>lr. i +@+ '': error in rule-removal proof'' +#+ shows_nl +@+ shows_rule lr +@+
         '' is a new pair not occurring in the previous set of pairs'' +#+ shows_nl);
     check_subseteq r' r
       <+? (\<lambda>lr. i +@+ '': error in rule-removal proof'' +#+ shows_nl +@+ shows_rule lr +@+
         '' is a new rule not occurring in the previous TRS'' +#+ shows_nl);
     check_dp_disproof (i +@+ shows_string ''.1'') p' r' prf
   })"

lemma check_dp_disproof:
  assumes "isOK(check_dp_disproof i p r prf)"
  shows "infinite_dpp (set p, set r)"
using assms 
proof(induct "prf" arbitrary: r p i)
  case (DPLoop s rseq \<sigma> C)
  hence "isOK(check_dp_loop s rseq \<sigma> C p r)" by simp
  with check_dp_loop show ?case by simp
next
  case (DPRuleRemoval p' r')
  hence p: "set p' \<subseteq> set p" and r: "set r' \<subseteq> set r" and inf: "infinite_dpp (set p', set r')" by auto
  from inf have "(\<not> SN (rstep (set r'))) \<or> (\<exists> s t \<sigma>. ichain (set p', set r') s t \<sigma>)" (is "?nSN \<or> ?ichain") by (simp del: ichain.simps)
  thus ?case
  proof
    assume ?nSN
    with SN_subset[OF _ rstep_mono[OF r]] have "\<not> SN (rstep (set r))" by auto
    thus ?thesis by simp
  next
    assume ?ichain
    then obtain s t \<sigma> where "ichain (set p',set r') s t \<sigma>" by auto
    with p have "ichain (set p, set r') s t \<sigma>" by auto
    with rtrancl_mono[OF rstep_mono[OF r]] have "ichain (set p, set r) s t \<sigma>" by auto
    thus ?thesis by (simp only: infinite_dpp.simps, blast)
  qed
qed


datatype ('f,'v)"TRSDisproof" = 
  Loop "('f,'v)term" "('f,'v)rseq" "('f,'v)subL" "('f,'v)ctxt" | 
  NotWellFormed | 
  NTRuleRemoval "('f,'v)rules" "('f,'v)TRSDisproof" | 
  NTStringReversal "('f,'v)rules" "('f,'v)TRSDisproof" |
  NTDPTrans "('f shp,'v)rules" "('f shp,'v)DPDisproof"

fun check_DPs :: "('f :: show,'v :: show)rules \<Rightarrow> ('f shp,'v)rules \<Rightarrow> ('f shp,'v)rule check"
  where "check_DPs r p = (check_all (\<lambda> (ll,rr). any (\<lambda> (l,r). l = map_term unsharp_fun ll \<and> r \<unrhd> map_term unsharp_fun rr) r) p)"

lemma check_DPs : assumes check: "isOK(check_DPs r p)" 
    and infinite: "infinite_dpp (set p, ID (set r))"
  shows "\<not> SN (rstep (set r))"
proof 
  assume sn: "SN (rstep (set r))"
  let ?ir = "ID (set r)"
  let ?ut = "map_term unsharp_fun"  
  have "(map_rule unsharp_fun) ` ?ir \<subseteq> set r"
  proof -
    {
      fix t
      have "?ut (plain t) = t" 
      proof (induct t, simp)
        case (Fun f ss)
        thus ?case 
          by (simp, induct ss, auto)
      qed
    } note id = this
    show ?thesis unfolding ID_def
      by (clarify, simp add: id)
  qed
  hence subset: "map_trs unsharp_fun ?ir \<subseteq> set r" (is "?uir \<subseteq> set r") by auto
  let ?rel = "supt \<union> rstep (set r)"
  have snuir: "SN (rstep ?uir)" using SN_subset[OF sn rstep_mono[OF subset]] .
  have snir: "SN (rstep ?ir)" unfolding SN_def
  proof
    fix t
    from snuir have "SN_elt (rstep ?uir) (?ut t)" unfolding SN_def by simp
    from SN_map_imp_SN[OF this] show "SN_elt (rstep ?ir) t" .
  qed
  with infinite obtain s t \<sigma> where ichain: "ichain (set p, ?ir) s t \<sigma>" by auto
  let ?t = "\<lambda> i. (?ut (t i) \<cdot> map_subst unsharp_fun (\<sigma> i))"
  have main: "\<And> i. (?t i, ?t (Suc i)) \<in> ?rel^+"
  proof -
    fix i    
    have steps: "(?ut (t i \<cdot> \<sigma> i), ?ut (s (Suc i) \<cdot> \<sigma> (Suc i))) \<in> (rstep (set r))^*"
    proof -
      from ichain have "(t i \<cdot> \<sigma> i, s (Suc i) \<cdot> \<sigma> (Suc i)) \<in> (rstep ?ir)^*" by simp
      from rewrites_imp_map_rewrites[OF this] rtrancl_mono[OF rstep_mono[OF subset]]
      show ?thesis by auto
    qed
    let ?sig = "\<lambda> i. map_subst unsharp_fun (\<sigma> i)"
    from steps have steps: "(?ut (t i) \<cdot> ?sig i, ?ut (s (Suc i)) \<cdot> ?sig (Suc i)) \<in> ?rel^*"
      using rtrancl_mono[of "rstep (set r)" "supt \<union> rstep (set r)"] by auto
    let ?i = "Suc i"
    from ichain have "(s ?i, t ?i) \<in> set p" by auto
    with check obtain l rr where l: "l = ?ut (s ?i)" and rr: "rr \<unrhd> ?ut (t ?i)" and lr: "(l,rr) \<in> set r" by auto
    have "(?ut (s ?i) \<cdot> ?sig ?i, rr \<cdot> ?sig ?i) \<in> ?rel"
      using rstepI[OF lr refl refl, of \<box>] by (auto simp: l)
    moreover from supteqp_stable[OF rr, of "?sig ?i"] have "(rr \<cdot> ?sig ?i, ?ut (t ?i) \<cdot> ?sig ?i) \<in> ?rel^=" 
      by (auto simp: supteq_def)      
    ultimately have "(?ut (s ?i) \<cdot> ?sig ?i, ?ut (t ?i) \<cdot> ?sig ?i) \<in> ?rel O ?rel^=" by auto
    with steps have "(?ut (t i) \<cdot> ?sig i, ?ut (t (Suc i)) \<cdot> ?sig (Suc i)) \<in> ?rel^* O ?rel O ?rel^=" (is "?pair \<in> _") by blast
    hence "?pair \<in> ?rel^+ O ?rel^=" by (simp only:  O_assoc[symmetric] trancl_to_rtrancl)
    thus "?pair \<in> ?rel^+" by (simp only: trancl_o_refl_is_trancl)
  qed
  hence "\<not> SN_elt (?rel^+) (?t 0)" unfolding SN_elt_def by auto
  hence "\<not> SN_elt ?rel (?t 0)" using SN_elt_imp_SN_elt_trancl[of ?rel] by auto
  hence "\<not> SN_elt (rstep (set r)) (?t 0)" using SN_elt_rstep_imp_SN_elt_supt_union_rstep by auto
  with sn show False unfolding SN_def by auto
qed
  
fun
  checkTRSDisproof :: "shows \<Rightarrow> ('f::show, 'v::show) rules \<Rightarrow> ('f, 'v) TRSDisproof \<Rightarrow> shows check"
where
  "checkTRSDisproof i r (Loop s rseq \<sigma> C) = debug (i []) ''Loop'' (check_loop s rseq \<sigma> C r <+? (\<lambda>s. i +@+ '': error in loop proof'' +#+ shows_nl +@+ s))" |
  "checkTRSDisproof i r NotWellFormed =
     debug (i []) ''not well-formed'' (check_not_wf_trs r <+? (\<lambda>s. i +@+ '': error in not-well-formed proof'' +#+ shows_nl +@+ s))" |
  "checkTRSDisproof i r (NTRuleRemoval r' prf) = debug (i []) ''rule removal'' (do {
     check_subseteq r' r 
       <+? (\<lambda>lr. i +@+ '': error in rule-removal proof'' +#+ shows_nl +@+ shows_rule lr +@+
         '' is a new rule not occurring in the previous TRS'' +#+ shows_nl);
     checkTRSDisproof (i +@+ shows_string ''.1'') r' prf
   })" |
  "checkTRSDisproof i r (NTStringReversal r' prf) = debug (i []) ''string reversal'' (do {
     check_string_rev_nt r r'                  
       <+? (\<lambda>s. i +@+ '': error in string reversal proof'' +#+ shows_nl +@+ s);
     checkTRSDisproof (i +@+ shows_string ''.1'') r' prf
   })" |
  "checkTRSDisproof i r (NTDPTrans p prf) = debug (i []) ''DP transformation'' (do {
     check_DPs r p 
       <+? (\<lambda>lr. i +@+ '': error in DP transformation'' +#+ shows_nl +@+ shows ''problematic rule '' +@+ shows_rule lr);
     check_dp_disproof (i +@+ shows_string ''.1'') p (ID_list r) prf
   })"

lemma checkTRSDisproofSound:
 assumes "isOK(checkTRSDisproof i r prf)"
 shows "\<not> SN (rstep (set r))"
using assms 
proof (induct "prf" arbitrary: r i)
  case NotWellFormed
  thus ?case using check_not_wf_trs by simp
next
  case (Loop s rseq \<sigma> C)
  hence "isOK(check_loop s rseq \<sigma> C r)" by simp
  thus ?case by (rule check_loop_sound)
next
  case (NTRuleRemoval r' prof)
  hence nSN: "\<not> SN (rstep (set r'))" and subset: "set r' \<subseteq> set r" by auto
  with SN_subset[OF _ rstep_mono] show ?case by auto
next
  case (NTStringReversal r' prof)
  hence nSN: "\<not> SN (rstep (set r'))" and check: "isOK(check_string_rev_nt r r')" by auto 
  show ?case by (rule check_string_rev_nt_sound[OF check nSN])
next
  case (NTDPTrans p prof)
  hence check: "isOK(check_DPs r p)" and infinite: "infinite_dpp (set p, set (ID_list r))" using check_dp_disproof[of _ p "ID_list r" prof] by auto
  from infinite have "infinite_dpp (set p, ID (set r))" by simp
  from check_DPs[OF check this]
  show ?case .
qed


end
