(*
Copyright 2009 Christian Sternagel, René Thiemann, Sarah Winkler, Harald Zankl

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory PolyCarrier imports MatrixCarrier PolyOrder MatrixPoly 
begin

instantiation arctic :: "show" begin
fun shows_arctic :: "arctic \<Rightarrow> shows"
  where "shows_arctic(Num_arc i) = shows i"
  | "shows_arctic(MinInfty)  = shows_string ''-inf''"
definition "shows_prec (d::nat) ai = shows_arctic ai"
definition "shows_list (as::arctic list) \<equiv> shows_list_aux shows as"
lemma assoc_arctic: "shows_prec d (a::arctic) r @ s = shows_prec d a (r @ s)" by (cases a) (auto simp: shows_prec_arctic_def)
lemma assoc_arctic_elt: "\<forall>(a::arctic)\<in>set as. shows_prec d a r @ s = shows_prec d a (r @ s)"
  using assoc_arctic by auto
instance 
  by (intro_classes,rule assoc_arctic,unfold shows_list_arctic_def)
     (rule shows_list_aux_assoc[OF assoc_arctic_elt])
end

instantiation arctic_rat :: "show" begin
fun shows_arctic_rat :: "arctic_rat \<Rightarrow> shows"
  where "shows_arctic_rat (Num_arc_rat i) = shows i"
  | "shows_arctic_rat(MinInfty_rat)  = shows_string ''-inf''"
definition "shows_prec (d::nat) ari = shows_arctic_rat ari"
definition "shows_list (as::arctic_rat list) \<equiv> shows_list_aux shows as"
lemma assoc_arctic_rat: "shows_prec d (a::arctic_rat) r @ s = shows_prec d a (r @ s)" by (cases a) (auto simp: shows_prec_arctic_rat_def)
lemma assoc_arctic_rat_elt: "\<forall>(a::arctic_rat)\<in>set as. shows_prec d a r @ s = shows_prec d a (r @ s)"
  using assoc_arctic_rat by auto
instance 
  by (intro_classes,rule assoc_arctic_rat,unfold shows_list_arctic_rat_def)
     (rule shows_list_aux_assoc[OF assoc_arctic_rat_elt])
end

definition nat_name :: "string" where "nat_name = ''naturals''"
definition int_name :: "string" where "int_name = ''naturals with negative constants''"
definition rat_name :: "string" where "rat_name = ''rationals (with negative constants)''"
definition arctic_name :: "string" where "arctic_name = ''arctic semiring''"
definition arctic_rat_name :: "string" where "arctic_rat_name = ''arctic rationals''"

interpretation nat_matrix_carrier: mono_matrix_carrier nat_mono "op >" 1 ..
interpretation int_matrix_carrier: mono_matrix_carrier int_mono "op >" 1 ..
interpretation rat_matrix_carrier: mono_matrix_carrier rat_mono "rat_gt \<delta>" "positive_delta \<delta>" ..
interpretation arctic_matrix_carrier: both_mono_matrix_carrier pos_arctic gt_arctic 1 ..
interpretation arctic_rat_matrix_carrier: both_mono_matrix_carrier pos_arctic_rat "gt_arctic_rat \<delta>" 1 ..

definition mat_components :: "nat \<Rightarrow> nat \<Rightarrow> 'a poly_components \<Rightarrow> 'a mat poly_components"
where "mat_components n sd c \<equiv> let sdp = Suc (sd - 1); np = max n sdp in
  (wf_mat np np, mat0I (get_comp_0 c) np np, mat1I (get_comp_0 c) (get_comp_1 c) np, mat1I (get_comp_0 c) (get_comp_def c) np, mat_plusI (get_comp_plus c), mat_multI (get_comp_0 c) (get_comp_plus c) (get_comp_times c) np,
   mat_max0I (get_comp_max0 c), mat_geI (get_comp_ge c), mat_gtI (get_comp_ge c) (get_comp_gt c) sdp, get_comp_mono c, 
  concat [to_string np, ''x'',to_string np,'' matrices with strict dimension '',to_string sdp,'' over '',get_comp_description c], 
  (\<lambda> x. True))"


definition mat_arctic_components :: "nat \<Rightarrow> 'a poly_components \<Rightarrow> 'a mat poly_components"
where "mat_arctic_components n c \<equiv> let np = Suc (n - 1) in (wf_mat np np, mat0I (get_comp_0 c) np np, mat1I (get_comp_0 c) (get_comp_1 c) np,
   mat1I (get_comp_0 c) (get_comp_def c) np, 
   mat_plusI (get_comp_plus c), mat_multI (get_comp_0 c) (get_comp_plus c) (get_comp_times c) np,
   id, mat_geI (get_comp_ge c), mat_geI (get_comp_gt c), False,
  concat [to_string np, ''x'',to_string np,'' matrices over '',get_comp_description c], 
  mat_arc_posI (get_comp_apos c))"

definition mat_mono :: "nat \<Rightarrow> ('a \<Rightarrow> bool) \<Rightarrow> ('a mat \<Rightarrow> bool)"
where "mat_mono sd m \<equiv> mat_monoI m (Suc (sd - 1))"


definition nat_components :: "nat poly_components"
where "nat_components \<equiv> ((\<lambda> x. True) , 0 , 1 , 1, op +, op *, max0, ge, op >, True, nat_name, \<lambda> x. True)"

definition int_components :: "int poly_components"
where "int_components \<equiv> ((\<lambda> x. True) , 0 , 1 , 1, op +, op *, max0, ge, op >, True, int_name, \<lambda> x. True)"

definition rat_components :: "rat \<Rightarrow> rat poly_components"
where "rat_components \<delta> \<equiv> ((\<lambda> x. True) , 0 , 1 , positive_delta \<delta>, op +, op *, max0, ge, rat_gt \<delta>, True, rat_name, \<lambda> x. True)"

definition arctic_components :: "arctic poly_components"
where "arctic_components \<equiv> ((\<lambda> x. True) , 0 , 1 , 1, op +, op *, id, ge, gt_arctic, False, arctic_name, pos_arctic)"

definition arctic_rat_components :: "rat \<Rightarrow> arctic_rat poly_components"
where "arctic_rat_components \<delta> \<equiv> ((\<lambda> x. True) , 0 , 1 , 1, op +, op *, id, ge, gt_arctic_rat \<delta>, False, arctic_rat_name, pos_arctic_rat)"

interpretation mat_arctic_poly_order_carrier : poly_order_carrier 
  "get_comp_wf_elem (mat_arctic_components n arctic_components)"
  "get_comp_0 (mat_arctic_components n arctic_components)"
  "get_comp_plus (mat_arctic_components n arctic_components)"
  "get_comp_times (mat_arctic_components n arctic_components)"
  "get_comp_ge (mat_arctic_components n arctic_components)"
  "get_comp_gt (mat_arctic_components n arctic_components)"
  "get_comp_mono (mat_arctic_components n arctic_components)"
  "get_comp_1 (mat_arctic_components n arctic_components)"
  "get_comp_def (mat_arctic_components n arctic_components)"
  "get_comp_max0 (mat_arctic_components n arctic_components)"
  "get_comp_apos (mat_arctic_components n arctic_components)"
  by (simp add: mat_arctic_components_def arctic_components_def Left_def, rule arctic_matrix_carrier.mat_poly_order_carrier[of "Suc (n - 1)", simplified])


interpretation int_mono_poly_order_carrier : mono_poly_order_carrier 
  "get_comp_wf_elem (mat_components n sd (int_components ))" 
  "get_comp_0 (mat_components n sd (int_components ))" 
  "get_comp_plus (mat_components n sd (int_components))" 
  "get_comp_times (mat_components n sd (int_components))" 
  "get_comp_ge (mat_components n sd (int_components ))" 
  "get_comp_gt (mat_components n sd (int_components ))" 
  "get_comp_mono (mat_components n sd (int_components))" 
  "get_comp_1 (mat_components n sd (int_components ))" 
  "get_comp_def (mat_components n sd (int_components ))" 
  "get_comp_max0 (mat_components n sd (int_components))" 
  "get_comp_apos (mat_components n sd (int_components))" 
  "mat_mono sd int_mono"
  by (simp add: mat_components_def int_components_def Let_def mat_mono_def, rule int_matrix_carrier.mat_mono_poly_order_carrier[of "Suc (sd - Suc 0)" "max n (Suc (sd - Suc 0))", simplified])


interpretation rat_mono_poly_order_carrier : mono_poly_order_carrier 
  "get_comp_wf_elem (mat_components n sd (rat_components d))" 
  "get_comp_0 (mat_components n sd (rat_components d))" 
  "get_comp_plus (mat_components n sd (rat_components d))" 
  "get_comp_times (mat_components n sd (rat_components d))" 
  "get_comp_ge (mat_components n sd (rat_components d))" 
  "get_comp_gt (mat_components n sd (rat_components d))" 
  "get_comp_mono (mat_components n sd (rat_components d))" 
  "get_comp_1 (mat_components n sd (rat_components d))" 
  "get_comp_def (mat_components n sd (rat_components d))" 
  "get_comp_max0 (mat_components n sd (rat_components d))" 
  "get_comp_apos (mat_components n sd (rat_components d))" 
  "mat_mono sd rat_mono"
  by (simp add: mat_components_def rat_components_def Let_def mat_mono_def, rule rat_matrix_carrier.mat_mono_poly_order_carrier[of "Suc (sd - Suc 0)" "max n (Suc (sd - Suc 0))", simplified]) 

interpretation mat_arctic_rat_poly_order_carrier : poly_order_carrier 
  "get_comp_wf_elem (mat_arctic_components n (arctic_rat_components d))"
  "get_comp_0 (mat_arctic_components n (arctic_rat_components d))"
  "get_comp_plus (mat_arctic_components n (arctic_rat_components d))"
  "get_comp_times (mat_arctic_components n (arctic_rat_components d))"
  "get_comp_ge (mat_arctic_components n (arctic_rat_components d))"
  "get_comp_gt (mat_arctic_components n (arctic_rat_components d))"
  "get_comp_mono (mat_arctic_components n (arctic_rat_components d))"
  "get_comp_1 (mat_arctic_components n (arctic_rat_components d))"
  "get_comp_def (mat_arctic_components n (arctic_rat_components d))"
  "get_comp_max0 (mat_arctic_components n (arctic_rat_components d))"
  "get_comp_apos (mat_arctic_components n (arctic_rat_components d))"
  by (simp add: mat_arctic_components_def arctic_rat_components_def Left_def, rule arctic_rat_matrix_carrier.mat_poly_order_carrier[of "Suc (n - 1)", simplified])

end
