module Ceta = 
struct

let rec id x = (fun xa -> xa) x;;

type 'a eq = {eq : 'a -> 'a -> bool};;
let eq _A = _A.eq;;

type nat = Zero_nat | Suc of nat;;

type ('a, 'b) scg = Null | Scg of 'a * 'a * ('b * 'b) list * ('b * 'b) list;;

type 'a shp = Sharp of 'a | Plain of 'a;;

type 'a spL = SimpProj of ('a * nat) list;;

let one_nat : nat = Suc Zero_nat;;

let rec minus_nat
  m n = match m, n with Suc m, Suc n -> minus_nat m n
    | Zero_nat, n -> Zero_nat
    | m, Zero_nat -> m;;

type nibble = Nibble0 | Nibble1 | Nibble2 | Nibble3 | Nibble4 | Nibble5 |
  Nibble6 | Nibble7 | Nibble8 | Nibble9 | NibbleA | NibbleB | NibbleC | NibbleD
  | NibbleE | NibbleF;;

let rec nat_aux
  i n = (if Big_int.le_big_int i (Big_int.big_int_of_int 0) then n
          else nat_aux (Big_int.sub_big_int i (Big_int.big_int_of_int 1))
                 (Suc n));;

let rec nat i = nat_aux i Zero_nat;;

type ('a, 'b) sum = Inl of 'a | Inr of 'b;;

let rec tl = function [] -> []
             | x :: xs -> xs;;

let rec append
  x0 ys = match x0, ys with [], ys -> ys
    | x :: xs, ys -> x :: append xs ys;;

let rec digit2int
  x = (if ((x : char) = '0') then Inr (Big_int.big_int_of_int 0)
        else (if ((x : char) = '1') then Inr (Big_int.big_int_of_int 1)
               else (if ((x : char) = '2') then Inr (Big_int.big_int_of_int 2)
                      else (if ((x : char) = '3')
                             then Inr (Big_int.big_int_of_int 3)
                             else (if ((x : char) = '4')
                                    then Inr (Big_int.big_int_of_int 4)
                                    else (if ((x : char) = '5')
   then Inr (Big_int.big_int_of_int 5)
   else (if ((x : char) = '6') then Inr (Big_int.big_int_of_int 6)
          else (if ((x : char) = '7') then Inr (Big_int.big_int_of_int 7)
                 else (if ((x : char) = '8') then Inr (Big_int.big_int_of_int 8)
                        else (if ((x : char) = '9')
                               then Inr (Big_int.big_int_of_int 9)
                               else Inl (append [x]
  [' '; 'i'; 's'; ' '; 'n'; 'o'; 't'; ' '; 'a'; ' '; 'd'; 'i'; 'g'; 'i';
    't'])))))))))));;

let rec sum_case
  f1 f2 x2 = match f1, f2, x2 with f1, f2, Inr b -> f2 b
    | f1, f2, Inl a -> f1 a;;

let rec bind m f = (match m with Inl a -> Inl a | Inr a -> f a);;

let rec string2int_aux
  n x1 = match n, x1 with n, [] -> Inr n
    | n, d :: s ->
        bind (digit2int d)
          (fun m ->
            string2int_aux
              (Big_int.add_big_int
                (Big_int.mult_big_int (Big_int.big_int_of_int 10) n) m)
              s);;

let rec nat_case
  f1 f2 x2 = match f1, f2, x2 with f1, f2, Suc nat -> f2 nat
    | f1, f2, Zero_nat -> f1;;

let rec take
  n x1 = match n, x1 with n, [] -> []
    | n, x :: xs -> (match n with Zero_nat -> [] | Suc m -> x :: take m xs);;

let eq_char = ({eq = (fun a b -> ((a : char) = b))} : char eq);;

let rec eqa _A a b = eq _A a b;;

let rec eq_list _A
  x0 x1 = match x0, x1 with [], [] -> true
    | a :: lista, a' :: list' -> eqa _A a a' && eq_list _A lista list'
    | [], a' :: list' -> false
    | a' :: list', [] -> false;;

let rec null = function [] -> true
               | x :: xs -> false;;

let rec string2int
  s = (if null s
        then Inl ['c'; 'a'; 'n'; 'n'; 'o'; 't'; ' '; 'c'; 'o'; 'n'; 'v'; 'e';
                   'r'; 't'; ' '; 'e'; 'm'; 'p'; 't'; 'y'; ' '; 's'; 't'; 'r';
                   'i'; 'n'; 'g'; ' '; 'i'; 'n'; 't'; 'o'; ' '; 'n'; 'u'; 'm';
                   'b'; 'e'; 'r']
        else (if eq_list eq_char (take one_nat s) ['-']
               then bind (string2int_aux (Big_int.big_int_of_int 0) (tl s))
                      (fun i ->
                        Inr (Big_int.sub_big_int (Big_int.big_int_of_int 0) i))
               else string2int_aux (Big_int.big_int_of_int 0) s));;

let newline : char = '\010';;

let rec the (Some x) = x;;

let rec replicate
  xa0 x = match xa0, x with Zero_nat, x -> []
    | Suc n, x -> x :: replicate n x;;

let rec comp f g = (fun x -> f (g x));;

let rec shows_map
  s x1 = match s, x1 with s, [] -> id
    | s, x :: xs -> comp (s x) (shows_map s xs);;

let rec is_none = function Some x -> false
                  | None -> true;;

let rec snd (a, b) = b;;

let dquote : char list = ['\034'];;

let rec shows_string x = append x;;

let rec fst (a, b) = a;;

let rec shows_list_char cs = shows_string cs;;

type 'a show =
  {shows_prec : nat -> 'a -> char list -> char list;
    shows_list : 'a list -> char list -> char list};;
let shows_prec _A = _A.shows_prec;;
let shows_list _A = _A.shows_list;;

let rec shows_prec_char d c = (fun a -> c :: a);;

let show_char =
  ({shows_prec = shows_prec_char; shows_list = shows_list_char} : char show);;

let rec shows_prec_list _A d l = shows_list _A l;;

let rec shows_attr
  av = comp (shows_prec_list show_char Zero_nat (fst av))
         (comp (shows_string ['='])
           (shows_string (append dquote (append (snd av) dquote))));;

let rec shows_attrs
  asa = shows_map (fun a -> comp (shows_string [' ']) (shows_attr a)) asa;;

type xml =
  Xml of
    char list * (char list * char list) list * xml list * (char list) option;;

let rec shows_XML_indent
  ind i (Xml (n, a, c, t)) =
    comp (shows_string ind)
      (comp (shows_string ['<'])
        (comp (shows_prec_list show_char Zero_nat n)
          (comp (shows_attrs a)
            (if null c && is_none t then shows_string ['/'; '>']
              else comp (shows_string ['>'])
                     (comp (if not (null c)
                             then comp (shows_map
 (shows_XML_indent (append ind (replicate i ' ')) i) c)
                                    (shows_string ind)
                             else id)
                       (comp (if is_none t then id else shows_string (the t))
                         (comp (shows_string ['<'; '/'])
                           (comp (shows_prec_list show_char Zero_nat n)
                             (shows_string ['>'])))))))));;

let rec shows_prec_xml
  d xml = shows_XML_indent [newline] (nat (Big_int.big_int_of_int 2)) xml;;

let rec concat = function [] -> []
                 | x :: xs -> append x (concat xs);;

let rec text
  tag (Xml (n, atts, cs, t)) =
    (if eq_list eq_char n tag &&
          (null atts && (null cs && (not (is_none t) && not (null (the t)))))
      then Inr (the t)
      else Inl (concat
                 [['c'; 'o'; 'u'; 'l'; 'd'; ' '; 'n'; 'o'; 't'; ' '; 'e'; 'x';
                    't'; 'r'; 'a'; 'c'; 't'; ' '; 't'; 'e'; 'x'; 't'; ' '; 'f';
                    'o'; 'r'; ' '];
                   tag; [' '; 'f'; 'r'; 'o'; 'm'; ' ']; [newline];
                   shows_prec_xml Zero_nat (Xml (n, atts, cs, t)) []]));;

let rec nata
  tag x = bind (bind (text tag x) string2int) (fun i -> Inr (nat i));;

let rec no_transform
  tag xml =
    Inl (concat
          [['c'; 'o'; 'u'; 'l'; 'd'; ' '; 'n'; 'o'; 't'; ' '; 't'; 'r'; 'a';
             'n'; 's'; 'f'; 'o'; 'r'; 'm'; ' '; 't'; 'h'; 'e'; ' '; 'f'; 'o';
             'l'; 'l'; 'o'; 'w'; 'i'; 'n'; 'g'; '\010'; ' '; ' '; ' '; ' '; ' ';
             'x'; 'm'; 'l'; ' '; 'e'; 'l'; 'e'; 'm'; 'e'; 'n'; 't'; ' '; '(';
             'e'; 'x'; 'p'; 'e'; 'c'; 't'; 'e'; 'd'; ' '];
            tag; [')']; [newline]; shows_prec_xml Zero_nat xml []]);;

let rec nth
  x0 x1 = match x0, x1 with x :: xs, Suc n -> nth xs n
    | x :: xs, Zero_nat -> x;;

let rec plus_nat
  x0 n = match x0, n with Suc m, n -> plus_nat m (Suc n)
    | Zero_nat, n -> n;;

let rec size_list
  = function [] -> Zero_nat
    | a :: lista -> plus_nat (size_list lista) (Suc Zero_nat);;

let rec eq_nat
  x0 x1 = match x0, x1 with Zero_nat, Zero_nat -> true
    | Suc nat, Suc nat' -> eq_nat nat nat'
    | Zero_nat, Suc nat' -> false
    | Suc nat', Zero_nat -> false;;

let rec xml3elements
  tag p1 p2 p3 f (Xml (name, atts, cs, text)) =
    (if eq_list eq_char name tag &&
          (null atts &&
            (eq_nat (size_list cs) (nat (Big_int.big_int_of_int 3)) &&
              is_none text))
      then bind (p1 (nth cs Zero_nat))
             (fun x1 ->
               bind (p2 (nth cs one_nat))
                 (fun x2 ->
                   bind (p3 (nth cs (nat (Big_int.big_int_of_int 2))))
                     (fun x3 -> Inr (f x1 x2 x3))))
      else no_transform tag (Xml (name, atts, cs, text)));;

let rec map
  f x1 = match f, x1 with f, [] -> []
    | f, x :: xs -> f x :: map f xs;;

let rec sequence
  = function [] -> Inr []
    | m :: ms ->
        bind m (fun x -> bind (sequence ms) (fun xs -> Inr (x :: xs)));;

let rec mapm f ms = sequence (map f ms);;

let rec mapa x = mapm x;;

let rec xml_many_elements
  tag p f (Xml (name, atts, cs, text)) =
    (if eq_list eq_char name tag && (null atts && is_none text)
      then bind (mapa p cs) (comp (fun a -> Inr a) f)
      else no_transform tag (Xml (name, atts, cs, text)));;

let rec sp
  xml2name =
    xml_many_elements
      ['a'; 'r'; 'g'; 'u'; 'm'; 'e'; 'n'; 't'; 'F'; 'i'; 'l'; 't'; 'e'; 'r']
      (xml3elements
        ['a'; 'r'; 'g'; 'u'; 'm'; 'e'; 'n'; 't'; 'F'; 'i'; 'l'; 't'; 'e'; 'r';
          'E'; 'n'; 't'; 'r'; 'y']
        xml2name (nata ['a'; 'r'; 'i'; 't'; 'y'])
        (nata ['c'; 'o'; 'l'; 'l'; 'a'; 'p'; 's'; 'i'; 'n'; 'g'])
        (fun f _ n -> (f, minus_nat n one_nat)))
      (fun a -> SimpProj a);;

let rec leta s f = f s;;

type 'a ord = {less_eq : 'a -> 'a -> bool; less : 'a -> 'a -> bool};;
let less_eq _A = _A.less_eq;;
let less _A = _A.less;;

let rec hd (x :: xs) = x;;

let rec split f (a, b) = f a b;;

let rec lex_ext_unbounded
  f x1 x2 = match f, x1, x2 with f, [], [] -> (false, true)
    | f, uu :: uv, [] -> (true, true)
    | f, [], uw :: ux -> (false, false)
    | f, a :: asa, b :: bs ->
        (match f a b with (true, nstri) -> (true, true)
          | (false, true) -> lex_ext_unbounded f asa bs
          | (false, false) -> (false, false));;

let rec less_eq_nat
  x0 n = match x0, n with Suc m, n -> less_nat m n
    | Zero_nat, n -> true
and less_nat
  m x1 = match m, x1 with m, Suc n -> less_eq_nat m n
    | n, Zero_nat -> false;;

let rec lex_ext
  f n ss ts =
    let lts = size_list ts in
    (if eq_nat (size_list ss) lts || less_eq_nat lts n
      then lex_ext_unbounded f ss ts else (false, false));;

let rec forall
  p x1 = match p, x1 with p, [] -> true
    | p, x :: xs -> p x && forall p xs;;

let rec any
  uu x1 = match uu, x1 with uu, [] -> false
    | p, x :: xs -> p x || any p xs;;

type ('a, 'b) term = Var of 'b | Fun of 'a * ('a, 'b) term list;;

let rec lpo _B
  uu uv x2 x3 = match uu, uv, x2, x3 with
    uu, uv, Var x, Var y -> (false, eqa _B x y)
    | pr, uw, Var x, Fun (g, ts) ->
        (false, null ts && eq_nat (pr g Zero_nat) Zero_nat)
    | pr, n, Fun (f, ss), Var y ->
        let con = any (fun s -> snd (lpo _B pr n s (Var y))) ss in
        (con, con)
    | pr, n, Fun (f, ss), Fun (g, ts) ->
        (if any (fun s -> snd (lpo _B pr n s (Fun (g, ts)))) ss
          then (true, true)
          else let prl = pr f (size_list ss) in
               let prr = pr g (size_list ts) in
               (if less_eq_nat prr prl &&
                     forall (fun t -> fst (lpo _B pr n (Fun (f, ss)) t)) ts
                 then (if less_nat prr prl then (true, true)
                        else lex_ext (lpo _B pr n) n ss ts)
                 else (false, false)));;

type ('a, 'b) sub = Sub of ('b -> ('a, 'b) term);;

type af_entry = Collapse of nat | AFList of nat list;;

let rec xml_change p f x = bind (p x) (fun a -> Inr (f a));;

let rec position
  x = xml_change (nata ['p'; 'o'; 's'; 'i'; 't'; 'i'; 'o'; 'n'])
        (fun n -> minus_nat n one_nat) x;;

let rec get_name (Xml (name, uu, uv, uw)) = name;;

let rec eq_lista _A = ({eq = eq_list _A} : ('a list) eq);;

let rec lookup _A
  uu x1 = match uu, x1 with uu, [] -> None
    | a, (a', b) :: xs -> (if eqa _A a a' then Some b else lookup _A a xs);;

let rec concat_map
  f x1 = match f, x1 with f, [] -> []
    | f, x :: xs -> append (f x) (concat_map f xs);;

let rec option_case
  f1 f2 x2 = match f1, f2, x2 with f1, f2, Some a -> f2 a
    | f1, f2, None -> f1;;

let rec xml_options
  ps x =
    (match lookup (eq_lista eq_char) (get_name x) ps
      with None ->
        Inl (concat
              [['e'; 'x'; 'p'; 'e'; 'c'; 't'; 'e'; 'd'; ' '; 'o'; 'n'; 'e'; ' ';
                 'o'; 'f'; ':'; ' '];
                concat_map (fun p -> append (fst p) [' ']) ps; [newline];
                ['b'; 'u'; 't'; ' '; 'f'; 'o'; 'u'; 'n'; 'd']; [newline];
                shows_prec_xml Zero_nat x []])
      | Some p -> p x);;

let rec afs
  xml2name =
    xml_many_elements
      ['a'; 'r'; 'g'; 'u'; 'm'; 'e'; 'n'; 't'; 'F'; 'i'; 'l'; 't'; 'e'; 'r']
      (xml3elements
        ['a'; 'r'; 'g'; 'u'; 'm'; 'e'; 'n'; 't'; 'F'; 'i'; 'l'; 't'; 'e'; 'r';
          'E'; 'n'; 't'; 'r'; 'y']
        xml2name (nata ['a'; 'r'; 'i'; 't'; 'y'])
        (xml_options
          [(['c'; 'o'; 'l'; 'l'; 'a'; 'p'; 's'; 'i'; 'n'; 'g'],
             xml_change
               (nata ['c'; 'o'; 'l'; 'l'; 'a'; 'p'; 's'; 'i'; 'n'; 'g'])
               (fun i -> Collapse (minus_nat i one_nat)));
            (['n'; 'o'; 'n'; 'C'; 'o'; 'l'; 'l'; 'a'; 'p'; 's'; 'i'; 'n'; 'g'],
              xml_many_elements
                ['n'; 'o'; 'n'; 'C'; 'o'; 'l'; 'l'; 'a'; 'p'; 's'; 'i'; 'n';
                  'g']
                position (fun a -> AFList a))])
        (fun f a -> (fun b -> ((f, a), b))))
      id;;

let rec shows_between l p r = comp l (comp p r);;

let rec quote
  s = shows_between (shows_prec_char Zero_nat '\039') s
        (shows_prec_char Zero_nat '\039');;

let rec expecting _A
  msg ts =
    Inl (append ['e'; 'x'; 'p'; 'e'; 'c'; 't'; 'i'; 'n'; 'g'; ':'; ' ']
          (append msg
            (append
              [','; ' '; 'b'; 'u'; 't'; ' '; 'f'; 'o'; 'u'; 'n'; 'd'; ':'; ' ']
              (quote
                (shows_prec_list _A Zero_nat
                  (take (nat (Big_int.big_int_of_int 30)) ts))
                []))));;

let rec eoi _A
  = function [] -> Inr ((), [])
    | v :: va ->
        expecting _A
          ['e'; 'n'; 'd'; ' '; 'o'; 'f'; ' '; 'i'; 'n'; 'p'; 'u'; 't']
          (v :: va);;

let rec int tag x = bind (text tag x) string2int;;

type 'a poly = PVar of nat | PNum of 'a | PSum of 'a poly list |
  PMult of 'a poly list;;

type pos = Empty | PCons of nat * pos;;

let rec list2position
  = function [] -> Empty
    | n :: ns -> PCons (n, list2position ns);;

let rec pos
  x = xml_many_elements
        ['p'; 'o'; 's'; 'i'; 't'; 'i'; 'o'; 'n'; 'I'; 'n'; 'T'; 'e'; 'r'; 'm']
        position list2position x;;

let rec xml2elements
  tag p1 p2 f (Xml (name, atts, cs, text)) =
    (if eq_list eq_char name tag &&
          (null atts &&
            (eq_nat (size_list cs) (nat (Big_int.big_int_of_int 2)) &&
              is_none text))
      then bind (p1 (nth cs Zero_nat))
             (fun x1 -> bind (p2 (nth cs one_nat)) (fun x2 -> Inr (f x1 x2)))
      else no_transform tag (Xml (name, atts, cs, text)));;

type rat = Fract of Big_int.big_int * Big_int.big_int;;

let rec rat
  x = xml_options
        [(['i'; 'n'; 't'; 'e'; 'g'; 'e'; 'r'],
           xml_change (int ['i'; 'n'; 't'; 'e'; 'g'; 'e'; 'r'])
             (fun xa -> Fract (xa, (Big_int.big_int_of_int 1))));
          (['r'; 'a'; 't'; 'i'; 'o'; 'n'; 'a'; 'l'],
            xml2elements ['r'; 'a'; 't'; 'i'; 'o'; 'n'; 'a'; 'l']
              (int ['n'; 'u'; 'm'; 'e'; 'r'; 'a'; 't'; 'o'; 'r'])
              (int ['d'; 'e'; 'n'; 'o'; 'm'; 'i'; 'n'; 'a'; 't'; 'o'; 'r'])
              (fun a b -> Fract (a, b)))]
        x;;

let rec filter
  p x1 = match p, x1 with p, [] -> []
    | p, x :: xs -> (if p x then x :: filter p xs else filter p xs);;

let rec string2bool
  s = (if eq_list eq_char s ['t'; 'r'; 'u'; 'e'] then Inr true
        else (if eq_list eq_char s ['f'; 'a'; 'l'; 's'; 'e'] then Inr false
               else Inl (append
                          ['c'; 'a'; 'n'; 'n'; 'o'; 't'; ' '; 'c'; 'o'; 'n';
                            'v'; 'e'; 'r'; 't'; ' ']
                          (append s
                            [' '; 'i'; 'n'; 't'; 'o'; ' '; 'B'; 'o'; 'o'; 'l';
                              'e'; 'a'; 'n']))));;

let rec boola tag node = bind (text tag node) string2bool;;

let rec scg_position x = nata ['p'; 'o'; 's'; 'i'; 't'; 'i'; 'o'; 'n'] x;;

let rec xml1element
  tag p1 f (Xml (name, atts, cs, text)) =
    (if eq_list eq_char name tag &&
          (null atts && (eq_nat (size_list cs) one_nat && is_none text))
      then bind (p1 (nth cs Zero_nat)) (fun x1 -> Inr (f x1))
      else no_transform tag (Xml (name, atts, cs, text)));;

let rec xml1many_elements_gen
  tag p1 p2 f (Xml (name, atts, cs, text)) =
    (if eq_list eq_char name tag &&
          (null atts && (is_none text && not (null cs)))
      then bind (p1 (hd cs))
             (fun x -> bind (mapa (p2 x) (tl cs)) (fun xs -> Inr (f x xs)))
      else no_transform tag (Xml (name, atts, cs, text)));;

let rec xml1many_elements
  tag p1 p2 = xml1many_elements_gen tag p1 (fun _ -> p2);;

let rec var
  x = bind (text ['v'; 'a'; 'r'] x) (comp (fun a -> Inr a) (fun a -> Var a));;

let rec term
  xml2name x =
    xml_options
      [(['v'; 'a'; 'r'], var);
        (['f'; 'u'; 'n'; 'a'; 'p'; 'p'],
          xml1many_elements ['f'; 'u'; 'n'; 'a'; 'p'; 'p'] xml2name
            (xml1element ['a'; 'r'; 'g'] (term xml2name) id)
            (fun a b -> Fun (a, b)))]
      x;;

let rec rule
  xml2name =
    xml2elements ['r'; 'u'; 'l'; 'e']
      (xml1element ['l'; 'h'; 's'] (term xml2name) id)
      (xml1element ['r'; 'h'; 's'] (term xml2name) id) (fun a b -> (a, b));;

let rec scg
  xml2name =
    xml1many_elements
      ['s'; 'i'; 'z'; 'e'; 'C'; 'h'; 'a'; 'n'; 'g'; 'e'; 'G'; 'r'; 'a'; 'p';
        'h']
      (rule xml2name)
      (xml3elements ['e'; 'd'; 'g'; 'e'] scg_position
        (boola ['s'; 't'; 'r'; 'i'; 'c'; 't']) scg_position
        (fun p s q -> (s, (p, q))))
      (fun lr edges ->
        (lr, (map snd (filter fst edges),
               map snd (filter (fun x -> not (fst x)) edges))));;

let rec sub
  xml2name =
    xml_many_elements
      ['s'; 'u'; 'b'; 's'; 't'; 'i'; 't'; 'u'; 't'; 'i'; 'o'; 'n']
      (xml2elements ['s'; 'u'; 'b'; 's'; 't'; 'E'; 'n'; 't'; 'r'; 'y']
        (text ['v'; 'a'; 'r']) (term xml2name) (fun a b -> (a, b)))
      id;;

let rec decomp
  x0 x1 = match x0, x1 with [], [] -> Some []
    | uu :: uv, [] -> None
    | [], uw :: ux -> None
    | s :: ss, t :: ts ->
        (match decomp ss ts with None -> None
          | Some pairs -> Some ((s, t) :: pairs));;

type ('a, 'b) hterm = HHole | HFun of 'a * ('a, 'b) hterm list;;

let rec gcMatchList _A
  = function [] -> Some []
    | (HHole, uu) :: ps -> gcMatchList _A ps
    | (HFun (f, hs), Fun (g, ss)) :: ps ->
        (if eqa _A f g
          then (match decomp hs ss with None -> None
                 | Some pairs -> gcMatchList _A (append pairs ps))
          else None)
    | (HFun (f, hs), Var x) :: ps ->
        (match gcMatchList _A ps with None -> None
          | Some pairs -> Some ((HFun (f, hs), x) :: pairs));;

let rec mergeHTerms _A
  x0 x1 = match x0, x1 with [], [] -> Some []
    | [], uu :: uv -> None
    | uw :: ux, [] -> None
    | HHole :: hsl, h :: hsr ->
        (match mergeHTerms _A hsl hsr with None -> None
          | Some res -> Some (h :: res))
    | HFun (f, hs) :: hsl, HHole :: hsr ->
        (match mergeHTerms _A hsl hsr with None -> None
          | Some res -> Some (HFun (f, hs) :: res))
    | HFun (f, fs) :: hsl, HFun (g, gs) :: hsr ->
        (if eqa _A f g
          then (match mergeHTerms _A fs gs with None -> None
                 | Some mhs ->
                   (match mergeHTerms _A hsl hsr with None -> None
                     | Some res -> Some (HFun (f, mhs) :: res)))
          else None);;

let rec merge _A
  hl hr =
    (match mergeHTerms _A [hl] [hr] with None -> None
      | Some res -> Some (nth res Zero_nat));;

let rec mergeX _A _B
  x h xa2 = match x, h, xa2 with x, h, [] -> Some []
    | x, h, (ht, y) :: ps ->
        (if eqa _A x y
          then (match merge _B h ht with None -> None
                 | Some m -> mergeX _A _B x m ps)
          else (match mergeX _A _B x h ps with None -> None
                 | Some pairs -> Some ((ht, y) :: pairs)));;

let rec mergeAll _A _B
  = function [] -> true
    | (h, x) :: ps ->
        (match mergeX _B _A x h ps with None -> false
          | Some a -> mergeAll _A _B a);;

let rec gcMatchI _A _B
  h t = (match gcMatchList _A [(h, t)] with None -> false
          | Some a -> mergeAll _A _B a);;

let rec capI _A _B
  uu x1 = match uu, x1 with uu, Var uv -> HHole
    | r, Fun (f, ts) ->
        let h = HFun (f, map (capI _A _B r) ts) in
        (if any (fun ra -> gcMatchI _A _B h (fst ra)) r then HHole else h);;

let rec foldl
  f a x2 = match f, a, x2 with f, a, [] -> a
    | f, a, x :: xs -> foldl f (f a x) xs;;

let rec rev xs = foldl (fun xsa x -> x :: xsa) [] xs;;

let rec upt i j = (if less_nat i j then i :: upt (Suc i) j else []);;

let rec list_case
  f1 f2 x2 = match f1, f2, x2 with f1, f2, a :: lista -> f2 a lista
    | f1, f2, [] -> f1;;

let rec zip
  xs x1 = match xs, x1 with xs, [] -> []
    | xs, y :: ys -> (match xs with [] -> [] | z :: zs -> (z, y) :: zip zs ys);;

let rec prod_case f1 (a, b) = f1 a b;;

let rec eqb _A _B (x1, y1) (x2, y2) = eqa _A x1 x2 && eqa _B y1 y2;;

let rec eqc _A _B = ({eq = eqb _A _B} : ('a * 'b) eq);;

let rec member _A
  x xa1 = match x, xa1 with x, [] -> false
    | x, y :: ys -> eqa _A x y || member _A x ys;;

let rec remdups _A
  = function [] -> []
    | x :: xs ->
        (if member _A x xs then remdups _A xs else x :: remdups _A xs);;

let rec compa _A
  es es' =
    remdups (eqc _A _A)
      (concat_map
        (fun (x, y) ->
          concat_map (fun (y', z) -> (if eqa _A y y' then [(x, z)] else []))
            es')
        es);;

let rec shows_paren
  p = comp (shows_prec_char Zero_nat '(')
        (comp p (shows_prec_char Zero_nat ')'));;

let rec shows_preca _A _B
  d p = shows_paren
          (comp (shows_prec _A Zero_nat (fst p))
            (comp (shows_prec_char Zero_nat ',')
              (shows_prec _B Zero_nat (snd p))));;

let rec shows_sep
  s sep x2 = match s, sep, x2 with s, sep, [] -> shows_string []
    | s, sep, [x] -> s x
    | s, sep, x :: v :: va ->
        comp (s x) (comp sep (shows_sep s sep (v :: va)));;

let rec shows_list_gen
  elt e l s r xs =
    (if null xs then shows_string e
      else shows_between (shows_string l) (shows_sep elt (shows_string s) xs)
             (shows_string r));;

let rec shows_list_aux s xs = shows_list_gen s ['['; ']'] ['['] [','] [']'] xs;;

let rec shows_lista _A _B ps = shows_list_aux (shows_preca _A _B Zero_nat) ps;;

let rec show _A _B =
  ({shows_prec = shows_preca _A _B; shows_list = shows_lista _A _B} :
    ('a * 'b) show);;

type ('a, 'b) ctxt = Hole |
  More of 'a * ('a, 'b) term list * ('a, 'b) ctxt * ('a, 'b) term list;;

type ('a, 'b, 'c, 'd, 'e, 'f, 'g, 'h) alter = Alter1 of 'a | Alter2 of 'b |
  Alter3 of 'c | Alter4 of 'd | Alter5 of 'e | Alter6 of 'f | Alter7 of 'g |
  Alter8 of 'h;;

let rec binda m f ts = bind (m ts) (fun (a, b) -> f a b);;

let rec xml4elements
  tag p1 p2 p3 p4 f (Xml (name, atts, cs, text)) =
    (if eq_list eq_char name tag &&
          (null atts &&
            (eq_nat (size_list cs) (nat (Big_int.big_int_of_int 4)) &&
              is_none text))
      then bind (p1 (nth cs Zero_nat))
             (fun x1 ->
               bind (p2 (nth cs one_nat))
                 (fun x2 ->
                   bind (p3 (nth cs (nat (Big_int.big_int_of_int 2))))
                     (fun x3 ->
                       bind (p4 (nth cs (nat (Big_int.big_int_of_int 3))))
                         (fun x4 -> Inr (f x1 x2 x3 x4)))))
      else no_transform tag (Xml (name, atts, cs, text)));;

let rec xml0elements
  tag x (Xml (name, atts, cs, text)) =
    (if eq_list eq_char name tag && (null atts && (null cs && is_none text))
      then Inr x else no_transform tag (Xml (name, atts, cs, text)));;

let rec ctxt
  xml2name x =
    xml_options
      [(['b'; 'o'; 'x'], xml0elements ['b'; 'o'; 'x'] Hole);
        (['f'; 'u'; 'n'; 'C'; 'o'; 'n'; 't'; 'e'; 'x'; 't'],
          xml4elements ['f'; 'u'; 'n'; 'C'; 'o'; 'n'; 't'; 'e'; 'x'; 't']
            xml2name
            (xml_many_elements ['b'; 'e'; 'f'; 'o'; 'r'; 'e'] (term xml2name)
              id)
            (ctxt xml2name)
            (xml_many_elements ['a'; 'f'; 't'; 'e'; 'r'] (term xml2name) id)
            (fun a b c d -> More (a, b, c, d)))]
      x;;

let rec xml3to4elements
  tag p1 p2 p3 p4 f (Xml (name, atts, cs, text)) =
    let l = size_list cs in
    (if eq_list eq_char name tag &&
          (null atts &&
            (less_eq_nat (nat (Big_int.big_int_of_int 3)) l &&
              (less_eq_nat l (nat (Big_int.big_int_of_int 4)) && is_none text)))
      then bind (p1 (nth cs Zero_nat))
             (fun x1 ->
               bind (p2 (nth cs one_nat))
                 (fun x2 ->
                   (if eq_nat l (nat (Big_int.big_int_of_int 4))
                     then bind (p3 (nth cs (nat (Big_int.big_int_of_int 2))))
                            (fun x3 ->
                              bind (p4 (nth cs
 (nat (Big_int.big_int_of_int 3))))
                                (fun x4 -> Inr (f x1 x2 (Some x3) x4)))
                     else bind (p4 (nth cs (nat (Big_int.big_int_of_int 2))))
                            (fun x4 -> Inr (f x1 x2 None x4)))))
      else no_transform tag (Xml (name, atts, cs, text)));;

let rec rstep
  xml2name =
    xml3to4elements ['r'; 'e'; 'w'; 'r'; 'i'; 't'; 'e'; 'S'; 't'; 'e'; 'p'] pos
      (rule xml2name) (xml0elements ['r'; 'e'; 'l'; 'a'; 't'; 'i'; 'v'; 'e'] ())
      (term xml2name) (fun p r rel t -> (p, (r, (is_none rel, t))));;

let rec rsteps
  xml2name =
    xml1many_elements
      ['r'; 'e'; 'w'; 'r'; 'i'; 't'; 'e'; 'S'; 'e'; 'q'; 'u'; 'e'; 'n'; 'c';
        'e']
      (xml1element ['s'; 't'; 'a'; 'r'; 't'; 'T'; 'e'; 'r'; 'm'] (term xml2name)
        id)
      (rstep xml2name) (fun a b -> (a, b));;

let rec loop
  xml2name =
    xml3elements ['l'; 'o'; 'o'; 'p'] (rsteps xml2name) (sub xml2name)
      (ctxt xml2name) (fun (s, rseq) sigma c -> (s, (rseq, (sigma, c))));;

let rec many
  p x1 = match p, x1 with p, [] -> Inr ([], [])
    | p, t :: ts ->
        (if p t then bind (many p ts) (fun (rs, ts') -> Inr (t :: rs, ts'))
          else Inr ([], t :: ts));;

let rec proj_term
  p x1 = match p, x1 with p, Var x -> Var x
    | p, Fun (f, ts) ->
        (if less_nat (p f) (size_list ts) then nth ts (p f) else Fun (f, ts));;

let rec list_all2
  p xs ys = match p, xs, ys with
    p, x :: xs, y :: ys -> p x y && list_all2 p xs ys
    | p, xs, [] -> null xs
    | p, [], ys -> null ys;;

let rec eq_term _A _B
  x0 x1 = match x0, x1 with
    Fun (f, ts), Fun (g, us) -> eq _A f g && list_all2 (eq_term _A _B) ts us
    | Var x, Var y -> eq _B x y
    | Var v, Fun (f', list') -> false
    | Fun (f', list'), Var v -> false;;

let rec start_term _A
  xml2name t =
    xml1element ['s'; 't'; 'a'; 'r'; 't'; 'T'; 'e'; 'r'; 'm']
      (fun xml ->
        bind (term xml2name xml)
          (fun s ->
            (if eq_term _A (eq_lista eq_char) s t then Inr t
              else Inl ['<'; 's'; 't'; 'a'; 'r'; 't'; 'T'; 'e'; 'r'; 'm'; '>';
                         ' '; 'd'; 'o'; 'e'; 's'; ' '; 'n'; 'o'; 't'; ' '; 'm';
                         'a'; 't'; 'c'; 'h'; ' '; 'l'; 'h'; 's'])))
      id;;

let rec rseq _A
  xml2name pi r =
    xml1many_elements
      ['r'; 'e'; 'w'; 'r'; 'i'; 't'; 'e'; 'S'; 'e'; 'q'; 'u'; 'e'; 'n'; 'c';
        'e']
      (start_term _A xml2name (proj_term pi (fst r))) (rstep xml2name)
      (fun _ rseq -> (r, map (fun (x, (y, (_, z))) -> (x, (y, z))) rseq));;

let wspace : char list = [newline; '\009'; '\013'; ' '];;

let rec trim
  = function [] -> []
    | c :: s -> (if member eq_char c wspace then trim s else c :: s);;

let rec capRM2 _A _B
  uu x1 = match uu, x1 with uu, Var uv -> HHole
    | rm, Fun (f, ts) ->
        let h = HFun (f, map (capRM2 _A _B rm) ts) in
        (if any (fun r -> gcMatchI _A _B h (fst r)) (rm f) then HHole else h);;

let rec capRM _A _B
  r = (match r with None -> (fun _ -> HHole) | Some a -> capRM2 _A _B a);;

let eq_int = ({eq = Big_int.eq_big_int} : Big_int.big_int eq);;

let rec drop
  n x1 = match n, x1 with n, [] -> []
    | n, x :: xs -> (match n with Zero_nat -> x :: xs | Suc m -> drop m xs);;

let rec last (x :: xs) = (if null xs then x else last xs);;

let eq_nata = ({eq = eq_nat} : nat eq);;

let rec combine _B
  x0 s = match x0, s with
    Scg (p, q, str, wk), Scg (p', q', str', wk') ->
      Scg (p, q, remdups (eqc _B _B) (append str str'),
            remdups (eqc _B _B) (append wk wk'))
    | Null, s -> Null
    | Scg (v, va, vb, vc), Null -> Null;;

let rec list_all
  p x1 = match p, x1 with p, [] -> true
    | p, x :: xs -> p x && list_all p xs;;

let rec subsumes _A _B
  g x1 = match g, x1 with
    Scg (p, q, str, wk), Scg (p', q', str', wk') ->
      eqa _A p p' &&
        (eqa _A q q' &&
          (list_all (fun x -> member (eqc _B _B) x str') str &&
            list_all (fun x -> member (eqc _B _B) x (append str' wk')) wk))
    | g, Null -> true
    | Null, Scg (v, va, vb, vc) -> false;;

let rec subtract_list _A xs ys = filter (fun x -> not (member _A x ys)) xs;;

let rec scg_comp _B
  conn x1 g = match conn, x1, g with
    conn, Scg (p, q, str, wk), Scg (p', q', str', wk') ->
      (if not (conn q p') then Null
        else let strs =
               remdups (eqc _B _B)
                 (append (compa _B str str')
                   (append (compa _B str wk') (compa _B wk str')))
               in
             let a =
               subtract_list (eqc _B _B) (remdups (eqc _B _B) (compa _B wk wk'))
                 strs
               in
             Scg (p, q', strs, a))
    | conn, Null, g -> Null
    | conn, Scg (v, va, vb, vc), Null -> Null;;

let rec list_ex
  p x1 = match p, x1 with p, [] -> false
    | p, x :: xs -> p x || list_ex p xs;;

let rec in_situ _B
  = function Null -> true
    | Scg (p, q, str, wk) -> list_ex (fun (a, b) -> eqa _B a b) str;;

let rec sagiv _A _B
  conn g =
    (if in_situ _B g then true
      else let gg = scg_comp _B conn g g in
           (if subsumes _A _B gg g then false
             else sagiv _A _B conn (combine _B g gg)));;

let rec args = function Var x -> []
               | Fun (f, ts) -> ts;;

let rec root = function Var x -> None
               | Fun (f, ts) -> Some f;;

let rec eq_shp _A
  x0 x1 = match x0, x1 with Sharp f, Sharp f' -> eqa _A f f'
    | Plain f, Plain f' -> eqa _A f f'
    | Sharp f, Plain f' -> false
    | Plain f', Sharp f -> false;;

let rec eq_shpa _A = ({eq = eq_shp _A} : 'a shp eq);;

type domain = Natural | Arctic | Arctic_rat of rat | Int_mat of nat * nat |
  Arc_mat of nat | Arc_rat_mat of rat * nat | Rational of rat |
  Rat_mat of rat * nat * nat;;

let rec label
  xml2name =
    xml_options
      [(['n'; 'u'; 'm'; 'b'; 'e'; 'r'; 'L'; 'a'; 'b'; 'e'; 'l'],
         xml_many_elements
           ['n'; 'u'; 'm'; 'b'; 'e'; 'r'; 'L'; 'a'; 'b'; 'e'; 'l']
           (nata ['n'; 'u'; 'm'; 'b'; 'e'; 'r']) (fun a -> Inl a));
        (['s'; 'y'; 'm'; 'b'; 'o'; 'l'; 'L'; 'a'; 'b'; 'e'; 'l'],
          xml_many_elements
            ['s'; 'y'; 'm'; 'b'; 'o'; 'l'; 'L'; 'a'; 'b'; 'e'; 'l'] xml2name
            (fun a -> Inr a))];;

let rec shows_list_list _A
  ls = shows_list_aux (shows_prec_list _A Zero_nat) ls;;

let rec oneof_aux
  allowed x1 ts = match allowed, x1, ts with
    allowed, [], ts ->
      expecting show_char
        (append ['o'; 'n'; 'e'; ' '; 'o'; 'f'; ' ']
          (shows_list_list show_char allowed []))
        ts
    | allowed, x :: xs, ts ->
        (if eq_list eq_char (map snd (zip x ts)) x
          then Inr (x, trim (drop (size_list x) ts))
          else oneof_aux allowed xs ts);;

let rec oneof xs = oneof_aux xs xs;;

let rec rules
  xml2name = xml_many_elements ['r'; 'u'; 'l'; 'e'; 's'] (rule xml2name) id;;

type xmldoc = Xmldoc of char list * xml;;

let ord_int =
  ({less_eq = Big_int.le_big_int; less = Big_int.lt_big_int} :
    Big_int.big_int ord);;

let rec foldr
  f x1 a = match f, x1, a with f, [], a -> a
    | f, x :: xs, a -> f x (foldr f xs a);;

let ord_nat = ({less_eq = less_eq_nat; less = less_nat} : nat ord);;

let rec eq_scg _A _B
  x0 x1 = match x0, x1 with Null, Null -> true
    | Scg (pp1, pp2, list1, list2), Scg (pp1', pp2', list1', list2') ->
        eqa _A pp1 pp1' &&
          (eqa _A pp2 pp2' &&
            (eq_list (eqc _B _B) list1 list1' &&
              eq_list (eqc _B _B) list2 list2'))
    | Null, Scg (pp1', pp2', list1', list2') -> false
    | Scg (pp1', pp2', list1', list2'), Null -> false;;

let rec eq_terma _A _B = ({eq = eq_term _A _B} : ('a, 'b) term eq);;

let rec term_case
  f1 f2 x2 = match f1, f2, x2 with f1, f2, Fun (f, lista) -> f2 f lista
    | f1, f2, Var v -> f1 v;;

let rec suptp _A _B
  s t = (match s with Var _ -> false
          | Fun (_, ts) ->
            member (eq_terma _A _B) t ts ||
              any (fun sa -> suptp _A _B sa t) ts);;

let rec plaina
  = function Var x -> Var x
    | Fun (f, ss) -> Fun (Plain f, map plaina ss);;

let rec sharpa
  = function Var x -> Var x
    | Fun (f, ss) -> Fun (Sharp f, map plaina ss);;

let rec debug i t x = x;;

let rec manyof cs = many (fun c -> member eq_char c cs);;

let rec spaces x = binda (manyof wspace) (fun _ ts -> Inr ((), ts)) x;;

let rec eq_hterm _A () = ({eq = eq_hterma _A} : ('a, 'b) hterm eq)
and eq_hterma _A
  x0 x1 = match x0, x1 with HHole, HHole -> true
    | HFun (f, lista), HFun (f', list') ->
        eqa _A f f' && eq_list (eq_hterm _A ()) lista list'
    | HHole, HFun (f', list') -> false
    | HFun (f', list'), HHole -> false;;

let rec fun_upd _A f a b = (fun x -> (if eqa _A x a then b else f x));;

let rec abs_int
  i = (if Big_int.lt_big_int i (Big_int.big_int_of_int 0)
        then Big_int.minus_big_int i else i);;

let rec sgn_int
  i = (if Big_int.eq_big_int i (Big_int.big_int_of_int 0)
        then (Big_int.big_int_of_int 0)
        else (if Big_int.lt_big_int (Big_int.big_int_of_int 0) i
               then (Big_int.big_int_of_int 1)
               else Big_int.minus_big_int (Big_int.big_int_of_int 1)));;

let rec apsnd f (x, y) = (x, f y);;

let rec divmod_int
  k l = (if Big_int.eq_big_int k (Big_int.big_int_of_int 0)
          then ((Big_int.big_int_of_int 0), (Big_int.big_int_of_int 0))
          else (if Big_int.eq_big_int l (Big_int.big_int_of_int 0)
                 then ((Big_int.big_int_of_int 0), k)
                 else apsnd (Big_int.mult_big_int (sgn_int l))
                        (if Big_int.eq_big_int (sgn_int k) (sgn_int l)
                          then (fun k -> fun l -> Big_int.quomod_big_int
                                 (Big_int.abs_big_int k)
                                 (Big_int.abs_big_int l))
                                 k l
                          else let (r, s) =
                                 (fun k -> fun l -> Big_int.quomod_big_int
                                   (Big_int.abs_big_int k)
                                   (Big_int.abs_big_int l))
                                   k l
                                 in
                               (if Big_int.eq_big_int s
                                     (Big_int.big_int_of_int 0)
                                 then (Big_int.minus_big_int r,
(Big_int.big_int_of_int 0))
                                 else (Big_int.sub_big_int
 (Big_int.minus_big_int r) (Big_int.big_int_of_int 1),
Big_int.sub_big_int (abs_int l) s)))));;

let rec mod_int a b = snd (divmod_int a b);;

let rec gcd_int
  k l = abs_int
          (if Big_int.eq_big_int l (Big_int.big_int_of_int 0) then k
            else gcd_int l (mod_int (abs_int k) (abs_int l)));;

let rec implies
  p pa = match p, pa with p, true -> true
    | p, false -> not p
    | true, p -> p
    | false, p -> true;;

let rec is_var = function Var x -> true
                 | Fun (f, ts) -> false;;

let rec shows_shp _A
  = function Sharp s -> comp (shows_prec _A Zero_nat s) (shows_string ['#'])
    | Plain s -> shows_prec _A Zero_nat s;;

let rec shows_prec_shp _A d f = shows_shp _A f;;

let rec shows_list_shp _A fs = shows_list_aux (shows_prec_shp _A Zero_nat) fs;;

let rec show_shp _A =
  ({shows_prec = shows_prec_shp _A; shows_list = shows_list_shp _A} :
    'a shp show);;

let rec shows_nl x = shows_prec_char Zero_nat newline x;;

let rec shows_term _A _B
  = function Var x -> shows_prec _B Zero_nat x
    | Fun (f, ts) ->
        comp (shows_prec _A Zero_nat f)
          (shows_list_gen id [] ['('] [','; ' '] [')']
            (map (shows_term _A _B) ts));;

let rec shows_prec_term _A _B d t = shows_term _A _B t;;

let rec shows_rule _A _B
  (l, r) =
    comp (shows_prec_term _A _B Zero_nat l)
      (comp (shows_string [' '; '-'; '>'; ' '])
        (shows_prec_term _A _B Zero_nat r));;

let rec shows_rules _A _B
  trs = comp (shows_list_gen (shows_rule _A _B) [] [] [newline] [] trs)
          shows_nl;;

let rec shows_p _A _B
  p = comp (shows_string
             ['d'; 'e'; 'p'; 'e'; 'n'; 'd'; 'e'; 'n'; 'c'; 'y'; ' '; 'p'; 'a';
               'i'; 'r'; 's'; ':'])
        (comp shows_nl (shows_rules _A _B p));;

let rec alter1a (Alter1 x) = x;;

let rec alter2a (Alter2 x) = x;;

let rec alter3a (Alter3 x) = x;;

let rec alter4a (Alter4 x) = x;;

let rec alter5a (Alter5 x) = x;;

let rec alter6a (Alter6 x) = x;;

let rec alter7a (Alter7 x) = x;;

let rec alter8a (Alter8 x) = x;;

let letters : char list
  = ['a'; 'b'; 'c'; 'd'; 'e'; 'f'; 'g'; 'h'; 'i'; 'j'; 'k'; 'l'; 'm'; 'n'; 'o';
      'p'; 'q'; 'r'; 's'; 't'; 'u'; 'v'; 'w'; 'x'; 'y'; 'z'; 'A'; 'B'; 'C'; 'D';
      'E'; 'F'; 'G'; 'H'; 'I'; 'J'; 'K'; 'L'; 'M'; 'N'; 'O'; 'P'; 'Q'; 'R'; 'S';
      'T'; 'U'; 'V'; 'W'; 'X'; 'Y'; 'Z'; '_'; '0'; '1'; '2'; '3'; '4'; '5'; '6';
      '7'; '8'; '9'; '&'; ';'; ':'];;

type arctic_rat = MinInfty_rat | Num_arc_rat of rat;;

type arctic = MinInfty | Num_arc of Big_int.big_int;;

type 'a redpair = Nat_carrier of (('a * nat) * (nat * nat list)) list |
  Int_carrier of (('a * nat) * (Big_int.big_int * Big_int.big_int list)) list |
  Arctic_carrier of (('a * nat) * (arctic * arctic list)) list |
  Arctic_rat_carrier of rat * (('a * nat) * (arctic_rat * arctic_rat list)) list
  | Int_mat_carrier of
      nat * nat *
        (('a * nat) *
          ((Big_int.big_int list) list *
            ((Big_int.big_int list) list) list)) list
  | Arc_mat_carrier of
      nat * (('a * nat) * ((arctic list) list * ((arctic list) list) list)) list
  | Arc_rat_mat_carrier of
      rat * nat *
        (('a * nat) *
          ((arctic_rat list) list * ((arctic_rat list) list) list)) list
  | Lpo of (('a * nat) * nat) list * (('a * nat) * af_entry) list |
  Rat_carrier of rat * (('a * nat) * (rat * rat list)) list |
  Rat_mat_carrier of
    rat * nat * nat *
      (('a * nat) * ((rat list) list * ((rat list) list) list)) list;;

let rec precedence
  xml2name =
    xml_many_elements
      ['s'; 't'; 'a'; 't'; 'u'; 's'; 'P'; 'r'; 'e'; 'c'; 'e'; 'd'; 'e'; 'n';
        'c'; 'e']
      (xml4elements
        ['s'; 't'; 'a'; 't'; 'u'; 's'; 'P'; 'r'; 'e'; 'c'; 'e'; 'd'; 'e'; 'n';
          'c'; 'e'; 'E'; 'n'; 't'; 'r'; 'y']
        xml2name (nata ['a'; 'r'; 'i'; 't'; 'y'])
        (nata ['p'; 'r'; 'e'; 'c'; 'e'; 'd'; 'e'; 'n'; 'c'; 'e'])
        (xml_options
          [(['l'; 'e'; 'x'], xml0elements ['l'; 'e'; 'x'] ());
            (['m'; 'u'; 'l'],
              (fun _ ->
                Inl ['m'; 'u'; 'l'; 't'; 'i'; 's'; 'e'; 't'; '-'; 'c'; 'o'; 'm';
                      'p'; 'a'; 'r'; 'i'; 's'; 'o'; 'n'; 's'; ' '; 'n'; 'o';
                      't'; ' '; 's'; 'u'; 'p'; 'p'; 'o'; 'r'; 't'; 'e'; 'd']))])
        (fun f a p _ -> ((f, a), p)))
      id;;

let rec xml1to2elements
  tag p1 p2 f (Xml (name, atts, cs, text)) =
    let l = size_list cs in
    (if eq_list eq_char name tag &&
          (null atts &&
            (less_eq_nat one_nat l &&
              (less_eq_nat l (nat (Big_int.big_int_of_int 2)) && is_none text)))
      then bind (p1 (nth cs Zero_nat))
             (fun x1 ->
               (if eq_nat l (nat (Big_int.big_int_of_int 2))
                 then bind (p2 (nth cs one_nat))
                        (fun x2 -> Inr (f x1 (Some x2)))
                 else Inr (f x1 None)))
      else no_transform tag (Xml (name, atts, cs, text)));;

let rec path_order
  xml2name =
    xml1to2elements ['p'; 'a'; 't'; 'h'; 'O'; 'r'; 'd'; 'e'; 'r']
      (precedence xml2name) (afs xml2name)
      (fun pr a ->
        (match a with None -> Lpo (pr, []) | Some aa -> Lpo (pr, aa)));;

let rat_name : char list
  = ['r'; 'a'; 't'; 'i'; 'o'; 'n'; 'a'; 'l'; 's'; ' '; '('; 'w'; 'i'; 't'; 'h';
      ' '; 'n'; 'e'; 'g'; 'a'; 't'; 'i'; 'v'; 'e'; ' '; 'c'; 'o'; 'n'; 's'; 't';
      'a'; 'n'; 't'; 's'; ')'];;

let rec div_int a b = fst (divmod_int a b);;

let rec fract_norm
  a b = (if Big_int.eq_big_int a (Big_int.big_int_of_int 0) ||
              Big_int.eq_big_int b (Big_int.big_int_of_int 0)
          then Fract ((Big_int.big_int_of_int 0), (Big_int.big_int_of_int 1))
          else let c = gcd_int a b in
               (if Big_int.lt_big_int (Big_int.big_int_of_int 0) b
                 then Fract (div_int a c, div_int b c)
                 else Fract (Big_int.minus_big_int (div_int a c),
                              Big_int.minus_big_int (div_int b c))));;

let rec minus_rat
  (Fract (a, b)) (Fract (c, d)) =
    (if Big_int.eq_big_int b (Big_int.big_int_of_int 0)
      then Fract (Big_int.minus_big_int c, d)
      else (if Big_int.eq_big_int d (Big_int.big_int_of_int 0) then Fract (a, b)
             else fract_norm
                    (Big_int.sub_big_int (Big_int.mult_big_int a d)
                      (Big_int.mult_big_int c b))
                    (Big_int.mult_big_int b d)));;

let rec less_eq_rat
  (Fract (a, b)) (Fract (c, d)) =
    (if Big_int.eq_big_int b (Big_int.big_int_of_int 0)
      then Big_int.le_big_int (Big_int.big_int_of_int 0)
             (Big_int.mult_big_int (sgn_int c) (sgn_int d))
      else (if Big_int.eq_big_int d (Big_int.big_int_of_int 0)
             then Big_int.le_big_int
                    (Big_int.mult_big_int (sgn_int a) (sgn_int b))
                    (Big_int.big_int_of_int 0)
             else Big_int.le_big_int
                    (Big_int.mult_big_int (Big_int.mult_big_int a (abs_int d))
                      (sgn_int b))
                    (Big_int.mult_big_int (Big_int.mult_big_int c (abs_int b))
                      (sgn_int d))));;

let rec less_rat
  (Fract (a, b)) (Fract (c, d)) =
    (if Big_int.eq_big_int b (Big_int.big_int_of_int 0)
      then Big_int.lt_big_int (Big_int.big_int_of_int 0)
             (Big_int.mult_big_int (sgn_int c) (sgn_int d))
      else (if Big_int.eq_big_int d (Big_int.big_int_of_int 0)
             then Big_int.lt_big_int
                    (Big_int.mult_big_int (sgn_int a) (sgn_int b))
                    (Big_int.big_int_of_int 0)
             else Big_int.lt_big_int
                    (Big_int.mult_big_int (Big_int.mult_big_int a (abs_int d))
                      (sgn_int b))
                    (Big_int.mult_big_int (Big_int.mult_big_int c (abs_int b))
                      (sgn_int d))));;

let rec positive_delta
  delta =
    (if less_rat
          (Fract ((Big_int.big_int_of_int 0), (Big_int.big_int_of_int 1))) delta
      then delta
      else Fract ((Big_int.big_int_of_int 1), (Big_int.big_int_of_int 1)));;

let rec rat_gt
  delta =
    let d = positive_delta delta in
    (fun x y -> less_eq_rat d (minus_rat x y));;

let rec ge_rat x y = less_eq_rat y x;;

let ord_rat = ({less_eq = less_eq_rat; less = less_rat} : rat ord);;

let rec maxa less_eq a b = (if less_eq a b then b else a);;

let rec max _A = maxa (less_eq _A);;

let rec max0_rat
  x = max ord_rat
        (Fract ((Big_int.big_int_of_int 0), (Big_int.big_int_of_int 1))) x;;

let rec times_rat
  (Fract (a, b)) (Fract (c, d)) =
    fract_norm (Big_int.mult_big_int a c) (Big_int.mult_big_int b d);;

let rec plus_rat
  (Fract (a, b)) (Fract (c, d)) =
    (if Big_int.eq_big_int b (Big_int.big_int_of_int 0) then Fract (c, d)
      else (if Big_int.eq_big_int d (Big_int.big_int_of_int 0) then Fract (a, b)
             else fract_norm
                    (Big_int.add_big_int (Big_int.mult_big_int a d)
                      (Big_int.mult_big_int c b))
                    (Big_int.mult_big_int b d)));;

let rec rat_components
  delta =
    ((fun _ -> true),
      (Fract ((Big_int.big_int_of_int 0), (Big_int.big_int_of_int 1)),
        (Fract ((Big_int.big_int_of_int 1), (Big_int.big_int_of_int 1)),
          (positive_delta delta,
            (plus_rat,
              (times_rat,
                (max0_rat,
                  (ge_rat,
                    (rat_gt delta,
                      (true, (rat_name, (fun _ -> true))))))))))));;

let rec eq_rat
  (Fract (a, b)) (Fract (c, d)) =
    (if Big_int.eq_big_int b (Big_int.big_int_of_int 0)
      then Big_int.eq_big_int c (Big_int.big_int_of_int 0) ||
             Big_int.eq_big_int d (Big_int.big_int_of_int 0)
      else (if Big_int.eq_big_int d (Big_int.big_int_of_int 0)
             then Big_int.eq_big_int a (Big_int.big_int_of_int 0) ||
                    Big_int.eq_big_int b (Big_int.big_int_of_int 0)
             else Big_int.eq_big_int (Big_int.mult_big_int a d)
                    (Big_int.mult_big_int b c)));;

let eq_rata = ({eq = eq_rat} : rat eq);;

let rec get_comp_apos
  (wfe, (ze, (on, (def, (pl, (ti, (mz, (ge, (gt, (mo, (de, ap))))))))))) = ap;;

let rec mat_arc_posI ap m = ap (nth (nth m Zero_nat) Zero_nat);;

let rec get_comp_description
  (wfe, (ze, (on, (def, (pl, (ti, (mz, (ge, (gt, (mo, (de, ap))))))))))) = de;;

let rec divmod_nat
  m n = (if eq_nat n Zero_nat || less_nat m n then (Zero_nat, m)
          else let (q, a) = divmod_nat (minus_nat m n) n in
               (Suc q, a));;

let rec mod_nat m n = snd (divmod_nat m n);;

let rec div_nat m n = fst (divmod_nat m n);;

let rec digit2string
  n = (if eq_nat n Zero_nat then ['0']
        else (if eq_nat n one_nat then ['1']
               else (if eq_nat n (nat (Big_int.big_int_of_int 2)) then ['2']
                      else (if eq_nat n (nat (Big_int.big_int_of_int 3))
                             then ['3']
                             else (if eq_nat n (nat (Big_int.big_int_of_int 4))
                                    then ['4']
                                    else (if eq_nat n
       (nat (Big_int.big_int_of_int 5))
   then ['5']
   else (if eq_nat n (nat (Big_int.big_int_of_int 6)) then ['6']
          else (if eq_nat n (nat (Big_int.big_int_of_int 7)) then ['7']
                 else (if eq_nat n (nat (Big_int.big_int_of_int 8)) then ['8']
                        else (if eq_nat n (nat (Big_int.big_int_of_int 9))
                               then ['9']
                               else ['n'; 'o'; 't'; ' '; 'a'; ' '; 'd'; 'i';
                                      'g'; 'i'; 't']))))))))));;

let rec shows_nat
  n = (if less_nat n (nat (Big_int.big_int_of_int 10))
        then shows_string (digit2string n)
        else comp (shows_nat (div_nat n (nat (Big_int.big_int_of_int 10))))
               (shows_string
                 (digit2string
                   (mod_nat n (nat (Big_int.big_int_of_int 10))))));;

let rec shows_prec_nat d n = shows_nat n;;

let rec get_comp_gt
  (wfe, (ze, (on, (def, (pl, (ti, (mz, (ge, (gt, (mo, (de, ap))))))))))) = gt;;

let rec get_comp_ge
  (wfe, (ze, (on, (def, (pl, (ti, (mz, (ge, (gt, (mo, (de, ap))))))))))) = ge;;

let rec vec_geI ge v w = forall (fun (a, b) -> ge a b) (zip v w);;

let rec mat_geI ge m1 m2 = forall (fun (a, b) -> vec_geI ge a b) (zip m1 m2);;

let rec get_comp_times
  (wfe, (ze, (on, (def, (pl, (ti, (mz, (ge, (gt, (mo, (de, ap))))))))))) = ti;;

let rec transpose
  nr x1 = match nr, x1 with nr, [] -> replicate nr []
    | nr, v :: m -> map (fun (a, b) -> a :: b) (zip v (transpose nr m));;

let rec scalar_prodI
  ze pl ti v w = foldr (fun (x, y) -> pl (ti x y)) (zip v w) ze;;

let rec matT_vec_multI
  ze pl ti m v = map (fun w -> scalar_prodI ze pl ti w v) m;;

let rec mat_multI
  ze pl ti nr m1 m2 = map (matT_vec_multI ze pl ti (transpose nr m1)) m2;;

let rec get_comp_plus
  (wfe, (ze, (on, (def, (pl, (ti, (mz, (ge, (gt, (mo, (de, ap))))))))))) = pl;;

let rec vec_plusI pl v w = map (fun xy -> pl (fst xy) (snd xy)) (zip v w);;

let rec mat_plusI
  pl m1 m2 = map (fun uv -> vec_plusI pl (fst uv) (snd uv)) (zip m1 m2);;

let rec get_comp_def
  (wfe, (ze, (on, (def, (pl, (ti, (mz, (ge, (gt, (mo, (de, ap))))))))))) = def;;

let rec get_comp_1
  (wfe, (ze, (on, (def, (pl, (ti, (mz, (ge, (gt, (mo, (de, ap))))))))))) = on;;

let rec vec1I
  ze on n i =
    append (replicate i ze)
      (on :: replicate (minus_nat (minus_nat n one_nat) i) ze);;

let rec mat1I ze on n = map (vec1I ze on n) (upt Zero_nat n);;

let rec get_comp_0
  (wfe, (ze, (on, (def, (pl, (ti, (mz, (ge, (gt, (mo, (de, ap))))))))))) = ze;;

let rec vec0I ze n = replicate n ze;;

let rec mat0I ze nr nc = replicate nc (vec0I ze nr);;

let rec wf_vec n x = eq_nat (size_list x) n;;

let rec wf_mat nr nc m = eq_nat (size_list m) nc && forall (wf_vec nr) m;;

let rec mat_arctic_components
  n c = let np = Suc (minus_nat n one_nat) in
        (wf_mat np np,
          (mat0I (get_comp_0 c) np np,
            (mat1I (get_comp_0 c) (get_comp_1 c) np,
              (mat1I (get_comp_0 c) (get_comp_def c) np,
                (mat_plusI (get_comp_plus c),
                  (mat_multI (get_comp_0 c) (get_comp_plus c) (get_comp_times c)
                     np,
                    (id, (mat_geI (get_comp_ge c),
                           (mat_geI (get_comp_gt c),
                             (false,
                               (concat
                                  [shows_prec_nat Zero_nat np []; ['x'];
                                    shows_prec_nat Zero_nat np [];
                                    [' '; 'm'; 'a'; 't'; 'r'; 'i'; 'c'; 'e';
                                      's'; ' '; 'o'; 'v'; 'e'; 'r'; ' '];
                                    get_comp_description c],
                                 mat_arc_posI (get_comp_apos c))))))))))));;

let rec get_comp_mono
  (wfe, (ze, (on, (def, (pl, (ti, (mz, (ge, (gt, (mo, (de, ap))))))))))) = mo;;

let rec sub_vec x = take x;;

let rec sub_mat nr nc m = map (sub_vec nr) (take nc m);;

let rec vec_pre_gtI gt v w = any (fun (a, b) -> gt a b) (zip v w);;

let rec mat_pre_gtI
  gt m1 m2 = any (fun (a, b) -> vec_pre_gtI gt a b) (zip m1 m2);;

let rec mat_gtI
  ge gt sd m1 m2 =
    mat_geI ge m1 m2 && mat_pre_gtI gt (sub_mat sd sd m1) (sub_mat sd sd m2);;

let rec get_comp_max0
  (wfe, (ze, (on, (def, (pl, (ti, (mz, (ge, (gt, (mo, (de, ap))))))))))) = mz;;

let rec vec_map x = map x;;

let rec mat_map f = map (vec_map f);;

let rec mat_max0I x = mat_map x;;

let rec mat_components
  n sd c =
    let sdp = Suc (minus_nat sd one_nat) in
    let np = max ord_nat n sdp in
    (wf_mat np np,
      (mat0I (get_comp_0 c) np np,
        (mat1I (get_comp_0 c) (get_comp_1 c) np,
          (mat1I (get_comp_0 c) (get_comp_def c) np,
            (mat_plusI (get_comp_plus c),
              (mat_multI (get_comp_0 c) (get_comp_plus c) (get_comp_times c) np,
                (mat_max0I (get_comp_max0 c),
                  (mat_geI (get_comp_ge c),
                    (mat_gtI (get_comp_ge c) (get_comp_gt c) sdp,
                      (get_comp_mono c,
                        (concat
                           [shows_prec_nat Zero_nat np []; ['x'];
                             shows_prec_nat Zero_nat np [];
                             [' '; 'm'; 'a'; 't'; 'r'; 'i'; 'c'; 'e'; 's'; ' ';
                               'w'; 'i'; 't'; 'h'; ' '; 's'; 't'; 'r'; 'i'; 'c';
                               't'; ' '; 'd'; 'i'; 'm'; 'e'; 'n'; 's'; 'i'; 'o';
                               'n'; ' '];
                             shows_prec_nat Zero_nat sdp [];
                             [' '; 'o'; 'v'; 'e'; 'r'; ' '];
                             get_comp_description c],
                          (fun _ -> true))))))))))));;

let rec vec_coeff
  xml2coeff =
    xml_many_elements ['v'; 'e'; 'c'; 't'; 'o'; 'r']
      (xml1element ['c'; 'o'; 'e'; 'f'; 'f'; 'i'; 'c'; 'i'; 'e'; 'n'; 't']
        xml2coeff id)
      id;;

let rec mat_coeff
  ze xml2coeff =
    xml_options
      [(['m'; 'a'; 't'; 'r'; 'i'; 'x'],
         xml_many_elements ['m'; 'a'; 't'; 'r'; 'i'; 'x'] (vec_coeff xml2coeff)
           id);
        (['v'; 'e'; 'c'; 't'; 'o'; 'r'],
          xml_change (vec_coeff xml2coeff)
            (fun v ->
              v :: replicate (minus_nat (size_list v) one_nat)
                     (replicate (size_list v) ze)))];;

let rec pos_arctic_rat
  = function MinInfty_rat -> false
    | Num_arc_rat n ->
        less_eq_rat
          (Fract ((Big_int.big_int_of_int 0), (Big_int.big_int_of_int 1))) n;;

let arctic_rat_name : char list
  = ['a'; 'r'; 'c'; 't'; 'i'; 'c'; ' '; 'r'; 'a'; 't'; 'i'; 'o'; 'n'; 'a'; 'l';
      's'];;

let rec gt_arctic_rat
  delta uu x2 = match delta, uu, x2 with delta, uu, MinInfty_rat -> true
    | delta, MinInfty_rat, Num_arc_rat uv -> false
    | delta, Num_arc_rat x, Num_arc_rat y -> rat_gt delta x y;;

let rec ge_arctic_rat
  x xa1 = match x, xa1 with x, MinInfty_rat -> true
    | MinInfty_rat, Num_arc_rat uu -> false
    | Num_arc_rat x, Num_arc_rat y -> less_eq_rat y x;;

let rec times_arctic_rat
  x0 y = match x0, y with MinInfty_rat, y -> MinInfty_rat
    | Num_arc_rat v, MinInfty_rat -> MinInfty_rat
    | Num_arc_rat x, Num_arc_rat y -> Num_arc_rat (plus_rat x y);;

let rec plus_arctic_rat
  x0 y = match x0, y with MinInfty_rat, y -> y
    | Num_arc_rat v, MinInfty_rat -> Num_arc_rat v
    | Num_arc_rat x, Num_arc_rat y -> Num_arc_rat (max ord_rat x y);;

let one_arctic_rat : arctic_rat
  = Num_arc_rat
      (Fract ((Big_int.big_int_of_int 0), (Big_int.big_int_of_int 1)));;

let zero_arctic_rat : arctic_rat = MinInfty_rat;;

let rec arctic_rat_components
  delta =
    ((fun _ -> true),
      (zero_arctic_rat,
        (one_arctic_rat,
          (one_arctic_rat,
            (plus_arctic_rat,
              (times_arctic_rat,
                (id, (ge_arctic_rat,
                       (gt_arctic_rat delta,
                         (false, (arctic_rat_name, pos_arctic_rat)))))))))));;

let rec eq_arctic_rat
  x0 x1 = match x0, x1 with MinInfty_rat, MinInfty_rat -> true
    | Num_arc_rat rat, Num_arc_rat rat' -> eq_rat rat rat'
    | MinInfty_rat, Num_arc_rat rat' -> false
    | Num_arc_rat rat', MinInfty_rat -> false;;

let eq_arctic_rata = ({eq = eq_arctic_rat} : arctic_rat eq);;

let rec arctic_rat_coeff
  x = xml_options
        [(['i'; 'n'; 't'; 'e'; 'g'; 'e'; 'r'],
           xml_change rat (fun a -> Num_arc_rat a));
          (['r'; 'a'; 't'; 'i'; 'o'; 'n'; 'a'; 'l'],
            xml_change rat (fun a -> Num_arc_rat a));
          (['m'; 'i'; 'n'; 'u'; 's'; 'I'; 'n'; 'f'; 'i'; 'n'; 'i'; 't'; 'y'],
            xml0elements
              ['m'; 'i'; 'n'; 'u'; 's'; 'I'; 'n'; 'f'; 'i'; 'n'; 'i'; 't'; 'y']
              MinInfty_rat)]
        x;;

let rec pos_arctic
  = function MinInfty -> false
    | Num_arc n -> Big_int.le_big_int (Big_int.big_int_of_int 0) n;;

let arctic_name : char list
  = ['a'; 'r'; 'c'; 't'; 'i'; 'c'; ' '; 's'; 'e'; 'm'; 'i'; 'r'; 'i'; 'n';
      'g'];;

let rec gt_arctic
  uu x1 = match uu, x1 with uu, MinInfty -> true
    | MinInfty, Num_arc uv -> false
    | Num_arc x, Num_arc y -> Big_int.lt_big_int y x;;

let rec ge_arctic
  x xa1 = match x, xa1 with x, MinInfty -> true
    | MinInfty, Num_arc uu -> false
    | Num_arc x, Num_arc y -> Big_int.le_big_int y x;;

let rec times_arctic
  x0 y = match x0, y with MinInfty, y -> MinInfty
    | Num_arc v, MinInfty -> MinInfty
    | Num_arc x, Num_arc y -> Num_arc (Big_int.add_big_int x y);;

let rec plus_arctic
  x0 y = match x0, y with MinInfty, y -> y
    | Num_arc v, MinInfty -> Num_arc v
    | Num_arc x, Num_arc y -> Num_arc (max ord_int x y);;

let one_arctic : arctic = Num_arc (Big_int.big_int_of_int 0);;

let zero_arctic : arctic = MinInfty;;

let arctic_components :
  (arctic -> bool) *
    (arctic *
      (arctic *
        (arctic *
          ((arctic -> arctic -> arctic) *
            ((arctic -> arctic -> arctic) *
              ((arctic -> arctic) *
                ((arctic -> arctic -> bool) *
                  ((arctic -> arctic -> bool) *
                    (bool * (char list * (arctic -> bool)))))))))))
  = ((fun _ -> true),
      (zero_arctic,
        (one_arctic,
          (one_arctic,
            (plus_arctic,
              (times_arctic,
                (id, (ge_arctic,
                       (gt_arctic, (false, (arctic_name, pos_arctic)))))))))));;

let rec eq_arctic
  x0 x1 = match x0, x1 with MinInfty, MinInfty -> true
    | Num_arc int, Num_arc int' -> Big_int.eq_big_int int int'
    | MinInfty, Num_arc int' -> false
    | Num_arc int', MinInfty -> false;;

let eq_arctica = ({eq = eq_arctic} : arctic eq);;

let rec arctic_coeff
  x = xml_options
        [(['i'; 'n'; 't'; 'e'; 'g'; 'e'; 'r'],
           xml_change (int ['i'; 'n'; 't'; 'e'; 'g'; 'e'; 'r'])
             (fun a -> Num_arc a));
          (['m'; 'i'; 'n'; 'u'; 's'; 'I'; 'n'; 'f'; 'i'; 'n'; 'i'; 't'; 'y'],
            xml0elements
              ['m'; 'i'; 'n'; 'u'; 's'; 'I'; 'n'; 'f'; 'i'; 'n'; 'i'; 't'; 'y']
              MinInfty)]
        x;;

let rec fit_length
  uu x1 uv = match uu, x1, uv with uu, Zero_nat, uv -> []
    | a, Suc v, [] -> replicate (Suc v) a
    | a, Suc n, b :: asa -> b :: fit_length a n asa;;

let int_name : char list
  = ['n'; 'a'; 't'; 'u'; 'r'; 'a'; 'l'; 's'; ' '; 'w'; 'i'; 't'; 'h'; ' '; 'n';
      'e'; 'g'; 'a'; 't'; 'i'; 'v'; 'e'; ' '; 'c'; 'o'; 'n'; 's'; 't'; 'a'; 'n';
      't'; 's'];;

let rec ge_int x y = Big_int.le_big_int y x;;

let rec max0_int x = max ord_int (Big_int.big_int_of_int 0) x;;

let int_components :
  (Big_int.big_int -> bool) *
    (Big_int.big_int *
      (Big_int.big_int *
        (Big_int.big_int *
          ((Big_int.big_int -> Big_int.big_int -> Big_int.big_int) *
            ((Big_int.big_int -> Big_int.big_int -> Big_int.big_int) *
              ((Big_int.big_int -> Big_int.big_int) *
                ((Big_int.big_int -> Big_int.big_int -> bool) *
                  ((Big_int.big_int -> Big_int.big_int -> bool) *
                    (bool * (char list * (Big_int.big_int -> bool)))))))))))
  = ((fun _ -> true),
      ((Big_int.big_int_of_int 0),
        ((Big_int.big_int_of_int 1),
          ((Big_int.big_int_of_int 1),
            (Big_int.add_big_int,
              (Big_int.mult_big_int,
                (max0_int,
                  (ge_int,
                    ((fun x y -> Big_int.lt_big_int y x),
                      (true, (int_name, (fun _ -> true))))))))))));;

let rec sum_lpoly
  uu x1 ys = match uu, x1, ys with uu, [], ys -> ys
    | uv, v :: va, [] -> v :: va
    | pl, x :: xs, y :: ys -> pl x y :: sum_lpoly pl xs ys;;

let rec poly_to_lpoly _A
  uu uv uw ux x4 = match uu, uv, uw, ux, x4 with
    uu, uv, uw, ux, PNum i -> Inr (i, [])
    | ze, on, pl, ti, PVar x -> Inr (ze, append (replicate x ze) [on])
    | ze, on, pl, ti, PSum [p] -> poly_to_lpoly _A ze on pl ti p
    | ze, on, pl, ti, PSum (p :: q :: ps) ->
        bind (poly_to_lpoly _A ze on pl ti p)
          (fun (cp, ncp) ->
            bind (poly_to_lpoly _A ze on pl ti (PSum (q :: ps)))
              (fun (cq, ncq) -> Inr (pl cp cq, sum_lpoly pl ncp ncq)))
    | ze, on, pl, ti, PMult [p] -> poly_to_lpoly _A ze on pl ti p
    | ze, on, pl, ti, PMult (p :: q :: ps) ->
        bind (poly_to_lpoly _A ze on pl ti p)
          (fun (cp, ncp) ->
            bind (poly_to_lpoly _A ze on pl ti (PMult (q :: ps)))
              (fun (cq, ncq) ->
                (if forall (eqa _A ze) ncp then Inr (ti cp cq, map (ti cp) ncq)
                  else (if forall (eqa _A ze) ncq
                         then Inr (ti cp cq, map (fun x -> ti x cq) ncp)
                         else Inl ['n'; 'o'; 'n'; '-'; 'l'; 'i'; 'n'; 'e'; 'a';
                                    'r'; ' '; 'p'; 'o'; 'l'; 'y'; 'n'; 'o'; 'm';
                                    'i'; 'a'; 'l'; 's'; ' '; 'n'; 'o'; 't'; ' ';
                                    's'; 'u'; 'p'; 'p'; 'o'; 'r'; 't'; 'e';
                                    'd']))));;

let rec int_coeff x = int ['i'; 'n'; 't'; 'e'; 'g'; 'e'; 'r'] x;;

let rec polynomial
  xml2coeff x =
    xml1element ['p'; 'o'; 'l'; 'y'; 'n'; 'o'; 'm'; 'i'; 'a'; 'l']
      (xml_options
        [(['c'; 'o'; 'e'; 'f'; 'f'; 'i'; 'c'; 'i'; 'e'; 'n'; 't'],
           xml1element ['c'; 'o'; 'e'; 'f'; 'f'; 'i'; 'c'; 'i'; 'e'; 'n'; 't']
             xml2coeff (fun a -> PNum a));
          (['v'; 'a'; 'r'; 'i'; 'a'; 'b'; 'l'; 'e'],
            xml_change (nata ['v'; 'a'; 'r'; 'i'; 'a'; 'b'; 'l'; 'e'])
              (fun n -> PVar (minus_nat n one_nat)));
          (['s'; 'u'; 'm'],
            xml_many_elements ['s'; 'u'; 'm'] (polynomial xml2coeff)
              (fun a -> PSum a));
          (['p'; 'r'; 'o'; 'd'; 'u'; 'c'; 't'],
            xml_many_elements ['p'; 'r'; 'o'; 'd'; 'u'; 'c'; 't']
              (polynomial xml2coeff) (fun a -> PMult a))])
      id x;;

let rec domain_case
  f1 f2 f3 f4 f5 f6 f7 f8 x8 = match f1, f2, f3, f4, f5, f6, f7, f8, x8 with
    f1, f2, f3, f4, f5, f6, f7, f8, Rat_mat (rat, nat1, nat2) ->
      f8 rat nat1 nat2
    | f1, f2, f3, f4, f5, f6, f7, f8, Rational rat -> f7 rat
    | f1, f2, f3, f4, f5, f6, f7, f8, Arc_rat_mat (rat, nat) -> f6 rat nat
    | f1, f2, f3, f4, f5, f6, f7, f8, Arc_mat nat -> f5 nat
    | f1, f2, f3, f4, f5, f6, f7, f8, Int_mat (nat1, nat2) -> f4 nat1 nat2
    | f1, f2, f3, f4, f5, f6, f7, f8, Arctic_rat rat -> f3 rat
    | f1, f2, f3, f4, f5, f6, f7, f8, Arctic -> f2
    | f1, f2, f3, f4, f5, f6, f7, f8, Natural -> f1;;

let rec rat_domain
  x = xml1element ['r'; 'a'; 't'; 'i'; 'o'; 'n'; 'a'; 'l'; 's']
        (xml1element ['d'; 'e'; 'l'; 't'; 'a'] rat id) id x;;

let rec basic_domain
  x = xml_options
        [(['n'; 'a'; 't'; 'u'; 'r'; 'a'; 'l'; 's'],
           xml0elements ['n'; 'a'; 't'; 'u'; 'r'; 'a'; 'l'; 's'] Natural);
          (['a'; 'r'; 'c'; 't'; 'i'; 'c'],
            xml1element ['a'; 'r'; 'c'; 't'; 'i'; 'c']
              (xml1element ['d'; 'o'; 'm'; 'a'; 'i'; 'n']
                (xml_options
                  [(['n'; 'a'; 't'; 'u'; 'r'; 'a'; 'l'; 's'],
                     xml0elements ['n'; 'a'; 't'; 'u'; 'r'; 'a'; 'l'; 's']
                       Arctic);
                    (['i'; 'n'; 't'; 'e'; 'g'; 'e'; 'r'; 's'],
                      xml0elements ['i'; 'n'; 't'; 'e'; 'g'; 'e'; 'r'; 's']
                        Arctic);
                    (['r'; 'a'; 't'; 'i'; 'o'; 'n'; 'a'; 'l'; 's'],
                      xml_change rat_domain (fun a -> Arctic_rat a))])
                id)
              id);
          (['r'; 'a'; 't'; 'i'; 'o'; 'n'; 'a'; 'l'; 's'],
            xml_change rat_domain (fun a -> Rational a))]
        x;;

let rec interpretation_type
  x = xml1element ['t'; 'y'; 'p'; 'e']
        (xml_options
          [(['p'; 'o'; 'l'; 'y'; 'n'; 'o'; 'm'; 'i'; 'a'; 'l'],
             xml2elements ['p'; 'o'; 'l'; 'y'; 'n'; 'o'; 'm'; 'i'; 'a'; 'l']
               (xml1element ['d'; 'o'; 'm'; 'a'; 'i'; 'n']
                 (xml_options
                   [(['n'; 'a'; 't'; 'u'; 'r'; 'a'; 'l'; 's'], basic_domain);
                     (['a'; 'r'; 'c'; 't'; 'i'; 'c'], basic_domain);
                     (['r'; 'a'; 't'; 'i'; 'o'; 'n'; 'a'; 'l'; 's'],
                       basic_domain);
                     (['m'; 'a'; 't'; 'r'; 'i'; 'c'; 'e'; 's'],
                       xml3elements ['m'; 'a'; 't'; 'r'; 'i'; 'c'; 'e'; 's']
                         (nata ['d'; 'i'; 'm'; 'e'; 'n'; 's'; 'i'; 'o'; 'n'])
                         (nata ['s'; 't'; 'r'; 'i'; 'c'; 't'; 'D'; 'i'; 'm';
                                 'e'; 'n'; 's'; 'i'; 'o'; 'n'])
                         (xml1element ['d'; 'o'; 'm'; 'a'; 'i'; 'n']
                           basic_domain id)
                         (fun di sd a ->
                           (match a with Natural -> Int_mat (di, sd)
                             | Arctic -> Arc_mat di
                             | Arctic_rat delta -> Arc_rat_mat (delta, di)
                             | Rational delta -> Rat_mat (delta, di, sd))))])
                 id)
               (nata ['d'; 'e'; 'g'; 'r'; 'e'; 'e']) (fun typea _ -> typea));
            (['m'; 'a'; 't'; 'r'; 'i'; 'x'; 'I'; 'n'; 't'; 'e'; 'r'; 'p'; 'r';
               'e'; 't'; 'a'; 't'; 'i'; 'o'; 'n'],
              xml3elements
                ['m'; 'a'; 't'; 'r'; 'i'; 'x'; 'I'; 'n'; 't'; 'e'; 'r'; 'p';
                  'r'; 'e'; 't'; 'a'; 't'; 'i'; 'o'; 'n']
                (xml1element ['d'; 'o'; 'm'; 'a'; 'i'; 'n'] basic_domain id)
                (nata ['d'; 'i'; 'm'; 'e'; 'n'; 's'; 'i'; 'o'; 'n'])
                (nata ['s'; 't'; 'r'; 'i'; 'c'; 't'; 'D'; 'i'; 'm'; 'e'; 'n';
                        's'; 'i'; 'o'; 'n'])
                (fun domain di sd ->
                  (match domain with Natural -> Int_mat (di, sd)
                    | Arctic -> Arc_mat di
                    | Arctic_rat delta -> Arc_rat_mat (delta, di)
                    | Rational delta -> Rat_mat (delta, di, sd))))])
        id x;;

let rec interpretation
  xml2name =
    xml1many_elements_gen
      ['i'; 'n'; 't'; 'e'; 'r'; 'p'; 'r'; 'e'; 't'; 'a'; 't'; 'i'; 'o'; 'n']
      interpretation_type
      (fun a ->
        (match a
          with Natural ->
            xml3elements ['i'; 'n'; 't'; 'e'; 'r'; 'p'; 'r'; 'e'; 't'] xml2name
              (nata ['a'; 'r'; 'i'; 't'; 'y'])
              (fun x ->
                bind (polynomial int_coeff x)
                  (poly_to_lpoly eq_int (get_comp_0 int_components)
                    (get_comp_1 int_components) (get_comp_plus int_components)
                    (get_comp_times int_components)))
              (fun f aa poly ->
                Alter1
                  ((f, aa),
                    (fst poly,
                      fit_length (get_comp_0 int_components) aa (snd poly))))
          | Arctic ->
            xml3elements ['i'; 'n'; 't'; 'e'; 'r'; 'p'; 'r'; 'e'; 't'] xml2name
              (nata ['a'; 'r'; 'i'; 't'; 'y'])
              (fun x ->
                bind (polynomial arctic_coeff x)
                  (poly_to_lpoly eq_arctica (get_comp_0 arctic_components)
                    (get_comp_1 arctic_components)
                    (get_comp_plus arctic_components)
                    (get_comp_times arctic_components)))
              (fun f aa poly ->
                Alter3
                  ((f, aa),
                    (fst poly,
                      fit_length (get_comp_0 arctic_components) aa (snd poly))))
          | Arctic_rat d ->
            xml3elements ['i'; 'n'; 't'; 'e'; 'r'; 'p'; 'r'; 'e'; 't'] xml2name
              (nata ['a'; 'r'; 'i'; 't'; 'y'])
              (fun x ->
                bind (polynomial arctic_rat_coeff x)
                  (poly_to_lpoly eq_arctic_rata
                    (get_comp_0 (arctic_rat_components d))
                    (get_comp_1 (arctic_rat_components d))
                    (get_comp_plus (arctic_rat_components d))
                    (get_comp_times (arctic_rat_components d))))
              (fun f aa poly ->
                Alter7
                  ((f, aa),
                    (fst poly,
                      fit_length (get_comp_0 (arctic_rat_components d)) aa
                        (snd poly))))
          | Int_mat (n, sd) ->
            xml3elements ['i'; 'n'; 't'; 'e'; 'r'; 'p'; 'r'; 'e'; 't'] xml2name
              (nata ['a'; 'r'; 'i'; 't'; 'y'])
              (fun x ->
                bind (polynomial
                       (mat_coeff (Big_int.big_int_of_int 0) int_coeff) x)
                  (poly_to_lpoly (eq_lista (eq_lista eq_int))
                    (get_comp_0 (mat_components n sd int_components))
                    (get_comp_1 (mat_components n sd int_components))
                    (get_comp_plus (mat_components n sd int_components))
                    (get_comp_times (mat_components n sd int_components))))
              (fun f aa poly ->
                Alter4
                  ((f, aa),
                    (fst poly,
                      fit_length
                        (get_comp_0 (mat_components n sd int_components)) aa
                        (snd poly))))
          | Arc_mat n ->
            xml3elements ['i'; 'n'; 't'; 'e'; 'r'; 'p'; 'r'; 'e'; 't'] xml2name
              (nata ['a'; 'r'; 'i'; 't'; 'y'])
              (fun x ->
                bind (polynomial (mat_coeff MinInfty arctic_coeff) x)
                  (poly_to_lpoly (eq_lista (eq_lista eq_arctica))
                    (get_comp_0 (mat_arctic_components n arctic_components))
                    (get_comp_1 (mat_arctic_components n arctic_components))
                    (get_comp_plus (mat_arctic_components n arctic_components))
                    (get_comp_times
                      (mat_arctic_components n arctic_components))))
              (fun f aa poly ->
                Alter6
                  ((f, aa),
                    (fst poly,
                      fit_length
                        (get_comp_0 (mat_arctic_components n arctic_components))
                        aa (snd poly))))
          | Arc_rat_mat (d, n) ->
            xml3elements ['i'; 'n'; 't'; 'e'; 'r'; 'p'; 'r'; 'e'; 't'] xml2name
              (nata ['a'; 'r'; 'i'; 't'; 'y'])
              (fun x ->
                bind (polynomial (mat_coeff MinInfty_rat arctic_rat_coeff) x)
                  (poly_to_lpoly (eq_lista (eq_lista eq_arctic_rata))
                    (get_comp_0
                      (mat_arctic_components n (arctic_rat_components d)))
                    (get_comp_1
                      (mat_arctic_components n (arctic_rat_components d)))
                    (get_comp_plus
                      (mat_arctic_components n (arctic_rat_components d)))
                    (get_comp_times
                      (mat_arctic_components n (arctic_rat_components d)))))
              (fun f aa poly ->
                Alter8
                  ((f, aa),
                    (fst poly,
                      fit_length
                        (get_comp_0
                          (mat_arctic_components n (arctic_rat_components d)))
                        aa (snd poly))))
          | Rational d ->
            xml3elements ['i'; 'n'; 't'; 'e'; 'r'; 'p'; 'r'; 'e'; 't'] xml2name
              (nata ['a'; 'r'; 'i'; 't'; 'y'])
              (fun x ->
                bind (polynomial rat x)
                  (poly_to_lpoly eq_rata (get_comp_0 (rat_components d))
                    (get_comp_1 (rat_components d))
                    (get_comp_plus (rat_components d))
                    (get_comp_times (rat_components d))))
              (fun f aa poly ->
                Alter2
                  ((f, aa),
                    (fst poly,
                      fit_length (get_comp_0 (rat_components d)) aa
                        (snd poly))))
          | Rat_mat (d, n, sd) ->
            xml3elements ['i'; 'n'; 't'; 'e'; 'r'; 'p'; 'r'; 'e'; 't'] xml2name
              (nata ['a'; 'r'; 'i'; 't'; 'y'])
              (fun x ->
                bind (polynomial
                       (mat_coeff
                         (Fract
                           ((Big_int.big_int_of_int 0),
                             (Big_int.big_int_of_int 1)))
                         rat)
                       x)
                  (poly_to_lpoly (eq_lista (eq_lista eq_rata))
                    (get_comp_0 (mat_components n sd (rat_components d)))
                    (get_comp_1 (mat_components n sd (rat_components d)))
                    (get_comp_plus (mat_components n sd (rat_components d)))
                    (get_comp_times (mat_components n sd (rat_components d)))))
              (fun f aa poly ->
                Alter5
                  ((f, aa),
                    (fst poly,
                      fit_length
                        (get_comp_0 (mat_components n sd (rat_components d))) aa
                        (snd poly))))))
      (fun typea pi ->
        (match typea with Natural -> Int_carrier (map alter1a pi)
          | Arctic -> Arctic_carrier (map alter3a pi)
          | Arctic_rat d -> Arctic_rat_carrier (d, map alter7a pi)
          | Int_mat (n, sd) -> Int_mat_carrier (n, sd, map alter4a pi)
          | Arc_mat n -> Arc_mat_carrier (n, map alter6a pi)
          | Arc_rat_mat (d, n) -> Arc_rat_mat_carrier (d, n, map alter8a pi)
          | Rational d -> Rat_carrier (d, map alter2a pi)
          | Rat_mat (d, n, sd) -> Rat_mat_carrier (d, n, sd, map alter5a pi)));;

let rec redpair
  xml2name =
    xml1element ['r'; 'e'; 'd'; 'P'; 'a'; 'i'; 'r']
      (xml_options
        [(['i'; 'n'; 't'; 'e'; 'r'; 'p'; 'r'; 'e'; 't'; 'a'; 't'; 'i'; 'o';
            'n'],
           interpretation xml2name);
          (['p'; 'a'; 't'; 'h'; 'O'; 'r'; 'd'; 'e'; 'r'], path_order xml2name)])
      id;;

let rec shows_prec_int
  d i = (if Big_int.lt_big_int i (Big_int.big_int_of_int 0)
          then comp (shows_prec_char Zero_nat '-')
                 (shows_prec_nat Zero_nat
                   (nat (Big_int.sub_big_int (Big_int.big_int_of_int 0) i)))
          else shows_prec_nat Zero_nat (nat i));;

let rec shows_list_int is = shows_list_aux (shows_prec_int Zero_nat) is;;

let show_int =
  ({shows_prec = shows_prec_int; shows_list = shows_list_int} :
    Big_int.big_int show);;

let rec shows_list_nat ns = shows_list_aux (shows_prec_nat Zero_nat) ns;;

let show_nat =
  ({shows_prec = shows_prec_nat; shows_list = shows_list_nat} : nat show);;

let rec normalize
  (Fract (a, b)) =
    (if Big_int.lt_big_int (Big_int.big_int_of_int 0) b
      then let g = gcd_int a b in
           (div_int a g, div_int b g)
      else (if Big_int.eq_big_int b (Big_int.big_int_of_int 0)
             then ((Big_int.big_int_of_int 0), (Big_int.big_int_of_int 1))
             else let g = Big_int.minus_big_int (gcd_int a b) in
                  (div_int a g, div_int b g)));;

let rec show_rat
  x = let (den, num) = normalize x in
      (if Big_int.eq_big_int num (Big_int.big_int_of_int 1)
        then shows_prec_int Zero_nat den []
        else comp (comp (shows_prec_int Zero_nat den)
                    (shows_prec_char Zero_nat '/'))
               (shows_prec_int Zero_nat num) []);;

let rec shows_prec_rat d x = shows_string (show_rat x);;

let rec shows_list_rat rs = shows_list_aux (shows_prec_rat Zero_nat) rs;;

let show_rata =
  ({shows_prec = shows_prec_rat; shows_list = shows_list_rat} : rat show);;

let rec eq_ctxt _A _B
  x0 x1 = match x0, x1 with Hole, Hole -> true
    | More (f, list1, ctxt, list2), More (f', list1', ctxt', list2') ->
        eqa _A f f' &&
          (eq_list (eq_terma _A _B) list1 list1' &&
            (eq_ctxt _A _B ctxt ctxt' && eq_list (eq_terma _A _B) list2 list2'))
    | Hole, More (f', list1', ctxt', list2') -> false
    | More (f', list1', ctxt', list2'), Hole -> false;;

let rec get_sub (Sub sigma) = sigma;;

let rec subt_at
  s x1 = match s, x1 with s, Empty -> s
    | Fun (f, ss), PCons (i, p) -> subt_at (nth ss i) p;;

let rec supteqp _A _B s t = eq_term _A _B s t || suptp _A _B s t;;

let rec the_var (Var x) = x;;

let rec map_term
  fg x1 = match fg, x1 with fg, Var x -> Var x
    | fg, Fun (f, ts) -> Fun (fg f, map (map_term fg) ts);;

let rec map_rule fg lr = (map_term fg (fst lr), map_term fg (snd lr));;

let rec map_trsI fg r = map (map_rule fg) r;;

type arithFun = Arg of nat | Const of nat | Sum of arithFun list |
  Max of arithFun list | Min of arithFun list | Prod of arithFun list;;

type ('a, 'b) lab = Lab of ('a, 'b) lab * 'b |
  FunLab of ('a, 'b) lab * ('a, 'b) lab list | UnLab of 'a;;

type ('a, 'b) sl_inter =
  SL_Inter of nat * ((('a, 'b) lab * nat) * arithFun) list;;

let rec arith_fun
  xml = xml1element
          ['a'; 'r'; 'i'; 't'; 'h'; 'F'; 'u'; 'n'; 'c'; 't'; 'i'; 'o'; 'n']
          (xml_options
            [(['n'; 'a'; 't'; 'u'; 'r'; 'a'; 'l'],
               xml_change (nata ['n'; 'a'; 't'; 'u'; 'r'; 'a'; 'l'])
                 (fun a -> Const a));
              (['v'; 'a'; 'r'; 'i'; 'a'; 'b'; 'l'; 'e'],
                xml_change (nata ['v'; 'a'; 'r'; 'i'; 'a'; 'b'; 'l'; 'e'])
                  (fun n -> Arg (minus_nat n one_nat)));
              (['s'; 'u'; 'm'],
                xml_many_elements ['s'; 'u'; 'm'] arith_fun (fun a -> Sum a));
              (['p'; 'r'; 'o'; 'd'; 'u'; 'c'; 't'],
                xml_many_elements ['p'; 'r'; 'o'; 'd'; 'u'; 'c'; 't'] arith_fun
                  (fun a -> Prod a));
              (['m'; 'i'; 'n'],
                xml_many_elements ['m'; 'i'; 'n'] arith_fun (fun a -> Min a));
              (['m'; 'a'; 'x'],
                xml_many_elements ['m'; 'a'; 'x'] arith_fun (fun a -> Max a))])
          id xml;;

let rec sl_inter
  xml2name =
    xml1element ['m'; 'o'; 'd'; 'e'; 'l']
      (xml_options
        [(['f'; 'i'; 'n'; 'i'; 't'; 'e'; 'M'; 'o'; 'd'; 'e'; 'l'],
           xml1many_elements
             ['f'; 'i'; 'n'; 'i'; 't'; 'e'; 'M'; 'o'; 'd'; 'e'; 'l']
             (xml_change
               (nata ['c'; 'a'; 'r'; 'r'; 'i'; 'e'; 'r'; 'S'; 'i'; 'z'; 'e'])
               (fun n -> minus_nat n one_nat))
             (xml3elements ['i'; 'n'; 't'; 'e'; 'r'; 'p'; 'r'; 'e'; 't']
               xml2name (nata ['a'; 'r'; 'i'; 't'; 'y']) arith_fun
               (fun f a -> (fun b -> ((f, a), b))))
             (fun n xs -> Some (SL_Inter (n, xs))));
          (['r'; 'o'; 'o'; 't'; 'L'; 'a'; 'b'; 'e'; 'l'; 'i'; 'n'; 'g'],
            xml0elements
              ['r'; 'o'; 'o'; 't'; 'L'; 'a'; 'b'; 'e'; 'l'; 'i'; 'n'; 'g']
              None)])
      id;;

let rec xml_term _A _B
  = function
    Var x -> Xml (['v'; 'a'; 'r'], [], [], Some (shows_prec _B Zero_nat x []))
    | Fun (f, ts) ->
        Xml (['f'; 'u'; 'n'; 'a'; 'p'; 'p'], [],
              Xml (['n'; 'a'; 'm'; 'e'], [], [],
                    Some (shows_prec _A Zero_nat f []))
                :: map (fun t ->
                         Xml (['a'; 'r'; 'g'], [], [xml_term _A _B t], None))
                     ts,
              None);;

let rec distinct _A
  = function [] -> true
    | x :: xs -> not (member _A x xs) && distinct _A xs;;

let rec times_nat
  x0 n = match x0, n with Zero_nat, n -> Zero_nat
    | Suc m, n -> plus_nat n (times_nat m n);;

type 'a preorder = {ord_preorder : 'a ord};;

type 'a bot = {preorder_bot : 'a preorder; bot : 'a};;
let bot _A = _A.bot;;

let rec mina less_eq a b = (if less_eq a b then a else b);;

let rec min _A = mina (less_eq _A);;

let rec insert_scg _A _B
  g gs =
    (if eq_scg _A _B g Null then None
      else (if list_ex (fun g' -> subsumes _A _B g' g) gs then None
             else Some (g :: filter (fun g' -> not (subsumes _A _B g g'))
                               gs)));;

let rec mk_trancl _A _B
  conn x1 s = match conn, x1, s with conn, [], s -> s
    | conn, g :: gs, s ->
        (match insert_scg _A _B g s with None -> mk_trancl _A _B conn gs s
          | Some s' ->
            mk_trancl _A _B conn
              (append (map (scg_comp _B conn g) s')
                (append (map (fun f -> scg_comp _B conn f g) s') gs))
              s');;

let rec check_SCT _A _B
  conn gs =
    let a = mk_trancl _A _B conn gs [] in
    list_all (sagiv _A _B conn) a;;

let rec show_list _A =
  ({shows_prec = shows_prec_list _A; shows_list = shows_list_list _A} :
    ('a list) show);;

let rec shows_list_term _A _B
  is = shows_list_aux (shows_prec_term _A _B Zero_nat) is;;

let rec show_term _A _B =
  ({shows_prec = shows_prec_term _A _B; shows_list = shows_list_term _A _B} :
    ('a, 'b) term show);;

let rec defs_list _A
  = function [] -> []
    | (Fun (f, ss), r) :: ts ->
        let fs = defs_list _A ts in
        (if member _A f fs then fs else f :: fs)
    | (Var x, r) :: ts -> defs_list _A ts;;

let rec shows_trs _A _B
  r = comp (shows_string
             ['r'; 'e'; 'w'; 'r'; 'i'; 't'; 'e'; ' '; 's'; 'y'; 's'; 't'; 'e';
               'm'; ':'])
        (comp shows_nl (shows_rules _A _B r));;

let rec shows_dpp _A _B p r = comp (shows_p _A _B p) (shows_trs _A _B r);;

let rec rev_append
  x0 ys = match x0, ys with [], ys -> ys
    | x :: xs, ys -> rev_append xs (x :: ys);;

let rec flat_map_aux
  f x1 ys = match f, x1, ys with f, [], ys -> rev ys
    | f, x :: xs, ys -> flat_map_aux f xs (rev_append (f x) ys);;

let rec flat_map f xs = flat_map_aux f xs [];;

let rec max_list
  = function [] -> Zero_nat
    | x :: xs -> max ord_nat x (max_list xs);;

let rec parse_text
  x = binda (many (fun c -> not (eq_list eq_char [c] ['<'])))
        (fun ts ->
          let text = trim ts in
          (if null text then (fun tsa -> Inr (None, tsa))
            else (fun tsa -> Inr (Some (rev (trim (rev text))), tsa))))
        x;;

let rec parse_attribute_value
  x = binda (oneof [dquote])
        (fun _ ->
          binda (many (fun c -> not (eq_list eq_char [c] dquote)))
            (fun v -> binda (oneof [dquote]) (fun _ ts -> Inr (v, ts))))
        x;;

let rec parse_name
  x = binda (manyof letters)
        (fun n ->
          binda spaces
            (fun _ ->
              (if null n then (fun _ -> Inl letters)
                else (fun ts -> Inr (n, ts)))))
        x;;

let rec parse_attributes
  = function [] -> Inr ([], [])
    | c :: s ->
        (if member eq_char c ['/'; '>'] then Inr ([], c :: s)
          else binda parse_name
                 (fun k ->
                   binda (oneof [['=']])
                     (fun _ ->
                       binda parse_attribute_value
                         (fun v ->
                           binda parse_attributes
                             (fun atts ts -> Inr ((k, v) :: atts, ts)))))
                 (c :: s));;

let rec update_tokens f ts = Inr (ts, f ts);;

let rec get_tokens x = update_tokens id x;;

let rec parse_nodes
  ts = binda get_tokens
         (fun tsa ->
           (if null tsa ||
                 eq_list eq_char (take (nat (Big_int.big_int_of_int 2)) tsa)
                   ['<'; '/']
             then (fun tsb -> Inr ([], tsb))
             else binda (oneof [['<']])
                    (fun _ ->
                      binda parse_name
                        (fun n ->
                          binda parse_attributes
                            (fun atts ->
                              binda (oneof [['/'; '>']; ['>']])
                                (fun e ->
                                  (if eq_list eq_char e ['/'; '>']
                                    then binda parse_nodes
   (fun cs tsb -> Inr (Xml (n, atts, [], None) :: cs, tsb))
                                    else binda parse_text
   (fun t ->
     binda parse_nodes
       (fun cs ->
         binda (oneof [append ['<'; '/'] (append n ['>'])])
           (fun _ ->
             binda parse_nodes
               (fun ns tsb -> Inr (Xml (n, atts, cs, t) :: ns, tsb))))))))))))
         ts;;

let rec parse_node
  x = binda (oneof [['<']])
        (fun _ ->
          binda parse_name
            (fun n ->
              binda parse_attributes
                (fun atts ->
                  binda (oneof [['/'; '>']; ['>']])
                    (fun e ->
                      (if eq_list eq_char e ['/'; '>']
                        then (fun ts -> Inr (Xml (n, atts, [], None), ts))
                        else binda parse_text
                               (fun t ->
                                 binda parse_nodes
                                   (fun cs ->
                                     binda (oneof
     [append ['<'; '/'] (append n ['>'])])
                                       (fun _ ts ->
 Inr (Xml (n, atts, cs, t), ts)))))))))
        x;;

let rec scan_upto
  enda x1 = match enda, x1 with
    enda, [] ->
      Inl (append
            ['d'; 'i'; 'd'; ' '; 'n'; 'o'; 't'; ' '; 'f'; 'i'; 'n'; 'd'; ' ';
              'e'; 'n'; 'd'; '-'; 'm'; 'a'; 'r'; 'k'; 'e'; 'r'; ' ']
            (quote (shows_prec_list show_char Zero_nat enda) []))
    | enda, t :: ts ->
        (if eq_list eq_char (map snd (zip enda (t :: ts))) enda
          then let l = size_list enda in
               Inr (enda, drop l (t :: ts))
          else bind (scan_upto enda ts)
                 (fun (res, ts') -> Inr (t :: res, ts')));;

let rec parse_header
  ts = binda get_tokens
         (fun tsa ->
           (if eq_list eq_char
                 (take (nat (Big_int.big_int_of_int 2)) (trim tsa)) ['<'; '?']
             then binda (scan_upto ['?'; '>'])
                    (fun h ->
                      binda parse_header (fun hs tsb -> Inr (append h hs, tsb)))
             else binda spaces (fun _ tsb -> Inr ([], tsb))))
         ts;;

let rec remove_comments_aux
  uu x1 = match uu, x1 with
    false, c :: cs ->
      (if ((c : char) = '<') &&
            eq_list eq_char (take (nat (Big_int.big_int_of_int 3)) cs)
              ['!'; '-'; '-']
        then remove_comments_aux true (tl cs)
        else c :: remove_comments_aux false cs)
    | true, c :: cs ->
        (if ((c : char) = '-') &&
              eq_list eq_char (take (nat (Big_int.big_int_of_int 2)) cs)
                ['-'; '>']
          then remove_comments_aux false
                 (drop (nat (Big_int.big_int_of_int 2)) cs)
          else remove_comments_aux true cs)
    | uu, [] -> [];;

let rec remove_comments x = remove_comments_aux false x;;

let rec parse_doc
  x = debug ['0']
        ['p'; 'a'; 'r'; 's'; 'i'; 'n'; 'g'; ' '; 't'; 'x'; 't'; ' '; 't'; 'o';
          ' '; 'x'; 'm'; 'l']
        (binda (update_tokens remove_comments)
          (fun _ ->
            binda parse_header
              (fun h ->
                binda parse_node
                  (fun xml ->
                    binda (eoi show_char)
                      (fun _ ts -> Inr (Xmldoc (h, xml), ts))))))
        x;;

let rec xml_rules _A _B
  rls = Xml (['t'; 'r'; 's'], [],
              [Xml (['r'; 'u'; 'l'; 'e'; 's'], [],
                     map (fun (l, r) ->
                           Xml (['r'; 'u'; 'l'; 'e'], [],
                                 [Xml (['l'; 'h'; 's'], [], [xml_term _A _B l],
None);
                                   Xml (['r'; 'h'; 's'], [], [xml_term _A _B r],
 None)],
                                 None))
                       rls,
                     None)],
              None);;

let rec partition
  p x1 = match p, x1 with p, [] -> ([], [])
    | p, x :: xs ->
        let (yes, no) = partition p xs in
        (if p x then (x :: yes, no) else (yes, x :: no));;

let rec eq_pos
  x0 x1 = match x0, x1 with Empty, Empty -> true
    | PCons (nat, pos), PCons (nat', pos') -> eq_nat nat nat' && eq_pos pos pos'
    | Empty, PCons (nat', pos') -> false
    | PCons (nat', pos'), Empty -> false;;

let eq_posa = ({eq = eq_pos} : pos eq);;

let rec check b e = (if b then Inr () else Inl e);;

let rec lpo_strict _A (_B1, _B2)
  pr n =
    (fun (s, t) ->
      check (fst (lpo _B1 pr n s t))
        (comp (shows_string
                ['c'; 'o'; 'u'; 'l'; 'd'; ' '; 'n'; 'o'; 't'; ' '; 'o'; 'r';
                  'i'; 'e'; 'n'; 't'; ' '])
          (comp (shows_prec_term _A _B2 Zero_nat s)
            (comp (shows_string [' '; '>'; 'L'; 'P'; 'O'; ' '])
              (comp (shows_prec_term _A _B2 Zero_nat t) shows_nl)))));;

let rec ctxt_case
  f1 f2 x2 = match f1, f2, x2 with
    f1, f2, More (f, list1, ctxt, list2) -> f2 f list1 ctxt list2
    | f1, f2, Hole -> f1;;

let rec poss_list
  = function Var x -> [Empty]
    | Fun (f, ss) ->
        Empty ::
          concat_map (fun (i, a) -> map (fun aa -> PCons (i, aa)) a)
            (zip (upt Zero_nat (size_list ss)) (map poss_list ss));;

let sub_empty : ('a, 'b) sub = Sub (fun a -> Var a);;

let rec list_diff _A
  x0 ys = match x0, ys with [], ys -> []
    | x :: xs, ys ->
        let zs = list_diff _A xs ys in
        (if member _A x ys then zs else x :: zs);;

let rec plain_name
  x = xml_options
        [(['n'; 'a'; 'm'; 'e'],
           xml_change (text ['n'; 'a'; 'm'; 'e']) (fun a -> UnLab a));
          (['l'; 'a'; 'b'; 'e'; 'l'; 'e'; 'd'; 'S'; 'y'; 'm'; 'b'; 'o'; 'l'],
            xml2elements
              ['l'; 'a'; 'b'; 'e'; 'l'; 'e'; 'd'; 'S'; 'y'; 'm'; 'b'; 'o'; 'l']
              plain_name (label plain_name)
              (fun f a ->
                (match a with Inl aa -> Lab (f, aa)
                  | Inr aa -> FunLab (f, aa))))]
        x;;

let rec sharp_name
  x = (if eq_list eq_char (get_name x) ['s'; 'h'; 'a'; 'r'; 'p']
        then xml1element ['s'; 'h'; 'a'; 'r'; 'p'] (text ['n'; 'a'; 'm'; 'e'])
               (fun a -> Sharp a) x
        else xml_change (text ['n'; 'a'; 'm'; 'e']) (fun a -> Plain a) x);;

let rec list_inter _A
  x0 bs = match x0, bs with [], bs -> []
    | a :: asa, bs ->
        (if member _A a bs then a :: list_inter _A asa bs
          else list_inter _A asa bs);;

let rec lpo_nstrict _A (_B1, _B2)
  pr n =
    (fun (s, t) ->
      check (snd (lpo _B1 pr n s t))
        (comp (shows_string
                ['c'; 'o'; 'u'; 'l'; 'd'; ' '; 'n'; 'o'; 't'; ' '; 'o'; 'r';
                  'i'; 'e'; 'n'; 't'; ' '])
          (comp (shows_prec_term _A _B2 Zero_nat s)
            (comp (shows_string [' '; '>'; '='; 'L'; 'P'; 'O'; ' '])
              (comp (shows_prec_term _A _B2 Zero_nat t) shows_nl)))));;

let rec ctxt_apply
  x0 s = match x0, s with Hole, s -> s
    | More (f, ss1, c, ss2), s -> Fun (f, append ss1 (ctxt_apply c s :: ss2));;

let rec shows_ctxt _A _B
  = function Hole -> shows_string ['['; ']']
    | More (f, ss1, d, ss2) ->
        comp (shows_prec _A Zero_nat f)
          (comp (shows_string ['('])
            (comp (shows_list_gen (shows_prec_term _A _B Zero_nat) [] []
                    [','; ' '] [','; ' '] ss1)
              (comp (shows_ctxt _A _B d)
                (shows_list_gen (shows_prec_term _A _B Zero_nat) [')']
                  [','; ' '] [','; ' '] [')'] ss2))));;

let rec unsharp_fun = function Plain f -> f
                      | Sharp f -> f;;

let rec list_union _A
  x0 ys = match x0, ys with [], ys -> ys
    | x :: xs, ys ->
        let zs = list_union _A xs ys in
        (if member _A x zs then zs else x :: zs);;

let rec id_labeling uu cs = Inl cs;;

type ('a, 'b, 'c) dPProof = PisEmpty |
  RedPairProc of
    ('a, 'b) lab redpair *
      ((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list *
      ('a, 'b, 'c) dPProof
  | RedPairUrProc of
      ('a, 'b) lab redpair *
        ((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list *
        ((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list *
        ('a, 'b, 'c) dPProof
  | DepGraphProc of
      (('a, 'b, 'c) dPProof option *
        ((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list) list
  | MonoRedPairProc of
      ('a, 'b) lab redpair *
        ((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list *
        ((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list *
        ('a, 'b, 'c) dPProof
  | MonoRedPairUrProc of
      ('a, 'b) lab redpair *
        ((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list *
        ((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list *
        ((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list *
        ('a, 'b, 'c) dPProof
  | UnlabProc of
      ((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list *
        ((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list *
        ('a, 'b, 'c) dPProof
  | SemlabProc of
      ('a, 'b) sl_inter *
        (('a, 'b) lab -> nat list -> ('b, (('a, 'b) lab list)) sum) *
        ((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list *
        ((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list *
        ('a, 'b, 'c) dPProof
  | RootlabProc of
      ((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list *
        ((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list *
        ('a, 'b, 'c) dPProof
  | SubtermCriterionProc of
      ('a, 'b) lab spL *
        (((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) *
          (pos *
            (((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) *
              (('a, 'b) lab, 'c) term)) list) list *
        ((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list *
        ('a, 'b, 'c) dPProof
  | SizeChangeSubTerm of
      (((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) *
        ((nat * nat) list * (nat * nat) list)) list
  | SizeChangeRedPair of
      ('a, 'b) lab redpair *
        ((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list *
        (((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) *
          ((nat * nat) list * (nat * nat) list)) list
  | FccProc of
      ('a, 'b) lab * (('a, 'b) lab, 'c) ctxt list *
        ((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list *
        ((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list *
        ('a, 'b, 'c) dPProof;;

let rec flat_contexts
  xml2name =
    xml_many_elements
      ['f'; 'l'; 'a'; 't'; 'C'; 'o'; 'n'; 't'; 'e'; 'x'; 't'; 's']
      (ctxt xml2name) id;;

let rec create_sp _A
  (SimpProj sp) f = (match lookup _A f sp with None -> Zero_nat | Some n -> n);;

let rec eq_lab _A _B
  x0 x1 = match x0, x1 with
    FunLab (f, ls), FunLab (g, ks) ->
      eq_lab _A _B f g && list_all2 (eq_lab _A _B) ls ks
    | Lab (lab, l), Lab (lab', l') -> eq_lab _A _B lab lab' && eqa _B l l'
    | UnLab f, UnLab f' -> eqa _A f f'
    | Lab (lab, l), FunLab (lab', list') -> false
    | FunLab (lab', list'), Lab (lab, l) -> false
    | Lab (lab, l), UnLab f' -> false
    | UnLab f', Lab (lab, l) -> false
    | FunLab (lab, lista), UnLab f' -> false
    | UnLab f', FunLab (lab, lista) -> false;;

let rec eq_laba _A _B = ({eq = eq_lab _A _B} : ('a, 'b) lab eq);;

let rec projected_rseq _A
  xml2name pi (Xml (name, atts, cs, text)) =
    let tag =
      ['p'; 'r'; 'o'; 'j'; 'e'; 'c'; 't'; 'e'; 'd'; 'R'; 'e'; 'w'; 'r'; 'i';
        't'; 'e'; 'S'; 'e'; 'q'; 'u'; 'e'; 'n'; 'c'; 'e']
      in
    (if eq_list eq_char name tag &&
          (null atts &&
            (is_none text &&
              eq_nat (size_list cs) (nat (Big_int.big_int_of_int 2))))
      then bind (rule xml2name (nth cs Zero_nat))
             (fun r -> rseq _A xml2name pi r (nth cs one_nat))
      else no_transform tag (Xml (name, atts, cs, text)));;

let rec xml1many2elements_gen
  tag p1 p2 p3 p4 f (Xml (name, atts, cs, text)) =
    let ds = rev cs in
    let l = size_list cs in
    (if eq_list eq_char name tag &&
          (null atts &&
            (is_none text && less_eq_nat (nat (Big_int.big_int_of_int 3)) l))
      then bind (p1 (nth cs Zero_nat))
             (fun x ->
               bind (mapa (p2 x)
                      (tl (take (minus_nat l (nat (Big_int.big_int_of_int 2)))
                            cs)))
                 (fun xs ->
                   bind (p3 (nth ds one_nat))
                     (fun y ->
                       bind (p4 (nth ds Zero_nat))
                         (fun z -> Inr (f x xs y z)))))
      else no_transform tag (Xml (name, atts, cs, text)));;

let rec xml5elements
  tag p1 p2 p3 p4 p5 f (Xml (name, atts, cs, text)) =
    (if eq_list eq_char name tag &&
          (null atts &&
            (eq_nat (size_list cs) (nat (Big_int.big_int_of_int 5)) &&
              is_none text))
      then bind (p1 (nth cs Zero_nat))
             (fun x1 ->
               bind (p2 (nth cs one_nat))
                 (fun x2 ->
                   bind (p3 (nth cs (nat (Big_int.big_int_of_int 2))))
                     (fun x3 ->
                       bind (p4 (nth cs (nat (Big_int.big_int_of_int 3))))
                         (fun x4 ->
                           bind (p5 (nth cs (nat (Big_int.big_int_of_int 4))))
                             (fun x5 -> Inr (f x1 x2 x3 x4 x5))))))
      else no_transform tag (Xml (name, atts, cs, text)));;

let rec ordering_constraint_proof
  xml2name =
    xml1element
      ['o'; 'r'; 'd'; 'e'; 'r'; 'i'; 'n'; 'g'; 'C'; 'o'; 'n'; 's'; 't'; 'r';
        'a'; 'i'; 'n'; 't'; 'P'; 'r'; 'o'; 'o'; 'f']
      (redpair xml2name) id;;

let rec xml2nd_choice
  tag p1 cn p2 f (Xml (name, atts, cs, text)) =
    let l = size_list cs in
    (if eq_list eq_char name tag &&
          (null atts &&
            (less_eq_nat (nat (Big_int.big_int_of_int 2)) l && is_none text))
      then bind (p1 (nth cs Zero_nat))
             (fun x1 ->
               bind (boola cn (nth cs one_nat))
                 (fun b ->
                   (if b then bind (p2 (nth cs (minus_nat l one_nat)))
                                (fun x2 -> Inr (f x1 (Some x2)))
                     else Inr (f x1 None))))
      else no_transform tag (Xml (name, atts, cs, text)));;

let rec xml2dpProof _A
  ld xml2name x =
    xml1element ['d'; 'p'; 'P'; 'r'; 'o'; 'o'; 'f']
      (xml_options
        [(['p'; 'I'; 's'; 'E'; 'm'; 'p'; 't'; 'y'],
           xml0elements ['p'; 'I'; 's'; 'E'; 'm'; 'p'; 't'; 'y'] PisEmpty);
          (['d'; 'e'; 'p'; 'G'; 'r'; 'a'; 'p'; 'h'; 'P'; 'r'; 'o'; 'c'],
            xml_many_elements
              ['d'; 'e'; 'p'; 'G'; 'r'; 'a'; 'p'; 'h'; 'P'; 'r'; 'o'; 'c']
              (xml2nd_choice ['c'; 'o'; 'm'; 'p'; 'o'; 'n'; 'e'; 'n'; 't']
                (xml1element ['d'; 'p'; 's'] (rules (xml2name ld)) id)
                ['r'; 'e'; 'a'; 'l'; 'S'; 'c'; 'c'] (xml2dpProof _A ld xml2name)
                (fun dps prfOpt -> (prfOpt, dps)))
              (fun a -> DepGraphProc a));
          (['r'; 'e'; 'd'; 'P'; 'a'; 'i'; 'r'; 'P'; 'r'; 'o'; 'c'],
            xml3elements ['r'; 'e'; 'd'; 'P'; 'a'; 'i'; 'r'; 'P'; 'r'; 'o'; 'c']
              (ordering_constraint_proof (xml2name ld))
              (xml1element ['d'; 'p'; 's'] (rules (xml2name ld)) id)
              (xml2dpProof _A ld xml2name)
              (fun a b c -> RedPairProc (a, b, c)));
          (['r'; 'e'; 'd'; 'P'; 'a'; 'i'; 'r'; 'U'; 'r'; 'P'; 'r'; 'o'; 'c'],
            xml4elements
              ['r'; 'e'; 'd'; 'P'; 'a'; 'i'; 'r'; 'U'; 'r'; 'P'; 'r'; 'o'; 'c']
              (ordering_constraint_proof (xml2name ld))
              (xml1element ['d'; 'p'; 's'] (rules (xml2name ld)) id)
              (xml1element
                ['u'; 's'; 'a'; 'b'; 'l'; 'e'; 'R'; 'u'; 'l'; 'e'; 's']
                (rules (xml2name ld)) id)
              (xml2dpProof _A ld xml2name)
              (fun a b c d -> RedPairUrProc (a, b, c, d)));
          (['m'; 'o'; 'n'; 'o'; 'R'; 'e'; 'd'; 'P'; 'a'; 'i'; 'r'; 'P'; 'r';
             'o'; 'c'],
            xml4elements
              ['m'; 'o'; 'n'; 'o'; 'R'; 'e'; 'd'; 'P'; 'a'; 'i'; 'r'; 'P'; 'r';
                'o'; 'c']
              (ordering_constraint_proof (xml2name ld))
              (xml1element ['d'; 'p'; 's'] (rules (xml2name ld)) id)
              (xml1element ['t'; 'r'; 's'] (rules (xml2name ld)) id)
              (xml2dpProof _A ld xml2name)
              (fun a b c d -> MonoRedPairProc (a, b, c, d)));
          (['m'; 'o'; 'n'; 'o'; 'R'; 'e'; 'd'; 'P'; 'a'; 'i'; 'r'; 'U'; 'r';
             'P'; 'r'; 'o'; 'c'],
            xml5elements
              ['m'; 'o'; 'n'; 'o'; 'R'; 'e'; 'd'; 'P'; 'a'; 'i'; 'r'; 'U'; 'r';
                'P'; 'r'; 'o'; 'c']
              (ordering_constraint_proof (xml2name ld))
              (xml1element ['d'; 'p'; 's'] (rules (xml2name ld)) id)
              (xml1element ['t'; 'r'; 's'] (rules (xml2name ld)) id)
              (xml1element
                ['u'; 's'; 'a'; 'b'; 'l'; 'e'; 'R'; 'u'; 'l'; 'e'; 's']
                (rules (xml2name ld)) id)
              (xml2dpProof _A ld xml2name)
              (fun a b c d e -> MonoRedPairUrProc (a, b, c, d, e)));
          (['s'; 'e'; 'm'; 'l'; 'a'; 'b'; 'P'; 'r'; 'o'; 'c'],
            xml4elements ['s'; 'e'; 'm'; 'l'; 'a'; 'b'; 'P'; 'r'; 'o'; 'c']
              (sl_inter (xml2name ld))
              (xml1element ['d'; 'p'; 's'] (rules (xml2name (Suc ld))) id)
              (xml1element ['t'; 'r'; 's'] (rules (xml2name (Suc ld))) id)
              (xml2dpProof _A (Suc ld) xml2name)
              (fun a ->
                (match a with None -> (fun aa b c -> RootlabProc (aa, b, c))
                  | Some i ->
                    (fun aa b c -> SemlabProc (i, id_labeling, aa, b, c)))));
          (['u'; 'n'; 'l'; 'a'; 'b'; 'P'; 'r'; 'o'; 'c'],
            xml3elements ['u'; 'n'; 'l'; 'a'; 'b'; 'P'; 'r'; 'o'; 'c']
              (xml1element ['d'; 'p'; 's']
                (rules (xml2name (minus_nat ld one_nat))) id)
              (xml1element ['t'; 'r'; 's']
                (rules (xml2name (minus_nat ld one_nat))) id)
              (xml2dpProof _A (minus_nat ld one_nat) xml2name)
              (fun a b c -> UnlabProc (a, b, c)));
          (['s'; 'u'; 'b'; 't'; 'e'; 'r'; 'm'; 'P'; 'r'; 'o'; 'c'],
            xml1many2elements_gen
              ['s'; 'u'; 'b'; 't'; 'e'; 'r'; 'm'; 'P'; 'r'; 'o'; 'c']
              (sp (xml2name ld))
              (fun pi ->
                projected_rseq (eq_laba _A (eq_lista eq_nata)) (xml2name ld)
                  (create_sp (eq_laba _A (eq_lista eq_nata)) pi))
              (xml1element ['d'; 'p'; 's'] (rules (xml2name ld)) id)
              (xml2dpProof _A ld xml2name)
              (fun a b c d -> SubtermCriterionProc (a, b, c, d)));
          (['f'; 'l'; 'a'; 't'; 'C'; 'o'; 'n'; 't'; 'e'; 'x'; 't'; 'C'; 'l';
             'o'; 's'; 'u'; 'r'; 'e'; 'P'; 'r'; 'o'; 'c'],
            xml5elements
              ['f'; 'l'; 'a'; 't'; 'C'; 'o'; 'n'; 't'; 'e'; 'x'; 't'; 'C'; 'l';
                'o'; 's'; 'u'; 'r'; 'e'; 'P'; 'r'; 'o'; 'c']
              (xml1element
                ['f'; 'r'; 'e'; 's'; 'h'; 'S'; 'y'; 'm'; 'b'; 'o'; 'l']
                (xml2name ld) id)
              (flat_contexts (xml2name ld))
              (xml1element ['d'; 'p'; 's'] (rules (xml2name ld)) id)
              (xml1element ['t'; 'r'; 's'] (rules (xml2name ld)) id)
              (xml2dpProof _A ld xml2name)
              (fun a b c d e -> FccProc (a, b, c, d, e)));
          (['s'; 'i'; 'z'; 'e'; 'C'; 'h'; 'a'; 'n'; 'g'; 'e'; 'P'; 'r'; 'o';
             'c'],
            xml1many_elements
              ['s'; 'i'; 'z'; 'e'; 'C'; 'h'; 'a'; 'n'; 'g'; 'e'; 'P'; 'r'; 'o';
                'c']
              (xml_options
                [(['s'; 'u'; 'b'; 't'; 'e'; 'r'; 'm'; 'C'; 'r'; 'i'; 't'; 'e';
                    'r'; 'i'; 'o'; 'n'],
                   xml0elements
                     ['s'; 'u'; 'b'; 't'; 'e'; 'r'; 'm'; 'C'; 'r'; 'i'; 't';
                       'e'; 'r'; 'i'; 'o'; 'n']
                     None);
                  (['r'; 'e'; 'd'; 'u'; 'c'; 't'; 'i'; 'o'; 'n'; 'P'; 'a'; 'i';
                     'r'],
                    xml2elements
                      ['r'; 'e'; 'd'; 'u'; 'c'; 't'; 'i'; 'o'; 'n'; 'P'; 'a';
                        'i'; 'r']
                      (ordering_constraint_proof (xml2name ld))
                      (xml1element
                        ['u'; 's'; 'a'; 'b'; 'l'; 'e'; 'R'; 'u'; 'l'; 'e'; 's']
                        (rules (xml2name ld)) id)
                      (fun redp ur -> Some (redp, ur)))])
              (scg (xml2name ld))
              (fun a ->
                (match a with None -> (fun aa -> SizeChangeSubTerm aa)
                  | Some redp_ur ->
                    (fun aa ->
                      SizeChangeRedPair (fst redp_ur, snd redp_ur, aa)))))])
      id x;;

let rec isOK m = (match m with Inl _ -> false | Inr _ -> true);;

let rec mat_monoI geq1 sd m = forall (any geq1) (sub_mat sd sd m);;

let rec mul_pvarsI _A
  uu uv a x3 = match uu, uv, a, x3 with uu, uv, a, [] -> []
    | ze, ti, a, (x, b) :: vas ->
        let p = ti a b in
        let res = mul_pvarsI _A ze ti a vas in
        (if eqa _A p ze then res else (x, p) :: res);;

type ('a, 'b) l_poly = LPoly of 'b * ('a * 'b) list;;

let rec mul_lpolyI _A
  ze ti a (LPoly (b, vas)) = LPoly (ti a b, mul_pvarsI _A ze ti a vas);;

let rec list_sumI
  ze pl x2 = match ze, pl, x2 with ze, pl, [] -> ze
    | ze, pl, x :: xs -> pl x (list_sumI ze pl xs);;

let rec add_varI _A _B
  uu uv x a xa4 = match uu, uv, x, a, xa4 with uu, uv, x, a, [] -> [(x, a)]
    | ze, pl, x, a, (y, b) :: vas ->
        (if eqa _B x y
          then let s = pl a b in
               (if eqa _A s ze then vas else (x, s) :: vas)
          else (y, b) :: add_varI _A _B ze pl x a vas);;

let rec sum_pvarsI _A _B
  uu uv x2 vbs = match uu, uv, x2, vbs with uu, uv, [], vbs -> vbs
    | ze, pl, (x, a) :: vas, vbs ->
        (if eqa _A a ze then sum_pvarsI _A _B ze pl vas vbs
          else sum_pvarsI _A _B ze pl vas (add_varI _A _B ze pl x a vbs));;

let rec sum_lpolyI _A _B
  ze pl (LPoly (a, vas)) (LPoly (b, vbs)) =
    LPoly (pl a b, sum_pvarsI _A _B ze pl vas vbs);;

let rec l_poly_case f1 (LPoly (a, lista)) = f1 a lista;;

let rec pleftI _A _C
  ze on pl ti mz pi x6 = match ze, on, pl, ti, mz, pi, x6 with
    ze, on, pl, ti, mz, pi, Var x -> LPoly (ze, [(x, on)])
    | ze, on, pl, ti, mz, pi, Fun (f, ts) ->
        let (c, asa) = pi f (size_list ts) in
        (match
          sum_lpolyI _A _C ze pl (LPoly (c, []))
            (list_sumI (LPoly (ze, [])) (sum_lpolyI _A _C ze pl)
              (map (fun at ->
                     mul_lpolyI _A ze ti (fst at)
                       (pleftI _A _C ze on pl ti mz pi (snd at)))
                (zip asa ts)))
          with LPoly (d, []) -> LPoly (mz d, [])
          | LPoly (d, ab :: lista) -> LPoly (d, ab :: lista));;

type ('a, 'b, 'c) tRSProof =
  DPTrans of
    ((('a, 'b) lab shp, 'c) term * (('a, 'b) lab shp, 'c) term) list *
      (('a, 'b) lab shp, 'b, 'c) dPProof
  | RuleRemoval of
      ('a, 'b) lab redpair *
        ((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list *
        ('a, 'b, 'c) tRSProof
  | Unlab of
      ((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list *
        ('a, 'b, 'c) tRSProof
  | Semlab of
      ('a, 'b) sl_inter *
        (('a, 'b) lab -> nat list -> ('b, (('a, 'b) lab list)) sum) *
        ((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list *
        ('a, 'b, 'c) tRSProof
  | Rootlab of
      ((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list *
        ('a, 'b, 'c) tRSProof
  | Fcc of
      (('a, 'b) lab, 'c) ctxt list *
        ((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list *
        ('a, 'b, 'c) tRSProof
  | StringReversal of
      ((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list *
        ('a, 'b, 'c) tRSProof
  | RisEmpty;;

let rec lpoS x = shows_string ['>'; '_'; 'L'; 'P'; 'O'; '+'; 'A'; 'F'] x;;

let rec linear_term_impl _A
  xs x1 = match xs, x1 with
    xs, Var x -> (if member _A x xs then None else Some (x :: xs))
    | xs, Fun (uu, []) -> Some xs
    | xs, Fun (f, t :: ts) ->
        (match linear_term_impl _A xs t with None -> None
          | Some ys -> linear_term_impl _A ys (Fun (f, ts)));;

let rec linear_term _B
  t = (match linear_term_impl _B [] t with None -> false | Some _ -> true);;

let rec xml2trsInput
  xml2name =
    xml1element ['i'; 'n'; 'p'; 'u'; 't']
      (xml1element ['t'; 'r'; 's'; 'I'; 'n'; 'p'; 'u'; 't']
        (xml1element ['t'; 'r'; 's'] (rules xml2name) id) id)
      id;;

let rec xml2trsProof _A
  xml2name xml2sname xml2lsname x =
    xml1element
      ['t'; 'r'; 's'; 'T'; 'e'; 'r'; 'm'; 'i'; 'n'; 'a'; 't'; 'i'; 'o'; 'n';
        'P'; 'r'; 'o'; 'o'; 'f']
      (xml_options
        [(['r'; 'I'; 's'; 'E'; 'm'; 'p'; 't'; 'y'],
           xml0elements ['r'; 'I'; 's'; 'E'; 'm'; 'p'; 't'; 'y'] RisEmpty);
          (['d'; 'p'; 'T'; 'r'; 'a'; 'n'; 's'],
            xml3elements ['d'; 'p'; 'T'; 'r'; 'a'; 'n'; 's']
              (xml1element ['d'; 'p'; 's'] (rules xml2sname) id)
              (boola
                ['m'; 'a'; 'r'; 'k'; 'e'; 'd'; 'S'; 'y'; 'm'; 'b'; 'o'; 'l';
                  's'])
              (xml2dpProof (eq_shpa (eq_laba _A (eq_lista eq_nata))) Zero_nat
                xml2lsname)
              (fun dps _ -> (fun a -> DPTrans (dps, a))));
          (['r'; 'u'; 'l'; 'e'; 'R'; 'e'; 'm'; 'o'; 'v'; 'a'; 'l'],
            xml3elements ['r'; 'u'; 'l'; 'e'; 'R'; 'e'; 'm'; 'o'; 'v'; 'a'; 'l']
              (ordering_constraint_proof xml2name)
              (xml1element ['t'; 'r'; 's'] (rules xml2name) id)
              (xml2trsProof _A xml2name xml2sname xml2lsname)
              (fun a b c -> RuleRemoval (a, b, c)));
          (['u'; 'n'; 'l'; 'a'; 'b'],
            xml2elements ['u'; 'n'; 'l'; 'a'; 'b']
              (xml1element ['t'; 'r'; 's'] (rules xml2name) id)
              (xml2trsProof _A xml2name xml2sname xml2lsname)
              (fun a b -> Unlab (a, b)));
          (['s'; 't'; 'r'; 'i'; 'n'; 'g'; 'R'; 'e'; 'v'; 'e'; 'r'; 's'; 'a';
             'l'],
            xml2elements
              ['s'; 't'; 'r'; 'i'; 'n'; 'g'; 'R'; 'e'; 'v'; 'e'; 'r'; 's'; 'a';
                'l']
              (xml1element ['t'; 'r'; 's'] (rules xml2name) id)
              (xml2trsProof _A xml2name xml2sname xml2lsname)
              (fun a b -> StringReversal (a, b)));
          (['s'; 'e'; 'm'; 'l'; 'a'; 'b'],
            xml3elements ['s'; 'e'; 'm'; 'l'; 'a'; 'b'] (sl_inter xml2name)
              (xml1element ['t'; 'r'; 's'] (rules xml2name) id)
              (xml2trsProof _A xml2name xml2sname xml2lsname)
              (fun a ->
                (match a with None -> (fun aa b -> Rootlab (aa, b))
                  | Some i -> (fun aa b -> Semlab (i, id_labeling, aa, b)))));
          (['f'; 'l'; 'a'; 't'; 'C'; 'o'; 'n'; 't'; 'e'; 'x'; 't'; 'C'; 'l';
             'o'; 's'; 'u'; 'r'; 'e'],
            xml3elements
              ['f'; 'l'; 'a'; 't'; 'C'; 'o'; 'n'; 't'; 'e'; 'x'; 't'; 'C'; 'l';
                'o'; 's'; 'u'; 'r'; 'e']
              (flat_contexts xml2name)
              (xml1element ['t'; 'r'; 's'] (rules xml2name) id)
              (xml2trsProof _A xml2name xml2sname xml2lsname)
              (fun a b c -> Fcc (a, b, c)))])
      id x;;

let rec catch m f = (match m with Inl a -> f a | Inr a -> Inr a);;

let rec forallm
  f x1 = match f, x1 with f, [] -> Inr ()
    | f, x :: xs ->
        bind (catch (f x) (fun xa -> Inl (x, xa))) (fun _ -> forallm f xs);;

let rec check_edges
  ss g c d =
    catch (forallm
            (fun n ->
              catch (forallm
                      (fun m ->
                        check (not (g (m, n)))
                          (comp (shows_string
                                  ['e'; 'd'; 'g'; 'e'; ' '; 'f'; 'r'; 'o'; 'm';
                                    ' '])
                            (comp (ss m)
                              (comp (shows_string [' '; 't'; 'o'; ' '])
                                (ss n)))))
                      d)
                (fun x -> Inl (snd x)))
            c)
      (fun x -> Inl (snd x));;

let bot_bool : bool = false;;

let rec less_bool x0 b = match x0, b with true, b -> false
                    | false, b -> b;;

let rec less_eq_bool x0 b = match x0, b with true, b -> b
                       | false, b -> true;;

let ord_bool = ({less_eq = less_eq_bool; less = less_bool} : bool ord);;

let preorder_bool = ({ord_preorder = ord_bool} : bool preorder);;

let bot_boola = ({preorder_bot = preorder_bool; bot = bot_bool} : bool bot);;

let rec bot_fun _B = (fun _ -> bot _B);;

let rec prightI _A _C
  ze on pl ti mz pi x6 = match ze, on, pl, ti, mz, pi, x6 with
    ze, on, pl, ti, mz, pi, Var x -> LPoly (ze, [(x, on)])
    | ze, on, pl, ti, mz, pi, Fun (f, ts) ->
        let (c, asa) = pi f (size_list ts) in
        let LPoly (d, a) =
          sum_lpolyI _A _C ze pl (LPoly (c, []))
            (list_sumI (LPoly (ze, [])) (sum_lpolyI _A _C ze pl)
              (map (fun at ->
                     mul_lpolyI _A ze ti (fst at)
                       (prightI _A _C ze on pl ti mz pi (snd at)))
                (zip asa ts)))
          in
        LPoly (mz d, a);;

let rec lpoNS x = shows_string ['>'; '='; '_'; 'L'; 'P'; 'O'; '+'; 'A'; 'F'] x;;

let rec funs_term_list'
  = function Var x -> []
    | Fun (f, ts) -> f :: flat_map funs_term_list' ts;;

let rec funs_rule_list'
  r = append (funs_term_list' (fst r)) (funs_term_list' (snd r));;

let rec funs_trs_list _A r = remdups _A (flat_map funs_rule_list' r);;

let rec vars_term_list'
  = function Var x -> [x]
    | Fun (f, ts) -> flat_map vars_term_list' ts;;

let rec vars_rule_list'
  r = append (vars_term_list' (fst r)) (vars_term_list' (snd r));;

let rec vars_trs_list _B trs = remdups _B (flat_map vars_rule_list' trs);;

let rec doc_of_string s = bind (parse_doc s) (fun (doc, _) -> Inr doc);;

let rec length_unique _A
  = function [] -> Zero_nat
    | x :: xs ->
        (if member _A x xs then length_unique _A xs
          else Suc (length_unique _A xs));;

type certificationResult = Certified of char list | Error of char list;;

let rec vars_term_list _B t = remdups _B (vars_term_list' t);;

let rec enum_vectors
  c x1 = match c, x1 with c, [] -> [[]]
    | c, x :: xs ->
        let a = enum_vectors c xs in
        concat_map (fun vec -> map (fun ca -> (x, ca) :: vec) c) a;;

let rec assign_from_vector _A vec x = the (lookup _A x vec);;

let rec shows_lab _A _B
  = function UnLab f -> shows_prec _A Zero_nat f
    | Lab (f, l) ->
        comp (shows_lab _A _B f)
          (comp (shows_string ['['])
            (comp (shows_prec _B Zero_nat l) (shows_string [']'])))
    | FunLab (f, l) ->
        comp (shows_lab _A _B f)
          (comp (shows_string ['['])
            (comp (shows_map (shows_lab _A _B) l) (shows_string [']'])));;

let rec shows_prec_lab _A _B d l = shows_lab _A _B l;;

let rec shows_list_lab _A _B
  fs = shows_list_aux (shows_prec_lab _A _B Zero_nat) fs;;

let rec show_lab _A _B =
  ({shows_prec = shows_prec_lab _A _B; shows_list = shows_list_lab _A _B} :
    ('a, 'b) lab show);;

let rec labela
  f x1 = match f, x1 with f, Inl l -> Lab (f, l)
    | f, Inr l -> FunLab (f, l);;

let rec eval_lab
  i l alpha x3 = match i, l, alpha, x3 with
    i, l, alpha, Var x -> (alpha x, Var x)
    | i, l, alpha, Fun (f, ts) ->
        let clts = map (eval_lab i l alpha) ts in
        let cs = map fst clts in
        let c = i f cs in
        let lts = map snd clts in
        let la = l f cs in
        (c, Fun (labela f la, lts));;

let rec check_sl_rule_ass (_A1, _A2) (_B1, _B2) (_C1, _C2) (_D1, _D2)
  mc i la lR alpha (l, r) =
    let cl_ll = eval_lab i la alpha l in
    let cr_lr = eval_lab i la alpha r in
    bind (check (implies mc (eqa _C1 (fst cl_ll) (fst cr_lr)))
           (comp (shows_string ['r'; 'u'; 'l'; 'e'; ' '])
             (comp (shows_rule (show_lab _A2 _B2) _D2 (l, r))
               (comp (shows_string
                       [' '; 'v'; 'i'; 'o'; 'l'; 'a'; 't'; 'e'; 's'; ' '; 't';
                         'h'; 'e'; ' '; 'm'; 'o'; 'd'; 'e'; 'l'; ' '; 'c'; 'o';
                         'n'; 'd'; 'i'; 't'; 'i'; 'o'; 'n'; ','; ' '; '['; 'l';
                         'h'; 's'; ']'; ' '; '='; ' '])
                 (comp (shows_prec _C2 Zero_nat (fst cl_ll))
                   (comp (shows_string
                           [','; ' '; '['; 'r'; 'h'; 's'; ']'; ' '; '='; ' '])
                     (shows_prec _C2 Zero_nat (fst cr_lr))))))))
      (fun _ ->
        check (member
                (eqc (eq_terma (eq_laba _A1 _B1) _D1)
                  (eq_terma (eq_laba _A1 _B1) _D1))
                (snd cl_ll, snd cr_lr) lR)
          (comp (shows_string
                  ['l'; 'a'; 'b'; 'e'; 'l'; 'e'; 'd'; ' '; 'r'; 'u'; 'l'; 'e';
                    ' '])
            (comp (shows_rule (show_lab _A2 _B2) _D2 (snd cl_ll, snd cr_lr))
              (shows_string [' '; 'm'; 'i'; 's'; 's'; 'i'; 'n'; 'g']))));;

let rec check_sl_rule (_A1, _A2) (_B1, _B2) (_C1, _C2) (_D1, _D2)
  i la c mc lR (l, r) =
    catch (forallm
            (fun alpha ->
              check_sl_rule_ass (_A1, _A2) (_B1, _B2) (_C1, _C2) (_D1, _D2) mc i
                la lR alpha (l, r))
            (map (assign_from_vector _D1)
              (enum_vectors c
                (list_union _D1 (vars_term_list _D1 l)
                  (vars_term_list _D1 r)))))
      (fun x -> Inl (snd x));;

let rec check_sl_model_lab_trs (_A1, _A2) (_B1, _B2) (_C1, _C2) (_D1, _D2)
  i l c lR r =
    catch (forallm
            (check_sl_rule (_A1, _A2) (_B1, _B2) (_C1, _C2) (_D1, _D2) i l c
              true lR)
            r)
      (fun x -> Inl (snd x));;

let rec rl_fun uu cs = Inr cs;;

let rec rl_inter_list _A fs d g uu = (if member _A g fs then g else d);;

let rec check_sl_lab_trs (_A1, _A2) (_B1, _B2) (_C1, _C2) (_D1, _D2)
  i l c lP p =
    catch (forallm
            (check_sl_rule (_A1, _A2) (_B1, _B2) (_C1, _C2) (_D1, _D2) i l c
              false lP)
            p)
      (fun x -> Inl (snd x));;

let rec get_lab = function Lab (f, l) -> Inl l
                  | FunLab (f, l) -> Inr l;;

let rec unlab
  = function Lab (f, l) -> f
    | FunLab (f, l) -> f
    | UnLab v -> UnLab v;;

let rec check_sl_only_lab_and_decr (_A1, _A2) (_B1, _B2) (_C1, _C2)
  gr r =
    (fun xs ->
      catch (forallm
              (fun (ll, lr) ->
                let l = map_term unlab ll in
                let ra = map_term unlab lr in
                check (if eq_term (eq_laba _B1 _A1) _C1 l ra
                        then not (is_var ll) &&
                               (not (is_var lr) &&
                                 gr (get_lab (the (root ll)),
                                      get_lab (the (root lr))))
                               || eq_term (eq_laba _B1 _A1) _C1 ll lr &&
                                    member
                                      (eqc (eq_terma (eq_laba _B1 _A1) _C1)
(eq_terma (eq_laba _B1 _A1) _C1))
                                      (l, ra) r
                        else member
                               (eqc (eq_terma (eq_laba _B1 _A1) _C1)
                                 (eq_terma (eq_laba _B1 _A1) _C1))
                               (l, ra) r)
                  (comp (shows_string
                          ['l'; 'a'; 'b'; 'e'; 'l'; 'e'; 'd'; ' '; 'r'; 'u';
                            'l'; 'e'; ' '])
                    (comp (shows_rule (show_lab _B2 _A2) _C2 (l, ra))
                      (shows_string
                        [' '; 'i'; 's'; ' '; 'n'; 'e'; 'i'; 't'; 'h'; 'e'; 'r';
                          ' '; 'd'; 'e'; 'c'; 'r'; 'e'; 'a'; 's'; 'i'; 'n'; 'g';
                          ' '; 'n'; 'o'; 'r'; ' '; 'a'; ' '; 'l'; 'a'; 'b'; 'e';
                          'l'; 'e'; 'd'; ' '; 'v'; 'e'; 'r'; 's'; 'i'; 'o'; 'n';
                          ' '; 'o'; 'f'; ' '; 'o'; 'r'; 'i'; 'g'; 'i'; 'n'; 'a';
                          'l'; ' '; 'r'; 'u'; 'l'; 'e']))))
              xs)
        (fun x -> Inl (snd x)));;

let rec check_sem_lab_tt (_A1, _A2) (_B1, _B2) (_C1, _C2)
  check_lab check_model_lab gr i l r lR =
    catch (bind (check_model_lab lR r)
            (fun _ ->
              check_sl_only_lab_and_decr (_B1, _B2) (_A1, _A2) (_C1, _C2) gr r
                lR))
      (fun x ->
        Inl (comp (shows_string
                    ['p'; 'r'; 'o'; 'b'; 'l'; 'e'; 'm'; ' '; 'w'; 'i'; 't'; 'h';
                      ' '; 'l'; 'a'; 'b'; 'e'; 'l'; 'e'; 'd'; ' '; 'T'; 'R';
                      'S'; ':'])
              (comp shows_nl x)));;

let rec rl_tt (_A1, _A2) (_B1, _B2)
  r lR =
    let fs = funs_trs_list (eq_laba _A1 _B1) r in
    bind (check (not (null fs))
           (comp (shows_string
                   ['r'; 'o'; 'o'; 't'; '-'; 'l'; 'a'; 'b'; 'e'; 'l'; 'i'; 'n';
                     'g'; ' '; 'r'; 'e'; 'q'; 'u'; 'i'; 'r'; 'e'; 's'; ' '; 'a';
                     't'; ' '; 'l'; 'e'; 'a'; 's'; 't'; ' '; 'o'; 'n'; 'e'; ' ';
                     'f'; 'u'; 'n'; 'c'; 't'; 'i'; 'o'; 'n'; ' '; 's'; 'y'; 'm';
                     'b'; 'o'; 'l'; ' '; 'i'; 'n'; ' '; 't'; 'h'; 'e'; ' '; 's';
                     'i'; 'g'; 'n'; 'a'; 't'; 'u'; 'r'; 'e'])
             shows_nl))
      (fun _ ->
        let d = hd fs in
        check_sem_lab_tt (_A1, _A2) (_B1, _B2)
          (eq_lista eq_char, show_list show_char)
          (check_sl_lab_trs (_A1, _A2) (_B1, _B2)
            (eq_laba _A1 _B1, show_lab _A2 _B2)
            (eq_lista eq_char, show_list show_char)
            (rl_inter_list (eq_laba _A1 _B1) fs d) rl_fun fs)
          (check_sl_model_lab_trs (_A1, _A2) (_B1, _B2)
            (eq_laba _A1 _B1, show_lab _A2 _B2)
            (eq_lista eq_char, show_list show_char)
            (rl_inter_list (eq_laba _A1 _B1) fs d) rl_fun fs)
          (bot_fun bot_boola) (rl_inter_list (eq_laba _A1 _B1) fs d) rl_fun r
          lR);;

let rec wf_terms_list _A
  uu x1 = match uu, x1 with uu, Var uv -> true
    | fa, Fun (f, ts) ->
        member (eqc _A eq_nata) (f, size_list ts) fa &&
          forall (wf_terms_list _A fa) ts;;

let rec funas_term_list'
  = function Var x -> []
    | Fun (f, ts) -> (f, size_list ts) :: flat_map funas_term_list' ts;;

let rec funas_rule_list'
  r = append (funas_term_list' (fst r)) (funas_term_list' (snd r));;

let rec funas_trs_list _A
  r = remdups (eqc _A eq_nata) (flat_map funas_rule_list' r);;

let rec sharp_lab_name
  x = (if eq_list eq_char (get_name x) ['s'; 'h'; 'a'; 'r'; 'p']
        then xml1element ['s'; 'h'; 'a'; 'r'; 'p'] plain_name (fun a -> Sharp a)
               x
        else xml_change plain_name (fun a -> Plain a) x);;

let rec lab_sharp_name
  = function Zero_nat -> xml_change sharp_lab_name (fun a -> UnLab a)
    | Suc n ->
        xml2elements
          ['l'; 'a'; 'b'; 'e'; 'l'; 'e'; 'd'; 'S'; 'y'; 'm'; 'b'; 'o'; 'l']
          (lab_sharp_name n) (label (lab_sharp_name n))
          (fun f a ->
            (match a with Inl aa -> Lab (f, aa) | Inr aa -> FunLab (f, aa)));;

type ('a, 'b) dPDisproof =
  DPLoop of
    ('a, 'b) term *
      (pos * ((('a, 'b) term * ('a, 'b) term) * (bool * ('a, 'b) term))) list *
      ('b * ('a, 'b) term) list * ('a, 'b) ctxt
  | DPRuleRemoval of
      (('a, 'b) term * ('a, 'b) term) list *
        (('a, 'b) term * ('a, 'b) term) list * ('a, 'b) dPDisproof;;

let rec xml2dpDisproof
  xml2name =
    xml1element
      ['d'; 'p'; 'N'; 'o'; 'n'; 't'; 'e'; 'r'; 'm'; 'i'; 'n'; 'a'; 't'; 'i';
        'o'; 'n'; 'P'; 'r'; 'o'; 'o'; 'f']
      (xml_options
        [(['d'; 'p'; 'R'; 'u'; 'l'; 'e'; 'R'; 'e'; 'm'; 'o'; 'v'; 'a'; 'l'],
           xml3elements
             ['d'; 'p'; 'R'; 'u'; 'l'; 'e'; 'R'; 'e'; 'm'; 'o'; 'v'; 'a'; 'l']
             (xml1element ['d'; 'p'; 's'] (rules xml2name) id)
             (xml1element ['t'; 'r'; 's'] (rules xml2name) id)
             (xml2dpDisproof xml2name) (fun a b c -> DPRuleRemoval (a, b, c)));
          (['l'; 'o'; 'o'; 'p'],
            xml_change (loop xml2name)
              (fun (s, (rseq, (a, b))) -> DPLoop (s, rseq, a, b)))])
      id;;

let rec supteq_list
  = function Var x -> [Var x]
    | Fun (f, ts) -> Fun (f, ts) :: concat_map supteq_list ts;;

let rec dP_list _A _B
  r = let d = defs_list _A r in
      concat_map
        (fun lr ->
          let l = fst lr in
          let s = sharpa l in
          map (fun u -> (s, sharpa u))
            (filter
              (fun u ->
                not (suptp _A _B l u) &&
                  (not (is_var u) && member _A (the (root u)) d))
              (supteq_list (snd lr))))
        r;;

let rec iD_list r = map (fun lr -> (plaina (fst lr), plaina (snd lr))) r;;

type ('a, 'b) tRSDisproof =
  Loop of
    ('a, 'b) term *
      (pos * ((('a, 'b) term * ('a, 'b) term) * ('a, 'b) term)) list *
      ('b * ('a, 'b) term) list * ('a, 'b) ctxt
  | NotWellFormed |
  NTRuleRemoval of (('a, 'b) term * ('a, 'b) term) list * ('a, 'b) tRSDisproof |
  NTStringReversal of
    (('a, 'b) term * ('a, 'b) term) list * ('a, 'b) tRSDisproof
  | NTDPTrans of
      (('a shp, 'b) term * ('a shp, 'b) term) list * ('a shp, 'b) dPDisproof;;

let rec indent
  p s = flat_map
          (fun c -> (if ((c : char) = newline) then [newline; ' '] else [c]))
          (p s);;

let rec check_rule_reflecting (_A1, _A2)
  fcs rs rule =
    check (any (fun (l, r) ->
                 any (fun c ->
                       eq_term _A1 (eq_lista eq_char) (fst rule)
                         (ctxt_apply c l)
                         && eq_term _A1 (eq_lista eq_char) (snd rule)
                              (ctxt_apply c r))
                   (Hole :: fcs))
            rs)
      (comp (shows_string ['t'; 'h'; 'e'; ' '; 'r'; 'u'; 'l'; 'e'; ' '])
        (comp (shows_rule _A2 (show_list show_char) rule)
          (comp (shows_string
                  [' '; 'i'; 's'; ' '; 'n'; 'e'; 'i'; 't'; 'h'; 'e'; 'r'; ' ';
                    'c'; 'o'; 'n'; 't'; 'a'; 'i'; 'n'; 'e'; 'd'; ' '; 'i'; 'n';
                    ' '; 't'; 'h'; 'e'; ' '; 'o'; 'r'; 'i'; 'g'; 'i'; 'n'; 'a';
                    'l'; ' '; 's'; 'e'; 't'; ' '; 'o'; 'f'; ' '; 'r'; 'u'; 'l';
                    'e'; 's'; ' '; 'n'; 'o'; 'r'; ' '; 'o'; 'b'; 't'; 'a'; 'i';
                    'n'; 'e'; 'd'; ' '; 'b'; 'y'; ' '; 'a'; 'p'; 'p'; 'l'; 'y';
                    'i'; 'n'; 'g'; ' '; 'a'; ' '; 'f'; 'l'; 'a'; 't'; ' '; 'c';
                    'o'; 'n'; 't'; 'e'; 'x'; 't'])
            shows_nl)));;

let rec check_rule_preserving (_A1, _A2)
  fcs rs' rule =
    check (member
             (eqc (eq_terma _A1 (eq_lista eq_char))
               (eq_terma _A1 (eq_lista eq_char)))
             rule rs'
            || forall
                 (fun c ->
                   member
                     (eqc (eq_terma _A1 (eq_lista eq_char))
                       (eq_terma _A1 (eq_lista eq_char)))
                     (ctxt_apply c (fst rule), ctxt_apply c (snd rule)) rs')
                 fcs)
      (comp (shows_string ['t'; 'h'; 'e'; ' '; 'r'; 'u'; 'l'; 'e'; ' '])
        (comp (shows_rule _A2 (show_list show_char) rule)
          (comp (shows_string
                  [' '; 'i'; 's'; ' '; 'n'; 'e'; 'i'; 't'; 'h'; 'e'; 'r'; ' ';
                    'c'; 'o'; 'n'; 't'; 'a'; 'i'; 'n'; 'e'; 'd'; ' '; 'i'; 'n';
                    ' '; 't'; 'h'; 'e'; ' '; 'r'; 'e'; 's'; 'u'; 'l'; 't'; 'i';
                    'n'; 'g'; ' '; 's'; 'e'; 't'; ' '; 'o'; 'f'; ' '; 'r'; 'u';
                    'l'; 'e'; 's'; ' '; 'n'; 'o'; 'r'; ' '; 'c'; 'l'; 'o'; 's';
                    'e'; 'd'; ' '; 'u'; 'n'; 'd'; 'e'; 'r'; ' '; 'a'; 'l'; 'l';
                    ' '; 'f'; 'l'; 'a'; 't'; ' '; 'c'; 'o'; 'n'; 't'; 'e'; 'x';
                    't'; 's'])
            shows_nl)));;

let rec hole_at _A
  n i f x3 = match n, i, f, x3 with
    n, i, f, More (g, ss1, Hole, ss2) ->
      eqa _A g f &&
        (eq_nat (size_list ss1) i &&
          eq_nat (size_list ss2) (minus_nat (minus_nat n i) one_nat))
    | n, i, f, Hole -> false
    | n, i, f, More (v, va, More (vd, ve, vf, vg), vc) -> false;;

let rec check_flat_ctxt_complete (_A1, _A2)
  fcs fa =
    check (forall (fun i -> any (hole_at _A1 (snd fa) i (fst fa)) fcs)
            (upt Zero_nat (snd fa)))
      (comp (shows_string
              ['t'; 'h'; 'e'; ' '; 'l'; 'i'; 's'; 't'; ' '; 'o'; 'f'; ' '; 'f';
                'l'; 'a'; 't'; ' '; 'c'; 'o'; 'n'; 't'; 'e'; 'x'; 't'; 's'; ' ';
                'i'; 's'; ' '; 'i'; 'n'; 'c'; 'o'; 'm'; 'p'; 'l'; 'e'; 't';
                'e'])
        shows_nl);;

let rec shows_prec_ctxt _A _B d c = shows_ctxt _A _B c;;

let rec is_flat_ctxt_list _A _B
  vs fas x2 = match vs, fas, x2 with
    vs, fas, More (f, ss1, Hole, ss2) ->
      member (eqc _B eq_nata) (f, Suc (size_list (append ss1 ss2))) fas &&
        (forall is_var (append ss1 ss2) &&
          (distinct (eq_terma _B _A) (append ss1 ss2) &&
            null (list_inter _A (map the_var (append ss1 ss2)) vs)))
    | vs, fas, Hole -> false
    | vs, fas, More (v, va, More (vd, ve, vf, vg), vc) -> false;;

let rec check_is_flat_ctxt (_A1, _A2)
  vs fas c =
    check (is_flat_ctxt_list (eq_lista eq_char) _A1 vs fas c)
      (comp (shows_prec_ctxt _A2 (show_list show_char) Zero_nat c)
        (comp (shows_string
                [' '; 'i'; 's'; ' '; 'n'; 'o'; 't'; ' '; 'a'; ' '; 'f'; 'l';
                  'a'; 't'; ' '; 'c'; 'o'; 'n'; 't'; 'e'; 'x'; 't'])
          shows_nl));;

let rec check_flat_ctxt (_A1, _A2)
  vs c =
    (match c
      with Hole ->
        Inl (comp (shows_prec_ctxt _A2 (show_list show_char) Zero_nat c)
              (comp (shows_string
                      [' '; 'i'; 's'; ' '; 'n'; 'o'; 't'; ' '; 'a'; ' '; 'f';
                        'l'; 'a'; 't'; ' '; 'c'; 'o'; 'n'; 't'; 'e'; 'x'; 't'])
                shows_nl))
      | More (_, ss1, Hole, ss2) ->
        bind (check
               (distinct (eq_terma _A1 (eq_lista eq_char)) (append ss1 ss2))
               (comp (shows_prec_ctxt _A2 (show_list show_char) Zero_nat c)
                 (comp (shows_string
                         [' '; 'c'; 'o'; 'n'; 't'; 'a'; 'i'; 'n'; 's'; ' '; 'd';
                           'u'; 'p'; 'l'; 'i'; 'c'; 'a'; 't'; 'e'; ' '; 'v';
                           'a'; 'r'; 'i'; 'a'; 'b'; 'l'; 'e'; 's'])
                   shows_nl)))
          (fun _ ->
            bind (check (forall is_var (append ss1 ss2))
                   (comp (shows_prec_ctxt _A2 (show_list show_char) Zero_nat c)
                     (comp (shows_string
                             [' '; 'i'; 's'; ' '; 'n'; 'o'; 't'; ' '; 'f'; 'l';
                               'a'; 't'; ','; ' '; 'i'; '.'; 'e'; '.'; ','; ' ';
                               'h'; 'a'; 's'; ' '; 'd'; 'e'; 'p'; 't'; 'h'; ' ';
                               'g'; 'r'; 'e'; 'a'; 't'; 'e'; 'r'; ' '; 't'; 'h';
                               'a'; 'n'; ' '; 'o'; 'n'; 'e'])
                       shows_nl)))
              (fun _ ->
                check (forall
                        (fun t ->
                          not (member (eq_lista eq_char) (the_var t) vs))
                        (append ss1 ss2))
                  (comp (shows_prec_ctxt _A2 (show_list show_char) Zero_nat c)
                    (comp (shows_string
                            [' '; 'h'; 'a'; 's'; ' '; 't'; 'o'; ' '; 'c'; 'o';
                              'n'; 't'; 'a'; 'i'; 'n'; ' '; 'o'; 'n'; 'l'; 'y';
                              ' '; 'f'; 'r'; 'e'; 's'; 'h'; ' '; 'v'; 'a'; 'r';
                              'i'; 'a'; 'b'; 'l'; 'e'; 's'])
                      shows_nl))))
      | More (_, ss1, More (_, _, _, _), ss2) ->
        Inl (comp (shows_prec_ctxt _A2 (show_list show_char) Zero_nat c)
              (comp (shows_string
                      [' '; 'i'; 's'; ' '; 'n'; 'o'; 't'; ' '; 'a'; ' '; 'f';
                        'l'; 'a'; 't'; ' '; 'c'; 'o'; 'n'; 't'; 'e'; 'x'; 't'])
                shows_nl)));;

let rec fcc_tt (_A1, _A2)
  r fcs lR =
    let vs = vars_trs_list (eq_lista eq_char) r in
    let fas = funas_trs_list _A1 r in
    bind (check (not (null fcs))
           (comp (shows_string
                   [' '; 'a'; 't'; ' '; 'l'; 'e'; 'a'; 's'; 't'; ' '; 'o'; 'n';
                     'e'; ' '; 'f'; 'l'; 'a'; 't'; ' '; 'c'; 'o'; 'n'; 't'; 'e';
                     'x'; 't'; ' '; 'i'; 's'; ' '; 'r'; 'e'; 'q'; 'u'; 'i'; 'r';
                     'e'; 'd'; ' '; 'f'; 'o'; 'r'; ' '; 't'; 'h'; 'e'; ' '; 'c';
                     'l'; 'o'; 's'; 'u'; 'r'; 'e'; ' '; 'u'; 'n'; 'd'; 'e'; 'r';
                     ' '; 'f'; 'l'; 'a'; 't'; ' '; 'c'; 'o'; 'n'; 't'; 'e'; 'x';
                     't'; 's'])
             shows_nl))
      (fun _ ->
        bind (catch (forallm (check_flat_ctxt (_A1, _A2) vs) fcs)
               (fun x -> Inl (snd x)))
          (fun _ ->
            bind (catch (forallm (check_is_flat_ctxt (_A1, _A2) vs fas) fcs)
                   (fun x -> Inl (snd x)))
              (fun _ ->
                bind (catch
                       (forallm (check_flat_ctxt_complete (_A1, _A2) fcs) fas)
                       (fun x -> Inl (snd x)))
                  (fun _ ->
                    bind (catch
                           (forallm (check_rule_preserving (_A1, _A2) fcs lR) r)
                           (fun x -> Inl (snd x)))
                      (fun _ ->
                        catch (forallm (check_rule_reflecting (_A1, _A2) fcs r)
                                lR)
                          (fun x -> Inl (snd x)))))));;

let rec sub_apply_term
  x0 sigma = match x0, sigma with Var x, sigma -> get_sub sigma x
    | Fun (f, ss), sigma -> Fun (f, map (fun t -> sub_apply_term t sigma) ss);;

let rec pi_full uu uv uw = true;;

type ('a, 'b, 'c) certificationProblem =
  TRSProof of (('a, 'c) term * ('a, 'c) term) list * ('a, 'b, 'c) tRSProof |
  TRSDisproof of (('a, 'c) term * ('a, 'c) term) list * ('a, 'c) tRSDisproof;;

let rec xml2trsDisproof
  xml2name xml2sname =
    xml1element
      ['t'; 'r'; 's'; 'N'; 'o'; 'n'; 't'; 'e'; 'r'; 'm'; 'i'; 'n'; 'a'; 't';
        'i'; 'o'; 'n'; 'P'; 'r'; 'o'; 'o'; 'f']
      (xml_options
        [(['v'; 'a'; 'r'; 'i'; 'a'; 'b'; 'l'; 'e'; 'C'; 'o'; 'n'; 'd'; 'i'; 't';
            'i'; 'o'; 'n'; 'V'; 'i'; 'o'; 'l'; 'a'; 't'; 'e'; 'd'],
           xml0elements
             ['v'; 'a'; 'r'; 'i'; 'a'; 'b'; 'l'; 'e'; 'C'; 'o'; 'n'; 'd'; 'i';
               't'; 'i'; 'o'; 'n'; 'V'; 'i'; 'o'; 'l'; 'a'; 't'; 'e'; 'd']
             NotWellFormed);
          (['l'; 'o'; 'o'; 'p'],
            xml_change (loop xml2name)
              (fun (s, (rseq, (a, b))) ->
                Loop (s, map (fun (x, (y, (_, z))) -> (x, (y, z))) rseq, a,
                       b)));
          (['r'; 'u'; 'l'; 'e'; 'R'; 'e'; 'm'; 'o'; 'v'; 'a'; 'l'],
            xml2elements ['r'; 'u'; 'l'; 'e'; 'R'; 'e'; 'm'; 'o'; 'v'; 'a'; 'l']
              (xml1element ['t'; 'r'; 's'] (rules xml2name) id)
              (xml2trsDisproof xml2name xml2sname)
              (fun a b -> NTRuleRemoval (a, b)));
          (['d'; 'p'; 'T'; 'r'; 'a'; 'n'; 's'],
            xml3elements ['d'; 'p'; 'T'; 'r'; 'a'; 'n'; 's']
              (xml1element ['d'; 'p'; 's'] (rules xml2sname) id)
              (boola
                ['m'; 'a'; 'r'; 'k'; 'e'; 'd'; 'S'; 'y'; 'm'; 'b'; 'o'; 'l';
                  's'])
              (xml2dpDisproof xml2sname)
              (fun p _ -> (fun a -> NTDPTrans (p, a))));
          (['s'; 't'; 'r'; 'i'; 'n'; 'g'; 'R'; 'e'; 'v'; 'e'; 'r'; 's'; 'a';
             'l'],
            xml2elements
              ['s'; 't'; 'r'; 'i'; 'n'; 'g'; 'R'; 'e'; 'v'; 'e'; 'r'; 's'; 'a';
                'l']
              (xml1element ['t'; 'r'; 's'] (rules xml2name) id)
              (xml2trsDisproof xml2name xml2sname)
              (fun a b -> NTStringReversal (a, b)))])
      id;;

let rec xml2certProblem _A
  xml2uname xml2usname xml2name xml2sname xml2lsname
    (Xml (name, egal, cs, texta)) =
    (if eq_list eq_char name
          ['c'; 'e'; 'r'; 't'; 'i'; 'f'; 'i'; 'c'; 'a'; 't'; 'i'; 'o'; 'n'; 'P';
            'r'; 'o'; 'b'; 'l'; 'e'; 'm']
          && (is_none texta &&
               less_eq_nat (nat (Big_int.big_int_of_int 3)) (size_list cs))
      then bind (xml2trsInput xml2uname (nth cs Zero_nat))
             (fun trs ->
               bind (text ['c'; 'p'; 'f'; 'V'; 'e'; 'r'; 's'; 'i'; 'o'; 'n']
                      (nth cs one_nat))
                 (fun version ->
                   (if eq_list eq_char
                         (take (nat (Big_int.big_int_of_int 2)) version)
                         ['2'; '.']
                     then xml1element ['p'; 'r'; 'o'; 'o'; 'f']
                            (xml_options
                              [(['t'; 'r'; 's'; 'T'; 'e'; 'r'; 'm'; 'i'; 'n';
                                  'a'; 't'; 'i'; 'o'; 'n'; 'P'; 'r'; 'o'; 'o';
                                  'f'],
                                 xml_change
                                   (xml2trsProof _A xml2name xml2sname
                                     xml2lsname)
                                   (fun a -> TRSProof (trs, a)));
                                (['t'; 'r'; 's'; 'N'; 'o'; 'n'; 't'; 'e'; 'r';
                                   'm'; 'i'; 'n'; 'a'; 't'; 'i'; 'o'; 'n'; 'P';
                                   'r'; 'o'; 'o'; 'f'],
                                  xml_change
                                    (xml2trsDisproof xml2uname xml2usname)
                                    (fun a -> TRSDisproof (trs, a)))])
                            id (nth cs (nat (Big_int.big_int_of_int 2)))
                     else Inl ['c'; 'p'; 'f'; '-'; 'v'; 'e'; 'r'; 's'; 'i'; 'o';
                                'n'; ' '; '2'; '.'; 'x'; ' '; 'r'; 'e'; 'q';
                                'u'; 'i'; 'r'; 'e'; 'd'])))
      else Inl ['c'; 'o'; 'u'; 'l'; 'd'; ' '; 'n'; 'o'; 't'; ' '; 'p'; 'a'; 'r';
                 's'; 'e'; ' '; 'r'; 'o'; 'o'; 't'; ' '; 'e'; 'l'; 'e'; 'm';
                 'e'; 'n'; 't']);;

type 'a filtered = FPair of 'a * nat;;

let rec check_subseteq _A
  xs ys =
    catch (forallm (fun x -> (if member _A x ys then Inr () else Inl x)) xs)
      (fun x -> Inl (snd x));;

let rec check_wf_trs _A (_B1, _B2)
  r = catch (catch
              (forallm
                (fun rule ->
                  catch (bind (check (not (is_var (fst rule)))
                                (shows_string
                                  ['v'; 'a'; 'r'; 'i'; 'a'; 'b'; 'l'; 'e'; ' ';
                                    'l'; 'e'; 'f'; 't'; '-'; 'h'; 'a'; 'n'; 'd';
                                    ' '; 's'; 'i'; 'd'; 'e'; ' '; 'i'; 'n']))
                          (fun _ ->
                            catch (check_subseteq _B1
                                    (vars_term_list _B1 (snd rule))
                                    (vars_term_list _B1 (fst rule)))
                              (fun x ->
                                Inl (comp (shows_string
    ['f'; 'r'; 'e'; 'e'; ' '; 'v'; 'a'; 'r'; 'i'; 'a'; 'b'; 'l'; 'e'; ' '])
                                      (comp (shows_prec _B2 Zero_nat x)
(shows_string
  [' '; 'i'; 'n'; ' '; 'r'; 'i'; 'g'; 'h'; 't'; '-'; 'h'; 'a'; 'n'; 'd'; ' ';
    's'; 'i'; 'd'; 'e'; ' '; 'o'; 'f']))))))
                    (fun x ->
                      Inl (comp x
                            (comp (shows_string [' '; 'r'; 'u'; 'l'; 'e'; ' '])
                              (comp (shows_rule _A _B2 rule) shows_nl)))))
                r)
              (fun x -> Inl (snd x)))
        (fun x ->
          Inl (comp (shows_string
                      ['t'; 'h'; 'e'; ' '; 'T'; 'R'; 'S'; ' '; 'i'; 's'; ' ';
                        'n'; 'o'; 't'; ' '; 'w'; 'e'; 'l'; 'l'; '-'; 'f'; 'o';
                        'r'; 'm'; 'e'; 'd'])
                (comp shows_nl x)));;

let rec check_DP (_A1, _A2) (_B1, _B2)
  r p = catch (bind (check_wf_trs _A2 (_B1, _B2) r)
                (fun _ ->
                  catch (check_subseteq
                          (eqc (eq_terma (eq_shpa _A1) _B1)
                            (eq_terma (eq_shpa _A1) _B1))
                          (dP_list _A1 _B1 r) p)
                    (fun x ->
                      Inl (comp (shows_string
                                  ['t'; 'h'; 'e'; ' '; 'D'; 'P'; ' '])
                            (comp (shows_rule (show_shp _A2) _B2 x)
                              (comp (shows_string
                                      [' '; 'd'; 'o'; 'e'; 's'; ' '; 'n'; 'o';
't'; ' '; 'a'; 'p'; 'p'; 'e'; 'a'; 'r'; ' '; 'i'; 'n'; ' '; 't'; 'h'; 'e'; ' ';
'D'; 'P'; ' '; 'p'; 'r'; 'o'; 'b'; 'l'; 'e'; 'm'])
                                shows_nl))))))
          (fun x ->
            Inl (comp (shows_string
                        ['t'; 'h'; 'e'; ' '; 'D'; 'P'; '-'; 't'; 'r'; 'a'; 'n';
                          's'; 'f'; 'o'; 'r'; 'm'; 'a'; 't'; 'i'; 'o'; 'n'; ' ';
                          'i'; 's'; ' '; 'n'; 'o'; 't'; ' '; 'a'; 'p'; 'p'; 'l';
                          'i'; 'e'; 'd'; ' '; 'c'; 'o'; 'r'; 'r'; 'e'; 'c'; 't';
                          'l'; 'y'; '.'])
                  (comp shows_nl x)));;

let rec ge_nat x y = less_eq_nat y x;;

let rec mat_mono sd m = mat_monoI m (Suc (minus_nat sd one_nat));;

let nat_name : char list = ['n'; 'a'; 't'; 'u'; 'r'; 'a'; 'l'; 's'];;

let rec create_afI _A _B
  ze x1 f n i = match ze, x1, f, n, i with ze, [], f, n, i -> true
    | ze, ((g, m), (c, coeffs)) :: rest, f, n, i ->
        (if eqa _B f g && eq_nat m n
          then less_nat i (size_list coeffs) && not (eqa _A (nth coeffs i) ze)
          else create_afI _A _B ze rest f n i);;

let rec lpo_repr x = shows_string ['L'; 'P'; 'O'; '+'; 'A'; 'F'] x;;

let rec check_sem_lab_proc (_A1, _A2) (_B1, _B2) (_C1, _C2)
  check_lab check_model_lab gr i l p r lP lR =
    bind (catch (check_lab lP p)
           (fun x ->
             Inl (comp (shows_string
                         ['n'; 'o'; 't'; ' '; 'a'; 'l'; 'l'; ' '; 'l'; 'a'; 'b';
                           'e'; 'l'; 'e'; 'd'; ' '; 'D'; 'P'; 's'; ' '; 'p';
                           'r'; 'e'; 's'; 'e'; 'n'; 't'; ':'])
                   (comp shows_nl x))))
      (fun _ ->
        bind (catch (check_model_lab lR r)
               (fun x ->
                 Inl (comp (shows_string
                             ['p'; 'r'; 'o'; 'b'; 'l'; 'e'; 'm'; ' '; 'w'; 'i';
                               't'; 'h'; ' '; 'l'; 'a'; 'b'; 'e'; 'l'; 'e'; 'd';
                               ' '; 'T'; 'R'; 'S'; ':'])
                       (comp shows_nl x))))
          (fun _ ->
            catch (check_sl_only_lab_and_decr (_B1, _B2) (_A1, _A2) (_C1, _C2)
                    gr r lR)
              (fun x -> Inl (comp x shows_nl))));;

let rec rl_proc (_A1, _A2) (_B1, _B2)
  p r lP lR =
    let fs =
      list_union (eq_laba _A1 _B1) (funs_trs_list (eq_laba _A1 _B1) r)
        (list_diff (eq_laba _A1 _B1) (funs_trs_list (eq_laba _A1 _B1) p)
          (defs_list (eq_laba _A1 _B1) p))
      in
    bind (check (not (null fs))
           (comp (shows_string
                   ['r'; 'o'; 'o'; 't'; '-'; 'l'; 'a'; 'b'; 'e'; 'l'; 'i'; 'n';
                     'g'; ' '; 'r'; 'e'; 'q'; 'u'; 'i'; 'r'; 'e'; 's'; ' '; 'a';
                     't'; ' '; 'l'; 'e'; 'a'; 's'; 't'; ' '; 'o'; 'n'; 'e'; ' ';
                     'f'; 'u'; 'n'; 'c'; 't'; 'i'; 'o'; 'n'; ' '; 's'; 'y'; 'm';
                     'b'; 'o'; 'l'; ' '; 'i'; 'n'; ' '; 't'; 'h'; 'e'; ' '; 's';
                     'i'; 'g'; 'n'; 'a'; 't'; 'u'; 'r'; 'e'])
             shows_nl))
      (fun _ ->
        let d = hd fs in
        check_sem_lab_proc (_A1, _A2) (_B1, _B2)
          (eq_lista eq_char, show_list show_char)
          (check_sl_lab_trs (_A1, _A2) (_B1, _B2)
            (eq_laba _A1 _B1, show_lab _A2 _B2)
            (eq_lista eq_char, show_list show_char)
            (rl_inter_list (eq_laba _A1 _B1) fs d) rl_fun fs)
          (check_sl_model_lab_trs (_A1, _A2) (_B1, _B2)
            (eq_laba _A1 _B1, show_lab _A2 _B2)
            (eq_lista eq_char, show_list show_char)
            (rl_inter_list (eq_laba _A1 _B1) fs d) rl_fun fs)
          (bot_fun bot_boola) (rl_inter_list (eq_laba _A1 _B1) fs d) rl_fun p r
          lP lR);;

let rec funas_term_list _A t = remdups (eqc _A eq_nata) (funas_term_list' t);;

let rec xmldoc2certProblem
  (Xmldoc (header, xml)) =
    debug ['0']
      ['p'; 'a'; 'r'; 's'; 'i'; 'n'; 'g'; ' '; 'x'; 'm'; 'l'; ' '; 't'; 'o';
        ' '; 'c'; 'e'; 'r'; 't'; 'i'; 'f'; 'i'; 'c'; 'a'; 't'; 'i'; 'o'; 'n';
        ' '; 'p'; 'r'; 'o'; 'b'; 'l'; 'e'; 'm']
      (xml2certProblem (eq_lista eq_char) (text ['n'; 'a'; 'm'; 'e']) sharp_name
        plain_name sharp_lab_name lab_sharp_name xml);;

let rec parseCertProblem s = bind (doc_of_string s) xmldoc2certProblem;;

let rec show_pvars (_A1, _A2) _B
  on x1 = match on, x1 with on, [] -> []
    | on, (x, c) :: vas ->
        append (if eqa _A1 on c then [] else shows_prec _A2 Zero_nat c [])
          (shows_prec _B Zero_nat x [])
          :: show_pvars (_A1, _A2) _B on vas;;

let rec af_entry_case
  f1 f2 x2 = match f1, f2, x2 with f1, f2, AFList lista -> f2 lista
    | f1, f2, Collapse nat -> f1 nat;;

let rec shows_afs _A
  af = shows_map
         (fun (a, b) ->
           let (f, n) = a in
           (fun e ->
             comp (shows_string ['p'; 'i'; '('])
               (comp (shows_prec _A Zero_nat f)
                 (comp (shows_string ['/'])
                   (comp (shows_prec_nat Zero_nat n)
                     (comp (shows_string [')'; ' '; '='; ' '])
                       (comp (match e
                               with Collapse i ->
                                 shows_prec_nat Zero_nat (Suc i)
                               | AFList ids ->
                                 shows_prec_list show_nat Zero_nat
                                   (map (fun aa -> Suc aa) ids))
                         shows_nl))))))
             b)
         af;;

let rec block_term
  f x1 = match f, x1 with f, Var x -> Var x
    | f, Fun (g, ts) -> Fun (g, map (fun t -> Fun (f, [t])) ts);;

let rec block_rule f r = (block_term f (fst r), block_term f (snd r));;

let rec check_superset_of_blocked (_A1, _A2)
  f r' r =
    catch (catch
            (catch
              (forallm
                (fun x ->
                  (if member
                        (eqc (eq_terma _A1 (eq_lista eq_char))
                          (eq_terma _A1 (eq_lista eq_char)))
                        (block_rule f x) r'
                    then Inr () else Inl x))
                r)
              (fun x -> Inl (snd x)))
            (fun x ->
              Inl (comp (shows_string
                          ['t'; 'h'; 'e'; ' '; 'r'; 'u'; 'l'; 'e'; ' '])
                    (comp (shows_rule _A2 (show_list show_char)
                            (block_rule f x))
                      (comp (shows_string
                              [' '; 'i'; 's'; ' '; 'm'; 'i'; 's'; 's'; 'i'; 'n';
                                'g'])
                        shows_nl)))))
      (fun x ->
        Inl (comp (shows_trs _A2 (show_list show_char) r)
              (comp (shows_string
                      ['i'; 's'; ' '; 'n'; 'o'; 't'; ' '; 'a'; ' '; 's'; 'u';
                        'b'; 's'; 'e'; 't'; ' '; 'o'; 'f'])
                (comp (shows_trs _A2 (show_list show_char) r')
                  (comp x shows_nl)))));;

let rec check_no_defined_root (_A1, _A2) _B
  fs t =
    check (not (member _A1 (the (root t)) fs))
      (comp (shows_string
              ['t'; 'h'; 'e'; ' '; 'r'; 'o'; 'o'; 't'; ' '; 'o'; 'f'; ' '])
        (comp (shows_prec_term _A2 _B Zero_nat t)
          (comp (shows_string
                  [' '; 'i'; 's'; ' '; 'd'; 'e'; 'f'; 'i'; 'n'; 'e'; 'd'; ' ';
                    'w'; '.'; 'r'; '.'; 't'; '.'; ' '; 't'; 'h'; 'e'; ' '; 's';
                    'i'; 'g'; 'n'; 'a'; 't'; 'u'; 'r'; 'e'])
            (comp shows_nl
              (comp (shows_prec_list _A2 Zero_nat fs) shows_nl)))));;

let rec check_no_var _A _B
  t = check (not (is_var t))
        (comp (shows_string
                ['v'; 'a'; 'r'; 'i'; 'a'; 'b'; 'l'; 'e'; ' '; 'f'; 'o'; 'u';
                  'n'; 'd'])
          shows_nl);;

let rec check_left_linear_trs _A
  trs = check (forall (fun r -> linear_term (eq_lista eq_char) (fst r)) trs)
          (comp (shows_prec_list
                  (show (show_term _A (show_list show_char))
                    (show_term _A (show_list show_char)))
                  Zero_nat trs)
            (comp shows_nl
              (comp (shows_string
                      ['i'; 's'; ' '; 'n'; 'o'; 't'; ' '; 'l'; 'e'; 'f'; 't';
                        '-'; 'l'; 'i'; 'n'; 'e'; 'a'; 'r'])
                shows_nl)));;

let rec funas_below_term_list' t = flat_map funas_term_list' (args t);;

let rec funas_below_rule_list'
  r = append (funas_below_term_list' (fst r)) (funas_below_term_list' (snd r));;

let rec funas_below_trs_list _A
  trs = remdups (eqc _A eq_nata) (flat_map funas_below_rule_list' trs);;

let rec fcc_proc (_A1, _A2)
  p r f fcs p' r' =
    let ds = defs_list _A1 r in
    let fa = (f, Suc Zero_nat) in
    let cf = More (f, [], Hole, []) in
    let fcs' = cf :: fcs in
    let vs = vars_trs_list (eq_lista eq_char) r in
    let fs =
      list_union (eqc _A1 eq_nata) (funas_below_trs_list _A1 p)
        (funas_trs_list _A1 r)
      in
    let fas = fa :: fs in
    bind (check (not (member (eqc _A1 eq_nata) fa fs))
           (comp (shows_preca _A2 show_nat Zero_nat fa)
             (comp (shows_string
                     [' '; 'i'; 's'; ' '; 'n'; 'o'; 't'; ' '; 'f'; 'r'; 'e';
                       's'; 'h'])
               shows_nl)))
      (fun _ ->
        bind (check (not (member _A1 f ds))
               (comp (shows_prec _A2 Zero_nat f)
                 (comp (shows_string
                         [' '; 'i'; 's'; ' '; 'd'; 'e'; 'f'; 'i'; 'n'; 'e'; 'd';
                           ' '; 'i'; 'n'; ' '; 't'; 'h'; 'e'; ' '; 'r'; 'u';
                           'l'; 'e'; 's'])
                   shows_nl)))
          (fun _ ->
            bind (check_left_linear_trs _A2 r)
              (fun _ ->
                bind (check_wf_trs _A2 (eq_lista eq_char, show_list show_char)
                       r)
                  (fun _ ->
                    bind (catch
                           (forallm
                             (fun ra ->
                               bind (bind (check_no_var _A2
    (show_list show_char) (fst ra))
                                      (fun _ ->
check_no_var _A2 (show_list show_char) (snd ra)))
                                 (fun _ ->
                                   check_no_defined_root (_A1, _A2)
                                     (show_list show_char) ds (snd ra)))
                             p)
                           (fun x -> Inl (snd x)))
                      (fun _ ->
                        bind (catch
                               (forallm (check_flat_ctxt (_A1, _A2) vs) fcs')
                               (fun x -> Inl (snd x)))
                          (fun _ ->
                            bind (catch
                                   (forallm
                                     (check_is_flat_ctxt (_A1, _A2) vs fas)
                                     fcs')
                                   (fun x -> Inl (snd x)))
                              (fun _ ->
                                bind (catch
                                       (forallm
 (check_flat_ctxt_complete (_A1, _A2) fcs') fas)
                                       (fun x -> Inl (snd x)))
                                  (fun _ ->
                                    bind (catch
   (forallm (check_rule_preserving (_A1, _A2) fcs' r') r)
   (fun x -> Inl (snd x)))
                                      (fun _ ->
bind (catch (forallm (check_rule_reflecting (_A1, _A2) fcs' r) r')
       (fun x -> Inl (snd x)))
  (fun _ -> check_superset_of_blocked (_A1, _A2) f p' p))))))))));;

let rec get_arg
  t x1 = match t, x1 with t, Zero_nat -> t
    | t, Suc n -> nth (args t) n;;

let rec ctxt_of_pos_term
  x0 t = match x0, t with Empty, t -> Hole
    | PCons (i, ps), Fun (f, ts) ->
        More (f, take i ts, ctxt_of_pos_term ps (nth ts i), drop (Suc i) ts);;

let rec int_mono x = Big_int.le_big_int (Big_int.big_int_of_int 1) x;;

let rec max0_nat x = x;;

let rec nat_mono x = not (eq_nat x Zero_nat);;

let rec rat_mono
  x = less_eq_rat
        (Fract ((Big_int.big_int_of_int 1), (Big_int.big_int_of_int 1))) x;;

let rec shows_arctic
  = function Num_arc i -> shows_prec_int Zero_nat i
    | MinInfty -> shows_string ['-'; 'i'; 'n'; 'f'];;

let rec shows_prec_arctic d ai = shows_arctic ai;;

let rec shows_list_arctic
  asa = shows_list_aux (shows_prec_arctic Zero_nat) asa;;

let show_arctic =
  ({shows_prec = shows_prec_arctic; shows_list = shows_list_arctic} :
    arctic show);;

let rec lookup_rest _A
  x xa1 = match x, xa1 with x, [] -> None
    | x, (y, c) :: ycs ->
        (if eqa _A x y then Some (c, ycs)
          else (match lookup_rest _A x ycs with None -> None
                 | Some (d, yccs) -> Some (d, (y, c) :: yccs)));;

let rec shows_lpoly (_A1, _A2) _B
  ze on (LPoly (c, cs)) =
    (match show_pvars (_A1, _A2) _B on cs with [] -> shows_prec _A2 Zero_nat c
      | a :: lista ->
        comp (if eqa _A1 c ze then id
               else comp (shows_prec _A2 Zero_nat c)
                      (shows_string [' '; '+'; ' ']))
          (shows_list_gen (shows_prec_list show_char Zero_nat) [] []
            [' '; '+'; ' '] [] (a :: lista)));;

let rec matchList _A
  = function [] -> Some []
    | (Var uu, Fun (uv, uw)) :: ps -> None
    | (Fun (f, ts), Fun (g, ls)) :: ps ->
        (if eqa _A f g
          then (match decomp ts ls with None -> None
                 | Some pairs -> matchList _A (append pairs ps))
          else None)
    | (t, Var x) :: ps ->
        (match matchList _A ps with None -> None
          | Some pairs -> Some ((x, t) :: pairs));;

let rec matchVars _A _B
  = function [] -> Some sub_empty
    | (x, t) :: ps ->
        (match lookup _A x ps
          with None ->
            (match matchVars _A _B ps with None -> None
              | Some sigma -> Some (Sub (fun_upd _A (get_sub sigma) x t)))
          | Some s ->
            (if eq_term _B _A t s then matchVars _A _B ps else None));;

let rec apply_af_entry
  uu x1 ts = match uu, x1, ts with uu, Collapse i, ts -> nth ts i
    | f, AFList is, ts -> Fun (f, map (nth ts) is);;

let rec afs_term
  pi x1 = match pi, x1 with
    pi, Fun (f, ts) ->
      let l = size_list ts in
      apply_af_entry (FPair (f, l)) (pi f l) (map (afs_term pi) ts)
    | pi, Var x -> Var x;;

let rec search_len
  t s n =
    (if eq_list eq_char (take n t) s then Some ([], drop n t)
      else (match t with [] -> None
             | h :: tt ->
               (match search_len tt s n with None -> None
                 | Some (bef, aft) -> Some (h :: bef, aft))));;

let rec reverse_trs r = map (fun lr -> (snd lr, fst lr)) r;;

let rec fresh_strings_list
  name offset used n =
    take n
      (filter
        (fun s ->
          not (member (eq_lista eq_char) s (remdups (eq_lista eq_char) used)))
        (map (fun i ->
               append name (shows_prec_nat Zero_nat (plus_nat i offset) []))
          (upt Zero_nat
            (plus_nat n (length_unique (eq_lista eq_char) used)))));;

let rec subst_of _A
  xts = Sub (fun x ->
              (match lookup _A x xts with None -> Var x | Some t -> t));;

let rec check_pvarsI _B
  ze ge vas x3 = match ze, ge, vas, x3 with
    ze, ge, vas, [] ->
      catch (forallm (fun va -> check (ge (snd va) ze) (fst va)) vas)
        (fun x -> Inl (snd x))
    | ze, ge, vas, (x, b) :: vbs ->
        let (a', vas') =
          (match lookup_rest _B x vas with None -> (ze, vas)
            | Some (a, ba) -> (a, ba))
          in
        bind (check (ge a' b) x) (fun _ -> check_pvarsI _B ze ge vas' vbs);;

let rec check_lpoly_sI (_A1, _A2) (_B1, _B2)
  mode ze on ge gt (LPoly (a, vas)) (LPoly (b, vbs)) =
    catch (bind (check (gt a b)
                  (shows_string
                    ['p'; 'r'; 'o'; 'b'; 'l'; 'e'; 'm'; ' '; 'w'; 'h'; 'e'; 'n';
                      ' '; 'c'; 'o'; 'm'; 'p'; 'a'; 'r'; 'i'; 'n'; 'g'; ' ';
                      'c'; 'o'; 'n'; 's'; 't'; 'a'; 'n'; 't'; ' '; 'p'; 'a';
                      'r'; 't']))
            (fun _ ->
              catch (check_pvarsI _B1 ze (if mode then ge else gt) vas vbs)
                (fun x ->
                  Inl (comp (shows_string
                              ['p'; 'r'; 'o'; 'b'; 'l'; 'e'; 'm'; ' '; 'w'; 'h';
                                'e'; 'n'; ' '; 'c'; 'o'; 'm'; 'p'; 'a'; 'r';
                                'i'; 'n'; 'g'; ' '; 'c'; 'o'; 'e'; 'f'; 'f';
                                'i'; 'c'; 'i'; 'e'; 'n'; 't'; 's'; ' '; 'o';
                                'f'; ' '; 'v'; 'a'; 'r'; 'i'; 'a'; 'b'; 'l';
                                'e'; ' '])
                        (shows_prec _B2 Zero_nat x)))))
      (fun x ->
        Inl (comp (shows_string
                    ['p'; 'r'; 'o'; 'b'; 'l'; 'e'; 'm'; ' '; 'w'; 'h'; 'e'; 'n';
                      ' '; 'c'; 'o'; 'm'; 'p'; 'a'; 'r'; 'i'; 'n'; 'g'; ' '])
              (comp (shows_lpoly (_A1, _A2) _B2 ze on (LPoly (a, vas)))
                (comp (shows_string [' '; '>'; ' '])
                  (comp (shows_lpoly (_A1, _A2) _B2 ze on (LPoly (b, vbs)))
                    (comp shows_nl x))))));;

let rec check_polo_sI (_A1, _A2) _B (_C1, _C2)
  mode ze on pl ti mz ge gt pi (s, t) =
    let left = pleftI _A1 _C1 ze on pl ti mz pi s in
    let right = prightI _A1 _C1 ze on pl ti mz pi t in
    catch (check_lpoly_sI (_A1, _A2) (_C1, _C2) mode ze on ge gt left right)
      (fun x ->
        Inl (comp (shows_string
                    ['c'; 'o'; 'u'; 'l'; 'd'; ' '; 'n'; 'o'; 't'; ' '; 'e'; 'n';
                      's'; 'u'; 'r'; 'e'; ' '])
              (comp (shows_prec_term _B _C2 Zero_nat s)
                (comp (shows_string [' '; '>'; ' '])
                  (comp (shows_prec_term _B _C2 Zero_nat t)
                    (comp shows_nl x))))));;

let rec rev_term
  c x1 = match c, x1 with c, Var x -> ctxt_apply c (Var x)
    | c, Fun (f, [t]) -> rev_term (More (f, [], c, [])) t;;

let rec rev_rule (l, r) = (rev_term Hole l, rev_term Hole r);;

let rec match_list _A _B
  tls = (match matchList _A tls with None -> None
          | Some a -> matchVars _B _A a);;

let rec filtered_fun (FPair (f, n)) = f;;

let rec shows_prec_filtered _A d ff = shows_prec _A Zero_nat (filtered_fun ff);;

let rec shows_list_filtered _A
  ffs = shows_list_aux (shows_prec_filtered _A Zero_nat) ffs;;

let rec show_filtered _A =
  ({shows_prec = shows_prec_filtered _A; shows_list = shows_list_filtered _A} :
    'a filtered show);;

let rec afs_check _A _B
  rel pi g lr =
    let pl = afs_term pi (fst lr) in
    let pr = afs_term pi (snd lr) in
    catch (g (pl, pr))
      (fun x ->
        Inl (comp (shows_string
                    ['c'; 'o'; 'u'; 'l'; 'd'; ' '; 'n'; 'o'; 't'; ' '; 'o'; 'r';
                      'i'; 'e'; 'n'; 't'; ' '])
              (comp (shows_prec_term _A _B Zero_nat (fst lr))
                (comp (shows_string [' '])
                  (comp rel
                    (comp (shows_string [' '])
                      (comp (shows_prec_term _A _B Zero_nat (snd lr))
                        (comp shows_nl
                          (comp (shows_string ['p'; 'i'; '('; ' '])
                            (comp (shows_prec_term _A _B Zero_nat (fst lr))
                              (comp (shows_string [' '; ')'; ' '; '='; ' '])
                                (comp (shows_prec_term (show_filtered _A) _B
Zero_nat pl)
                                  (comp shows_nl
                                    (comp (shows_string ['p'; 'i'; '('; ' '])
                                      (comp
(shows_prec_term _A _B Zero_nat (snd lr))
(comp (shows_string [' '; ')'; ' '; '='; ' '])
  (comp (shows_prec_term (show_filtered _A) _B Zero_nat pr)
    (comp shows_nl x))))))))))))))))));;

let rec starts_with
  t s = (if not (null t) && member eq_char (hd t) wspace
          then starts_with (tl t) s
          else (match search_len t s (size_list s) with None -> false
                 | Some _ -> true));;

let rec check_pairwise
  c x1 = match c, x1 with c, [] -> Inr ()
    | c, x :: xs ->
        bind (catch (forallm (c x) xs) (fun xa -> Inl (snd xa)))
          (fun _ -> check_pairwise c xs);;

let rec check_graph_decomp
  ss g rcs =
    bind (check_pairwise (check_edges ss g) (map snd rcs))
      (fun _ ->
        catch (forallm (fun c -> check_edges ss g c c)
                (map snd (filter (fun rc -> not (fst rc)) rcs)))
          (fun x -> Inl (snd x)));;

let rec check_DPs (_A1, _A2) (_B1, _B2)
  r p = catch (forallm
                (fun x ->
                  (if let (ll, rr) = x in
                      any (fun (l, ra) ->
                            eq_term _A1 _B1 l (map_term unsharp_fun ll) &&
                              supteqp _A1 _B1 ra (map_term unsharp_fun rr))
                        r
                    then Inr () else Inl x))
                p)
          (fun x -> Inl (snd x));;

let rec check_lpoly_nsI (_A1, _A2) (_B1, _B2)
  ze on ge (LPoly (a, vas)) (LPoly (b, vbs)) =
    catch (bind (check (ge a b)
                  (shows_string
                    ['p'; 'r'; 'o'; 'b'; 'l'; 'e'; 'm'; ' '; 'w'; 'h'; 'e'; 'n';
                      ' '; 'c'; 'o'; 'm'; 'p'; 'a'; 'r'; 'i'; 'n'; 'g'; ' ';
                      'c'; 'o'; 'n'; 's'; 't'; 'a'; 'n'; 't'; ' '; 'p'; 'a';
                      'r'; 't'; 's']))
            (fun _ ->
              catch (check_pvarsI _B1 ze ge vas vbs)
                (fun x ->
                  Inl (comp (shows_string
                              ['p'; 'r'; 'o'; 'b'; 'l'; 'e'; 'm'; ' '; 'w'; 'h';
                                'e'; 'n'; ' '; 'c'; 'o'; 'm'; 'p'; 'a'; 'r';
                                'i'; 'n'; 'g'; ' '; 'c'; 'o'; 'e'; 'f'; 'f';
                                'i'; 'c'; 'i'; 'e'; 'n'; 't'; 's'; ' '; 'o';
                                'f'; ' '; 'v'; 'a'; 'r'; 'i'; 'a'; 'b'; 'l';
                                'e'; ' '])
                        (shows_prec _B2 Zero_nat x)))))
      (fun x ->
        Inl (comp (shows_string
                    ['p'; 'r'; 'o'; 'b'; 'l'; 'e'; 'm'; ' '; 'w'; 'h'; 'e'; 'n';
                      ' '; 'c'; 'o'; 'm'; 'p'; 'a'; 'r'; 'i'; 'n'; 'g'; ' '])
              (comp (shows_lpoly (_A1, _A2) _B2 ze on (LPoly (a, vas)))
                (comp (shows_string [' '; '>'; '='; ' '])
                  (comp (shows_lpoly (_A1, _A2) _B2 ze on (LPoly (b, vbs)))
                    (comp shows_nl x))))));;

let rec check_polo_nsI (_A1, _A2) _B (_C1, _C2)
  ze on pl ti mz ge pi (s, t) =
    let left = pleftI _A1 _C1 ze on pl ti mz pi s in
    let right = prightI _A1 _C1 ze on pl ti mz pi t in
    catch (check_lpoly_nsI (_A1, _A2) (_C1, _C2) ze on ge left right)
      (fun x ->
        Inl (comp (shows_string
                    ['c'; 'o'; 'u'; 'l'; 'd'; ' '; 'n'; 'o'; 't'; ' '; 'e'; 'n';
                      's'; 'u'; 'r'; 'e'; ' '])
              (comp (shows_prec_term _B _C2 Zero_nat s)
                (comp (shows_string [' '; '>'; '='; ' '])
                  (comp (shows_prec_term _B _C2 Zero_nat t)
                    (comp shows_nl x))))));;

let rec to_lpoly_inter _B
  d on i f n =
    (match lookup (eqc _B eq_nata) (f, n) i with None -> (d, replicate n on)
      | Some cas -> cas);;

let rec shows_lpo_repr _A
  prs = comp (shows_string
               ['L'; 'P'; 'O'; ' '; 'w'; 'i'; 't'; 'h'; ' '; 't'; 'h'; 'e'; ' ';
                 'f'; 'o'; 'l'; 'l'; 'o'; 'w'; 'i'; 'n'; 'g'; ' '; 'p'; 'r';
                 'e'; 'c'; 'e'; 'd'; 'e'; 'n'; 'c'; 'e'])
          (comp shows_nl
            (comp (shows_map
                    (fun (a, b) ->
                      let (f, n) = a in
                      (fun pr ->
                        comp (shows_string
                               ['p'; 'r'; 'e'; 'c'; 'e'; 'd'; 'e'; 'n'; 'c';
                                 'e'; '('])
                          (comp (shows_prec _A Zero_nat f)
                            (comp (shows_string ['['])
                              (comp (shows_prec_nat Zero_nat n)
                                (comp (shows_string [']'; ')'; ' '; '='; ' '])
                                  (comp (shows_prec_nat Zero_nat pr)
                                    shows_nl))))))
                        b)
                    prs)
              (comp shows_nl
                (comp (shows_string
                        ['p'; 'r'; 'e'; 'c'; 'e'; 'd'; 'e'; 'n'; 'c'; 'e'; '(';
                          '_'; ')'; ' '; '='; ' '; '0'])
                  shows_nl))));;

let rec shows_lpo_af _A
  repr af =
    comp (shows_lpo_repr _A repr)
      (comp shows_nl
        (comp (shows_string
                ['i'; 'n'; ' '; 'c'; 'o'; 'm'; 'b'; 'i'; 'n'; 'a'; 't'; 'i';
                  'o'; 'n'; ' '; 'w'; 'i'; 't'; 'h'; ' '; 't'; 'h'; 'e'; ' ';
                  'f'; 'o'; 'l'; 'l'; 'o'; 'w'; 'i'; 'n'; 'g'; ' '; 'a'; 'r';
                  'g'; 'u'; 'm'; 'e'; 'n'; 't'; ' '; 'f'; 'i'; 'l'; 't'; 'e';
                  'r'])
          (comp shows_nl (shows_afs _A af))));;

let rec extract_input_proof
  s = (match
        search_len s ['<'; 'i'; 'n'; 'p'; 'u'; 't'; '>']
          (size_list ['<'; 'i'; 'n'; 'p'; 'u'; 't'; '>'])
        with None -> None
        | Some (_, minput) ->
          (match
            search_len minput ['<'; '/'; 'i'; 'n'; 'p'; 'u'; 't'; '>']
              (size_list ['<'; '/'; 'i'; 'n'; 'p'; 'u'; 't'; '>'])
            with None -> None
            | Some (the_input, s') ->
              (match
                search_len s' ['<'; 'p'; 'r'; 'o'; 'o'; 'f'; '>']
                  (size_list ['<'; 'p'; 'r'; 'o'; 'o'; 'f'; '>'])
                with None -> None
                | Some (version, prf) -> Some (the_input, (version, prf)))));;

let rec shows_pos
  = function PCons (i, Empty) -> shows_prec_nat Zero_nat i
    | PCons (i, PCons (v, va)) ->
        comp (shows_prec_nat Zero_nat i) (shows_pos (PCons (v, va)))
    | Empty -> shows_string ['e'];;

let rec shows_prec_pos d p = shows_pos p;;

let rec check_rstep (_A1, _A2) (_B1, _B2)
  ra p r s t =
    let sps = poss_list s in
    let tps = poss_list t in
    bind (check (member (eqc (eq_terma _A1 _B1) (eq_terma _A1 _B1)) r ra)
           (comp (shows_rule _A2 _B2 r)
             (comp (shows_string
                     [' '; 'i'; 's'; ' '; 'n'; 'o'; 't'; ' '; 'a'; ' '; 'r';
                       'u'; 'l'; 'e'; ' '; 'o'; 'f'])
               (comp shows_nl (comp (shows_trs _A2 _B2 ra) shows_nl)))))
      (fun _ ->
        bind (check (member eq_posa p sps)
               (comp (shows_prec_pos Zero_nat p)
                 (comp (shows_string
                         [' '; 'i'; 's'; ' '; 'n'; 'o'; 't'; ' '; 'a'; ' '; 'p';
                           'o'; 's'; 'i'; 't'; 'i'; 'o'; 'n'; ' '; 'o'; 'f';
                           ' '])
                   (comp (shows_prec_term _A2 _B2 Zero_nat s) shows_nl))))
          (fun _ ->
            bind (check (member eq_posa p tps)
                   (comp (shows_prec_pos Zero_nat p)
                     (comp (shows_string
                             [' '; 'i'; 's'; ' '; 'n'; 'o'; 't'; ' '; 'a'; ' ';
                               'p'; 'o'; 's'; 'i'; 't'; 'i'; 'o'; 'n'; ' '; 'o';
                               'f'; ' '])
                       (comp (shows_prec_term _A2 _B2 Zero_nat t) shows_nl))))
              (fun _ ->
                let c = ctxt_of_pos_term p s in
                let d = ctxt_of_pos_term p t in
                let u = subt_at s p in
                let v = subt_at t p in
                (match match_list _A1 _B1 [(u, fst r); (v, snd r)]
                  with None ->
                    Inl (comp (shows_string
                                ['t'; 'h'; 'e'; ' '; 't'; 'e'; 'r'; 'm'; ' '])
                          (comp (shows_prec_term _A2 _B2 Zero_nat t)
                            (comp (shows_string
                                    [' '; 'd'; 'o'; 'e'; 's'; ' '; 'n'; 'o';
                                      't'; ' '; 'r'; 'e'; 's'; 'u'; 'l'; 't';
                                      ' '; 'f'; 'r'; 'o'; 'm'; ' '; 'a'; ' ';
                                      'p'; 'r'; 'o'; 'p'; 'e'; 'r'; ' '; 'a';
                                      'p'; 'p'; 'l'; 'i'; 'c'; 'a'; 't'; 'i';
                                      'o'; 'n'; ' '; 'o'; 'f'; ' '; 'r'; 'u';
                                      'l'; 'e'; ' '])
                              (comp shows_nl
                                (comp (shows_rule _A2 _B2 r)
                                  (comp (shows_string
  [' '; 'a'; 't'; ' '; 'p'; 'o'; 's'; 'i'; 't'; 'i'; 'o'; 'n'; ' '])
                                    (comp (shows_prec_pos Zero_nat p)
                                      shows_nl)))))))
                  | Some _ ->
                    check (eq_ctxt _A1 _B1 c d)
                      (comp (shows_string
                              ['t'; 'h'; 'e'; ' '; 't'; 'e'; 'r'; 'm'; ' '])
                        (comp (shows_prec_term _A2 _B2 Zero_nat t)
                          (comp (shows_string
                                  [' '; 'd'; 'o'; 'e'; 's'; ' '; 'n'; 'o'; 't';
                                    ' '; 'r'; 'e'; 's'; 'u'; 'l'; 't'; ' '; 'f';
                                    'r'; 'o'; 'm'; ' '; 'a'; ' '; 'p'; 'r'; 'o';
                                    'p'; 'e'; 'r'; ' '; 'a'; 'p'; 'p'; 'l'; 'i';
                                    'c'; 'a'; 't'; 'i'; 'o'; 'n'; ' '; 'o'; 'f';
                                    ' '; 'r'; 'u'; 'l'; 'e'; ' '])
                            (comp shows_nl
                              (comp (shows_rule _A2 _B2 r)
                                (comp (shows_string
[' '; 'a'; 't'; ' '; 'p'; 'o'; 's'; 'i'; 't'; 'i'; 'o'; 'n'; ' '])
                                  (comp (shows_prec_pos Zero_nat p)
                                    shows_nl)))))))))));;

let rec check_rrstep (_A1, _A2) (_B1, _B2)
  ra r s t = check_rstep (_A1, _A2) (_B1, _B2) ra Empty r s t;;

let rec check_steps (_A1, _A2) (_B1, _B2)
  p r x2 s u = match p, r, x2, s, u with
    p, r, [], s, u ->
      check (eq_term _A1 _B1 s u)
        (comp (shows_string
                ['t'; 'h'; 'e'; ' '; 'l'; 'a'; 's'; 't'; ' '; 't'; 'e'; 'r';
                  'm'; ' '; 'o'; 'f'; ' '; 't'; 'h'; 'e'; ' '; 'r'; 'e'; 'w';
                  'r'; 'i'; 't'; 'e'; ' '; 's'; 'e'; 'q'; 'u'; 'e'; 'n'; 'c';
                  'e'])
          (comp shows_nl
            (comp (shows_prec_term _A2 _B2 Zero_nat s)
              (comp (shows_string
                      (shows_nl
                        ['d'; 'o'; 'e'; 's'; ' '; 'n'; 'o'; 't'; ' '; 'c'; 'o';
                          'r'; 'r'; 'e'; 's'; 'p'; 'o'; 'n'; 'd'; ' '; 't'; 'o';
                          ' '; 't'; 'h'; 'e'; ' '; 'g'; 'o'; 'a'; 'l'; ' '; 't';
                          'e'; 'r'; 'm']))
                (comp shows_nl
                  (comp (shows_prec_term _A2 _B2 Zero_nat u) shows_nl))))))
    | p, ra, (uu, (r, (true, t))) :: prts, s, u ->
        bind (check_rrstep (_A1, _A2) (_B1, _B2) p r s t)
          (fun _ -> check_steps (_A1, _A2) (_B1, _B2) p ra prts t u)
    | pa, ra, (p, (r, (false, t))) :: prts, s, u ->
        bind (check_rstep (_A1, _A2) (_B1, _B2) ra p r s t)
          (fun _ -> check_steps (_A1, _A2) (_B1, _B2) pa ra prts t u);;

let rec check_rsteps (_A1, _A2) (_B1, _B2)
  r prts s u =
    check_steps (_A1, _A2) (_B1, _B2) [] r
      (map (fun (p, (ra, t)) -> (p, (ra, (false, t)))) prts) s u;;

let rec check_loop (_A1, _A2) (_B1, _B2)
  s rseq sigma c r =
    bind (check (not (null rseq))
           (shows_prec_list show_char Zero_nat
             ['r'; 'e'; 'w'; 'r'; 'i'; 't'; 'e'; ' '; 's'; 'e'; 'q'; 'u'; 'e';
               'n'; 'c'; 'e'; ' '; 'm'; 'u'; 's'; 't'; ' '; 'b'; 'e'; ' '; 'n';
               'o'; 'n'; '-'; 'e'; 'm'; 'p'; 't'; 'y']))
      (fun _ ->
        check_rsteps (_A1, _A2) (_B1, _B2) r rseq s
          (ctxt_apply c (sub_apply_term s (subst_of _B1 sigma))));;

let rec check_dp_loop (_A1, _A2) (_B1, _B2)
  s prseq sigma c p r =
    (if eq_ctxt _A1 _B1 c Hole
      then bind (check (not (null prseq))
                  (shows_prec_list show_char Zero_nat
                    ['r'; 'e'; 'w'; 'r'; 'i'; 't'; 'e'; ' '; 's'; 'e'; 'q'; 'u';
                      'e'; 'n'; 'c'; 'e'; ' '; 'm'; 'u'; 's'; 't'; ' '; 'b';
                      'e'; ' '; 'n'; 'o'; 'n'; '-'; 'e'; 'm'; 'p'; 't'; 'y']))
             (fun _ ->
               check_steps (_A1, _A2) (_B1, _B2) p r prseq s
                 (sub_apply_term s (subst_of _B1 sigma)))
      else check_loop (_A1, _A2) (_B1, _B2) s
             (map (fun (x, (y, (_, z))) -> (x, (y, z))) prseq) sigma c r);;

let rec check_dp_disproof (_A1, _A2) (_B1, _B2)
  i p r x3 = match i, p, r, x3 with
    i, p, r, DPLoop (s, rseq, sigma, c) ->
      debug (i []) ['L'; 'o'; 'o'; 'p']
        (catch (check_dp_loop (_A1, _A2) (_B1, _B2) s rseq sigma c p r)
          (fun x ->
            Inl (comp i
                  (comp (shows_string
                          [':'; ' '; 'e'; 'r'; 'r'; 'o'; 'r'; ' '; 'i'; 'n';
                            ' '; 'l'; 'o'; 'o'; 'p'; ' '; 'p'; 'r'; 'o'; 'o';
                            'f'])
                    (comp shows_nl x)))))
    | i, p, r, DPRuleRemoval (p', r', prf) ->
        debug (i [])
          ['R'; 'u'; 'l'; 'e'; ' '; 'R'; 'e'; 'm'; 'o'; 'v'; 'a'; 'l']
          (bind (catch
                  (check_subseteq (eqc (eq_terma _A1 _B1) (eq_terma _A1 _B1)) p'
                    p)
                  (fun x ->
                    Inl (comp i
                          (comp (shows_string
                                  [':'; ' '; 'e'; 'r'; 'r'; 'o'; 'r'; ' '; 'i';
                                    'n'; ' '; 'r'; 'u'; 'l'; 'e'; '-'; 'r'; 'e';
                                    'm'; 'o'; 'v'; 'a'; 'l'; ' '; 'p'; 'r'; 'o';
                                    'o'; 'f'])
                            (comp shows_nl
                              (comp (shows_rule _A2 _B2 x)
                                (comp (shows_string
[' '; 'i'; 's'; ' '; 'a'; ' '; 'n'; 'e'; 'w'; ' '; 'p'; 'a'; 'i'; 'r'; ' '; 'n';
  'o'; 't'; ' '; 'o'; 'c'; 'c'; 'u'; 'r'; 'r'; 'i'; 'n'; 'g'; ' '; 'i'; 'n';
  ' '; 't'; 'h'; 'e'; ' '; 'p'; 'r'; 'e'; 'v'; 'i'; 'o'; 'u'; 's'; ' '; 's';
  'e'; 't'; ' '; 'o'; 'f'; ' '; 'p'; 'a'; 'i'; 'r'; 's'])
                                  shows_nl)))))))
            (fun _ ->
              bind (catch
                     (check_subseteq (eqc (eq_terma _A1 _B1) (eq_terma _A1 _B1))
                       r' r)
                     (fun x ->
                       Inl (comp i
                             (comp (shows_string
                                     [':'; ' '; 'e'; 'r'; 'r'; 'o'; 'r'; ' ';
                                       'i'; 'n'; ' '; 'r'; 'u'; 'l'; 'e'; '-';
                                       'r'; 'e'; 'm'; 'o'; 'v'; 'a'; 'l'; ' ';
                                       'p'; 'r'; 'o'; 'o'; 'f'])
                               (comp shows_nl
                                 (comp (shows_rule _A2 _B2 x)
                                   (comp (shows_string
   [' '; 'i'; 's'; ' '; 'a'; ' '; 'n'; 'e'; 'w'; ' '; 'r'; 'u'; 'l'; 'e'; ' ';
     'n'; 'o'; 't'; ' '; 'o'; 'c'; 'c'; 'u'; 'r'; 'r'; 'i'; 'n'; 'g'; ' '; 'i';
     'n'; ' '; 't'; 'h'; 'e'; ' '; 'p'; 'r'; 'e'; 'v'; 'i'; 'o'; 'u'; 's'; ' ';
     'T'; 'R'; 'S'])
                                     shows_nl)))))))
                (fun _ ->
                  check_dp_disproof (_A1, _A2) (_B1, _B2)
                    (comp i (shows_string ['.'; '1'])) p' r' prf)));;

let rec unary_term
  = function Var x -> true
    | Fun (f, [t]) -> unary_term t
    | Fun (v, []) -> false
    | Fun (v, vb :: vd :: ve) -> false;;

let rec check_unary_signature
  r = catch (catch
              (forallm
                (fun x ->
                  (if let (l, ra) = x in
                      unary_term l && unary_term ra
                    then Inr () else Inl x))
                r)
              (fun x -> Inl (snd x)))
        (fun _ ->
          Inl (shows_string
                ['t'; 'h'; 'e'; ' '; 't'; 'r'; 's'; ' '; 'd'; 'o'; 'e'; 's';
                  ' '; 'n'; 'o'; 't'; ' '; 'h'; 'a'; 'v'; 'e'; ' '; 'a'; ' ';
                  'u'; 'n'; 'a'; 'r'; 'y'; ' '; 's'; 'i'; 'g'; 'n'; 'a'; 't';
                  'u'; 'r'; 'e']));;

let rec check_string_rev_nt (_A1, _A2) (_B1, _B2)
  r rR =
    bind (check_unary_signature r)
      (fun _ ->
        catch (check_subseteq (eqc (eq_terma _A1 _B1) (eq_terma _A1 _B1)) rR
                (map rev_rule r))
          (fun x ->
            Inl (comp (shows_string
                        ['t'; 'h'; 'e'; ' '; 'r'; 'u'; 'l'; 'e'; ' '])
                  (comp (shows_rule _A2 _B2 x)
                    (comp (shows_string
                            [' '; 'i'; 's'; ' '; 'n'; 'o'; 't'; ' '; 'a'; 'l';
                              'l'; 'o'; 'w'; 'e'; 'd'])
                      shows_nl)))));;

let rec check_not_wf_trs _A (_B1, _B2)
  r = check (not (isOK (check_wf_trs _A (_B1, _B2) r)))
        (comp (shows_string
                ['T'; 'h'; 'e'; ' '; 'T'; 'R'; 'S'; ' '; 'i'; 's'; ' '; 'w';
                  'e'; 'l'; 'l'; ' '; 'f'; 'o'; 'r'; 'm'; 'e'; 'd'])
          shows_nl);;

let rec checkTRSDisproof (_A1, _A2) (_B1, _B2)
  i r x2 = match i, r, x2 with
    i, r, Loop (s, rseq, sigma, c) ->
      debug (i []) ['L'; 'o'; 'o'; 'p']
        (catch (check_loop (_A1, _A2) (_B1, _B2) s rseq sigma c r)
          (fun x ->
            Inl (comp i
                  (comp (shows_string
                          [':'; ' '; 'e'; 'r'; 'r'; 'o'; 'r'; ' '; 'i'; 'n';
                            ' '; 'l'; 'o'; 'o'; 'p'; ' '; 'p'; 'r'; 'o'; 'o';
                            'f'])
                    (comp shows_nl x)))))
    | i, r, NotWellFormed ->
        debug (i [])
          ['n'; 'o'; 't'; ' '; 'w'; 'e'; 'l'; 'l'; '-'; 'f'; 'o'; 'r'; 'm'; 'e';
            'd']
          (catch (check_not_wf_trs _A2 (_B1, _B2) r)
            (fun x ->
              Inl (comp i
                    (comp (shows_string
                            [':'; ' '; 'e'; 'r'; 'r'; 'o'; 'r'; ' '; 'i'; 'n';
                              ' '; 'n'; 'o'; 't'; '-'; 'w'; 'e'; 'l'; 'l'; '-';
                              'f'; 'o'; 'r'; 'm'; 'e'; 'd'; ' '; 'p'; 'r'; 'o';
                              'o'; 'f'])
                      (comp shows_nl x)))))
    | i, r, NTRuleRemoval (r', prf) ->
        debug (i [])
          ['r'; 'u'; 'l'; 'e'; ' '; 'r'; 'e'; 'm'; 'o'; 'v'; 'a'; 'l']
          (bind (catch
                  (check_subseteq (eqc (eq_terma _A1 _B1) (eq_terma _A1 _B1)) r'
                    r)
                  (fun x ->
                    Inl (comp i
                          (comp (shows_string
                                  [':'; ' '; 'e'; 'r'; 'r'; 'o'; 'r'; ' '; 'i';
                                    'n'; ' '; 'r'; 'u'; 'l'; 'e'; '-'; 'r'; 'e';
                                    'm'; 'o'; 'v'; 'a'; 'l'; ' '; 'p'; 'r'; 'o';
                                    'o'; 'f'])
                            (comp shows_nl
                              (comp (shows_rule _A2 _B2 x)
                                (comp (shows_string
[' '; 'i'; 's'; ' '; 'a'; ' '; 'n'; 'e'; 'w'; ' '; 'r'; 'u'; 'l'; 'e'; ' '; 'n';
  'o'; 't'; ' '; 'o'; 'c'; 'c'; 'u'; 'r'; 'r'; 'i'; 'n'; 'g'; ' '; 'i'; 'n';
  ' '; 't'; 'h'; 'e'; ' '; 'p'; 'r'; 'e'; 'v'; 'i'; 'o'; 'u'; 's'; ' '; 'T';
  'R'; 'S'])
                                  shows_nl)))))))
            (fun _ ->
              checkTRSDisproof (_A1, _A2) (_B1, _B2)
                (comp i (shows_string ['.'; '1'])) r' prf))
    | i, r, NTStringReversal (r', prf) ->
        debug (i [])
          ['s'; 't'; 'r'; 'i'; 'n'; 'g'; ' '; 'r'; 'e'; 'v'; 'e'; 'r'; 's'; 'a';
            'l']
          (bind (catch (check_string_rev_nt (_A1, _A2) (_B1, _B2) r r')
                  (fun x ->
                    Inl (comp i
                          (comp (shows_string
                                  [':'; ' '; 'e'; 'r'; 'r'; 'o'; 'r'; ' '; 'i';
                                    'n'; ' '; 's'; 't'; 'r'; 'i'; 'n'; 'g'; ' ';
                                    'r'; 'e'; 'v'; 'e'; 'r'; 's'; 'a'; 'l'; ' ';
                                    'p'; 'r'; 'o'; 'o'; 'f'])
                            (comp shows_nl x)))))
            (fun _ ->
              checkTRSDisproof (_A1, _A2) (_B1, _B2)
                (comp i (shows_string ['.'; '1'])) r' prf))
    | i, r, NTDPTrans (p, prf) ->
        debug (i [])
          ['D'; 'P'; ' '; 't'; 'r'; 'a'; 'n'; 's'; 'f'; 'o'; 'r'; 'm'; 'a'; 't';
            'i'; 'o'; 'n']
          (bind (catch (check_DPs (_A1, _A2) (_B1, _B2) r p)
                  (fun x ->
                    Inl (comp i
                          (comp (shows_string
                                  [':'; ' '; 'e'; 'r'; 'r'; 'o'; 'r'; ' '; 'i';
                                    'n'; ' '; 'D'; 'P'; ' '; 't'; 'r'; 'a'; 'n';
                                    's'; 'f'; 'o'; 'r'; 'm'; 'a'; 't'; 'i'; 'o';
                                    'n'])
                            (comp shows_nl
                              (comp (shows_prec_list show_char Zero_nat
                                      ['p'; 'r'; 'o'; 'b'; 'l'; 'e'; 'm'; 'a';
't'; 'i'; 'c'; ' '; 'r'; 'u'; 'l'; 'e'; ' '])
                                (shows_rule (show_shp _A2) _B2 x)))))))
            (fun _ ->
              check_dp_disproof (eq_shpa _A1, show_shp _A2) (_B1, _B2)
                (comp i (shows_string ['.'; '1'])) p (iD_list r) prf));;

let rec init_lab f = UnLab f;;

let rec check_string_rev (_A1, _A2) (_B1, _B2)
  r rR =
    bind (check_unary_signature r)
      (fun _ ->
        catch (check_subseteq (eqc (eq_terma _A1 _B1) (eq_terma _A1 _B1))
                (map rev_rule r) rR)
          (fun x ->
            Inl (comp (shows_string
                        ['t'; 'h'; 'e'; ' '; 'r'; 'u'; 'l'; 'e'; ' '])
                  (comp (shows_rule _A2 _B2 x)
                    (comp (shows_string
                            [' '; 'i'; 's'; ' '; 'm'; 'i'; 's'; 's'; 'i'; 'n';
                              'g'])
                      shows_nl)))));;

let rec shows_arithFun
  = function
    Arg i -> comp (shows_string ['x'; '_']) (shows_prec_nat Zero_nat (Suc i))
    | Const n -> shows_prec_nat Zero_nat n
    | Sum fs -> shows_list_gen shows_arithFun [] [] [' '; '+'; ' '] [] fs
    | Prod fs -> shows_list_gen shows_arithFun [] [] [' '; '*'; ' '] [] fs
    | Min fs ->
        shows_list_gen shows_arithFun [] ['m'; 'i'; 'n'; '('] [','] [')'] fs
    | Max fs ->
        shows_list_gen shows_arithFun [] ['m'; 'a'; 'x'; '('] [','] [')'] fs;;

let rec shows_prec_arithFun d f = shows_arithFun f;;

let rec shows_sl_inter _A _B
  (SL_Inter (n, fnas)) =
    comp (shows_string
           ['c'; 'a'; 'r'; 'r'; 'i'; 'e'; 'r'; ' '; '{'; '0'; ','; '.'; '.';
             '.'; ','])
      (comp (shows_prec_nat Zero_nat n)
        (comp (shows_string ['}'])
          (comp shows_nl
            (shows_map
              (fun (a, b) ->
                let (f, na) = a in
                (fun aa ->
                  comp (shows_string ['['])
                    (comp (shows_prec_lab _A _B Zero_nat f)
                      (comp (shows_string ['/'])
                        (comp (shows_prec_nat Zero_nat na)
                          (comp (shows_string [']'; ' '; '='; ' '])
                            (comp (shows_prec_arithFun Zero_nat aa)
                              shows_nl))))))
                  b)
              fnas))));;

let rec shows_prec_sl_inter _A _B d sli = shows_sl_inter _A _B sli;;

let rec take_default
  def x1 uu = match def, x1, uu with def, [], uu -> def
    | uv, x :: xs, Zero_nat -> x
    | def, uw :: xs, Suc i -> take_default def xs i;;

let rec eval_arithFun_unbound
  nats x1 = match nats, x1 with nats, Arg i -> take_default Zero_nat nats i
    | nats, Const n -> n
    | nats, Sum [] -> Zero_nat
    | nats, Sum (f :: fs) ->
        plus_nat (eval_arithFun_unbound nats f)
          (eval_arithFun_unbound nats (Sum fs))
    | nats, Prod [] -> one_nat
    | nats, Prod (f :: fs) ->
        times_nat (eval_arithFun_unbound nats f)
          (eval_arithFun_unbound nats (Prod fs))
    | nats, Max [f] -> eval_arithFun_unbound nats f
    | nats, Max (f :: v :: va) ->
        max ord_nat (eval_arithFun_unbound nats f)
          (eval_arithFun_unbound nats (Max (v :: va)))
    | nats, Min [f] -> eval_arithFun_unbound nats f
    | nats, Min (f :: v :: va) ->
        min ord_nat (eval_arithFun_unbound nats f)
          (eval_arithFun_unbound nats (Min (v :: va)));;

let rec eval_arithFun c f nats = mod_nat (eval_arithFun_unbound nats f) c;;

let rec sl_inter_to_inter _A _B
  (SL_Inter (c, ls)) fl cs =
    (match lookup (eqc (eq_laba _A _B) eq_nata) (fl, size_list cs) ls
      with None -> Zero_nat | Some f -> eval_arithFun (Suc c) f cs);;

let rec get_largest_element (SL_Inter (n, uu)) = n;;

let rec create_sem_lab_tt (_A1, _A2) (_B1, _B2) (_C1, _C2)
  sli l r lr =
    let c = get_largest_element sli in
    let ca = upt Zero_nat (Suc c) in
    let i = sl_inter_to_inter _A1 _B1 sli in
    catch (check_sem_lab_tt (_A1, _A2) (_B1, _B2) (_C1, _C2)
            (check_sl_lab_trs (_A1, _A2) (_B1, _B2) (eq_nata, show_nat)
              (_C1, _C2) i l ca)
            (check_sl_model_lab_trs (_A1, _A2) (_B1, _B2) (eq_nata, show_nat)
              (_C1, _C2) i l ca)
            (fun _ -> false) i l r lr)
      (fun x ->
        Inl (comp (shows_prec_sl_inter _A2 _B2 Zero_nat sli)
              (comp shows_nl x)));;

let rec unlab_of_trsI _A _B _C
  r = remdups (eqc (eq_terma (eq_laba _A _B) _C) (eq_terma (eq_laba _A _B) _C))
        (map_trsI unlab
          (filter
            (fun (l, ra) ->
              eq_term (eq_laba _A _B) _C l ra ||
                not (eq_term (eq_laba _A _B) _C (map_term unlab l)
                      (map_term unlab ra)))
            r));;

let rec unlab_trs_tt (_A1, _A2) (_B1, _B2) (_C1, _C2)
  r rR =
    let u = unlab_of_trsI _A1 _B1 _C1 r in
    catch (bind (check_subseteq
                  (eqc (eq_terma (eq_laba _A1 _B1) _C1)
                    (eq_terma (eq_laba _A1 _B1) _C1))
                  rR u)
            (fun _ ->
              check_subseteq
                (eqc (eq_terma (eq_laba _A1 _B1) _C1)
                  (eq_terma (eq_laba _A1 _B1) _C1))
                u rR))
      (fun _ ->
        Inl (comp (shows_rules (show_lab _A2 _B2) _C2 rR)
              (comp shows_nl
                (comp (shows_string
                        [' '; 'd'; 'o'; 'e'; 's'; ' '; 'n'; 'o'; 't'; ' '; 'c';
                          'o'; 'r'; 'r'; 'e'; 's'; 'p'; 'o'; 'n'; 'd'; ' '; 't';
                          'o'; ' '; 'u'; 'n'; 'l'; 'a'; 'b'; 'e'; 'l'; 'e'; 'd';
                          ' '; 'T'; 'R'; 'S'; ' '])
                  (comp shows_nl
                    (comp (shows_rules (show_lab _A2 _B2) _C2 u)
                      shows_nl))))));;

let rec check_mono_afs_help _A
  ((f, a), e) =
    comp (shows_string ['e'; 'n'; 't'; 'r'; 'y'; ' '; 'f'; 'o'; 'r'; ' '])
      (comp (shows_prec _A Zero_nat f)
        (comp (shows_string
                [' '; 'n'; 'o'; 't'; ' '; 'm'; 'o'; 'n'; 'o'; 't'; 'o'; 'n';
                  'i'; 'c'])
          shows_nl));;

let rec mono_af_entry
  n x1 = match n, x1 with n, Collapse i -> less_eq_nat n one_nat
    | n, AFList ids -> forall (fun i -> member eq_nata i ids) (upt Zero_nat n);;

let rec check_mono_afs _A
  pi = catch (catch
               (forallm
                 (fun x ->
                   (if mono_af_entry (snd (fst x)) (snd x) then Inr ()
                     else Inl x))
                 pi)
               (fun x -> Inl (snd x)))
         (fun x -> Inl (check_mono_afs_help _A x));;

let rec internal_error
  x = comp (shows_string
             ['i'; 'n'; 't'; 'e'; 'r'; 'n'; 'a'; 'l'; ' '; 'e'; 'r'; 'r'; 'o';
               'r'])
        (comp shows_nl
          (comp (shows_string
                  ['p'; 'l'; 'e'; 'a'; 's'; 'e'; ' '; 'c'; 'o'; 'n'; 't'; 'a';
                    'c'; 't'; ' '; 'o'; 'n'; 'e'; ' '; 'o'; 'f'; ' '; 't'; 'h';
                    'e'; ' '; 'C'; 'e'; 'T'; 'A'; ' '; 'm'; 'e'; 'm'; 'b'; 'e';
                    'r'; 's'])
            (comp shows_nl
              (comp (shows_string
                      ['h'; 't'; 't'; 'p'; ':'; '/'; '/'; 'c'; 'l'; '-'; 'i';
                        'n'; 'f'; 'o'; 'r'; 'm'; 'a'; 't'; 'i'; 'k'; '.'; 'u';
                        'i'; 'b'; 'k'; '.'; 'a'; 'c'; '.'; 'a'; 't'; '/'; 's';
                        'o'; 'f'; 't'; 'w'; 'a'; 'r'; 'e'; '/'; 'c'; 'e'; 't';
                        'a'])
                (comp shows_nl
                  (comp (shows_string
                          ['c'; 'e'; 't'; 'a'; '@'; 'i'; 'n'; 'f'; 'o'; 'r';
                            'm'; 'a'; 't'; 'i'; 'k'; '.'; 'u'; 'i'; 'b'; 'k';
                            '.'; 'a'; 'c'; '.'; 'a'; 't'])
                    shows_nl))))))
        x;;

let rec check_rule_removal_proc _A _B
  checkMono checkSS checkNSS name representation r rR =
    catch (let (_, nSp) = partition (fun foo -> isOK (checkSS foo)) r in
           bind (catch checkMono
                  (fun x ->
                    Inl (comp (shows_string
                                ['p'; 'r'; 'o'; 'b'; 'l'; 'e'; 'm'; ' '; 'w';
                                  'i'; 't'; 'h'; ' '; 'm'; 'o'; 'n'; 'o'; 't';
                                  'o'; 'n'; 'i'; 'c'; 'i'; 't'; 'y'; ' '; 'o';
                                  'f'; ' '; 's'; 't'; 'r'; 'i'; 'c'; 't'; ' ';
                                  'o'; 'r'; 'd'; 'e'; 'r'])
                          (comp shows_nl x))))
             (fun _ ->
               bind (catch (catch (forallm checkNSS r) (fun x -> Inl (snd x)))
                      (fun x ->
                        Inl (comp (shows_string
                                    ['p'; 'r'; 'o'; 'b'; 'l'; 'e'; 'm'; ' ';
                                      'w'; 'h'; 'e'; 'n'; ' '; 'o'; 'r'; 'i';
                                      'e'; 'n'; 't'; 'i'; 'n'; 'g'; ' '; 'r';
                                      'u'; 'l'; 'e'; 's'])
                              (comp shows_nl x))))
                 (fun _ ->
                   bind (catch
                          (check_subseteq
                            (eqc (eq_terma _A _B) (eq_terma _A _B)) nSp rR)
                          (fun x ->
                            Inl (match checkSS x with Inl err -> err
                                  | Inr _ -> internal_error)))
                     (fun _ ->
                       catch (check_subseteq
                               (eqc (eq_terma _A _B) (eq_terma _A _B)) rR r)
                         (fun x ->
                           Inl (match checkSS x with Inl err -> err
                                 | Inr _ -> internal_error))))))
      (fun x ->
        Inl (comp (shows_string
                    ['c'; 'o'; 'u'; 'l'; 'd'; ' '; 'n'; 'o'; 't'; ' '; 'a'; 'p';
                      'p'; 'l'; 'y'; ' '; 't'; 'h'; 'e'; ' '; 'r'; 'u'; 'l';
                      'e'; ' '; 'r'; 'e'; 'm'; 'o'; 'v'; 'a'; 'l'; ' '; 'p';
                      'r'; 'o'; 'c'; 'e'; 's'; 's'; 'o'; 'r'; ' '; 'w'; 'i';
                      't'; 'h'; ' '; 't'; 'h'; 'e'; ' '; 'f'; 'o'; 'l'; 'l';
                      'o'; 'w'; 'i'; 'n'; 'g'])
              (comp shows_nl (comp representation (comp shows_nl x)))));;

let rec check_wf_afs _A
  pi = catch (forallm
               (fun (a, b) ->
                 let (f, n) = a in
                 (fun e ->
                   catch (match e with Collapse i -> check (less_nat i n) i
                           | AFList ids ->
                             catch (forallm
                                     (fun x ->
                                       (if less_nat x n then Inr () else Inl x))
                                     ids)
                               (fun x -> Inl (snd x)))
                     (fun x ->
                       Inl (comp (shows_string ['i'; 'n'; 'd'; 'e'; 'x'; ' '])
                             (comp (shows_prec_nat Zero_nat x)
                               (comp (shows_string
                                       [' '; 't'; 'o'; 'o'; ' '; 'h'; 'i'; 'g';
 'h'; ' '; 'f'; 'o'; 'r'; ' '; 'A'; 'F'; '-'; 'e'; 'n'; 't'; 'r'; 'y'; ' '; 'o';
 'f'])
                                 (comp (shows_preca _A show_nat Zero_nat (f, n))
                                   shows_nl))))))
                   b)
               pi)
         (fun x -> Inl (snd x));;

let rec default_af_entry n = AFList (upt Zero_nat n);;

let rec afs_list_to_afs _A
  pi f n =
    (match lookup (eqc _A eq_nata) (f, n) pi with None -> default_af_entry n
      | Some entry -> entry);;

let rec lpo_repr_to_pr _A
  prs (FPair (f, a)) ignored =
    (match lookup (eqc _A eq_nata) (f, a) prs with None -> Zero_nat
      | Some x -> x);;

let rec get_max_arity _A
  r = max_list
        (map snd
          (concat_map
            (fun (l, ra) ->
              append (funas_term_list _A l) (funas_term_list _A ra))
            r));;

let nat_components :
  (nat -> bool) *
    (nat *
      (nat *
        (nat *
          ((nat -> nat -> nat) *
            ((nat -> nat -> nat) *
              ((nat -> nat) *
                ((nat -> nat -> bool) *
                  ((nat -> nat -> bool) *
                    (bool * (char list * (nat -> bool)))))))))))
  = ((fun _ -> true),
      (Zero_nat,
        (one_nat,
          (one_nat,
            (plus_nat,
              (times_nat,
                (max0_nat,
                  (ge_nat,
                    ((fun x y -> less_nat y x),
                      (true, (nat_name, (fun _ -> true))))))))))));;

let rec create_lpoly_repr (_A1, _A2) (_B1, _B2)
  c i = let pi = to_lpoly_inter _B1 (get_comp_def c) (get_comp_1 c) i in
        comp (shows_string
               ['p'; 'o'; 'l'; 'y'; 'n'; 'o'; 'm'; 'i'; 'a'; 'l'; ' '; 'i'; 'n';
                 't'; 'e'; 'r'; 'p'; 'r'; 'e'; 't'; 'r'; 'a'; 't'; 'i'; 'o';
                 'n'; ' '; 'o'; 'v'; 'e'; 'r'; ' '])
          (comp (shows_string (get_comp_description c))
            (comp shows_nl
              (shows_sep
                (fun (f, n) ->
                  let t =
                    Fun (f, map (fun a -> Var a)
                              (fresh_strings_list ['x'; '_'] one_nat [] n))
                    in
                  comp (shows_string ['P'; 'o'; 'l'; '('])
                    (comp (shows_prec_term _B2 (show_list show_char) Zero_nat t)
                      (comp (shows_string [')'; ' '; '='; ' '])
                        (shows_lpoly (_A1, _A2) (show_list show_char)
                          (get_comp_0 c) (get_comp_1 c)
                          (pleftI _A1 (eq_lista eq_char) (get_comp_0 c)
                            (get_comp_1 c) (get_comp_plus c) (get_comp_times c)
                            (get_comp_max0 c) pi t)))))
                shows_nl (remdups (eqc _B1 eq_nata) (map fst i)))));;

let rec check_poly_monoI _A _B
  ze ge cm =
    (fun xs ->
      catch (forallm
              (fun (a, b) ->
                let (f, n) = a in
                (fun (c, cs) ->
                  catch (bind (check (ge c ze)
                                (comp (shows_string
['c'; 'o'; 'n'; 's'; 't'; 'a'; 'n'; 't'; ' '; 'p'; 'a'; 'r'; 't'; ' '])
                                  (comp (shows_prec _A Zero_nat c)
                                    (comp (shows_string
    [' '; 'm'; 'u'; 's'; 't'; ' '; 'b'; 'e'; ' '; 'a'; 't'; ' '; 'l'; 'e'; 'a';
      's'; 't'; ' '])
                                      (comp (shows_prec _A Zero_nat ze)
shows_nl)))))
                          (fun _ ->
                            bind (check (less_eq_nat n (size_list cs))
                                   (comp (shows_string
   ['t'; 'h'; 'e'; ' '; 'l'; 'a'; 's'; 't'; ' '; 'a'; 'r'; 'g'; 'u'; 'm'; 'e';
     'n'; 't'; ' '; 'i'; 's'; ' '; 'i'; 'g'; 'n'; 'o'; 'r'; 'e'; 'd'])
                                     shows_nl))
                              (fun _ ->
                                catch (forallm
(fun d ->
  check (cm d)
    (comp (shows_string
            ['c'; 'o'; 'e'; 'f'; 'f'; 'i'; 'c'; 'i'; 'e'; 'n'; 't'; ' '])
      (comp (shows_prec _A Zero_nat d)
        (comp (shows_string
                [' '; 'i'; 's'; ' '; 'n'; 'o'; 't'; ' '; 'a'; 'l'; 'l'; 'o';
                  'w'; 'e'; 'd'])
          shows_nl))))
cs)
                                  (fun x -> Inl (snd x)))))
                    (fun x ->
                      Inl (comp (shows_string
                                  ['p'; 'r'; 'o'; 'b'; 'l'; 'e'; 'm'; ' '; 'w';
                                    'i'; 't'; 'h'; ' '; 'm'; 'o'; 'n'; 'o'; 't';
                                    'o'; 'n'; 'i'; 'c'; 'i'; 't'; 'y'; ' '; 'd';
                                    'u'; 'e'; ' '; 't'; 'o'; ' '; 'i'; 'n'; 't';
                                    'e'; 'r'; 'p'; 'r'; 'e'; 't'; 'a'; 't'; 'i';
                                    'o'; 'n'; ' '; 'o'; 'f'; ' '])
                            (comp (shows_prec _B Zero_nat f)
                              (comp (shows_string ['/'])
                                (comp (shows_prec_nat Zero_nat n)
                                  (comp shows_nl x)))))))
                  b)
              xs)
        (fun x -> Inl (snd x)));;

let rec get_comp_wf_elem
  (wfe, (ze, (on, (def, (pl, (ti, (mz, (ge, (gt, (mo, (de, ap))))))))))) = wfe;;

let rec check_lpoly_coeffs _A _B
  wfe apos ch i =
    catch (forallm
            (fun (a, b) ->
              let (f, n) = a in
              (fun (c, cs) ->
                catch (bind (check (wfe c)
                              (comp (shows_string
                                      ['c'; 'o'; 'n'; 's'; 't'; 'a'; 'n'; 't';
' '; 'p'; 'a'; 'r'; 't'; ' '])
                                (comp (shows_prec _A Zero_nat c)
                                  (comp (shows_string
  [' '; 'i'; 's'; ' '; 'n'; 'o'; 't'; ' '; 'w'; 'e'; 'l'; 'l'; '-'; 'f'; 'o';
    'r'; 'm'; 'e'; 'd'])
                                    shows_nl))))
                        (fun _ ->
                          bind (check (apos c || any apos (take n cs))
                                 (comp (shows_string
 ['c'; 'o'; 'u'; 'l'; 'd'; ' '; 'n'; 'o'; 't'; ' '; 'f'; 'i'; 'n'; 'd'; ' ';
   'p'; 'o'; 's'; 'i'; 't'; 'i'; 'v'; 'e'; ' '; 'e'; 'n'; 't'; 'r'; 'y'; ' ';
   'w'; 'h'; 'i'; 'c'; 'h'; ' '; 'i'; 's'; ' '; 'r'; 'e'; 'q'; 'u'; 'i'; 'r';
   'e'; 'd'; ' '; 'f'; 'o'; 'r'; ' '; 'a'; 'r'; 'c'; 't'; 'i'; 'c'; ' '; 'i';
   'n'; 't'; 'e'; 'r'; 'p'; 'r'; 'e'; 't'; 'a'; 't'; 'i'; 'o'; 'n'; 's'])
                                   shows_nl))
                            (fun _ ->
                              catch (forallm
                                      (fun aa ->
check (ch aa && wfe aa)
  (comp (shows_string
          ['c'; 'o'; 'e'; 'f'; 'f'; 'i'; 'c'; 'i'; 'e'; 'n'; 't'; ' '])
    (comp (shows_prec _A Zero_nat aa)
      (comp (shows_string
              [' '; 'i'; 's'; ' '; 'n'; 'o'; 't'; ' '; 'a'; 'l'; 'l'; 'o'; 'w';
                'e'; 'd'])
        shows_nl))))
                                      cs)
                                (fun x -> Inl (snd x)))))
                  (fun x ->
                    Inl (comp (shows_string
                                ['p'; 'r'; 'o'; 'b'; 'l'; 'e'; 'm'; ' '; 'w';
                                  'i'; 't'; 'h'; ' '; 'i'; 'n'; 't'; 'e'; 'r';
                                  'p'; 'r'; 'e'; 't'; 'a'; 't'; 'i'; 'o'; 'n';
                                  ' '; 'o'; 'f'; ' '])
                          (comp (shows_prec _B Zero_nat f)
                            (comp (shows_string ['/'])
                              (comp (shows_prec_nat Zero_nat n)
                                (comp shows_nl x)))))))
                b)
            i)
      (fun x -> Inl (snd x));;

let rec create_rule_removal_proc_comp (_A1, _A2) (_B1, _B2) (_C1, _C2)
  c cm i r rR =
    let pi = to_lpoly_inter _B1 (get_comp_def c) (get_comp_1 c) i in
    bind (check_lpoly_coeffs _A2 _B2 (get_comp_wf_elem c) (get_comp_apos c)
           (fun x -> get_comp_ge c x (get_comp_0 c)) i)
      (fun _ ->
        check_rule_removal_proc _B1 _C1
          (check_poly_monoI _A2 _B2 (get_comp_0 c) (get_comp_ge c) cm i)
          (check_polo_sI (_A1, _A2) _B2 (_C1, _C2) (get_comp_mono c)
            (get_comp_0 c) (get_comp_1 c) (get_comp_plus c) (get_comp_times c)
            (get_comp_max0 c) (get_comp_ge c) (get_comp_gt c) pi)
          (check_polo_nsI (_A1, _A2) _B2 (_C1, _C2) (get_comp_0 c)
            (get_comp_1 c) (get_comp_plus c) (get_comp_times c)
            (get_comp_max0 c) (get_comp_ge c) pi)
          (comp (shows_string
                  ['p'; 'o'; 'l'; 'y'; 'n'; 'o'; 'm'; 'i'; 'a'; 'l'; ' '; 'o';
                    'r'; 'd'; 'e'; 'r'; ' '; '('])
            (comp (shows_string (get_comp_description c)) (shows_string [')'])))
          (create_lpoly_repr (_A1, _A2) (_B1, _B2) c i) r rR);;

let rec create_rule_removal_proc (_A1, _A2) (_B1, _B2)
  = function
    Int_carrier i ->
      create_rule_removal_proc_comp (eq_int, show_int) (_A1, _A2) (_B1, _B2)
        int_components int_mono i
    | Nat_carrier i ->
        create_rule_removal_proc_comp (eq_nata, show_nat) (_A1, _A2) (_B1, _B2)
          nat_components nat_mono i
    | Rat_carrier (delta, i) ->
        create_rule_removal_proc_comp (eq_rata, show_rata) (_A1, _A2) (_B1, _B2)
          (rat_components delta) rat_mono i
    | Arctic_carrier i ->
        (fun _ _ ->
          Inl (comp (shows_string
                      ['a'; 'r'; 'c'; 't'; 'i'; 'c'; ' '; 'n'; 'o'; 't'; ' ';
                        'm'; 'o'; 'n'; 'o'; 't'; 'o'; 'n'; 'i'; 'c'; ' '; 'f';
                        'o'; 'r'; ' '; 'T'; 'R'; 'S'; 's'])
                shows_nl))
    | Arctic_rat_carrier (delta, i) ->
        (fun _ _ ->
          Inl (comp (shows_string
                      ['a'; 'r'; 'c'; 't'; 'i'; 'c'; ' '; 'n'; 'o'; 't'; ' ';
                        'm'; 'o'; 'n'; 'o'; 't'; 'o'; 'n'; 'i'; 'c'; ' '; 'f';
                        'o'; 'r'; ' '; 'T'; 'R'; 'S'; 's'])
                shows_nl))
    | Arc_mat_carrier (n, i) ->
        (fun _ _ ->
          Inl (comp (shows_string
                      ['a'; 'r'; 'c'; 't'; 'i'; 'c'; ' '; 'n'; 'o'; 't'; ' ';
                        'm'; 'o'; 'n'; 'o'; 't'; 'o'; 'n'; 'i'; 'c'; ' '; 'f';
                        'o'; 'r'; ' '; 'T'; 'R'; 'S'; 's'])
                shows_nl))
    | Arc_rat_mat_carrier (delta, n, i) ->
        (fun _ _ ->
          Inl (comp (shows_string
                      ['a'; 'r'; 'c'; 't'; 'i'; 'c'; ' '; 'n'; 'o'; 't'; ' ';
                        'm'; 'o'; 'n'; 'o'; 't'; 'o'; 'n'; 'i'; 'c'; ' '; 'f';
                        'o'; 'r'; ' '; 'T'; 'R'; 'S'; 's'])
                shows_nl))
    | Int_mat_carrier (n, sd, i) ->
        create_rule_removal_proc_comp
          (eq_lista (eq_lista eq_int), show_list (show_list show_int))
          (_A1, _A2) (_B1, _B2) (mat_components n sd int_components)
          (mat_mono sd int_mono) i
    | Rat_mat_carrier (delta, n, sd, i) ->
        create_rule_removal_proc_comp
          (eq_lista (eq_lista eq_rata), show_list (show_list show_rata))
          (_A1, _A2) (_B1, _B2) (mat_components n sd (rat_components delta))
          (mat_mono sd rat_mono) i
    | Lpo (pri, pi) ->
        (fun r rR ->
          let m = get_max_arity _A1 r in
          let pr = lpo_repr_to_pr _A1 pri in
          let pia = afs_list_to_afs _A1 pi in
          bind (check_wf_afs _A2 pi)
            (fun _ ->
              check_rule_removal_proc _A1 _B1 (check_mono_afs _A2 pi)
                (afs_check _A2 _B2 lpoS pia
                  (lpo_strict (show_filtered _A2) (_B1, _B2) pr m))
                (afs_check _A2 _B2 lpoNS pia
                  (lpo_nstrict (show_filtered _A2) (_B1, _B2) pr m))
                lpo_repr (shows_lpo_af _A2 pri pi) r rR));;

let rec computeRuleMap _A
  = function [] -> Some (fun _ -> [])
    | (Fun (f, ts), r) :: rules ->
        (match computeRuleMap _A rules with None -> None
          | Some rm -> Some (fun_upd _A rm f ((Fun (f, ts), r) :: rm f)))
    | (Var uu, uv) :: rules -> None;;

let rec afs_list_to_af _A
  pi f n =
    (fun i ->
      (match lookup (eqc _A eq_nata) (f, n) pi with None -> true
        | Some (Collapse a) -> eq_nat i a
        | Some (AFList a) -> member eq_nata i a));;

let rec matchCapRMBelow _A _B
  l rm (Fun (f, ts)) = gcMatchI _A _B (HFun (f, map (capRM _A _B rm) ts)) l;;

let rec forallm_index_aux
  p i x2 = match p, i, x2 with p, i, [] -> Inr ()
    | p, i, x :: xs ->
        bind (catch (p x i) (fun xa -> Inl ((x, i), xa)))
          (fun _ -> forallm_index_aux p (Suc i) xs);;

let rec forallm_index p xs = forallm_index_aux p Zero_nat xs;;

let rec check_ur_closed_term_rm_af (_A1, _A2) (_B1, _B2)
  uu uv uw x3 = match uu, uv, uw, x3 with uu, uv, uw, Var x -> Inr ()
    | rm, ur, pi, Fun (f, ts) ->
        let n = size_list ts in
        bind (catch
               (forallm_index
                 (fun t i ->
                   (if pi f n i
                     then check_ur_closed_term_rm_af (_A1, _A2) (_B1, _B2) rm ur
                            pi t
                     else Inr ()))
                 ts)
               (fun x -> Inl (snd x)))
          (fun _ ->
            catch (forallm
                    (fun lr ->
                      check (member (eqc (eq_terma _A1 _B1) (eq_terma _A1 _B1))
                               lr ur
                              || not (matchCapRMBelow _A1 _B1 (fst lr) (Some rm)
                                       (Fun (f, ts))))
                        (comp (shows_string
                                ['d'; 'u'; 'e'; ' '; 't'; 'o'; ' '; 't'; 'h';
                                  'e'; ' '; 's'; 'u'; 'b'; 't'; 'e'; 'r'; 'm';
                                  ' '])
                          (comp (shows_prec_term _A2 _B2 Zero_nat (Fun (f, ts)))
                            (comp (shows_string
                                    [' '; 'o'; 'f'; ' '; 's'; 'o'; 'm'; 'e';
                                      ' '; 'u'; 's'; 'a'; 'b'; 'l'; 'e'; ' ';
                                      'r'; 'h'; 's'; ','; ' '; 'r'; 'u'; 'l';
                                      'e'; ' '])
                              (comp (shows_rule _A2 _B2 lr)
                                (shows_string
                                  [' '; 's'; 'h'; 'o'; 'u'; 'l'; 'd'; ' '; 'b';
                                    'e'; ' '; 'u'; 's'; 'a'; 'b'; 'l'; 'e';
                                    '.']))))))
                    (rm f))
              (fun x -> Inl (snd x)));;

let rec check_sct_ur_af_proc (_A1, _A2) (_B1, _B2)
  checkSS checkNSS representation p gs rm rrm ur pi =
    (match rm
      with None ->
        Inl (comp (shows_string
                    ['T'; 'R'; 'S'; ' '; 'h'; 'a'; 's'; ' '; 'v'; 'a'; 'r'; 'i';
                      'a'; 'b'; 'l'; 'e'; 's'; ' '; 'a'; 's'; ' '; 'l'; 'e';
                      'f'; 't'; '-'; 'h'; 'a'; 'n'; 'd'; ' '; 's'; 'i'; 'd';
                      'e'; 's'])
              shows_nl)
      | Some rma ->
        catch (let ggs =
                 filter
                   (fun g ->
                     member (eqc (eq_terma _A1 _B1) (eq_terma _A1 _B1)) (fst g)
                       p)
                   gs
                 in
               bind (catch
                      (catch
                        (forallm
                          (fun lr ->
                            check_ur_closed_term_rm_af (_A1, _A2) (_B1, _B2) rma
                              ur pi (snd lr))
                          ur)
                        (fun x -> Inl (snd x)))
                      (fun x ->
                        Inl (comp (shows_string
                                    ['e'; 'r'; 'r'; 'o'; 'r'; ' '; 'w'; 'h';
                                      'e'; 'n'; ' '; 'c'; 'h'; 'e'; 'c'; 'k';
                                      'i'; 'n'; 'g'; ' '; 'c'; 'l'; 'o'; 's';
                                      'u'; 'r'; 'e'; ' '; 'p'; 'r'; 'o'; 'p';
                                      'e'; 'r'; 't'; 'i'; 'e'; 's'; ' '; 'o';
                                      'f'; ' '; 'r'; 'h'; 's'; ' '; 'o'; 'f';
                                      ' '; 'u'; 's'; 'a'; 'b'; 'l'; 'e'; ' ';
                                      'r'; 'u'; 'l'; 'e'; 's'])
                              (comp shows_nl x))))
                 (fun _ ->
                   bind (catch
                          (check_subseteq
                            (eqc (eq_terma _A1 _B1) (eq_terma _A1 _B1)) p
                            (map fst ggs))
                          (fun x ->
                            Inl (comp (shows_string
['t'; 'h'; 'e'; 'r'; 'e'; ' '; 'i'; 's'; ' '; 'n'; 'o'; ' '; 's'; 'i'; 'z'; 'e';
  '-'; 'c'; 'h'; 'a'; 'n'; 'g'; 'e'; ' '; 'g'; 'r'; 'a'; 'p'; 'h'; ' '; 'f';
  'o'; 'r'; ' '; 'D'; 'P'; ' '])
                                  (shows_rule _A2 _B2 x))))
                     (fun _ ->
                       bind (catch
                              (forallm
                                (fun (a, b) ->
                                  let (s, t) = a in
                                  (fun (stri, nstri) ->
                                    catch (bind
    (check (not (is_var s) && (not (is_var t) && null (rma (the (root t)))))
      (comp (shows_string
              ['r'; 'o'; 'o'; 't'; 's'; ' '; 'o'; 'f'; ' '; 'D'; 'P'; ' '; 'm';
                'u'; 's'; 't'; ' '; 'b'; 'e'; ' '; 'c'; 'o'; 'n'; 's'; 't'; 'r';
                'u'; 'c'; 't'; 'o'; 'r'; 's'])
        shows_nl))
    (fun _ ->
      bind (catch
             (forallm
               (fun i ->
                 check (less_eq_nat i (size_list (args s)))
                   (comp (shows_string
                           ['l'; 'e'; 'f'; 't'; '-'; 'i'; 'n'; 'd'; 'e'; 'x';
                             ' '; 't'; 'o'; ' '; 'l'; 'a'; 'r'; 'g'; 'e'])
                     (comp (shows_prec_nat Zero_nat i) shows_nl)))
               (remdups eq_nata (map fst (append stri nstri))))
             (fun x -> Inl (snd x)))
        (fun _ ->
          bind (catch
                 (forallm
                   (fun j ->
                     check (less_eq_nat j (size_list (args t)) &&
                             isOK (check_ur_closed_term_rm_af (_A1, _A2)
                                    (_B1, _B2) rma ur pi (get_arg t j)))
                       (comp (shows_string
                               ['r'; 'i'; 'g'; 'h'; 't'; '-'; 'i'; 'n'; 'd';
                                 'e'; 'x'; ' '; 't'; 'o'; ' '; 'l'; 'a'; 'r';
                                 'g'; 'e'; ' '; 'o'; 'r'; ' '; 'a'; 'r'; 'g';
                                 'u'; 'm'; 'e'; 'n'; 't'; ' '; 'v'; 'i'; 'o';
                                 'l'; 'a'; 't'; 'e'; 's'; ' '; 'u'; 's'; 'a';
                                 'b'; 'l'; 'e'; '-'; 'r'; 'u'; 'l'; 'e'; 's';
                                 ' '; 'c'; 'o'; 'n'; 'd'; 'i'; 't'; 'i'; 'o';
                                 'n'])
                         (shows_prec_nat Zero_nat j)))
                   (remdups eq_nata (map snd (append stri nstri))))
                 (fun x -> Inl (snd x)))
            (fun _ ->
              let _ = args s in
              let _ = args t in
              bind (catch
                     (forallm
                       (fun (i, j) ->
                         check (isOK (checkSS (get_arg s i, get_arg t j)))
                           (comp (shows_string
                                   ['p'; 'r'; 'o'; 'b'; 'l'; 'e'; 'm'; ' '; 'w';
                                     'i'; 't'; 'h'; ' '; 'e'; 'd'; 'g'; 'e';
                                     ' '])
                             (comp (shows_prec_nat Zero_nat i)
                               (comp (shows_string
                                       [' '; '-'; 'S'; '-'; '>'; ' '])
                                 (shows_prec_nat Zero_nat j)))))
                       stri)
                     (fun x -> Inl (snd x)))
                (fun _ ->
                  catch (forallm
                          (fun (i, j) ->
                            check (isOK (checkNSS (get_arg s i, get_arg t j)))
                              (comp (shows_string
                                      ['p'; 'r'; 'o'; 'b'; 'l'; 'e'; 'm'; ' ';
'w'; 'i'; 't'; 'h'; ' '; 'e'; 'd'; 'g'; 'e'; ' '])
                                (comp (shows_prec_nat Zero_nat i)
                                  (comp (shows_string
  [' '; '-'; 'N'; 'S'; '-'; '>'; ' '])
                                    (shows_prec_nat Zero_nat j)))))
                          nstri)
                    (fun x -> Inl (snd x)))))))
                                      (fun x ->
Inl (comp (shows_string
            ['p'; 'r'; 'o'; 'b'; 'l'; 'e'; 'm'; 's'; ' '; 'w'; 'i'; 't'; 'h';
              ' '; 'D'; 'P'; ' '])
      (comp (shows_rule _A2 _B2 (s, t)) (comp shows_nl x)))))
                                    b)
                                ggs)
                              (fun x -> Inl (snd x)))
                         (fun _ ->
                           bind (catch
                                  (catch (forallm checkNSS ur)
                                    (fun x -> Inl (snd x)))
                                  (fun x ->
                                    Inl (comp
  (shows_string
    ['p'; 'r'; 'o'; 'b'; 'l'; 'e'; 'm'; ' '; 'w'; 'h'; 'e'; 'n'; ' '; 'o'; 'r';
      'i'; 'e'; 'n'; 't'; 'i'; 'n'; 'g'; ' '; 'u'; 's'; 'a'; 'b'; 'l'; 'e'; ' ';
      'r'; 'u'; 'l'; 'e'; 's'])
  (comp shows_nl x))))
                             (fun _ ->
                               check (check_SCT
                                       (eqc (eq_terma _A1 _B1)
 (eqc (eq_terma _A1 _B1) (eqc (eq_hterm _A1 ()) (eq_hterm _A1 ()))))
                                       eq_nata
                                       (fun (_, (t, (_, ct))) (u, (_, (cu, _)))
 -> gcMatchI _A1 _B1 ct u && gcMatchI _A1 _B1 cu t)
                                       (map
 (fun (a, b) ->
   let (s, t) = a in
   (fun (stri, nstri) ->
     let e = (s, (t, (capRM _A1 _B1 rrm s, capRM _A1 _B1 rm t))) in
     Scg (e, e, stri, nstri))
     b)
 ggs))
                                 (comp (shows_string
 ['s'; 'i'; 'z'; 'e'; '-'; 'c'; 'h'; 'a'; 'n'; 'g'; 'e'; ' '; 'a'; 'n'; 'a';
   'l'; 'y'; 's'; 'i'; 's'; ' '; 'f'; 'a'; 'i'; 'l'; 'e'; 'd'])
                                   shows_nl))))))
          (fun x ->
            Inl (comp (shows_string
                        ['c'; 'o'; 'u'; 'l'; 'd'; ' '; 'n'; 'o'; 't'; ' '; 'a';
                          'p'; 'p'; 'l'; 'y'; ' '; 't'; 'h'; 'e'; ' '; 's'; 'i';
                          'z'; 'e'; '-'; 'c'; 'h'; 'a'; 'n'; 'g'; 'e'; ' '; 'p';
                          'r'; 'o'; 'c'; 'e'; 's'; 's'; 'o'; 'r'; ' '; 'w'; 'i';
                          't'; 'h'; ' '; 't'; 'h'; 'e'; ' '; 'f'; 'o'; 'l'; 'l';
                          'o'; 'w'; 'i'; 'n'; 'g'; ' '])
                  (comp shows_nl
                    (comp representation
                      (comp shows_nl
                        (comp (shows_string
                                ['f'; 'o'; 'r'; ' '; 't'; 'h'; 'e'; ' '; 'f';
                                  'o'; 'l'; 'l'; 'o'; 'w'; 'i'; 'n'; 'g'; ' ';
                                  'r'; 'e'; 'a'; 's'; 'o'; 'n'])
                          (comp shows_nl x))))))));;

let rec shows_arctic_rat
  = function Num_arc_rat i -> shows_prec_rat Zero_nat i
    | MinInfty_rat -> shows_string ['-'; 'i'; 'n'; 'f'];;

let rec shows_prec_arctic_rat d ari = shows_arctic_rat ari;;

let rec shows_list_arctic_rat
  asa = shows_list_aux (shows_prec_arctic_rat Zero_nat) asa;;

let show_arctic_rat =
  ({shows_prec = shows_prec_arctic_rat; shows_list = shows_list_arctic_rat} :
    arctic_rat show);;

let rec create_sct_ur_af_proc_comp (_A1, _A2) (_B1, _B2) (_C1, _C2)
  c i p rm rrm ur graphs =
    let pi = to_lpoly_inter _B1 (get_comp_def c) (get_comp_1 c) i in
    bind (check_lpoly_coeffs _A2 _B2 (get_comp_wf_elem c) (get_comp_apos c)
           (fun x -> get_comp_ge c x (get_comp_0 c)) i)
      (fun _ ->
        check_sct_ur_af_proc (_B1, _B2) (_C1, _C2)
          (check_polo_sI (_A1, _A2) _B2 (_C1, _C2) (get_comp_mono c)
            (get_comp_0 c) (get_comp_1 c) (get_comp_plus c) (get_comp_times c)
            (get_comp_max0 c) (get_comp_ge c) (get_comp_gt c) pi)
          (check_polo_nsI (_A1, _A2) _B2 (_C1, _C2) (get_comp_0 c)
            (get_comp_1 c) (get_comp_plus c) (get_comp_times c)
            (get_comp_max0 c) (get_comp_ge c) pi)
          (comp (shows_string
                  ['p'; 'o'; 'l'; 'y'; 'n'; 'o'; 'm'; 'i'; 'a'; 'l'; ' '; 'o';
                    'r'; 'd'; 'e'; 'r'; ' '; '('])
            (comp (shows_string (get_comp_description c))
              (comp (shows_string [')'])
                (comp shows_nl
                  (comp (create_lpoly_repr (_A1, _A2) (_B1, _B2) c i)
                    shows_nl)))))
          p graphs rm rrm ur (create_afI _A1 _B1 (get_comp_0 c) i));;

let rec create_sct_ur_af_proc (_A1, _A2) (_B1, _B2)
  = function
    Int_carrier i ->
      create_sct_ur_af_proc_comp (eq_int, show_int) (_A1, _A2) (_B1, _B2)
        int_components i
    | Nat_carrier i ->
        create_sct_ur_af_proc_comp (eq_nata, show_nat) (_A1, _A2) (_B1, _B2)
          nat_components i
    | Rat_carrier (delta, i) ->
        create_sct_ur_af_proc_comp (eq_rata, show_rata) (_A1, _A2) (_B1, _B2)
          (rat_components delta) i
    | Arctic_carrier i ->
        create_sct_ur_af_proc_comp (eq_arctica, show_arctic) (_A1, _A2)
          (_B1, _B2) arctic_components i
    | Arctic_rat_carrier (delta, i) ->
        create_sct_ur_af_proc_comp (eq_arctic_rata, show_arctic_rat) (_A1, _A2)
          (_B1, _B2) (arctic_rat_components delta) i
    | Int_mat_carrier (n, sd, i) ->
        create_sct_ur_af_proc_comp
          (eq_lista (eq_lista eq_int), show_list (show_list show_int))
          (_A1, _A2) (_B1, _B2) (mat_components n sd int_components) i
    | Rat_mat_carrier (delta, n, sd, i) ->
        create_sct_ur_af_proc_comp
          (eq_lista (eq_lista eq_rata), show_list (show_list show_rata))
          (_A1, _A2) (_B1, _B2) (mat_components n sd (rat_components delta)) i
    | Arc_mat_carrier (n, i) ->
        create_sct_ur_af_proc_comp
          (eq_lista (eq_lista eq_arctica), show_list (show_list show_arctic))
          (_A1, _A2) (_B1, _B2) (mat_arctic_components n arctic_components) i
    | Arc_rat_mat_carrier (delta, n, i) ->
        create_sct_ur_af_proc_comp
          (eq_lista (eq_lista eq_arctic_rata),
            show_list (show_list show_arctic_rat))
          (_A1, _A2) (_B1, _B2)
          (mat_arctic_components n (arctic_rat_components delta)) i
    | Lpo (pri, pi) ->
        (fun p rm rrm ur graphs ->
          let m = max ord_nat (get_max_arity _A1 ur) (get_max_arity _A1 p) in
          let pr = lpo_repr_to_pr _A1 pri in
          let pia = afs_list_to_afs _A1 pi in
          bind (check_wf_afs _A2 pi)
            (fun _ ->
              check_sct_ur_af_proc (_A1, _A2) (_B1, _B2)
                (afs_check _A2 _B2 lpoS pia
                  (lpo_strict (show_filtered _A2) (_B1, _B2) pr m))
                (afs_check _A2 _B2 lpoNS pia
                  (lpo_nstrict (show_filtered _A2) (_B1, _B2) pr m))
                (shows_lpo_af _A2 pri pi) p graphs rm rrm ur
                (afs_list_to_af _A1 pi)));;

let rec check_supteq (_A1, _A2) (_B1, _B2)
  s t = check (supteqp _A1 _B1 s t)
          (comp (shows_prec_term _A2 _B2 Zero_nat t)
            (comp (shows_string
                    [' '; 'i'; 's'; ' '; 'n'; 'o'; 't'; ' '; 'a'; ' '; 's'; 'u';
                      'b'; 't'; 'e'; 'r'; 'm'; ' '; 'o'; 'f'; ' '])
              (shows_prec_term _A2 _B2 Zero_nat s)));;

let rec check_supt (_A1, _A2) (_B1, _B2)
  s t = check (suptp _A1 _B1 s t)
          (comp (shows_prec_term _A2 _B2 Zero_nat t)
            (comp (shows_string
                    [' '; 'i'; 's'; ' '; 'n'; 'o'; 't'; ' '; 'a'; ' '; 'p'; 'r';
                      'o'; 'p'; 'e'; 'r'; ' '; 's'; 'u'; 'b'; 't'; 'e'; 'r';
                      'm'; ' '; 'o'; 'f'; ' '])
              (shows_prec_term _A2 _B2 Zero_nat s)));;

let rec check_sct_subterm_proc (_A1, _A2) (_B1, _B2)
  p gs rm rrm =
    (match rm
      with None ->
        Inl (comp (shows_string
                    ['T'; 'R'; 'S'; ' '; 'h'; 'a'; 's'; ' '; 'v'; 'a'; 'r'; 'i';
                      'a'; 'b'; 'l'; 'e'; 's'; ' '; 'a'; 's'; ' '; 'l'; 'e';
                      'f'; 't'; '-'; 'h'; 'a'; 'n'; 'd'; ' '; 's'; 'i'; 'd';
                      'e'; 's'])
              shows_nl)
      | Some irm ->
        catch (bind (catch
                      (check_subseteq
                        (eqc (eq_terma _A1 _B1) (eq_terma _A1 _B1)) p
                        (map fst gs))
                      (fun x ->
                        Inl (comp (shows_string
                                    ['t'; 'h'; 'e'; 'r'; 'e'; ' '; 'i'; 's';
                                      ' '; 'n'; 'o'; ' '; 's'; 'i'; 'z'; 'e';
                                      '-'; 'c'; 'h'; 'a'; 'n'; 'g'; 'e'; ' ';
                                      'g'; 'r'; 'a'; 'p'; 'h'; ' '; 'f'; 'o';
                                      'r'; ' '; 'D'; 'P'; ' '])
                              (comp (shows_rule _A2 _B2 x) shows_nl))))
                (fun _ ->
                  bind (catch
                         (forallm
                           (fun (a, b) ->
                             let (s, t) = a in
                             (fun (stri, nstri) ->
                               catch (bind (check
     (not (is_var s) && (not (is_var t) && null (irm (the (root t)))))
     (comp (shows_string
             ['r'; 'o'; 'o'; 't'; 's'; ' '; 'o'; 'f'; ' '; 'D'; 'P'; ' '; 'm';
               'u'; 's'; 't'; ' '; 'b'; 'e'; ' '; 'c'; 'o'; 'n'; 's'; 't'; 'r';
               'u'; 'c'; 't'; 'o'; 'r'; 's'])
       shows_nl))
                                       (fun _ ->
 let n = size_list (args s) in
 let m = size_list (args t) in
 bind (catch
        (forallm
          (fun (i, j) ->
            check (less_eq_nat i n &&
                    (less_eq_nat j m &&
                      isOK (check_supt (_A1, _A2) (_B1, _B2) (get_arg s i)
                             (get_arg t j))))
              (comp (shows_string
                      ['p'; 'r'; 'o'; 'b'; 'l'; 'e'; 'm'; ' '; 'w'; 'i'; 't';
                        'h'; ' '; 'e'; 'd'; 'g'; 'e'; ' '])
                (comp (shows_prec_nat Zero_nat i)
                  (comp (shows_string [' '; '|'; '>'; ' '])
                    (comp (shows_prec_nat Zero_nat j) shows_nl)))))
          stri)
        (fun x -> Inl (snd x)))
   (fun _ ->
     catch (forallm
             (fun (i, j) ->
               check (less_eq_nat i n &&
                       (less_eq_nat j m &&
                         isOK (check_supteq (_A1, _A2) (_B1, _B2) (get_arg s i)
                                (get_arg t j))))
                 (comp (shows_string
                         ['p'; 'r'; 'o'; 'b'; 'l'; 'e'; 'm'; ' '; 'w'; 'i'; 't';
                           'h'; ' '; 'e'; 'd'; 'g'; 'e'; ' '])
                   (comp (shows_prec_nat Zero_nat i)
                     (comp (shows_string [' '; '|'; '>'; '='; ' '])
                       (comp (shows_prec_nat Zero_nat j) shows_nl)))))
             nstri)
       (fun x -> Inl (snd x)))))
                                 (fun x ->
                                   Inl (comp
 (shows_string
   ['p'; 'r'; 'o'; 'b'; 'l'; 'e'; 'm'; 's'; ' '; 'w'; 'i'; 't'; 'h'; ' '; 'D';
     'P'; ' '])
 (comp (shows_rule _A2 _B2 (s, t)) (comp shows_nl x)))))
                               b)
                           gs)
                         (fun x -> Inl (snd x)))
                    (fun _ ->
                      check (check_SCT
                              (eqc (eq_terma _A1 _B1)
                                (eqc (eq_terma _A1 _B1)
                                  (eqc (eq_hterm _A1 ()) (eq_hterm _A1 ()))))
                              eq_nata
                              (fun (_, (t, (_, ct))) (u, (_, (cu, _))) ->
                                gcMatchI _A1 _B1 ct u && gcMatchI _A1 _B1 cu t)
                              (map (fun (a, b) ->
                                     let (s, t) = a in
                                     (fun (stri, nstri) ->
                                       let e =
 (s, (t, (capRM _A1 _B1 rrm s, capRM _A1 _B1 rm t))) in
                                       Scg (e, e, stri, nstri))
                                       b)
                                gs))
                        (comp (shows_string
                                ['s'; 'i'; 'z'; 'e'; '-'; 'c'; 'h'; 'a'; 'n';
                                  'g'; 'e'; ' '; 'a'; 'n'; 'a'; 'l'; 'y'; 's';
                                  'i'; 's'; ' '; 'f'; 'a'; 'i'; 'l'; 'e'; 'd'])
                          shows_nl))))
          (fun x ->
            Inl (comp (shows_string
                        ['c'; 'o'; 'u'; 'l'; 'd'; ' '; 'n'; 'o'; 't'; ' '; 'a';
                          'p'; 'p'; 'l'; 'y'; ' '; 't'; 'h'; 'e'; ' '; 's'; 'i';
                          'z'; 'e'; '-'; 'c'; 'h'; 'a'; 'n'; 'g'; 'e'; ' '; 'p';
                          'r'; 'o'; 'c'; 'e'; 's'; 's'; 'o'; 'r'; ' '; 'w'; 'i';
                          't'; 'h'; ' '; 's'; 'u'; 'b'; 't'; 'e'; 'r'; 'm'; '-';
                          'r'; 'e'; 'l'; 'a'; 't'; 'i'; 'o'; 'n'; ' '; 'f'; 'o';
                          'r'; ' '; 't'; 'h'; 'e'; ' '; 'f'; 'o'; 'l'; 'l'; 'o';
                          'w'; 'i'; 'n'; 'g'; ' '; 'r'; 'e'; 'a'; 's'; 'o';
                          'n'])
                  (comp shows_nl x))));;

let rec create_sp_representation _A
  (SimpProj sp) =
    shows_map
      (fun (f, n) ->
        comp (shows_string ['p'; 'i'; '('])
          (comp (shows_prec _A Zero_nat f)
            (comp (shows_string [')'; ' '; '='; ' '])
              (comp (shows_prec_nat Zero_nat n) shows_nl))))
      sp;;

let rec check_weak (_A1, _A2) (_B1, _B2)
  p r = catch (check
                (eq_term _A1 _B1 (proj_term p (fst r)) (proj_term p (snd r)))
                (comp (shows_string
                        ['t'; 'h'; 'e'; ' '; 'p'; 'r'; 'o'; 'j'; 'e'; 'c'; 't';
                          'e'; 'd'; ' '; 'l'; 'h'; 's'; ' '; 'i'; 's'; ' '; 'n';
                          'o'; 't'; ' '; 'e'; 'q'; 'u'; 'a'; 'l'; ' '; 't'; 'o';
                          ' '; 't'; 'h'; 'e'; ' '; 'p'; 'r'; 'o'; 'j'; 'e'; 'c';
                          't'; 'e'; 'd'; ' '; 'r'; 'h'; 's'])
                  shows_nl))
          (fun x ->
            Inl (comp (shows_string
                        ['C'; 'o'; 'u'; 'l'; 'd'; ' '; 'n'; 'o'; 't'; ' '; 'o';
                          'r'; 'i'; 'e'; 'n'; 't'; ' '; 'r'; 'u'; 'l'; 'e';
                          ' '])
                  (comp (shows_rule _A2 _B2 r)
                    (comp (shows_string [','; ' '; 's'; 'i'; 'n'; 'c'; 'e'])
                      (comp shows_nl
                        (comp (shows_prec_term _A2 _B2 Zero_nat
                                (proj_term p (fst r)))
                          (comp (shows_string [' '; '!'; '='; ' '])
                            (comp (shows_prec_term _A2 _B2 Zero_nat
                                    (proj_term p (snd r)))
                              (comp shows_nl x)))))))));;

let rec check_strict (_A1, _A2) (_B1, _B2)
  ra rseqm p r =
    let s = proj_term p (fst r) in
    let t = proj_term p (snd r) in
    (match rseqm r with None -> check_supt (_A1, _A2) (_B1, _B2) s t
      | Some rseq ->
        (if eq_nat (size_list rseq) Zero_nat
          then check_supt (_A1, _A2) (_B1, _B2) s t
          else let u = snd (snd (last rseq)) in
               bind (check_rsteps (_A1, _A2) (_B1, _B2) ra rseq s u)
                 (fun _ -> check_supteq (_A1, _A2) (_B1, _B2) u t)));;

let rec check_subterm_criterion_proc (_A1, _A2) (_B1, _B2)
  pa rep rseqm p p' r rm =
    let ds = defs_list _A1 r in
    bind (catch
           (forallm
             (fun (l, ra) ->
               bind (bind (check_no_var _A2 _B2 l)
                      (fun _ -> check_no_var _A2 _B2 ra))
                 (fun _ -> check_no_defined_root (_A1, _A2) _B2 ds ra))
             p)
           (fun x -> Inl (snd x)))
      (fun _ ->
        bind (catch
               (forallm (check_strict (_A1, _A2) (_B1, _B2) r rseqm pa)
                 (list_diff (eqc (eq_terma _A1 _B1) (eq_terma _A1 _B1)) p p'))
               (fun x -> Inl (snd x)))
          (fun _ ->
            catch (forallm (check_weak (_A1, _A2) (_B1, _B2) pa) p')
              (fun x -> Inl (snd x))));;

let rec create_rseq_map _A _B
  rseqs r = lookup (eqc (eq_terma _A _B) (eq_terma _A _B)) r rseqs;;

let rec create_subterm_criterion_proc (_A1, _A2) (_B1, _B2)
  sp rseqs p rP r rm =
    let pi = create_sp _A1 sp in
    let rseqm = create_rseq_map _A1 _B1 rseqs in
    check_subterm_criterion_proc (_A1, _A2) (_B1, _B2) pi
      (create_sp_representation _A2 sp) rseqm p rP r rm;;

let rec create_sem_lab_proc (_A1, _A2) (_B1, _B2) (_C1, _C2)
  sli l r1 r2 r3 r4 =
    let c = get_largest_element sli in
    let ca = upt Zero_nat (Suc c) in
    let i = sl_inter_to_inter _A1 _B1 sli in
    catch (check_sem_lab_proc (_A1, _A2) (_B1, _B2) (_C1, _C2)
            (check_sl_lab_trs (_A1, _A2) (_B1, _B2) (eq_nata, show_nat)
              (_C1, _C2) i l ca)
            (check_sl_model_lab_trs (_A1, _A2) (_B1, _B2) (eq_nata, show_nat)
              (_C1, _C2) i l ca)
            (fun _ -> false) i l r1 r2 r3 r4)
      (fun x ->
        Inl (comp (shows_prec_sl_inter _A2 _B2 Zero_nat sli)
              (comp shows_nl x)));;

let rec check_unlab_proc (_A1, _A2) (_B1, _B2) (_C1, _C2)
  p r ulP ulR =
    catch (let ulr = unlab_of_trsI _A1 _B1 _C1 r in
           bind (check_subseteq
                  (eqc (eq_terma (eq_laba _A1 _B1) _C1)
                    (eq_terma (eq_laba _A1 _B1) _C1))
                  (map_trsI unlab p) ulP)
             (fun _ ->
               bind (check_subseteq
                      (eqc (eq_terma (eq_laba _A1 _B1) _C1)
                        (eq_terma (eq_laba _A1 _B1) _C1))
                      ulr ulR)
                 (fun _ ->
                   check_subseteq
                     (eqc (eq_terma (eq_laba _A1 _B1) _C1)
                       (eq_terma (eq_laba _A1 _B1) _C1))
                     ulR ulr)))
      (fun x ->
        Inl (comp (shows_string
                    ['p'; 'r'; 'o'; 'b'; 'l'; 'e'; 'm'; ' '; 'w'; 'i'; 't'; 'h';
                      ' '; 'r'; 'u'; 'l'; 'e'; ' '])
              (comp (shows_rule (show_lab _A2 _B2) _C2 x) shows_nl)));;

let rec matchCapBelowI _A _B
  l r (Fun (f, ts)) = gcMatchI _A _B (HFun (f, map (capI _A _B r) ts)) l;;

let rec check_ur_closed_term_af (_A1, _A2) (_B1, _B2)
  uu uv uw x3 = match uu, uv, uw, x3 with uu, uv, uw, Var x -> Inr ()
    | r, ur, pi, Fun (f, ts) ->
        let n = size_list ts in
        bind (catch
               (forallm_index
                 (fun t i ->
                   (if pi f n i
                     then check_ur_closed_term_af (_A1, _A2) (_B1, _B2) r ur pi
                            t
                     else Inr ()))
                 ts)
               (fun x -> Inl (snd x)))
          (fun _ ->
            catch (forallm
                    (fun lr ->
                      check (member (eqc (eq_terma _A1 _B1) (eq_terma _A1 _B1))
                               lr ur
                              || not (matchCapBelowI _A1 _B1 (fst lr) r
                                       (Fun (f, ts))))
                        (comp (shows_string
                                ['d'; 'u'; 'e'; ' '; 't'; 'o'; ' '; 't'; 'h';
                                  'e'; ' '; 's'; 'u'; 'b'; 't'; 'e'; 'r'; 'm';
                                  ' '])
                          (comp (shows_prec_term _A2 _B2 Zero_nat (Fun (f, ts)))
                            (comp (shows_string
                                    [' '; 'o'; 'f'; ' '; 's'; 'o'; 'm'; 'e';
                                      ' '; 'u'; 's'; 'a'; 'b'; 'l'; 'e'; ' ';
                                      'r'; 'h'; 's'; ','; ' '; 'r'; 'u'; 'l';
                                      'e'; ' '])
                              (comp (shows_rule _A2 _B2 lr)
                                (shows_string
                                  [' '; 's'; 'h'; 'o'; 'u'; 'l'; 'd'; ' '; 'b';
                                    'e'; ' '; 'u'; 's'; 'a'; 'b'; 'l'; 'e';
                                    '.']))))))
                    r)
              (fun x -> Inl (snd x)));;

let rec check_ur_P_closed_af (_A1, _A2) (_B1, _B2)
  r ur p pi =
    bind (catch
           (catch
             (forallm
               (fun lr ->
                 check_ur_closed_term_af (_A1, _A2) (_B1, _B2) r ur pi (snd lr))
               ur)
             (fun x -> Inl (snd x)))
           (fun x ->
             Inl (comp (shows_string
                         ['e'; 'r'; 'r'; 'o'; 'r'; ' '; 'w'; 'h'; 'e'; 'n'; ' ';
                           'c'; 'h'; 'e'; 'c'; 'k'; 'i'; 'n'; 'g'; ' '; 'c';
                           'l'; 'o'; 's'; 'u'; 'r'; 'e'; ' '; 'p'; 'r'; 'o';
                           'p'; 'e'; 'r'; 't'; 'i'; 'e'; 's'; ' '; 'o'; 'f';
                           ' '; 'r'; 'h'; 's'; ' '; 'o'; 'f'; ' '; 'u'; 's';
                           'a'; 'b'; 'l'; 'e'; ' '; 'r'; 'u'; 'l'; 'e'; 's'])
                   (comp shows_nl x))))
      (fun _ ->
        catch (catch
                (forallm
                  (fun st ->
                    check_ur_closed_term_af (_A1, _A2) (_B1, _B2) r ur pi
                      (snd st))
                  p)
                (fun x -> Inl (snd x)))
          (fun x ->
            Inl (comp (shows_string
                        ['e'; 'r'; 'r'; 'o'; 'r'; ' '; 'w'; 'h'; 'e'; 'n'; ' ';
                          'c'; 'h'; 'e'; 'c'; 'k'; 'i'; 'n'; 'g'; ' '; 'c'; 'l';
                          'o'; 's'; 'u'; 'r'; 'e'; ' '; 'p'; 'r'; 'o'; 'p'; 'e';
                          'r'; 't'; 'i'; 'e'; 's'; ' '; 'o'; 'f'; ' '; 'r'; 'h';
                          's'; ' '; 'o'; 'f'; ' '; 'D'; 'P'; 's'])
                  (comp shows_nl x))));;

let rec check_ur_P_closed_rm_af (_A1, _A2) (_B1, _B2)
  r rmo ur p pi =
    (match rmo with None -> check_ur_P_closed_af (_A1, _A2) (_B1, _B2) r ur p pi
      | Some rm ->
        bind (catch
               (catch
                 (forallm
                   (fun lr ->
                     check_ur_closed_term_rm_af (_A1, _A2) (_B1, _B2) rm ur pi
                       (snd lr))
                   ur)
                 (fun x -> Inl (snd x)))
               (fun x ->
                 Inl (comp (shows_string
                             ['e'; 'r'; 'r'; 'o'; 'r'; ' '; 'w'; 'h'; 'e'; 'n';
                               ' '; 'c'; 'h'; 'e'; 'c'; 'k'; 'i'; 'n'; 'g'; ' ';
                               'c'; 'l'; 'o'; 's'; 'u'; 'r'; 'e'; ' '; 'p'; 'r';
                               'o'; 'p'; 'e'; 'r'; 't'; 'i'; 'e'; 's'; ' '; 'o';
                               'f'; ' '; 'r'; 'h'; 's'; ' '; 'o'; 'f'; ' '; 'u';
                               's'; 'a'; 'b'; 'l'; 'e'; ' '; 'r'; 'u'; 'l'; 'e';
                               's'])
                       (comp shows_nl x))))
          (fun _ ->
            catch (catch
                    (forallm
                      (fun st ->
                        check_ur_closed_term_rm_af (_A1, _A2) (_B1, _B2) rm ur
                          pi (snd st))
                      p)
                    (fun x -> Inl (snd x)))
              (fun x ->
                Inl (comp (shows_string
                            ['e'; 'r'; 'r'; 'o'; 'r'; ' '; 'w'; 'h'; 'e'; 'n';
                              ' '; 'c'; 'h'; 'e'; 'c'; 'k'; 'i'; 'n'; 'g'; ' ';
                              'c'; 'l'; 'o'; 's'; 'u'; 'r'; 'e'; ' '; 'p'; 'r';
                              'o'; 'p'; 'e'; 'r'; 't'; 'i'; 'e'; 's'; ' '; 'o';
                              'f'; ' '; 'r'; 'h'; 's'; ' '; 'o'; 'f'; ' '; 'D';
                              'P'; 's'])
                      (comp shows_nl x)))));;

let rec check_mono_redpair_ur_proc (_A1, _A2) (_B1, _B2)
  checkMono checkSS checkNSS name representation p rP r rm rR ur =
    catch (let us = concat_map (comp (funas_term_list _A1) snd) (append p ur) in
           let filt = filter (fun lr -> wf_terms_list _A1 us (fst lr)) in
           let (_, nSp) = partition (fun foo -> isOK (checkSS foo)) p in
           let (_, nSr) = partition (fun foo -> isOK (checkSS foo)) ur in
           bind (catch checkMono
                  (fun x ->
                    Inl (comp (shows_string
                                ['p'; 'r'; 'o'; 'b'; 'l'; 'e'; 'm'; ' '; 'w';
                                  'i'; 't'; 'h'; ' '; 'm'; 'o'; 'n'; 'o'; 't';
                                  'o'; 'n'; 'i'; 'c'; 'i'; 't'; 'y'; ' '; 'o';
                                  'f'; ' '; 's'; 't'; 'r'; 'i'; 'c'; 't'; ' ';
                                  'o'; 'r'; 'd'; 'e'; 'r'])
                          (comp shows_nl x))))
             (fun _ ->
               bind (catch
                      (check_ur_P_closed_rm_af (_A1, _A2) (_B1, _B2) r rm ur p
                        pi_full)
                      (fun x ->
                        Inl (comp (shows_string
                                    ['p'; 'r'; 'o'; 'b'; 'l'; 'e'; 'm'; ' ';
                                      'w'; 'i'; 't'; 'h'; ' '; 'c'; 'l'; 'o';
                                      's'; 'u'; 'r'; 'e'; ' '; 'p'; 'r'; 'o';
                                      'p'; 'e'; 'r'; 't'; 'i'; 'e'; 's'; ' ';
                                      'o'; 'f'; ' '; 'u'; 's'; 'a'; 'b'; 'l';
                                      'e'; ' '; 'r'; 'u'; 'l'; 'e'; 's'])
                              (comp shows_nl x))))
                 (fun _ ->
                   bind (catch
                          (catch (forallm checkNSS nSr) (fun x -> Inl (snd x)))
                          (fun x ->
                            Inl (comp (shows_string
['p'; 'r'; 'o'; 'b'; 'l'; 'e'; 'm'; ' '; 'w'; 'h'; 'e'; 'n'; ' '; 'o'; 'r'; 'i';
  'e'; 'n'; 't'; 'i'; 'n'; 'g'; ' '; 'u'; 's'; 'a'; 'b'; 'l'; 'e'; ' '; 'r';
  'u'; 'l'; 'e'; 's'])
                                  (comp shows_nl x))))
                     (fun _ ->
                       bind (catch
                              (catch (forallm checkNSS nSp)
                                (fun x -> Inl (snd x)))
                              (fun x ->
                                Inl (comp (shows_string
    ['p'; 'r'; 'o'; 'b'; 'l'; 'e'; 'm'; ' '; 'w'; 'h'; 'e'; 'n'; ' '; 'o'; 'r';
      'i'; 'e'; 'n'; 't'; 'i'; 'n'; 'g'; ' '; 'D'; 'P'; 's'])
                                      (comp shows_nl x))))
                         (fun _ ->
                           bind (catch
                                  (check_subseteq
                                    (eqc (eq_terma _A1 _B1) (eq_terma _A1 _B1))
                                    rR r)
                                  (fun _ ->
                                    Inl (comp
  (shows_string
    ['p'; 'r'; 'o'; 'b'; 'l'; 'e'; 'm'; ':'; ' '; 'a'; 'd'; 'd'; 'i'; 't'; 'i';
      'o'; 'n'; 'a'; 'l'; ' '; 'r'; 'u'; 'l'; 'e'; '('; 's'; ')'; ' '; 't'; 'o';
      ' '; 'T'; 'R'; 'S'; ' '; 'a'; 'd'; 'd'; 'e'; 'd'])
  shows_nl)))
                             (fun _ ->
                               bind (catch
                                      (check_subseteq
(eqc (eq_terma _A1 _B1) (eq_terma _A1 _B1)) (filt nSr) rR)
                                      (fun x ->
Inl (match checkSS x with Inl err -> err | Inr _ -> internal_error)))
                                 (fun _ ->
                                   catch (check_subseteq
   (eqc (eq_terma _A1 _B1) (eq_terma _A1 _B1)) (filt nSp) rP)
                                     (fun x ->
                                       Inl
 (match checkSS x with Inl err -> err | Inr _ -> internal_error)))))))))
      (fun x ->
        Inl (comp (shows_string
                    ['c'; 'o'; 'u'; 'l'; 'd'; ' '; 'n'; 'o'; 't'; ' '; 'a'; 'p';
                      'p'; 'l'; 'y'; ' '; 't'; 'h'; 'e'; ' '; '('; 'm'; 'o';
                      'n'; 'o'; 't'; 'o'; 'n'; 'i'; 'c'; ')'; ' '; 'r'; 'e';
                      'd'; 'u'; 'c'; 't'; 'i'; 'o'; 'n'; ' '; 'p'; 'a'; 'i';
                      'r'; ' '; 'p'; 'r'; 'o'; 'c'; 'e'; 's'; 's'; 'o'; 'r';
                      ' '; 'w'; 'i'; 't'; 'h'; ' '; 't'; 'h'; 'e'; ' '; 'f';
                      'o'; 'l'; 'l'; 'o'; 'w'; 'i'; 'n'; 'g'])
              (comp shows_nl
                (comp representation
                  (comp shows_nl
                    (comp (shows_string
                            ['a'; 'n'; 'd'; ' '; 't'; 'h'; 'e'; ' '; 'f'; 'o';
                              'l'; 'l'; 'o'; 'w'; 'i'; 'n'; 'g'; ' '; 'u'; 's';
                              'a'; 'b'; 'l'; 'e'; ' '; 'r'; 'u'; 'l'; 'e'; 's'])
                      (comp (shows_rules _A2 _B2 ur) (comp shows_nl x))))))));;

let rec create_mono_redpair_ur_proc_comp (_A1, _A2) (_B1, _B2) (_C1, _C2)
  c cm i p rP r rm rR ur =
    let pi = to_lpoly_inter _B1 (get_comp_def c) (get_comp_1 c) i in
    bind (check_lpoly_coeffs _A2 _B2 (get_comp_wf_elem c) (get_comp_apos c)
           (fun x -> get_comp_ge c x (get_comp_0 c)) i)
      (fun _ ->
        check_mono_redpair_ur_proc (_B1, _B2) (_C1, _C2)
          (check_poly_monoI _A2 _B2 (get_comp_0 c) (get_comp_ge c) cm i)
          (check_polo_sI (_A1, _A2) _B2 (_C1, _C2) (get_comp_mono c)
            (get_comp_0 c) (get_comp_1 c) (get_comp_plus c) (get_comp_times c)
            (get_comp_max0 c) (get_comp_ge c) (get_comp_gt c) pi)
          (check_polo_nsI (_A1, _A2) _B2 (_C1, _C2) (get_comp_0 c)
            (get_comp_1 c) (get_comp_plus c) (get_comp_times c)
            (get_comp_max0 c) (get_comp_ge c) pi)
          (comp (shows_string
                  ['p'; 'o'; 'l'; 'y'; 'n'; 'o'; 'm'; 'i'; 'a'; 'l'; ' '; 'o';
                    'r'; 'd'; 'e'; 'r'; ' '; '('])
            (comp (shows_string (get_comp_description c)) (shows_string [')'])))
          (create_lpoly_repr (_A1, _A2) (_B1, _B2) c i) p rP r rm rR ur);;

let rec create_mono_redpair_ur_proc (_A1, _A2) (_B1, _B2)
  = function
    Int_carrier i ->
      create_mono_redpair_ur_proc_comp (eq_int, show_int) (_A1, _A2) (_B1, _B2)
        int_components int_mono i
    | Nat_carrier i ->
        create_mono_redpair_ur_proc_comp (eq_nata, show_nat) (_A1, _A2)
          (_B1, _B2) nat_components nat_mono i
    | Rat_carrier (delta, i) ->
        create_mono_redpair_ur_proc_comp (eq_rata, show_rata) (_A1, _A2)
          (_B1, _B2) (rat_components delta) rat_mono i
    | Arctic_carrier i ->
        (fun _ _ _ _ _ _ ->
          Inl (comp (shows_string
                      ['a'; 'r'; 'c'; 't'; 'i'; 'c'; ' '; 'n'; 'o'; 't'; ' ';
                        'm'; 'o'; 'n'; 'o'; 't'; 'o'; 'n'; 'i'; 'c'; ' '; 'f';
                        'o'; 'r'; ' '; 'T'; 'R'; 'S'; 's'])
                shows_nl))
    | Arctic_rat_carrier (delta, i) ->
        (fun _ _ _ _ _ _ ->
          Inl (comp (shows_string
                      ['a'; 'r'; 'c'; 't'; 'i'; 'c'; ' '; 'n'; 'o'; 't'; ' ';
                        'm'; 'o'; 'n'; 'o'; 't'; 'o'; 'n'; 'i'; 'c'; ' '; 'f';
                        'o'; 'r'; ' '; 'T'; 'R'; 'S'; 's'])
                shows_nl))
    | Arc_mat_carrier (n, i) ->
        (fun _ _ _ _ _ _ ->
          Inl (comp (shows_string
                      ['a'; 'r'; 'c'; 't'; 'i'; 'c'; ' '; 'n'; 'o'; 't'; ' ';
                        'm'; 'o'; 'n'; 'o'; 't'; 'o'; 'n'; 'i'; 'c'; ' '; 'f';
                        'o'; 'r'; ' '; 'T'; 'R'; 'S'; 's'])
                shows_nl))
    | Arc_rat_mat_carrier (delta, n, i) ->
        (fun _ _ _ _ _ _ ->
          Inl (comp (shows_string
                      ['a'; 'r'; 'c'; 't'; 'i'; 'c'; ' '; 'n'; 'o'; 't'; ' ';
                        'm'; 'o'; 'n'; 'o'; 't'; 'o'; 'n'; 'i'; 'c'; ' '; 'f';
                        'o'; 'r'; ' '; 'T'; 'R'; 'S'; 's'])
                shows_nl))
    | Int_mat_carrier (n, sd, i) ->
        create_mono_redpair_ur_proc_comp
          (eq_lista (eq_lista eq_int), show_list (show_list show_int))
          (_A1, _A2) (_B1, _B2) (mat_components n sd int_components)
          (mat_mono sd int_mono) i
    | Rat_mat_carrier (d, n, sd, i) ->
        create_mono_redpair_ur_proc_comp
          (eq_lista (eq_lista eq_rata), show_list (show_list show_rata))
          (_A1, _A2) (_B1, _B2) (mat_components n sd (rat_components d))
          (mat_mono sd rat_mono) i
    | Lpo (pri, pi) ->
        (fun p rP r rR rm ur ->
          let m = max ord_nat (get_max_arity _A1 r) (get_max_arity _A1 p) in
          let pr = lpo_repr_to_pr _A1 pri in
          let pia = afs_list_to_afs _A1 pi in
          bind (check_wf_afs _A2 pi)
            (fun _ ->
              check_mono_redpair_ur_proc (_A1, _A2) (_B1, _B2)
                (check_mono_afs _A2 pi)
                (afs_check _A2 _B2 lpoS pia
                  (lpo_strict (show_filtered _A2) (_B1, _B2) pr m))
                (afs_check _A2 _B2 lpoNS pia
                  (lpo_nstrict (show_filtered _A2) (_B1, _B2) pr m))
                lpo_repr (shows_lpo_af _A2 pri pi) p rP r rR rm ur));;

let rec check_mono_redpair_proc _A _B
  checkMono checkSS checkNSS name representation p rP r rR =
    catch (let (_, nSp) = partition (fun foo -> isOK (checkSS foo)) p in
           let (_, nSr) = partition (fun foo -> isOK (checkSS foo)) r in
           bind (catch checkMono
                  (fun x ->
                    Inl (comp (shows_string
                                ['p'; 'r'; 'o'; 'b'; 'l'; 'e'; 'm'; ' '; 'w';
                                  'i'; 't'; 'h'; ' '; 'm'; 'o'; 'n'; 'o'; 't';
                                  'o'; 'n'; 'i'; 'c'; 'i'; 't'; 'y'; ' '; 'o';
                                  'f'; ' '; 's'; 't'; 'r'; 'i'; 'c'; 't'; ' ';
                                  'o'; 'r'; 'd'; 'e'; 'r'])
                          (comp shows_nl x))))
             (fun _ ->
               bind (catch (catch (forallm checkNSS nSr) (fun x -> Inl (snd x)))
                      (fun x ->
                        Inl (comp (shows_string
                                    ['p'; 'r'; 'o'; 'b'; 'l'; 'e'; 'm'; ' ';
                                      'w'; 'h'; 'e'; 'n'; ' '; 'o'; 'r'; 'i';
                                      'e'; 'n'; 't'; 'i'; 'n'; 'g'; ' '; 'T';
                                      'R'; 'S'])
                              (comp shows_nl x))))
                 (fun _ ->
                   bind (catch
                          (catch (forallm checkNSS nSp) (fun x -> Inl (snd x)))
                          (fun x ->
                            Inl (comp (shows_string
['p'; 'r'; 'o'; 'b'; 'l'; 'e'; 'm'; ' '; 'w'; 'h'; 'e'; 'n'; ' '; 'o'; 'r'; 'i';
  'e'; 'n'; 't'; 'i'; 'n'; 'g'; ' '; 'D'; 'P'; 's'])
                                  (comp shows_nl x))))
                     (fun _ ->
                       bind (catch
                              (check_subseteq
                                (eqc (eq_terma _A _B) (eq_terma _A _B)) rR r)
                              (fun _ ->
                                Inl (shows_string
                                      ['p'; 'r'; 'o'; 'b'; 'l'; 'e'; 'm'; ':';
' '; 'a'; 'd'; 'd'; 'i'; 't'; 'i'; 'o'; 'n'; 'a'; 'l'; ' '])))
                         (fun _ ->
                           bind (catch
                                  (check_subseteq
                                    (eqc (eq_terma _A _B) (eq_terma _A _B)) nSr
                                    rR)
                                  (fun x ->
                                    Inl (match checkSS x with Inl err -> err
  | Inr _ -> internal_error)))
                             (fun _ ->
                               catch (check_subseteq
                                       (eqc (eq_terma _A _B) (eq_terma _A _B))
                                       nSp rP)
                                 (fun x ->
                                   Inl (match checkSS x with Inl err -> err
 | Inr _ -> internal_error))))))))
      (fun x ->
        Inl (comp (shows_string
                    ['c'; 'o'; 'u'; 'l'; 'd'; ' '; 'n'; 'o'; 't'; ' '; 'a'; 'p';
                      'p'; 'l'; 'y'; ' '; 't'; 'h'; 'e'; ' '; 'r'; 'e'; 'd';
                      'u'; 'c'; 't'; 'i'; 'o'; 'n'; ' '; 'p'; 'a'; 'i'; 'r';
                      ' '; 'p'; 'r'; 'o'; 'c'; 'e'; 's'; 's'; 'o'; 'r'; ' ';
                      'w'; 'i'; 't'; 'h'; ' '; 't'; 'h'; 'e'; ' '; 'f'; 'o';
                      'l'; 'l'; 'o'; 'w'; 'i'; 'n'; 'g'])
              (comp shows_nl (comp representation (comp shows_nl x)))));;

let rec create_mono_redpair_proc_comp (_A1, _A2) (_B1, _B2) (_C1, _C2)
  c cm i p rP r rR =
    let pi = to_lpoly_inter _B1 (get_comp_def c) (get_comp_1 c) i in
    bind (check_lpoly_coeffs _A2 _B2 (get_comp_wf_elem c) (get_comp_apos c)
           (fun x -> get_comp_ge c x (get_comp_0 c)) i)
      (fun _ ->
        check_mono_redpair_proc _B1 _C1
          (check_poly_monoI _A2 _B2 (get_comp_0 c) (get_comp_ge c) cm i)
          (check_polo_sI (_A1, _A2) _B2 (_C1, _C2) (get_comp_mono c)
            (get_comp_0 c) (get_comp_1 c) (get_comp_plus c) (get_comp_times c)
            (get_comp_max0 c) (get_comp_ge c) (get_comp_gt c) pi)
          (check_polo_nsI (_A1, _A2) _B2 (_C1, _C2) (get_comp_0 c)
            (get_comp_1 c) (get_comp_plus c) (get_comp_times c)
            (get_comp_max0 c) (get_comp_ge c) pi)
          (comp (shows_string
                  ['p'; 'o'; 'l'; 'y'; 'n'; 'o'; 'm'; 'i'; 'a'; 'l'; ' '; 'o';
                    'r'; 'd'; 'e'; 'r'; ' '; '('])
            (comp (shows_string (get_comp_description c)) (shows_string [')'])))
          (create_lpoly_repr (_A1, _A2) (_B1, _B2) c i) p rP r rR);;

let rec create_mono_redpair_proc (_A1, _A2) (_B1, _B2)
  = function
    Int_carrier i ->
      create_mono_redpair_proc_comp (eq_int, show_int) (_A1, _A2) (_B1, _B2)
        int_components int_mono i
    | Nat_carrier i ->
        create_mono_redpair_proc_comp (eq_nata, show_nat) (_A1, _A2) (_B1, _B2)
          nat_components nat_mono i
    | Rat_carrier (delta, i) ->
        create_mono_redpair_proc_comp (eq_rata, show_rata) (_A1, _A2) (_B1, _B2)
          (rat_components delta) rat_mono i
    | Arctic_carrier i ->
        (fun _ _ _ _ ->
          Inl (comp (shows_string
                      ['a'; 'r'; 'c'; 't'; 'i'; 'c'; ' '; 'n'; 'o'; 't'; ' ';
                        'm'; 'o'; 'n'; 'o'; 't'; 'o'; 'n'; 'i'; 'c'; ' '; 'f';
                        'o'; 'r'; ' '; 'T'; 'R'; 'S'; 's'])
                shows_nl))
    | Arctic_rat_carrier (d, i) ->
        (fun _ _ _ _ ->
          Inl (comp (shows_string
                      ['a'; 'r'; 'c'; 't'; 'i'; 'c'; ' '; 'n'; 'o'; 't'; ' ';
                        'm'; 'o'; 'n'; 'o'; 't'; 'o'; 'n'; 'i'; 'c'; ' '; 'f';
                        'o'; 'r'; ' '; 'T'; 'R'; 'S'; 's'])
                shows_nl))
    | Arc_mat_carrier (n, i) ->
        (fun _ _ _ _ ->
          Inl (comp (shows_string
                      ['a'; 'r'; 'c'; 't'; 'i'; 'c'; ' '; 'n'; 'o'; 't'; ' ';
                        'm'; 'o'; 'n'; 'o'; 't'; 'o'; 'n'; 'i'; 'c'; ' '; 'f';
                        'o'; 'r'; ' '; 'T'; 'R'; 'S'; 's'])
                shows_nl))
    | Arc_rat_mat_carrier (d, n, i) ->
        (fun _ _ _ _ ->
          Inl (comp (shows_string
                      ['a'; 'r'; 'c'; 't'; 'i'; 'c'; ' '; 'n'; 'o'; 't'; ' ';
                        'm'; 'o'; 'n'; 'o'; 't'; 'o'; 'n'; 'i'; 'c'; ' '; 'f';
                        'o'; 'r'; ' '; 'T'; 'R'; 'S'; 's'])
                shows_nl))
    | Int_mat_carrier (n, sd, i) ->
        create_mono_redpair_proc_comp
          (eq_lista (eq_lista eq_int), show_list (show_list show_int))
          (_A1, _A2) (_B1, _B2) (mat_components n sd int_components)
          (mat_mono sd int_mono) i
    | Rat_mat_carrier (d, n, sd, i) ->
        create_mono_redpair_proc_comp
          (eq_lista (eq_lista eq_rata), show_list (show_list show_rata))
          (_A1, _A2) (_B1, _B2) (mat_components n sd (rat_components d))
          (mat_mono sd rat_mono) i
    | Lpo (pri, pi) ->
        (fun p rP r rR ->
          let m = max ord_nat (get_max_arity _A1 r) (get_max_arity _A1 p) in
          let pr = lpo_repr_to_pr _A1 pri in
          let pia = afs_list_to_afs _A1 pi in
          bind (check_wf_afs _A2 pi)
            (fun _ ->
              check_mono_redpair_proc _A1 _B1 (check_mono_afs _A2 pi)
                (afs_check _A2 _B2 lpoS pia
                  (lpo_strict (show_filtered _A2) (_B1, _B2) pr m))
                (afs_check _A2 _B2 lpoNS pia
                  (lpo_nstrict (show_filtered _A2) (_B1, _B2) pr m))
                lpo_repr (shows_lpo_af _A2 pri pi) p rP r rR));;

let rec check_allm_gen_index
  g f n xs =
    snd (foldl (fun (i, m) x -> (g x i, bind m (fun _ -> f x i))) (n, Inr ())
          xs);;

let rec check_dep_graph_proc (_A1, _A2) (_B1, _B2)
  p dps rm rrm =
    bind (catch
           (check_subseteq (eqc (eq_terma _A1 _B1) (eq_terma _A1 _B1)) p
             (concat_map snd dps))
           (fun x ->
             Inl (comp (shows_string
                         ['D'; 'e'; 'p'; 'e'; 'n'; 'd'; 'e'; 'n'; 'c'; 'y'; ' ';
                           'P'; 'a'; 'i'; 'r'; ' '])
                   (comp (shows_rule _A2 _B2 x)
                     (comp (shows_string
                             [' '; 'i'; 's'; ' '; 'm'; 'i'; 's'; 's'; 'i'; 'n';
                               'g'; ' '; 'i'; 'n'; ' '; 'd'; 'e'; 'c'; 'o'; 'm';
                               'p'; 'o'; 's'; 'i'; 't'; 'i'; 'o'; 'n'])
                       shows_nl)))))
      (fun _ ->
        catch (check_graph_decomp
                (comp (shows_preca (show_term _A2 _B2) (show_term _A2 _B2)
                        Zero_nat)
                  fst)
                (fun (a, b) ->
                  let (aa, ba) = a in
                  let (_, t) = aa in
                  (fun (_, ct) (ab, bb) ->
                    let (u, _) = ab in
                    (fun (cu, _) ->
                      gcMatchI _A1 _B1 ct u && gcMatchI _A1 _B1 cu t)
                      bb)
                    ba
                    b)
                (map (fun (real, cs) ->
                       (real,
                         map (fun (s, t) ->
                               ((s, t),
                                 (capRM _A1 _B1 rrm s, capRM _A1 _B1 rm t)))
                           cs))
                  dps))
          (fun x ->
            Inl (comp (shows_string
                        ['o'; 'u'; 'r'; ' '; 'e'; 's'; 't'; 'i'; 'm'; 'a'; 't';
                          'i'; 'o'; 'n'; ' '; '('; 'E'; 'D'; 'G'; '*'; '*'; '*';
                          ')'; ' '; 'c'; 'o'; 'u'; 'l'; 'd'; ' '; 'n'; 'o'; 't';
                          ' '; 's'; 'h'; 'o'; 'w'; ' '; 't'; 'h'; 'a'; 't'; ' ';
                          'y'; 'o'; 'u'; ' '; 'h'; 'a'; 'v'; 'e'; ' '; 'a'; ' ';
                          'v'; 'a'; 'l'; 'i'; 'd'; ' '; 'd'; 'e'; 'c'; 'o'; 'm';
                          'p'; 'o'; 's'; 'i'; 't'; 'i'; 'o'; 'n'; ' '])
                  (comp (shows_string
                          ['d'; 'u'; 'e'; ' '; 't'; 'o'; ' '; 't'; 'h'; 'e';
                            ' '; 'f'; 'o'; 'l'; 'l'; 'o'; 'w'; 'i'; 'n'; 'g';
                            ' '; 'r'; 'e'; 'a'; 's'; 'o'; 'n'])
                    (comp shows_nl x)))));;

let rec check_redpair_ur_af_proc (_A1, _A2) (_B1, _B2)
  checkSS checkNSS name representation p rP r rm ur pi =
    catch (let (_, nSp) = partition (fun foo -> isOK (checkSS foo)) p in
           bind (catch
                  (check_ur_P_closed_rm_af (_A1, _A2) (_B1, _B2) r rm ur p pi)
                  (fun x ->
                    Inl (comp (shows_string
                                ['p'; 'r'; 'o'; 'b'; 'l'; 'e'; 'm'; ' '; 'w';
                                  'i'; 't'; 'h'; ' '; 'c'; 'l'; 'o'; 's'; 'u';
                                  'r'; 'e'; ' '; 'p'; 'r'; 'o'; 'p'; 'e'; 'r';
                                  't'; 'i'; 'e'; 's'; ' '; 'o'; 'f'; ' '; 'u';
                                  's'; 'a'; 'b'; 'l'; 'e'; ' '; 'r'; 'u'; 'l';
                                  'e'; 's'; ' '; 'w'; '.'; 'r'; '.'; 't'; '.';
                                  ' '; 'a'; 'r'; 'g'; 'u'; 'm'; 'e'; 'n'; 't';
                                  ' '; 'f'; 'i'; 'l'; 't'; 'e'; 'r'])
                          (comp shows_nl x))))
             (fun _ ->
               bind (catch (catch (forallm checkNSS ur) (fun x -> Inl (snd x)))
                      (fun x ->
                        Inl (comp (shows_string
                                    ['p'; 'r'; 'o'; 'b'; 'l'; 'e'; 'm'; ' ';
                                      'w'; 'h'; 'e'; 'n'; ' '; 'o'; 'r'; 'i';
                                      'e'; 'n'; 't'; 'i'; 'n'; 'g'; ' '; 'u';
                                      's'; 'a'; 'b'; 'l'; 'e'; ' '; 'r'; 'u';
                                      'l'; 'e'; 's'])
                              (comp shows_nl x))))
                 (fun _ ->
                   bind (catch
                          (catch (forallm checkNSS nSp) (fun x -> Inl (snd x)))
                          (fun x ->
                            Inl (comp (shows_string
['p'; 'r'; 'o'; 'b'; 'l'; 'e'; 'm'; ' '; 'w'; 'h'; 'e'; 'n'; ' '; 'o'; 'r'; 'i';
  'e'; 'n'; 't'; 'i'; 'n'; 'g'; ' '; 'D'; 'P'; 's'])
                                  (comp shows_nl x))))
                     (fun _ ->
                       catch (check_subseteq
                               (eqc (eq_terma _A1 _B1) (eq_terma _A1 _B1)) nSp
                               rP)
                         (fun x ->
                           Inl (match checkSS x with Inl err -> err
                                 | Inr _ -> internal_error))))))
      (fun x ->
        Inl (comp (shows_string
                    ['c'; 'o'; 'u'; 'l'; 'd'; ' '; 'n'; 'o'; 't'; ' '; 'a'; 'p';
                      'p'; 'l'; 'y'; ' '; 't'; 'h'; 'e'; ' '; 'r'; 'e'; 'd';
                      'u'; 'c'; 't'; 'i'; 'o'; 'n'; ' '; 'p'; 'a'; 'i'; 'r';
                      ' '; 'p'; 'r'; 'o'; 'c'; 'e'; 's'; 's'; 'o'; 'r'; ' ';
                      'w'; 'i'; 't'; 'h'; ' '; 't'; 'h'; 'e'; ' '; 'f'; 'o';
                      'l'; 'l'; 'o'; 'w'; 'i'; 'n'; 'g'])
              (comp shows_nl
                (comp representation
                  (comp shows_nl
                    (comp (shows_string
                            ['a'; 'n'; 'd'; ' '; 't'; 'h'; 'e'; ' '; 'f'; 'o';
                              'l'; 'l'; 'o'; 'w'; 'i'; 'n'; 'g'; ' '; 'u'; 's';
                              'a'; 'b'; 'l'; 'e'; ' '; 'r'; 'u'; 'l'; 'e'; 's'])
                      (comp shows_nl
                        (comp (shows_rules _A2 _B2 ur)
                          (comp shows_nl x)))))))));;

let rec create_redpair_ur_af_proc_comp (_A1, _A2) (_B1, _B2) (_C1, _C2)
  c i p rP r rm ur =
    let pi = to_lpoly_inter _B1 (get_comp_def c) (get_comp_1 c) i in
    bind (check_lpoly_coeffs _A2 _B2 (get_comp_wf_elem c) (get_comp_apos c)
           (fun x -> get_comp_ge c x (get_comp_0 c)) i)
      (fun _ ->
        check_redpair_ur_af_proc (_B1, _B2) (_C1, _C2)
          (check_polo_sI (_A1, _A2) _B2 (_C1, _C2) (get_comp_mono c)
            (get_comp_0 c) (get_comp_1 c) (get_comp_plus c) (get_comp_times c)
            (get_comp_max0 c) (get_comp_ge c) (get_comp_gt c) pi)
          (check_polo_nsI (_A1, _A2) _B2 (_C1, _C2) (get_comp_0 c)
            (get_comp_1 c) (get_comp_plus c) (get_comp_times c)
            (get_comp_max0 c) (get_comp_ge c) pi)
          (comp (shows_string
                  ['p'; 'o'; 'l'; 'y'; 'n'; 'o'; 'm'; 'i'; 'a'; 'l'; ' '; 'o';
                    'r'; 'd'; 'e'; 'r'; ' '; '('])
            (comp (shows_string (get_comp_description c)) (shows_string [')'])))
          (create_lpoly_repr (_A1, _A2) (_B1, _B2) c i) p rP r rm ur
          (create_afI _A1 _B1 (get_comp_0 c) i));;

let rec create_redpair_ur_af_proc (_A1, _A2) (_B1, _B2)
  = function
    Int_carrier i ->
      create_redpair_ur_af_proc_comp (eq_int, show_int) (_A1, _A2) (_B1, _B2)
        int_components i
    | Nat_carrier i ->
        create_redpair_ur_af_proc_comp (eq_nata, show_nat) (_A1, _A2) (_B1, _B2)
          nat_components i
    | Rat_carrier (delta, i) ->
        create_redpair_ur_af_proc_comp (eq_rata, show_rata) (_A1, _A2)
          (_B1, _B2) (rat_components delta) i
    | Arctic_carrier i ->
        create_redpair_ur_af_proc_comp (eq_arctica, show_arctic) (_A1, _A2)
          (_B1, _B2) arctic_components i
    | Arctic_rat_carrier (delta, i) ->
        create_redpair_ur_af_proc_comp (eq_arctic_rata, show_arctic_rat)
          (_A1, _A2) (_B1, _B2) (arctic_rat_components delta) i
    | Int_mat_carrier (n, sd, i) ->
        create_redpair_ur_af_proc_comp
          (eq_lista (eq_lista eq_int), show_list (show_list show_int))
          (_A1, _A2) (_B1, _B2) (mat_components n sd int_components) i
    | Rat_mat_carrier (delta, n, sd, i) ->
        create_redpair_ur_af_proc_comp
          (eq_lista (eq_lista eq_rata), show_list (show_list show_rata))
          (_A1, _A2) (_B1, _B2) (mat_components n sd (rat_components delta)) i
    | Arc_mat_carrier (n, i) ->
        create_redpair_ur_af_proc_comp
          (eq_lista (eq_lista eq_arctica), show_list (show_list show_arctic))
          (_A1, _A2) (_B1, _B2) (mat_arctic_components n arctic_components) i
    | Arc_rat_mat_carrier (delta, n, i) ->
        create_redpair_ur_af_proc_comp
          (eq_lista (eq_lista eq_arctic_rata),
            show_list (show_list show_arctic_rat))
          (_A1, _A2) (_B1, _B2)
          (mat_arctic_components n (arctic_rat_components delta)) i
    | Lpo (pri, pi) ->
        (fun p rP r rm ur ->
          let m = max ord_nat (get_max_arity _A1 r) (get_max_arity _A1 p) in
          let pr = lpo_repr_to_pr _A1 pri in
          let pia = afs_list_to_afs _A1 pi in
          bind (check_wf_afs _A2 pi)
            (fun _ ->
              check_redpair_ur_af_proc (_A1, _A2) (_B1, _B2)
                (afs_check _A2 _B2 lpoS pia
                  (lpo_strict (show_filtered _A2) (_B1, _B2) pr m))
                (afs_check _A2 _B2 lpoNS pia
                  (lpo_nstrict (show_filtered _A2) (_B1, _B2) pr m))
                lpo_repr (shows_lpo_af _A2 pri pi) p rP r rm ur
                (afs_list_to_af _A1 pi)));;

let rec check_redpair_proc _A _B
  checkSS checkNSS name representation p rP r =
    catch (let (_, nSp) = partition (fun foo -> isOK (checkSS foo)) p in
           bind (catch (catch (forallm checkNSS r) (fun x -> Inl (snd x)))
                  (fun x ->
                    Inl (comp (shows_string
                                ['p'; 'r'; 'o'; 'b'; 'l'; 'e'; 'm'; ' '; 'w';
                                  'h'; 'e'; 'n'; ' '; 'o'; 'r'; 'i'; 'e'; 'n';
                                  't'; 'i'; 'n'; 'g'; ' '; 'T'; 'R'; 'S'])
                          (comp shows_nl x))))
             (fun _ ->
               bind (catch (catch (forallm checkNSS nSp) (fun x -> Inl (snd x)))
                      (fun x ->
                        Inl (comp (shows_string
                                    ['p'; 'r'; 'o'; 'b'; 'l'; 'e'; 'm'; ' ';
                                      'w'; 'h'; 'e'; 'n'; ' '; 'o'; 'r'; 'i';
                                      'e'; 'n'; 't'; 'i'; 'n'; 'g'; ' '; 'D';
                                      'P'; 's'])
                              (comp shows_nl x))))
                 (fun _ ->
                   catch (check_subseteq (eqc (eq_terma _A _B) (eq_terma _A _B))
                           nSp rP)
                     (fun x ->
                       Inl (match checkSS x with Inl err -> err
                             | Inr _ -> internal_error)))))
      (fun x ->
        Inl (comp (shows_string
                    ['c'; 'o'; 'u'; 'l'; 'd'; ' '; 'n'; 'o'; 't'; ' '; 'a'; 'p';
                      'p'; 'l'; 'y'; ' '; 't'; 'h'; 'e'; ' '; 'r'; 'e'; 'd';
                      'u'; 'c'; 't'; 'i'; 'o'; 'n'; ' '; 'p'; 'a'; 'i'; 'r';
                      ' '; 'p'; 'r'; 'o'; 'c'; 'e'; 's'; 's'; 'o'; 'r'; ' ';
                      'w'; 'i'; 't'; 'h'; ' '; 't'; 'h'; 'e'; ' '; 'f'; 'o';
                      'l'; 'l'; 'o'; 'w'; 'i'; 'n'; 'g'])
              (comp shows_nl (comp representation (comp shows_nl x)))));;

let rec create_redpair_proc_comp (_A1, _A2) (_B1, _B2) (_C1, _C2)
  c i p rP r =
    let pi = to_lpoly_inter _B1 (get_comp_def c) (get_comp_1 c) i in
    bind (check_lpoly_coeffs _A2 _B2 (get_comp_wf_elem c) (get_comp_apos c)
           (fun x -> get_comp_ge c x (get_comp_0 c)) i)
      (fun _ ->
        check_redpair_proc _B1 _C1
          (check_polo_sI (_A1, _A2) _B2 (_C1, _C2) (get_comp_mono c)
            (get_comp_0 c) (get_comp_1 c) (get_comp_plus c) (get_comp_times c)
            (get_comp_max0 c) (get_comp_ge c) (get_comp_gt c) pi)
          (check_polo_nsI (_A1, _A2) _B2 (_C1, _C2) (get_comp_0 c)
            (get_comp_1 c) (get_comp_plus c) (get_comp_times c)
            (get_comp_max0 c) (get_comp_ge c) pi)
          (comp (shows_string
                  ['p'; 'o'; 'l'; 'y'; 'n'; 'o'; 'm'; 'i'; 'a'; 'l'; ' '; 'o';
                    'r'; 'd'; 'e'; 'r'; ' '; '('])
            (comp (shows_string (get_comp_description c)) (shows_string [')'])))
          (create_lpoly_repr (_A1, _A2) (_B1, _B2) c i) p rP r);;

let rec create_redpair_proc (_A1, _A2) (_B1, _B2)
  = function
    Int_carrier i ->
      create_redpair_proc_comp (eq_int, show_int) (_A1, _A2) (_B1, _B2)
        int_components i
    | Nat_carrier i ->
        create_redpair_proc_comp (eq_nata, show_nat) (_A1, _A2) (_B1, _B2)
          nat_components i
    | Rat_carrier (delta, i) ->
        create_redpair_proc_comp (eq_rata, show_rata) (_A1, _A2) (_B1, _B2)
          (rat_components delta) i
    | Arctic_carrier i ->
        create_redpair_proc_comp (eq_arctica, show_arctic) (_A1, _A2) (_B1, _B2)
          arctic_components i
    | Arctic_rat_carrier (d, i) ->
        create_redpair_proc_comp (eq_arctic_rata, show_arctic_rat) (_A1, _A2)
          (_B1, _B2) (arctic_rat_components d) i
    | Int_mat_carrier (n, sd, i) ->
        create_redpair_proc_comp
          (eq_lista (eq_lista eq_int), show_list (show_list show_int))
          (_A1, _A2) (_B1, _B2) (mat_components n sd int_components) i
    | Rat_mat_carrier (delta, n, sd, i) ->
        create_redpair_proc_comp
          (eq_lista (eq_lista eq_rata), show_list (show_list show_rata))
          (_A1, _A2) (_B1, _B2) (mat_components n sd (rat_components delta)) i
    | Arc_mat_carrier (n, i) ->
        create_redpair_proc_comp
          (eq_lista (eq_lista eq_arctica), show_list (show_list show_arctic))
          (_A1, _A2) (_B1, _B2) (mat_arctic_components n arctic_components) i
    | Arc_rat_mat_carrier (d, n, i) ->
        create_redpair_proc_comp
          (eq_lista (eq_lista eq_arctic_rata),
            show_list (show_list show_arctic_rat))
          (_A1, _A2) (_B1, _B2)
          (mat_arctic_components n (arctic_rat_components d)) i
    | Lpo (pri, pi) ->
        (fun p rP r ->
          let m = max ord_nat (get_max_arity _A1 r) (get_max_arity _A1 p) in
          let pr = lpo_repr_to_pr _A1 pri in
          let pia = afs_list_to_afs _A1 pi in
          bind (check_wf_afs _A2 pi)
            (fun _ ->
              check_redpair_proc _A1 _B1
                (afs_check _A2 _B2 lpoS pia
                  (lpo_strict (show_filtered _A2) (_B1, _B2) pr m))
                (afs_check _A2 _B2 lpoNS pia
                  (lpo_nstrict (show_filtered _A2) (_B1, _B2) pr m))
                lpo_repr (shows_lpo_af _A2 pri pi) p rP r));;

let rec checkDPProof (_A1, _A2) (_B1, _B2)
  i p r m rm rrm x6 = match i, p, r, m, rm, rrm, x6 with
    i, p, r, m, rm, rrm, PisEmpty ->
      debug (i []) ['P'; ' '; 'i'; 's'; ' '; 'e'; 'm'; 'p'; 't'; 'y']
        (match p with [] -> Inr ()
          | _ :: _ ->
            Inl (comp i
                  (comp (shows_string
                          [':'; ' '; 'P'; ' '; 'i'; 's'; ' '; 'n'; 'o'; 't';
                            ' '; 'e'; 'm'; 'p'; 't'; 'y'; ' '; 'i'; 'n'; ' ';
                            't'; 'h'; 'e'; ' '; 'f'; 'o'; 'l'; 'l'; 'o'; 'w';
                            'i'; 'n'; 'g'; ' '; 'D'; 'P'; '-'; 'p'; 'r'; 'o';
                            'b'; 'l'; 'e'; 'm'; ' '])
                    (comp shows_nl
                      (shows_dpp (show_lab _A2 _B2) (show_list show_char) p
                        r)))))
    | i, p, r, m, rm, rrm, RedPairProc (redp, rP, prf) ->
        debug (i []) ['R'; 'e'; 'd'; 'P'; 'a'; 'i'; 'r'; 'P'; 'r'; 'o'; 'c']
          (bind (catch
                  (create_redpair_proc (eq_laba _A1 _B1, show_lab _A2 _B2)
                    (eq_lista eq_char, show_list show_char) redp p rP r)
                  (fun x ->
                    Inl (comp i
                          (comp (shows_string
                                  [':'; ' '; 'e'; 'r'; 'r'; 'o'; 'r'; ' '; 'w';
                                    'h'; 'e'; 'n'; ' '; 'a'; 'p'; 'p'; 'l'; 'y';
                                    'i'; 'n'; 'g'; ' '; 't'; 'h'; 'e'; ' '; 'r';
                                    'e'; 'd'; 'u'; 'c'; 't'; 'i'; 'o'; 'n'; ' ';
                                    'p'; 'a'; 'i'; 'r'; ' '; 'p'; 'r'; 'o'; 'c';
                                    'e'; 's'; 's'; 'o'; 'r'; ' '; 't'; 'o'; ' ';
                                    's'; 'w'; 'i'; 't'; 'c'; 'h'; ' '; 'f'; 'r';
                                    'o'; 'm'])
                            (comp shows_nl
                              (comp (shows_dpp (show_lab _A2 _B2)
                                      (show_list show_char) p r)
                                (comp shows_nl
                                  (comp (shows_string
  ['t'; 'o'; ' '; 't'; 'h'; 'e'; ' '; 'D'; 'P'; ' '; 'p'; 'r'; 'o'; 'b'; 'l';
    'e'; 'm'])
                                    (comp shows_nl
                                      (comp
(shows_dpp (show_lab _A2 _B2) (show_list show_char) rP r)
(comp shows_nl x)))))))))))
            (fun _ ->
              catch (checkDPProof (_A1, _A2) (_B1, _B2)
                      (comp i (shows_string ['.'; '1'])) rP r m rm rrm prf)
                (fun x ->
                  Inl (comp i
                        (comp (shows_string
                                [':'; ' '; 'e'; 'r'; 'r'; 'o'; 'r'; ' '; 'b';
                                  'e'; 'l'; 'o'; 'w'; ' '; 't'; 'h'; 'e'; ' ';
                                  'r'; 'e'; 'd'; 'u'; 'c'; 't'; 'i'; 'o'; 'n';
                                  ' '; 'p'; 'a'; 'i'; 'r'; ' '; 'p'; 'r'; 'o';
                                  'c'; 'e'; 's'; 's'; 'o'; 'r'])
                          (comp shows_nl (indent x)))))))
    | i, p, r, m, rm, rrm, RedPairUrProc (redp, rP, ur, prf) ->
        debug (i [])
          ['R'; 'e'; 'd'; 'P'; 'a'; 'i'; 'r'; 'P'; 'r'; 'o'; 'c'; ' '; '('; 'u';
            's'; 'a'; 'b'; 'l'; 'e'; ' '; 'r'; 'u'; 'l'; 'e'; 's'; ' '; '+';
            ' '; 'a'; 'f'; ')']
          (bind (catch
                  (create_redpair_ur_af_proc (eq_laba _A1 _B1, show_lab _A2 _B2)
                    (eq_lista eq_char, show_list show_char) redp p rP r rm ur)
                  (fun x ->
                    Inl (comp i
                          (comp (shows_string
                                  [':'; ' '; 'e'; 'r'; 'r'; 'o'; 'r'; ' '; 'w';
                                    'h'; 'e'; 'n'; ' '; 'a'; 'p'; 'p'; 'l'; 'y';
                                    'i'; 'n'; 'g'; ' '; 't'; 'h'; 'e'; ' '; 'r';
                                    'e'; 'd'; 'u'; 'c'; 't'; 'i'; 'o'; 'n'; ' ';
                                    'p'; 'a'; 'i'; 'r'; ' '])
                            (comp (shows_string
                                    ['p'; 'r'; 'o'; 'c'; 'e'; 's'; 's'; 'o';
                                      'r'; ' '; '('; 'w'; 'i'; 't'; 'h'; ' ';
                                      'u'; 's'; 'a'; 'b'; 'l'; 'e'; ' '; 'r';
                                      'u'; 'l'; 'e'; 's'; ' '; 'a'; 'n'; 'd';
                                      ' '; 'a'; 'f'; ')'; ' '; 't'; 'o'; ' ';
                                      's'; 'w'; 'i'; 't'; 'c'; 'h'; ' '; 'f';
                                      'r'; 'o'; 'm'; ' '])
                              (comp shows_nl
                                (comp (shows_dpp (show_lab _A2 _B2)
(show_list show_char) p r)
                                  (comp shows_nl
                                    (comp (shows_string
    ['t'; 'o'; ' '; 't'; 'h'; 'e'; ' '; 'D'; 'P'; ' '; 'p'; 'r'; 'o'; 'b'; 'l';
      'e'; 'm'; ' '])
                                      (comp shows_nl
(comp (shows_dpp (show_lab _A2 _B2) (show_list show_char) rP r)
  (comp shows_nl x))))))))))))
            (fun _ ->
              catch (checkDPProof (_A1, _A2) (_B1, _B2)
                      (comp i (shows_string ['.'; '1'])) rP r m rm rrm prf)
                (fun x ->
                  Inl (comp i
                        (comp (shows_string
                                [':'; ' '; 'e'; 'r'; 'r'; 'o'; 'r'; ' '; 'b';
                                  'e'; 'l'; 'o'; 'w'; ' '; 't'; 'h'; 'e'; ' ';
                                  'r'; 'e'; 'd'; 'u'; 'c'; 't'; 'i'; 'o'; 'n';
                                  ' '; 'p'; 'a'; 'i'; 'r'; ' '; 'p'; 'r'; 'o';
                                  'c'; 'e'; 's'; 's'; 'o'; 'r'])
                          (comp shows_nl (indent x)))))))
    | i, p, r, m, rm, rrm, DepGraphProc edpts ->
        debug (i [])
          ['D'; 'e'; 'p'; 'G'; 'r'; 'a'; 'p'; 'h'; 'P'; 'r'; 'o'; 'c']
          (bind (catch
                  (check_dep_graph_proc (eq_laba _A1 _B1, show_lab _A2 _B2)
                    (eq_lista eq_char, show_list show_char) p
                    (map (fun x ->
                           ((match fst x with None -> false | Some _ -> true),
                             snd x))
                      edpts)
                    rm rrm)
                  (fun x ->
                    Inl (comp i
                          (comp (shows_string
                                  [':'; ' '; 'e'; 'r'; 'r'; 'o'; 'r'; ' '; 'w';
                                    'h'; 'i'; 'l'; 'e'; ' '; 't'; 'r'; 'y'; 'i';
                                    'n'; 'g'; ' '; 't'; 'o'; ' '; 'p'; 'e'; 'r';
                                    'f'; 'o'; 'r'; 'm'; ' '; 'S'; 'C'; 'C'; '-';
                                    'd'; 'e'; 'c'; 'o'; 'm'; 'p'; 'o'; 's'; 'i';
                                    't'; 'i'; 'o'; 'n'; ' '; ' '; 'o'; 'n';
                                    ' '])
                            (comp shows_nl
                              (comp (shows_dpp (show_lab _A2 _B2)
                                      (show_list show_char) p r)
                                (comp shows_nl x)))))))
            (fun _ ->
              catch (check_allm_gen_index
                      (fun rp j ->
                        (match fst rp with None -> j | Some _ -> Suc j))
                      (fun rp j ->
                        (match fst rp with None -> Inr ()
                          | Some a ->
                            checkDPProof (_A1, _A2) (_B1, _B2)
                              (comp i
                                (comp (shows_string ['.'])
                                  (shows_prec_nat Zero_nat j)))
                              (snd rp) r m rm rrm a))
                      one_nat edpts)
                (fun x ->
                  Inl (comp i
                        (comp (shows_string
                                [':'; ' '; 'e'; 'r'; 'r'; 'o'; 'r'; ' '; 'b';
                                  'e'; 'l'; 'o'; 'w'; ' '; 't'; 'h'; 'e'; ' ';
                                  'd'; 'e'; 'p'; 'e'; 'n'; 'd'; 'e'; 'n'; 'c';
                                  'y'; ' '; 'g'; 'r'; 'a'; 'p'; 'h'; ' '; 'p';
                                  'r'; 'o'; 'c'; 'e'; 's'; 's'; 'o'; 'r'])
                          (comp shows_nl (indent x)))))))
    | i, p, r, m, rm, rrm, MonoRedPairProc (redp, rP, rR, prf) ->
        debug (i [])
          ['M'; 'o'; 'n'; 'o'; 'R'; 'e'; 'd'; 'P'; 'a'; 'i'; 'r'; 'P'; 'r'; 'o';
            'c']
          (bind (catch
                  (create_mono_redpair_proc (eq_laba _A1 _B1, show_lab _A2 _B2)
                    (eq_lista eq_char, show_list show_char) redp p rP r rR)
                  (fun x ->
                    Inl (comp i
                          (comp (shows_string
                                  [':'; ' '; 'e'; 'r'; 'r'; 'o'; 'r'; ' '; 'w';
                                    'h'; 'e'; 'n'; ' '; 'a'; 'p'; 'p'; 'l'; 'y';
                                    'i'; 'n'; 'g'; ' '; 't'; 'h'; 'e'; ' '; 'm';
                                    'o'; 'n'; 'o'; 't'; 'o'; 'n'; 'i'; 'c'; ' ';
                                    'r'; 'e'; 'd'; 'u'; 'c'; 't'; 'i'; 'o'; 'n';
                                    ' '; 'p'; 'a'; 'i'; 'r'; ' '; 'p'; 'r'; 'o';
                                    'c'; 'e'; 's'; 's'; 'o'; 'r'; ' '])
                            (comp (shows_string
                                    ['t'; 'o'; ' '; 's'; 'w'; 'i'; 't'; 'c';
                                      'h'; ' '; 'f'; 'r'; 'o'; 'm'; ' '])
                              (comp shows_nl
                                (comp (shows_dpp (show_lab _A2 _B2)
(show_list show_char) p r)
                                  (comp shows_nl
                                    (comp (shows_string
    ['t'; 'o'; ' '; 't'; 'h'; 'e'; ' '; 'D'; 'P'; ' '; 'p'; 'r'; 'o'; 'b'; 'l';
      'e'; 'm'; ' '])
                                      (comp shows_nl
(comp (shows_dpp (show_lab _A2 _B2) (show_list show_char) rP rR)
  (comp shows_nl x))))))))))))
            (fun _ ->
              catch (checkDPProof (_A1, _A2) (_B1, _B2)
                      (comp i (shows_string ['.'; '1'])) rP rR m
                      (computeRuleMap (eq_laba _A1 _B1) rR)
                      (computeRuleMap (eq_laba _A1 _B1) (reverse_trs rR)) prf)
                (fun x ->
                  Inl (comp i
                        (comp (shows_string
                                [':'; ' '; 'e'; 'r'; 'r'; 'o'; 'r'; ' '; 'b';
                                  'e'; 'l'; 'o'; 'w'; ' '; 't'; 'h'; 'e'; ' ';
                                  'm'; 'o'; 'n'; 'o'; 't'; 'o'; 'n'; 'i'; 'c';
                                  ' '; 'r'; 'e'; 'd'; 'u'; 'c'; 't'; 'i'; 'o';
                                  'n'; ' '; 'p'; 'a'; 'i'; 'r'; ' '; 'p'; 'r';
                                  'o'; 'c'; 'e'; 's'; 's'; 'o'; 'r'])
                          (comp shows_nl (indent x)))))))
    | i, p, r, m, rm, rrm, MonoRedPairUrProc (redp, rP, rR, ur, prf) ->
        debug (i [])
          ['M'; 'o'; 'n'; 'o'; 'R'; 'e'; 'd'; 'P'; 'a'; 'i'; 'r'; 'P'; 'r'; 'o';
            'c'; ' '; '('; 'u'; 's'; 'a'; 'b'; 'l'; 'e'; ' '; 'r'; 'u'; 'l';
            'e'; 's'; ')']
          (bind (catch
                  (create_mono_redpair_ur_proc
                    (eq_laba _A1 _B1, show_lab _A2 _B2)
                    (eq_lista eq_char, show_list show_char) redp p rP r rm rR
                    ur)
                  (fun x ->
                    Inl (comp i
                          (comp (shows_string
                                  [':'; ' '; 'e'; 'r'; 'r'; 'o'; 'r'; ' '; 'w';
                                    'h'; 'e'; 'n'; ' '; 'a'; 'p'; 'p'; 'l'; 'y';
                                    'i'; 'n'; 'g'; ' '; 't'; 'h'; 'e'; ' '; 'm';
                                    'o'; 'n'; 'o'; 't'; 'o'; 'n'; 'i'; 'c'; ' ';
                                    'r'; 'e'; 'd'; 'u'; 'c'; 't'; 'i'; 'o'; 'n';
                                    ' '; 'p'; 'a'; 'i'; 'r'; ' '])
                            (comp (shows_string
                                    ['p'; 'r'; 'o'; 'c'; 'e'; 's'; 's'; 'o';
                                      'r'; ' '; '('; 'w'; 'i'; 't'; 'h'; ' ';
                                      'u'; 's'; 'a'; 'b'; 'l'; 'e'; ' '; 'r';
                                      'u'; 'l'; 'e'; 's'; ')'; ' '; 't'; 'o';
                                      ' '; 's'; 'w'; 'i'; 't'; 'c'; 'h'; ' ';
                                      'f'; 'r'; 'o'; 'm'; ' '])
                              (comp shows_nl
                                (comp (shows_dpp (show_lab _A2 _B2)
(show_list show_char) p r)
                                  (comp shows_nl
                                    (comp (shows_string
    ['t'; 'o'; ' '; 't'; 'h'; 'e'; ' '; 'D'; 'P'; ' '; 'p'; 'r'; 'o'; 'b'; 'l';
      'e'; 'm'])
                                      (comp shows_nl
(comp (shows_dpp (show_lab _A2 _B2) (show_list show_char) rP rR)
  (comp shows_nl x))))))))))))
            (fun _ ->
              catch (checkDPProof (_A1, _A2) (_B1, _B2)
                      (comp i (shows_string ['.'; '1'])) rP rR m
                      (computeRuleMap (eq_laba _A1 _B1) rR)
                      (computeRuleMap (eq_laba _A1 _B1) (reverse_trs rR)) prf)
                (fun x ->
                  Inl (comp i
                        (comp (shows_string
                                [':'; ' '; 'e'; 'r'; 'r'; 'o'; 'r'; ' '; 'b';
                                  'e'; 'l'; 'o'; 'w'; ' '; 't'; 'h'; 'e'; ' ';
                                  'm'; 'o'; 'n'; 'o'; 't'; 'o'; 'n'; 'i'; 'c';
                                  ' '; 'r'; 'e'; 'd'; 'u'; 'c'; 't'; 'i'; 'o';
                                  'n'; ' '; 'p'; 'a'; 'i'; 'r'; ' '])
                          (comp (shows_string
                                  ['('; 'w'; 'i'; 't'; 'h'; ' '; 'u'; 's'; 'a';
                                    'b'; 'l'; 'e'; ' '; 'r'; 'u'; 'l'; 'e'; 's';
                                    ')'; ' '; 'p'; 'r'; 'o'; 'c'; 'e'; 's'; 's';
                                    'o'; 'r'])
                            (comp shows_nl (indent x))))))))
    | i, p, r, Suc m, rm, rrm, UnlabProc (ulP, ulR, prf) ->
        debug (i []) ['U'; 'n'; 'l'; 'a'; 'b'; 'P'; 'r'; 'o'; 'c']
          (bind (check_unlab_proc (_A1, _A2) (_B1, _B2)
                  (eq_lista eq_char, show_list show_char) p r ulP ulR)
            (fun _ ->
              catch (checkDPProof (_A1, _A2) (_B1, _B2)
                      (comp i (shows_string ['.'; '1'])) ulP ulR m
                      (computeRuleMap (eq_laba _A1 _B1) ulR)
                      (computeRuleMap (eq_laba _A1 _B1) (reverse_trs ulR)) prf)
                (fun x ->
                  Inl (comp i
                        (comp (shows_string
                                [':'; ' '; 'e'; 'r'; 'r'; 'o'; 'r'; ' '; 'b';
                                  'e'; 'l'; 'o'; 'w'; ' '; 't'; 'h'; 'e'; ' ';
                                  'u'; 'n'; 'l'; 'a'; 'b'; 'e'; 'l'; 'i'; 'n';
                                  'g'; ' '; 'p'; 'r'; 'o'; 'c'; 'e'; 's'; 's';
                                  'o'; 'r'])
                          (comp shows_nl (indent x)))))))
    | i, p, r, Zero_nat, rm, rrm, UnlabProc (ulP, ulR, prf) ->
        Inl (comp (shows_string
                    ['u'; 'n'; 'l'; 'a'; 'b'; 'e'; 'l'; 'i'; 'n'; 'g'; ' '; 'n';
                      'o'; 't'; ' '; 'a'; 'l'; 'l'; 'o'; 'w'; 'e'; 'd'; ' ';
                      'd'; 'u'; 'e'; ' '; 't'; 'o'; ' '; 'l'; 'a'; 'b'; 'e';
                      'l'; '-'; 'd'; 'e'; 'p'; 't'; 'h'; ' '; 'o'; 'f'; ' ';
                      '0'])
              shows_nl)
    | i, p, r, m, rm, rrm, SemlabProc (sli, l, lP, lR, prf) ->
        debug (i []) ['S'; 'e'; 'm'; 'l'; 'a'; 'b'; 'P'; 'r'; 'o'; 'c']
          (bind (catch
                  (create_sem_lab_proc (_A1, _A2) (_B1, _B2)
                    (eq_lista eq_char, show_list show_char) sli l p r lP lR)
                  (fun x ->
                    Inl (comp i
                          (comp (shows_string
                                  [':'; ' '; 'e'; 'r'; 'r'; 'o'; 'r'; ' '; 'w';
                                    'h'; 'e'; 'n'; ' '; 'a'; 'p'; 'p'; 'l'; 'y';
                                    'i'; 'n'; 'g'; ' '; 't'; 'h'; 'e'; ' '; 's';
                                    'e'; 'm'; 'a'; 'n'; 't'; 'i'; 'c'; ' '; 'l';
                                    'a'; 'b'; 'e'; 'l'; 'i'; 'n'; 'g'; ' '; 'p';
                                    'r'; 'o'; 'c'; 'e'; 's'; 's'; 'o'; 'r'])
                            (comp shows_nl x)))))
            (fun _ ->
              catch (checkDPProof (_A1, _A2) (_B1, _B2)
                      (comp i (shows_string ['.'; '1'])) lP lR (Suc m)
                      (computeRuleMap (eq_laba _A1 _B1) lR)
                      (computeRuleMap (eq_laba _A1 _B1) (reverse_trs lR)) prf)
                (fun x ->
                  Inl (comp i
                        (comp (shows_string
                                [':'; ' '; 'e'; 'r'; 'r'; 'o'; 'r'; ' '; 'b';
                                  'e'; 'l'; 'o'; 'w'; ' '; 't'; 'h'; 'e'; ' ';
                                  's'; 'e'; 'm'; 'a'; 'n'; 't'; 'i'; 'c'; ' ';
                                  'l'; 'a'; 'b'; 'e'; 'l'; 'i'; 'n'; 'g'; ' ';
                                  'p'; 'r'; 'o'; 'c'; 'e'; 's'; 's'; 'o'; 'r'])
                          (comp shows_nl (indent x)))))))
    | i, p, r, m, rm, rrm, RootlabProc (lP, lR, prf) ->
        debug (i []) ['R'; 'o'; 'o'; 't'; 'l'; 'a'; 'b'; 'P'; 'r'; 'o'; 'c']
          (bind (catch (rl_proc (_A1, _A2) (_B1, _B2) p r lP lR)
                  (fun x ->
                    Inl (comp i
                          (comp (shows_string
                                  [':'; ' '; 'e'; 'r'; 'r'; 'o'; 'r'; ' '; 'w';
                                    'h'; 'e'; 'n'; ' '; 'a'; 'p'; 'p'; 'l'; 'y';
                                    'i'; 'n'; 'g'; ' '; 't'; 'h'; 'e'; ' '; 'r';
                                    'o'; 'o'; 't'; ' '; 'l'; 'a'; 'b'; 'e'; 'l';
                                    'i'; 'n'; 'g'; ' '; 'p'; 'r'; 'o'; 'c'; 'e';
                                    's'; 's'; 'o'; 'r'])
                            (comp shows_nl x)))))
            (fun _ ->
              catch (checkDPProof (_A1, _A2) (_B1, _B2)
                      (comp i (shows_string ['.'; '1'])) lP lR (Suc m)
                      (computeRuleMap (eq_laba _A1 _B1) lR)
                      (computeRuleMap (eq_laba _A1 _B1) (reverse_trs lR)) prf)
                (fun x ->
                  Inl (comp i
                        (comp (shows_string
                                [':'; ' '; 'e'; 'r'; 'r'; 'o'; 'r'; ' '; 'b';
                                  'e'; 'l'; 'o'; 'w'; ' '; 't'; 'h'; 'e'; ' ';
                                  'r'; 'o'; 'o'; 't'; '-'; 'l'; 'a'; 'b'; 'e';
                                  'l'; 'i'; 'n'; 'g'; ' '; 'p'; 'r'; 'o'; 'c';
                                  'e'; 's'; 's'; 'o'; 'r'])
                          (comp shows_nl (indent x)))))))
    | i, p, r, m, rm, rrm, SubtermCriterionProc (sp, rseq, rP, prf) ->
        debug (i []) ['S'; 'u'; 'b'; 't'; 'e'; 'r'; 'm'; 'P'; 'r'; 'o'; 'c']
          (bind (catch
                  (create_subterm_criterion_proc
                    (eq_laba _A1 _B1, show_lab _A2 _B2)
                    (eq_lista eq_char, show_list show_char) sp rseq p rP r rm)
                  (fun x ->
                    Inl (comp i
                          (comp (shows_string
                                  [':'; ' '; 'e'; 'r'; 'r'; 'o'; 'r'; ' '; 'w';
                                    'h'; 'e'; 'n'; ' '; 'a'; 'p'; 'p'; 'l'; 'y';
                                    'i'; 'n'; 'g'; ' '; 't'; 'h'; 'e'; ' '; 's';
                                    'u'; 'b'; 't'; 'e'; 'r'; 'm'; ' '; 'c'; 'r';
                                    'i'; 't'; 'e'; 'r'; 'i'; 'o'; 'n'; ' '; 'p';
                                    'r'; 'o'; 'c'; 'e'; 's'; 's'; 'o'; 'r'])
                            (comp shows_nl x)))))
            (fun _ ->
              catch (checkDPProof (_A1, _A2) (_B1, _B2)
                      (comp i (shows_string ['.'; '1'])) rP r m rm rrm prf)
                (fun x ->
                  Inl (comp i
                        (comp (shows_string
                                [':'; ' '; 'e'; 'r'; 'r'; 'o'; 'r'; ' '; 'b';
                                  'e'; 'l'; 'o'; 'w'; ' '; 't'; 'h'; 'e'; ' ';
                                  's'; 'u'; 'b'; 't'; 'e'; 'r'; 'm'; ' '; 'c';
                                  'r'; 'i'; 't'; 'e'; 'r'; 'i'; 'o'; 'n'; ' ';
                                  'p'; 'r'; 'o'; 'c'; 'e'; 's'; 's'; 'o'; 'r'])
                          (comp shows_nl (indent x)))))))
    | i, p, r, m, rm, rrm, SizeChangeSubTerm graphs ->
        debug (i [])
          ['S'; 'i'; 'z'; 'e'; 'C'; 'h'; 'a'; 'n'; 'g'; 'e'; 'S'; 'u'; 'b'; 'T';
            'e'; 'r'; 'm']
          (check_sct_subterm_proc (eq_laba _A1 _B1, show_lab _A2 _B2)
            (eq_lista eq_char, show_list show_char) p graphs rm rrm)
    | i, p, r, m, rm, rrm, SizeChangeRedPair (redp, ur, graphs) ->
        debug (i [])
          ['S'; 'i'; 'z'; 'e'; 'C'; 'h'; 'a'; 'n'; 'g'; 'e'; 'R'; 'e'; 'd'; 'P';
            'a'; 'i'; 'r']
          (create_sct_ur_af_proc (eq_laba _A1 _B1, show_lab _A2 _B2)
            (eq_lista eq_char, show_list show_char) redp p rm rrm ur graphs)
    | i, p, r, m, rm, rrm, FccProc (f, fcs, rP, rR, prf) ->
        debug (i []) ['F'; 'c'; 'c'; 'P'; 'r'; 'o'; 'c']
          (bind (catch
                  (fcc_proc (eq_laba _A1 _B1, show_lab _A2 _B2) p r f fcs rP rR)
                  (fun x ->
                    Inl (comp i
                          (comp (shows_string
                                  [':'; ' '; 'e'; 'r'; 'r'; 'o'; 'r'; ' '; 'w';
                                    'h'; 'e'; 'n'; ' '; 'a'; 'p'; 'p'; 'l'; 'y';
                                    'i'; 'n'; 'g'; ' '; 't'; 'h'; 'e'; ' '; 'c';
                                    'l'; 'o'; 's'; 'u'; 'r'; 'e'; ' '; 'u'; 'n';
                                    'd'; 'e'; 'r'; ' '; 'f'; 'l'; 'a'; 't'; ' ';
                                    'c'; 'o'; 'n'; 't'; 'e'; 'x'; 't'; 's'])
                            (comp shows_nl x)))))
            (fun _ ->
              catch (checkDPProof (_A1, _A2) (_B1, _B2)
                      (comp i (shows_string ['.'; '1'])) rP rR Zero_nat
                      (computeRuleMap (eq_laba _A1 _B1) rR)
                      (computeRuleMap (eq_laba _A1 _B1) (reverse_trs rR)) prf)
                (fun x ->
                  Inl (comp i
                        (comp (shows_string
                                [':'; ' '; 'e'; 'r'; 'r'; 'o'; 'r'; ' '; 'b';
                                  'e'; 'l'; 'o'; 'w'; ' '; 'f'; 'l'; 'a'; 't';
                                  ' '; 'c'; 'o'; 'n'; 't'; 'e'; 'x'; 't'; ' ';
                                  'c'; 'l'; 'o'; 's'; 'u'; 'r'; 'e'; ' '; 'p';
                                  'r'; 'o'; 'c'; 'e'; 's'; 's'; 'o'; 'r'])
                          (comp shows_nl (indent x)))))));;

let rec checkTRSProof (_A1, _A2) (_B1, _B2)
  i r m x3 = match i, r, m, x3 with
    i, r, m, DPTrans (p, prf) ->
      debug (i []) ['D'; 'P'; ' '; 't'; 'r'; 'a'; 'n'; 's']
        (bind (catch
                (check_DP (eq_laba _A1 _B1, show_lab _A2 _B2)
                  (eq_lista eq_char, show_list show_char) r p)
                (fun x ->
                  Inl (comp i
                        (comp (shows_string
                                [':'; ' '; 'e'; 'r'; 'r'; 'o'; 'r'; ' '; 'w';
                                  'h'; 'e'; 'n'; ' '; 's'; 'w'; 'i'; 't'; 'c';
                                  'h'; 'i'; 'n'; 'g'; ' '; 'f'; 'r'; 'o'; 'm';
                                  ' '; 't'; 'h'; 'e'; ' '; 'T'; 'R'; 'S'])
                          (comp shows_nl
                            (comp (shows_trs (show_lab _A2 _B2)
                                    (show_list show_char) r)
                              (comp shows_nl
                                (comp (shows_string
['t'; 'o'; ' '; 't'; 'h'; 'e'; ' '; 'i'; 'n'; 'i'; 't'; 'i'; 'a'; 'l'; ' '; 'D';
  'P'; ' '; 'p'; 'r'; 'o'; 'b'; 'l'; 'e'; 'm'])
                                  (comp (shows_dpp (show_shp (show_lab _A2 _B2))
  (show_list show_char) p (iD_list r))
                                    (comp shows_nl x))))))))))
          (fun _ ->
            catch (checkDPProof
                    (eq_shpa (eq_laba _A1 _B1), show_shp (show_lab _A2 _B2))
                    (_B1, _B2) (comp i (shows_string ['.'; '1']))
                    (map (map_rule init_lab) p)
                    (map (map_rule init_lab) (iD_list r)) Zero_nat
                    (computeRuleMap (eq_laba (eq_shpa (eq_laba _A1 _B1)) _B1)
                      (map (map_rule init_lab) (iD_list r)))
                    (computeRuleMap (eq_laba (eq_shpa (eq_laba _A1 _B1)) _B1)
                      (reverse_trs (map (map_rule init_lab) (iD_list r))))
                    prf)
              (fun x ->
                Inl (comp i
                      (comp (shows_string
                              [':'; ' '; 'e'; 'r'; 'r'; 'o'; 'r'; ' '; 'b'; 'e';
                                'l'; 'o'; 'w'; ' '; 's'; 'w'; 'i'; 't'; 'c';
                                'h'; ' '; 't'; 'o'; ' '; 'd'; 'e'; 'p'; 'e';
                                'n'; 'd'; 'e'; 'n'; 'c'; 'y'; ' '; 'p'; 'a';
                                'i'; 'r'; 's'])
                        (comp shows_nl (indent x)))))))
    | i, r, m, RuleRemoval (rp, rR, prf) ->
        debug (i [])
          ['r'; 'u'; 'l'; 'e'; ' '; 'r'; 'e'; 'm'; 'o'; 'v'; 'a'; 'l']
          (bind (catch
                  (create_rule_removal_proc (eq_laba _A1 _B1, show_lab _A2 _B2)
                    (eq_lista eq_char, show_list show_char) rp r rR)
                  (fun x ->
                    Inl (comp i
                          (comp (shows_string
                                  [':'; ' '; 'e'; 'r'; 'r'; 'o'; 'r'; ' '; 'i';
                                    'n'; ' '; 'r'; 'u'; 'l'; 'e'; ' '; 'r'; 'e';
                                    'm'; 'o'; 'v'; 'a'; 'l'; ' '; 'w'; 'h'; 'e';
                                    'n'; ' '; 's'; 'w'; 'i'; 't'; 'c'; 'h'; 'i';
                                    'n'; 'g'; ' '; 'f'; 'r'; 'o'; 'm'; ' '; 't';
                                    'h'; 'e'; ' '; 'T'; 'R'; 'S'])
                            (comp shows_nl
                              (comp (shows_trs (show_lab _A2 _B2)
                                      (show_list show_char) r)
                                (comp shows_nl
                                  (comp (shows_string
  ['t'; 'o'; ' '; 't'; 'h'; 'e'; ' '; 'T'; 'R'; 'S'])
                                    (comp shows_nl
                                      (comp
(shows_trs (show_lab _A2 _B2) (show_list show_char) rR)
(comp shows_nl x)))))))))))
            (fun _ ->
              catch (checkTRSProof (_A1, _A2) (_B1, _B2)
                      (comp i (shows_string ['.'; '1'])) rR m prf)
                (fun x ->
                  Inl (comp i
                        (comp (shows_string
                                [':'; ' '; 'e'; 'r'; 'r'; 'o'; 'r'; ' '; 'b';
                                  'e'; 'l'; 'o'; 'w'; ' '; 't'; 'h'; 'e'; ' ';
                                  'r'; 'u'; 'l'; 'e'; ' '; 'r'; 'e'; 'm'; 'o';
                                  'v'; 'a'; 'l'; ' '; 'p'; 'r'; 'o'; 'c'; 'e';
                                  's'; 's'; 'o'; 'r'])
                          (comp shows_nl (indent x)))))))
    | i, r, Suc m, Unlab (uR, prf) ->
        debug (i []) ['u'; 'n'; 'l'; 'a'; 'b'; 'e'; 'l']
          (bind (catch
                  (unlab_trs_tt (_A1, _A2) (_B1, _B2)
                    (eq_lista eq_char, show_list show_char) r uR)
                  (fun x ->
                    Inl (comp i
                          (comp (shows_string
                                  [':'; ' '; 'e'; 'r'; 'r'; 'o'; 'r'; ' '; 'i';
                                    'n'; ' '; 'u'; 'n'; 'l'; 'a'; 'b'; 'e'; 'l';
                                    'i'; 'n'; 'g'; ' '; 'o'; 'f'; ' '; 'T'; 'R';
                                    'S'; ' '; 'w'; 'h'; 'e'; 'n'; ' '; 's'; 'w';
                                    'i'; 't'; 'c'; 'h'; 'i'; 'n'; 'g'; ' '; 'f';
                                    'r'; 'o'; 'm'; ' '; 't'; 'h'; 'e'; ' '; 'T';
                                    'R'; 'S'])
                            (comp shows_nl
                              (comp (shows_trs (show_lab _A2 _B2)
                                      (show_list show_char) r)
                                (comp shows_nl
                                  (comp (shows_string
  ['t'; 'o'; ' '; 't'; 'h'; 'e'; ' '; 'T'; 'R'; 'S'])
                                    (comp shows_nl
                                      (comp
(shows_trs (show_lab _A2 _B2) (show_list show_char) uR)
(comp shows_nl x)))))))))))
            (fun _ ->
              catch (checkTRSProof (_A1, _A2) (_B1, _B2)
                      (comp i (shows_string ['.'; '1'])) uR m prf)
                (fun x ->
                  Inl (comp i
                        (comp (shows_string
                                [':'; ' '; 'e'; 'r'; 'r'; 'o'; 'r'; ' '; 'b';
                                  'e'; 'l'; 'o'; 'w'; ' '; 't'; 'h'; 'e'; ' ';
                                  'u'; 'n'; 'l'; 'a'; 'b'; 'e'; 'l'; 'i'; 'n';
                                  'g'])
                          (comp shows_nl (indent x)))))))
    | i, r, Zero_nat, Unlab (uR, prf) ->
        Inl (comp (shows_string
                    ['u'; 'n'; 'l'; 'a'; 'b'; 'e'; 'l'; 'i'; 'n'; 'g'; ' '; 'n';
                      'o'; 't'; ' '; 'a'; 'l'; 'l'; 'o'; 'w'; 'e'; 'd'; ' ';
                      'f'; 'o'; 'r'; ' '; 'l'; 'a'; 'b'; 'e'; 'l'; ' '; 'd';
                      'e'; 'p'; 't'; 'h'; ' '; '0'])
              shows_nl)
    | i, r, m, RisEmpty ->
        debug (i []) ['R'; ' '; 'i'; 's'; ' '; 'e'; 'm'; 'p'; 't'; 'y']
          (check (null r)
            (comp i
              (comp (shows_string
                      [':'; ' '; 'R'; ' '; 'i'; 's'; ' '; 'n'; 'o'; 't'; ' ';
                        'e'; 'm'; 'p'; 't'; 'y'; ','; ' '; 'b'; 'u'; 't'; ' ';
                        'i'; 't'; ' '; 'i'; 's'; ' '; 't'; 'h'; 'e'; ' '; 'T';
                        'R'; 'S'])
                (comp (shows_trs (show_lab _A2 _B2) (show_list show_char) r)
                  shows_nl))))
    | i, r, m, Semlab (sli, l, lR, prf) ->
        debug (i []) ['S'; 'e'; 'm'; 'l'; 'a'; 'b']
          (bind (catch
                  (create_sem_lab_tt (_A1, _A2) (_B1, _B2)
                    (eq_lista eq_char, show_list show_char) sli l r lR)
                  (fun x ->
                    Inl (comp i
                          (comp (shows_string
                                  [':'; ' '; 'e'; 'r'; 'r'; 'o'; 'r'; ' '; 'w';
                                    'h'; 'e'; 'n'; ' '; 'a'; 'p'; 'p'; 'l'; 'y';
                                    'i'; 'n'; 'g'; ' '; 's'; 'e'; 'm'; 'a'; 'n';
                                    't'; 'i'; 'c'; ' '; 'l'; 'a'; 'b'; 'e'; 'l';
                                    'i'; 'n'; 'g'])
                            (comp shows_nl x)))))
            (fun _ ->
              catch (checkTRSProof (_A1, _A2) (_B1, _B2)
                      (comp i (shows_string ['.'; '1'])) lR (Suc m) prf)
                (fun x ->
                  Inl (comp i
                        (comp (shows_string
                                [':'; ' '; 'e'; 'r'; 'r'; 'o'; 'r'; ' '; 'b';
                                  'e'; 'l'; 'o'; 'w'; ' '; 's'; 'e'; 'm'; 'a';
                                  'n'; 't'; 'i'; 'c'; ' '; 'l'; 'a'; 'b'; 'e';
                                  'l'; 'i'; 'n'; 'g'])
                          (comp shows_nl (indent x)))))))
    | i, r, m, Rootlab (lR, prf) ->
        debug (i []) ['R'; 'o'; 'o'; 't'; 'l'; 'a'; 'b']
          (bind (catch (rl_tt (_A1, _A2) (_B1, _B2) r lR)
                  (fun x ->
                    Inl (comp i
                          (comp (shows_string
                                  [':'; ' '; 'e'; 'r'; 'r'; 'o'; 'r'; ' '; 'w';
                                    'h'; 'e'; 'n'; ' '; 'a'; 'p'; 'p'; 'l'; 'y';
                                    'i'; 'n'; 'g'; ' '; 'r'; 'o'; 'o'; 't'; '-';
                                    'l'; 'a'; 'b'; 'e'; 'l'; 'i'; 'n'; 'g'])
                            (comp shows_nl x)))))
            (fun _ ->
              catch (checkTRSProof (_A1, _A2) (_B1, _B2)
                      (comp i (shows_string ['.'; '1'])) lR (Suc m) prf)
                (fun x ->
                  Inl (comp i
                        (comp (shows_string
                                [':'; ' '; 'e'; 'r'; 'r'; 'o'; 'r'; ' '; 'b';
                                  'e'; 'l'; 'o'; 'w'; ' '; 'r'; 'o'; 'o'; 't';
                                  '-'; 'l'; 'a'; 'b'; 'e'; 'l'; 'i'; 'n'; 'g'])
                          (comp shows_nl (indent x)))))))
    | i, r, m, Fcc (fcs, rR, prf) ->
        debug (i []) ['F'; 'c'; 'c']
          (bind (catch (fcc_tt (eq_laba _A1 _B1, show_lab _A2 _B2) r fcs rR)
                  (fun x ->
                    Inl (comp i
                          (comp (shows_string
                                  [':'; ' '; 'e'; 'r'; 'r'; 'o'; 'r'; ' '; 'w';
                                    'h'; 'e'; 'n'; ' '; 'a'; 'p'; 'p'; 'l'; 'y';
                                    'i'; 'n'; 'g'; ' '; 'f'; 'l'; 'a'; 't'; ' ';
                                    'c'; 'o'; 'n'; 't'; 'e'; 'x'; 't'; ' '; 'c';
                                    'l'; 'o'; 's'; 'u'; 'r'; 'e'])
                            (comp shows_nl x)))))
            (fun _ ->
              catch (checkTRSProof (_A1, _A2) (_B1, _B2)
                      (comp i (shows_string ['.'; '1'])) rR Zero_nat prf)
                (fun x ->
                  Inl (comp i
                        (comp (shows_string
                                [':'; ' '; 'e'; 'r'; 'r'; 'o'; 'r'; ' '; 'b';
                                  'e'; 'l'; 'o'; 'w'; ' '; 'f'; 'l'; 'a'; 't';
                                  ' '; 'c'; 'o'; 'n'; 't'; 'e'; 'x'; 't'; ' ';
                                  'c'; 'l'; 'o'; 's'; 'u'; 'r'; 'e'])
                          (comp shows_nl (indent x)))))))
    | i, r, m, StringReversal (rR, prf) ->
        debug (i [])
          ['S'; 't'; 'r'; 'i'; 'n'; 'g'; ' '; 'R'; 'e'; 'v'; 'e'; 'r'; 's'; 'a';
            'l']
          (bind (catch
                  (check_string_rev (eq_laba _A1 _B1, show_lab _A2 _B2)
                    (eq_lista eq_char, show_list show_char) r rR)
                  (fun x ->
                    Inl (comp i
                          (comp (shows_string
                                  [':'; ' '; 'e'; 'r'; 'r'; 'o'; 'r'; ' '; 'w';
                                    'h'; 'e'; 'n'; ' '; 'a'; 'p'; 'p'; 'l'; 'y';
                                    'i'; 'n'; 'g'; ' '; 's'; 't'; 'r'; 'i'; 'n';
                                    'g'; ' '; 'r'; 'e'; 'v'; 'e'; 'r'; 's'; 'a';
                                    'l'])
                            (comp shows_nl x)))))
            (fun _ ->
              catch (checkTRSProof (_A1, _A2) (_B1, _B2)
                      (comp i (shows_string ['.'; '1'])) rR Zero_nat prf)
                (fun x ->
                  Inl (comp i
                        (comp (shows_string
                                [':'; ' '; 'e'; 'r'; 'r'; 'o'; 'r'; ' '; 'b';
                                  'e'; 'l'; 'o'; 'w'; ' '; 's'; 't'; 'r'; 'i';
                                  'n'; 'g'; ' '; 'r'; 'e'; 'v'; 'e'; 'r'; 's';
                                  'a'; 'l'])
                          (comp shows_nl (indent x)))))));;

let rec certifyCertProblem (_A1, _A2) (_B1, _B2)
  = function
    TRSProof (trs, prf) ->
      (match
        checkTRSProof (_A1, _A2) (_B1, _B2) (shows_string ['1'])
          (map (map_rule init_lab) trs) Zero_nat prf
        with Inl err ->
          Error (comp (shows_string
                        ['t'; 'e'; 'r'; 'm'; 'i'; 'n'; 'a'; 't'; 'i'; 'o'; 'n';
                          ' '; 'p'; 'r'; 'o'; 'o'; 'f'; ' '; 'n'; 'o'; 't'; ' ';
                          'a'; 'c'; 'c'; 'e'; 'p'; 't'; 'e'; 'd'])
                  (comp shows_nl err) [])
        | Inr _ ->
          Certified
            ['<'; 't'; 'r'; 's'; 'T'; 'e'; 'r'; 'm'; 'i'; 'n'; 'a'; 't'; 'i';
              'o'; 'n'; 'P'; 'r'; 'o'; 'o'; 'f'; '>'])
    | TRSDisproof (trs, prf) ->
        (match
          checkTRSDisproof (_A1, _A2) (eq_lista eq_char, show_list show_char)
            (shows_string ['1']) trs prf
          with Inl err ->
            Error (comp (shows_string
                          ['n'; 'o'; 'n'; '-'; 't'; 'e'; 'r'; 'm'; 'i'; 'n';
                            'a'; 't'; 'i'; 'o'; 'n'; ' '; 'p'; 'r'; 'o'; 'o';
                            'f'; ' '; 'n'; 'o'; 't'; ' '; 'a'; 'c'; 'c'; 'e';
                            'p'; 't'; 'e'; 'd'])
                    (comp shows_nl err) [])
          | Inr _ ->
            Certified
              ['<'; 't'; 'r'; 's'; 'N'; 'o'; 'n'; 't'; 'e'; 'r'; 'm'; 'i'; 'n';
                'a'; 't'; 'i'; 'o'; 'n'; 'P'; 'r'; 'o'; 'o'; 'f'; '>']);;

let rec certificationResult_case
  f1 f2 x2 = match f1, f2, x2 with f1, f2, Error lista -> f2 lista
    | f1, f2, Certified lista -> f1 lista;;

let rec certificationProblem_case
  f1 f2 x2 = match f1, f2, x2 with
    f1, f2, TRSDisproof (lista, tRSDisproof) -> f2 lista tRSDisproof
    | f1, f2, TRSProof (lista, tRSProof) -> f1 lista tRSProof;;

let rec xml_certificationProblem _A _C
  cp = let rls =
         (match cp with TRSProof (lr, _) -> lr | TRSDisproof (lr, _) -> lr) in
       Xml (['t'; 'r'; 's'; 'I'; 'n'; 'p'; 'u'; 't'], [], [xml_rules _A _C rls],
             None);;

let rec eqWhiteSpace
  s1 s2 =
    debug ['0']
      ['c'; 'o'; 'm'; 'p'; 'a'; 'r'; 'i'; 'n'; 'g'; ' '; 'w'; 'h'; 'e'; 't';
        'h'; 'e'; 'r'; ' '; 'p'; 'a'; 'r'; 's'; 'e'; 'd'; ' '; 'i'; 'n'; 'p';
        'u'; 't'; ' '; 'c'; 'o'; 'r'; 'r'; 'e'; 's'; 'p'; 'o'; 'n'; 'd'; 's';
        ' '; 't'; 'o'; ' '; 'r'; 'e'; 'a'; 'l'; ' '; 'i'; 'n'; 'p'; 'u'; 't']
      (eq_list eq_char (filter (fun c -> not (member eq_char c wspace)) s1)
        (filter (fun c -> not (member eq_char c wspace)) s2));;

let rec certifyProof
  problemString =
    (match parseCertProblem problemString
      with Inl err ->
        Error (append
                ['e'; 'r'; 'r'; 'o'; 'r'; ' '; 'w'; 'h'; 'i'; 'l'; 'e'; ' ';
                  'p'; 'a'; 'r'; 's'; 'i'; 'n'; 'g']
                (append [newline] err))
      | Inr cp ->
        (match extract_input_proof problemString
          with None ->
            Error ['c'; 'o'; 'u'; 'l'; 'd'; ' '; 'n'; 'o'; 't'; ' '; 'f'; 'i';
                    'n'; 'd'; ' '; 'i'; 'n'; 'p'; 'u'; 't'; ' '; 'a'; 'n'; 'd';
                    ' '; 'p'; 'r'; 'o'; 'o'; 'f'; ' '; 'i'; 'n'; ' '; 's'; 't';
                    'r'; 'i'; 'n'; 'g'; ' '; 'f'; 'o'; 'r'; ' '; 'c'; 'o'; 'm';
                    'p'; 'a'; 'r'; 'i'; 's'; 'o'; 'n'; ' '; 'w'; 'i'; 't'; 'h';
                    ' '; 'p'; 'a'; 'r'; 's'; 'e'; 'd'; ' '; 'r'; 'e'; 's'; 'u';
                    'l'; 't']
          | Some (the_input, (_, the_proof)) ->
            (if eqWhiteSpace the_input
                  (shows_prec_xml Zero_nat
                    (xml_certificationProblem (show_list show_char)
                      (show_list show_char) cp)
                    [])
              then (match
                     certifyCertProblem (eq_lista eq_char, show_list show_char)
                       (eq_lista eq_nata, show_list show_nat) cp
                     with Certified tag ->
                       (if starts_with the_proof tag then Certified tag
                         else Error (append
                                      ['p'; 'r'; 'o'; 'v'; 'e'; 'n'; ' '; 'p';
'r'; 'o'; 'p'; 'e'; 'r'; 't'; 'y'; ' ']
                                      (append tag
(append
  [' '; 'd'; 'o'; 'e'; 's'; ' '; 'n'; 'o'; 't'; ' '; 'c'; 'o'; 'r'; 'r'; 'e';
    's'; 'p'; 'o'; 'n'; 'd'; ' '; 't'; 'o'; ' '; 'p'; 'r'; 'o'; 'o'; 'f'; ' ';
    'i'; 'n'; ' '; 'i'; 'n'; 'p'; 'u'; 't'; ' '; 's'; 't'; 'r'; 'i'; 'n'; 'g';
    ':'; ' ']
  (take (nat (Big_int.big_int_of_int 20)) the_proof)))))
                     | Error a -> Error a)
              else Error (concat
                           [['p'; 'a'; 'r'; 's'; 'e'; 'd'; ' '; 'p'; 'r'; 'o';
                              'b'; 'l'; 'e'; 'm'; ' '; 'd'; 'o'; 'e'; 's'; ' ';
                              'n'; 'o'; 't'; ' '; 'c'; 'o'; 'r'; 'r'; 'e'; 's';
                              'p'; 'o'; 'n'; 'd'; ' '; 't'; 'o'; ' '; 'i'; 'n';
                              'p'; 'u'; 't'];
                             [newline]; ['i'; 'n'; 'p'; 'u'; 't'; ':'];
                             [newline];
                             filter (fun c -> not (member eq_char c wspace))
                               the_input;
                             [newline]; [newline];
                             ['p'; 'a'; 'r'; 's'; 'e'; 'd'; ':']; [newline];
                             filter (fun c -> not (member eq_char c wspace))
                               (shows_prec_xml Zero_nat
                                 (xml_certificationProblem (show_list show_char)
                                   (show_list show_char) cp)
                                 []);
                             [newline]]))));;

end;; (*struct Ceta*)
